unit Unit1;
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ActnList, ToolWin, ImgList, UMultiThread, CepUtils,
  StdCtrls, Spin, ExtCtrls;

type
  TMainForm = class(TForm)
    ImageList1: TImageList;
    ToolBar1: TToolBar;
    ActionList1: TActionList;
    ActExit: TAction;
    ActMultiThread: TAction;
    ActTraditional: TAction;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    StatusBar1: TStatusBar;
    CheckBox1: TCheckBox;
    ToolButton4: TToolButton;
    Panel1: TPanel;
    seThreadCount: TSpinEdit;
    procedure ActExitExecute(Sender: TObject);
    procedure ActMultiThreadExecute(Sender: TObject);
    procedure ActMultiThreadUpdate(Sender: TObject);
    procedure MultiThreadFor1BeforeStart(Sender: TMultiThreadFor);
    procedure MultiThreadFor1AfterFinish(Sender: TMultiThreadFor;
      const ErrorStack: array of TErrorStack);
    procedure FormPaint(Sender: TObject);
    procedure MultiThreadFor1Percent(Sender: TMultiThreadFor;
      const Percent: Integer);
    procedure MultiThreadFor1Execute(Sender: TMultiThreadFor; ThreadIndex: Word;
      const Index: Cardinal; const VectorIndex: array of Cardinal;
      const Vector: array of Extended);
    procedure ActTraditionalExecute(Sender: TObject);
    procedure seThreadCountChange(Sender: TObject);
  private
    { Private declarations }
    fMultiThreadFor: TMultiThreadFor;
    fW: integer;
    fH: integer;
    fA: array of integer;
    fT: Int64;
    fBitmap: TBitmap;
    fTraditionalStarted: boolean;
    procedure UpdateStatus;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateBitmap;
  end;

var
  MainForm: TMainForm;

implementation

const
  TextStatus: array [TMultiThreadStatus] of string = (
    '-',
    'Performing MultiThread',
    'MultiThread: Execution terminated',
    'MultiThread: in the implementation of any errors',
    'MultiThread: execution completed successfully');

resourcestring
  StrStop = 'End calculation';
  StrStartMultithread = 'Start calculations by using TMultiThreadFor';
  StrTraditionalCalc = 'In the main stream';

{$R *.dfm}


procedure TMainForm.ActExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.ActMultiThreadExecute(Sender: TObject);
begin
  if fMultiThreadFor.Started then
    fMultiThreadFor.Stop
  else
    fMultiThreadFor.Start;
end;

procedure TMainForm.ActMultiThreadUpdate(Sender: TObject);
begin
  if fMultiThreadFor.Started then
  begin
    ActMultiThread.Hint := StrStop;
    ActMultiThread.ImageIndex := 3
  end
  else
  begin
    ActMultiThread.Hint := StrStartMultithread;
    ActMultiThread.ImageIndex := 1;
  end;
end;


constructor TMainForm.Create(AOwner: TComponent);
  procedure AddOrder(ADisplayName: string; AStartValue, AFinishValue: Extended);
  begin
    fMultiThreadFor.Orders.Add;
    with fMultiThreadFor.Orders[fMultiThreadFor.Orders.Count - 1] do
    begin
      DisplayName := ADisplayName;
      StartValue := AStartValue;
      FinishValue := AFinishValue;
    end;
  end;
begin
  inherited;
  InitFreqCPU;
  fMultiThreadFor := TMultiThreadFor.Create(self);
  fMultiThreadFor.onExecute := MultiThreadFor1Execute;
  fMultiThreadFor.onPercent := MultiThreadFor1Percent;
  fMultiThreadFor.onBeforeStart := MultiThreadFor1BeforeStart;
  fMultiThreadFor.onAfterFinish := MultiThreadFor1AfterFinish;
  seThreadCount.Value := fMultiThreadFor.ThreadCount;
  AddOrder('X', -2, 0.5);
  AddOrder('Y', -1.2, 1.2);
end;

destructor TMainForm.Destroy;
begin
  FreeAndNil(fBitmap);
  inherited;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  if fBitmap <> nil then
  begin
    canvas.Draw(0, ToolBar1.Height, fBitmap);
  end;
end;

procedure TMainForm.UpdateStatus;
begin
  if fTraditionalStarted then
    StatusBar1.Panels[0].Text := StrTraditionalCalc
  else
    StatusBar1.Panels[0].Text := TextStatus[fMultiThreadFor.Status];
end;

procedure TMainForm.UpdateBitmap;
var I, x, y: integer;
begin
  if CheckBox1.Checked then Exit;
  if fBitmap = nil then
  begin
    fBitmap := TBitmap.Create;
    fBitmap.PixelFormat := pf24Bit;
  end;
  fBitmap.Width := fW;
  fBitmap.Height := fH;
  I := 0;
  for y := 0 to fH - 1 do
    for x := 0 to fW - 1 do
    begin
      fBitmap.Canvas.Pixels[x, y] := fA[I];
      inc(I);
    end;
  invalidate;
end;

//Events sufficient for the work component MultiThreadFor.
//------------------------------------------------------------

//Before the calculations, allocates memory for an array with the results
//of calculations. The size of the resulting image will correspond to
//the size of the client side of the form at the moment.
procedure TMainForm.MultiThreadFor1BeforeStart(Sender: TMultiThreadFor);
begin
  ActTraditional.Enabled := false;
  seThreadCount.Enabled := False;
  fW := ClientWidth;
  fH := ClientHeight - StatusBar1.Height - ToolBar1.Height;
  SetLength(fA, fW * fH);
  FillChar(fA[0], fW * fH * SizeOf(fA[0]), 0);
  fMultiThreadFor.Orders[0].Count := fW;
  fMultiThreadFor.Orders[1].Count := fH;
  UpdateStatus;
  CountCPU(fT);
end;

//Basic mathematical computations. Will determine whether the point belongs
//coordinates Vector, the Mandelbrot set.
//Not what optimization is not, because the main task to start your computer calculations
procedure TMainForm.MultiThreadFor1Execute(Sender: TMultiThreadFor;
  ThreadIndex: Word; const Index: Cardinal;
  const VectorIndex: array of Cardinal; const Vector: array of Extended);
const MaxCount = 65535;
      MaxColor = 255 * 255 * 255;
var N: integer;
    X, Y, x2, y2, p, q: Extended;
begin
  N := 0;
  X := 0;
  Y := 0;
  p := Vector[0];
  q := Vector[1];
  while N < MaxCount do
  begin
    x2 := sqr(X);
    y2 := sqr(Y);
    if X2 + Y2 > 1E50 then break;
    x2 := x2 - y2 + p;
    y := 2 * X * Y + q;
    x := x2;
    inc(N);
  end;
  if N = MaxCount then fA[Index] := 0
                  else fA[Index] := Round(ln(N - 5) / ln(MaxCount - 5) * MaxColor);
end;

// Periodically update the status line and the image
procedure TMainForm.MultiThreadFor1Percent(Sender: TMultiThreadFor;
  const Percent: Integer);
begin
  StatusBar1.Panels[1].Text := Format('%f%%', [Percent/10]);
  UpdateBitmap;
  UpdateStatus;
end;

//After calculations, we obtain the duration of execution,
//and write information to log
procedure TMainForm.MultiThreadFor1AfterFinish(Sender: TMultiThreadFor;
  const ErrorStack: array of TErrorStack);
var List: TStringList;
    S, FileName: string;
begin
  fT := CountCPU(fT);
  seThreadCount.Enabled := True;
  StatusBar1.Panels[1].Text := CountCPUtoSTR(fT);
  ActTraditional.Enabled := true;
  UpdateStatus;
  try
    FileName := ChangeFileExt(ParamStr(0), '.log');
    List := TStringList.Create;
    try
      if fTraditionalStarted then
        S := StrTraditionalCalc
      else
        S := TextStatus[fMultiThreadFor.Status];
      S := Format('%s  Width: %d; Height: %d; Time: %s', [S, fW, fH, CountCPUtoSTR(fT)]);
      try
        List.LoadFromFile(FileName);
      except
      end;
      List.Add(S);
      List.SaveToFile(FileName);
    finally
      List.Free;
    end;
  except
  end;
end;
// Late events MultiThreadFor
//------------------------------------------------------------

//Performing the same calculations without the use of MultiThreadFor
procedure TMainForm.ActTraditionalExecute(Sender: TObject);
var Dx, Dy: Extended;
    i, j, Index: integer;
    Vector: array of Extended;
    VectorIndex: array of DWORD;
    LastTick1, LastTick2: DWORD;
begin
  fTraditionalStarted := true;
  ActMultiThread.Enabled := false;
  MultiThreadFor1BeforeStart(nil);
  try
    Dx := (fMultiThreadFor.Orders[0].FinishValue - fMultiThreadFor.Orders[0].StartValue) / (fMultiThreadFor.Orders[0].Count - 1);
    Dy := (fMultiThreadFor.Orders[1].FinishValue - fMultiThreadFor.Orders[1].StartValue) / (fMultiThreadFor.Orders[1].Count - 1);
    SetLength(Vector, fMultiThreadFor.Orders.Count);
    SetLength(VectorIndex, fMultiThreadFor.Orders.Count);
    Index := 0;
    LastTick1 := GetTickCount;
    LastTick2 := LastTick1;
    for J := 0 to fMultiThreadFor.Orders[1].Count - 1 do
    begin
      Vector[1] := fMultiThreadFor.Orders[1].StartValue + Dy * J;
      VectorIndex[1] := J;
      for I := 0 to fMultiThreadFor.Orders[0].Count - 1 do
      begin
        Vector[0] := fMultiThreadFor.Orders[0].StartValue + Dx * I;
        VectorIndex[0] := I;
        MultiThreadFor1Execute(nil, 0, Index, VectorIndex, Vector);
        inc(Index);
        if (Index mod 100 = 0) and (GetTickCount - LastTick1 > 40) then
        begin
          LastTick1 := GetTickCount;
          Application.ProcessMessages;
          if GetTickCount - LastTick2 > 1000 then
          begin
            LastTick2 := GetTickCount;
            MultiThreadFor1Percent(nil, Round(1000 * index / (fW * fH)));
          end;
        end;
      end;
    end;
  finally
    MultiThreadFor1Percent(nil, 1000);
    MultiThreadFor1AfterFinish(nil, []);
    fTraditionalStarted := false;
    ActMultiThread.Enabled := True;;
  end;
end;

procedure TMainForm.seThreadCountChange(Sender: TObject);
begin
  fMultiThreadFor.DefThreadCount := seThreadCount.Value;
end;

end.
