unit ActionTreeView;

interface
{$I Directives.Inc}
uses SysUtils, Windows, Messages, Classes, Graphics, ImgList,
  {$IFDEF DELPHI_6_UP}RTLConsts,{$ELSE}Consts,{$ENDIF}
  Controls, ActnList, Commctrl, ComCtrls, SpeedButtonCep, BitOp,
  {$IFDEF DELPHI_6_UP} ActnMan, {$ENDIF}
  forms;

resourcestring
  StrECreateNode = '     TActionNode';

type
  TNodeClient = class; //        
  TNodeClients = class; //       
  TActionNode = class; //       
  TActionTreeView = class; //  ,   ,  

  TNodeClientClass = class of TNodeClient;
  TNodeClientsClass = class of TNodeClients;
  TActionNodeClass = class of TActionNode;


  //        
  TNodeClient = class(TCollectionItem)
  private
    fActionLink: TActionLink;
    fParentItem: TNodeClient;
    fItems: TNodeClients;
    fActionNode: TActionNode;
    fVisible: boolean;
    fEnabled: boolean;
    fChecked: boolean;
    fCaption: TCaption;
    fImageIndex: TImageIndex;
    fExpanded: boolean;
    fActionTreeView: TActionTreeView;
    procedure SetItems(const Value: TNodeClients);
    function GetActionTreeView: TActionTreeView;
    procedure SetVisible(const Value: boolean);
    function GetAction: TCustomAction;
    procedure SetAction(const Value: TCustomAction);
    procedure SetChecked(const Value: boolean);
    procedure SetEnabled(const Value: boolean);
    procedure Invalidate;
    procedure SetCaption(const Value: TCaption);
    procedure SetImageIndex(const Value: TImageIndex);
    procedure SetExpanded(const Value: boolean);
    procedure SetDefaultValues;
    function VisibleStored: Boolean;
    function EnabledStored: Boolean;
    function CheckedStored: Boolean;
    function ImageIndexStored: Boolean;
  protected
    function GetDisplayName: string; override;
    function NeedUpdate: boolean;
    procedure AssignValues(Source: TPersistent); virtual;
    class procedure AssignError(Source: TPersistent; ClassName: string);
    {$IFDEF DELPHI_2009_UP}
    procedure SetCollection(Value: TCollection); override;
    {$ENDIF}
    //  ItemsClass     Items.
    //  Nil,    TNodeClients
    function ItemsClass: TNodeClientsClass; virtual;
    //  ItemClass      Items
    //  Nil,    TNodeClient
    function ItemClass: TNodeClientClass; virtual;
    procedure DoUpdateNode; virtual;
    function CaptionStored: boolean;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    property ActionNode: TActionNode read fActionNode;
    property ActionTreeView: TActionTreeView read fActionTreeView;
    //   True,         
    function RootVisible: boolean;
    procedure UpdateNode;
    property ParentItem: TNodeClient read fParentItem;
    procedure Assign(Source: TPersistent); override;
  published
    property Items: TNodeClients read fItems write SetItems;
    property Visible: boolean read fVisible write SetVisible stored VisibleStored;
    property Enabled: boolean read fEnabled write SetEnabled stored EnabledStored;
    property Checked: boolean read fChecked write SetChecked stored CheckedStored;
    property Action: TCustomAction read GetAction write SetAction;
    property Caption: TCaption read fCaption write SetCaption stored CaptionStored;
    property Expanded: boolean read fExpanded write SetExpanded default true;
    property ImageIndex: TImageIndex read fImageIndex write SetImageIndex stored ImageIndexStored;
  end;

  TIterateProc = procedure(Item: TNodeClient) of object;

  //       
  TNodeClients = class(TOwnedCollection)
  private
    function GetItem(Index: Integer): TNodeClient;
    procedure SetItem(Index: Integer; const Value: TNodeClient);
  protected
    procedure Update(Item: TCollectionItem); override;
    procedure UpdateNode(Item: TNodeClient);
  public
    constructor Create(AOwner: TPersistent; ItemClass: TNodeClientClass);
    procedure Assign(Source: TPersistent); override;
    class function GetActionTreeView(NodeClients: TNodeClients): TActionTreeView;
    class procedure IterateClients(NodeClients: TNodeClients; ActionProc: TIterateProc);
    property Items[Index: Integer]: TNodeClient read GetItem write SetItem; default;
  end;

  TMarkedNodeType = (mntNone, mntBefore, mntAfter, mntChild, mntRoot);

  //       
  TActionNode = class(TTreeNode)
  private
    fNodeClient: TNodeClient;
    function GetAction: TCustomAction;
  protected
    procedure DrawMarked(TreeView: TActionTreeView; R: TRect); virtual;
    procedure DrawImage(TreeView: TActionTreeView; State: TCustomDrawState;
      Image: TCustomImageList; Index: integer; R: TRect); virtual;
    procedure DoDraw(Canvas: TCanvas; TreeView: TActionTreeView; State: TCustomDrawState; var PaintImages, DefaultDraw:
      boolean); virtual;
  public
    constructor Create(AOwner: TTreeNodes); {$IFDEF DELPHI_2009_UP} override; {$ENDIF}
    destructor Destroy; override;
    function RootEnabled: boolean;
    property NodeClient: TNodeClient read fNodeClient;
    property Action: TCustomAction read GetAction;
  end;

  TGliphState = class(TCustomGliphs)
  private
    fChecked: TImageIndex;
    fUnchecked: TImageIndex;
    function GetActionTreeView: TActionTreeView;
    function GetChecked: TImageIndex;
    function GetUnchecked: TImageIndex;
    procedure SetChecked(const Value: TImageIndex);
    procedure SetUnchecked(const Value: TImageIndex);
  protected
    procedure Changed; override;
    function GetHandle: THandle; override;
  public
    constructor Create(AOwner: TActionTreeView);
    property ActionTreeView: TActionTreeView read GetActionTreeView;
    procedure Assign(Source: TPersistent); override;
  published
    property Checked: TImageIndex read GetChecked write SetChecked default 2;
    property Unchecked: TImageIndex read GetUnchecked write SetUnchecked default 1;
  end;

  //  ,   ,  
  TActionTreeView = class(TCustomTreeView)
  private
    fNodeClients: TNodeClients;
    fInCreateNode: boolean;
    fInCreateNodeClient: boolean;
    {$IFDEF DELPHI_2005_UP}
    fNeedFreeEmptyNodes: boolean;
    {$ENDIF}
    fBtnFrameColor: TColor;
    fSelectedColor: TColor;
    fDisabledFontColor: TColor;
    fStateIndex: array[boolean] of TImageIndex;
    fHighlightTextColor: TColor;
    fDragged: boolean;
    fDragNode: TActionNode;
    fTargetNode: TActionNode;
    fMarkedChange: boolean;
    fMarkedNodeType: TMarkedNodeType;
    fUpdateCount: integer;
    fGliphState: TGliphState;
    fIsDesigning: boolean;
    fRichImageList: TRichImageList;
    FImageChangeLink: TChangeLink;
    FStateChangeLink: TChangeLink;
    procedure SetNodeClients(const Value: TNodeClients);
    procedure SetBtnFrameColor(const Value: TColor);
    procedure SetSelectedColor(const Value: TColor);
    procedure SetDisabledFontColor(const Value: TColor);
    function GetStateIndex(Enabled: boolean): TImageIndex;
    procedure SetStateIndex(Enabled: boolean; const Value: TImageIndex);
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure SetHighlightTextColor(const Value: TColor);
    function GetColorDef(Color, DefaultColor: TColor): TColor;
    procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
    procedure SetDragged(const Value: boolean);
    procedure SetTargetNode(const Value: TActionNode);
    procedure SetMarkedNodeType(const Value: TMarkedNodeType);
    {$IFDEF DELPHI_2005_UP}
    procedure WMSETREDRAW(var Message: TWMSETREDRAW); message WM_SETREDRAW;
    {$ENDIF}
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
    procedure SetGliphState(const Value: TGliphState);
  protected
    function ActionNodeClass: TActionNodeClass; virtual;
    function NodeClientsClass: TNodeClientsClass; virtual;
    function NodeClientsItemClass: TNodeClientClass; virtual;
    function CreateRichImageList: TRichImageList; virtual;
    property InCreateNode: boolean read fInCreateNode;
    property InCreateNodeClient: boolean read fInCreateNodeClient;
    function CreateNode: TTreeNode; override;
    {$IFDEF DELPHI_6_UP}
    function CreateNodes: TTreeNodes; override;
    {$ENDIF}
    procedure ImageListChange(Sender: TObject); virtual;
    procedure SetImages(Value: TCustomImageList); virtual;
    procedure SetStateImages(Value: TCustomImageList); virtual;
    procedure UpdateCanvasColor(Canvas: TCanvas; State: TCustomDrawState);
    procedure Expand(Node: TTreeNode); override;
    procedure Collapse(Node: TTreeNode); override;
    procedure BeginUpdate;
    procedure EndUpdate;
    property UpdateCount: integer read fUpdateCount;
    procedure KeyPress(var Key: Char); override;
    procedure DblClick; override;

    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean); override;
    procedure DoEndDrag(Target: TObject; X, Y: Integer); override;
    procedure DoStartDrag(var DragObject: TDragObject); override;
    procedure DoDragOver(Source: TDragObject; X, Y: Integer; CanDrop: Boolean); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property StateIndex[Enabled: boolean]: TImageIndex read GetStateIndex write SetStateIndex;
    function CopyNodeClient(Sourc: TPersistent; Dest: TNodeClient;
      MarkedNodeType: TMarkedNodeType; DelSource: boolean = false): TNodeClient;
    function ExecuteSelected: boolean;
    procedure DropSelectedNode(DelChilds: boolean);

    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    property Dragged: boolean read fDragged write SetDragged;
    property DragNode: TActionNode read fDragNode;
    property TargetNode: TActionNode read fTargetNode write SetTargetNode;
    property MarkedNodeType: TMarkedNodeType read fMarkedNodeType write SetMarkedNodeType;
    property IsDesigning: boolean read fIsDesigning write fIsDesigning;
  published
    property NodeClients: TNodeClients read fNodeClients write SetNodeClients;
    property BtnFrameColor: TColor read fBtnFrameColor write SetBtnFrameColor default $00C66931;
    property SelectedColor: TColor read fSelectedColor write SetSelectedColor default $00EFD3C6;
    property HighlightTextColor: TColor read fHighlightTextColor write SetHighlightTextColor default clNone;
    property DisabledFontColor: TColor read fDisabledFontColor write SetDisabledFontColor default clGrayText;
    property GliphState: TGliphState read fGliphState write SetGliphState;

    property Align;
    property Anchors;
    property AutoExpand;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property ChangeDelay;
    property Color;
    property Ctl3D;
    property Constraints;
    property DragKind;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property HotTrack;
    property Images write SetImages;
    property Indent;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RightClickSelect;
    property RowSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property StateImages write SetStateImages;
    property TabOrder;
    property TabStop default True;
    property ToolTips;
    property Visible;

    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

{$R ActionTreeView.res}

procedure Register;
begin
  RegisterComponents('Cep', [TActionTreeView]);
end;

type
  //    TActionLink  
  //      
  TNodeClientLink = class(TActionLink)
  private
    FClient: TNodeClient;
  protected
    procedure AssignClient(AClient: TObject); override;
    function IsCaptionLinked: Boolean; override;
    function IsEnabledLinked: Boolean; override;
    function IsImageIndexLinked: Boolean; override;
    function IsCheckedLinked: Boolean; override;
    function IsVisibleLinked: Boolean; override;
    procedure SetAction(Value: TBasicAction); override;
    procedure SetCaption(const Value: string); override;
    procedure SetChecked(Value: Boolean); override;
    procedure SetEnabled(Value: Boolean); override;
    procedure SetImageIndex(Value: Integer); override;
    procedure SetVisible(Value: Boolean); override;
  end;

  { TNodeClientLink }

procedure TNodeClientLink.AssignClient(AClient: TObject);
var SourceName: string;
begin
  inherited;
  if AClient is TNodeClient then
  begin
    FClient := TNodeClient(AClient);
  end
  else
  begin
    if AClient <> nil then
      SourceName := AClient.ClassName
    else
      SourceName := 'nil';
    raise EConvertError.CreateFmt(SAssignError, [SourceName, FClient.ClassName]);
  end;
end;

procedure TNodeClientLink.SetAction(Value: TBasicAction);
var NeedCaption: boolean;
begin
  if Action <> Value then
  begin
    NeedCaption := (FClient <> nil) and
      ((FClient.Caption = '')
      or ((Action is TCustomAction) and (TCustomAction(Action).Caption = FClient.Caption)));
    if Value is TCustomAction then
    begin
      inherited;
      if FClient <> nil then
      begin
        if NeedCaption then FClient.fCaption := TCustomAction(Action).Caption;
        if FClient.fImageIndex = -1 then
          FClient.fImageIndex := TCustomAction(Action).ImageIndex;
        FClient.fVisible := TCustomAction(Action).Visible;
        FClient.fEnabled := TCustomAction(Action).Enabled;
        FClient.fChecked := TCustomAction(Action).Checked;
        FClient.UpdateNode;
      end;
    end
    else
    begin
      if Value <> nil then
        raise EConvertError.CreateFmt(SAssignError, [Value.ClassName, TCustomAction.ClassName]);
      inherited;
      if NeedCaption then
        FClient.Caption := '';
    end;
  end;
end;

function TNodeClientLink.IsCaptionLinked: Boolean;
begin
  Result := inherited IsCaptionLinked and (FClient <> nil)
    and (TCustomAction(Action).Caption = FClient.Caption);
end;

function TNodeClientLink.IsCheckedLinked: Boolean;
begin
  Result := inherited IsCheckedLinked and (FClient <> nil);
end;

function TNodeClientLink.IsEnabledLinked: Boolean;
begin
  Result := inherited IsEnabledLinked and (FClient <> nil);
end;

function TNodeClientLink.IsImageIndexLinked: Boolean;
var ActionImages, ClientImages: TCustomImageList;
begin
  Result := inherited IsImageIndexLinked and (FClient <> nil) and
          (TCustomAction(Action).ImageIndex = FClient.ImageIndex);
  if Result then
  begin
    ActionImages := nil;
    ClientImages := nil;
    if TCustomAction(Action).ActionList <> nil then
      ActionImages := TCustomAction(Action).ActionList.Images;
    if ActionImages <> nil then
    begin
      if FClient.ActionTreeView <> nil then
        ClientImages := FClient.ActionTreeView.Images;
      Result := ClientImages = ActionImages;
    end;
  end;
end;

function TNodeClientLink.IsVisibleLinked: Boolean;
begin
  Result := inherited IsVisibleLinked and (FClient <> nil);
end;

procedure TNodeClientLink.SetCaption(const Value: string);
begin
  inherited;
  if IsCaptionLinked then FClient.Caption := Value;
end;

procedure TNodeClientLink.SetChecked(Value: Boolean);
begin
  inherited;
  if IsCheckedLinked then FClient.Checked := Value;
end;

procedure TNodeClientLink.SetEnabled(Value: Boolean);
begin
  inherited;
  if IsEnabledLinked then FClient.Enabled := Value;
end;

procedure TNodeClientLink.SetImageIndex(Value: Integer);
begin
  inherited;
  if IsImageIndexLinked then FClient.ImageIndex := Value;
end;

procedure TNodeClientLink.SetVisible(Value: Boolean);
begin
  inherited;
  if IsVisibleLinked then FClient.Visible := Value;
end;

type TDummyNodes = class(TTreeNodes)
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    procedure Assign(Source: TPersistent); override;
  end;

  { TDummyNodes }

procedure TDummyNodes.Assign(Source: TPersistent);
begin
  //inherited;
end;

procedure TDummyNodes.DefineProperties(Filer: TFiler);
begin
  //inherited;
end;

{ TNodeClient }

class procedure TNodeClient.AssignError(Source: TPersistent;
  ClassName: string);
var
  SourceName: string;
begin
  if Source <> nil then
    SourceName := Source.ClassName
  else
    SourceName := 'nil';
  raise EConvertError.CreateFmt(SAssignError, [SourceName, ClassName]);
end;

procedure TNodeClient.SetDefaultValues;
begin
  fVisible := true;
  fEnabled := true;
  fChecked := false;
  fExpanded := true;
  fImageIndex := -1;
  if fActionLink <> nil then
  begin
    fActionLink.Action := nil;
    fCaption := '';
  end;
end;

constructor TNodeClient.Create(Collection: TCollection);
var ANodeClientsClass: TNodeClientsClass;
  ANodeClientClass: TNodeClientClass;
begin
  {$IFNDEF DELPHI_2009_UP}
  if (Collection <> nil) and (not (Collection.InheritsFrom(TNodeClients))) then
    AssignError(Collection, TNodeClientsClass.ClassName);
  {$ENDIF}
  inherited Create(Collection);
  {$IFNDEF DELPHI_2009_UP}
  if (Collection is TNodeClients) and
    (TNodeClients(Collection).GetOwner is TNodeClient) then
    fParentItem := TNodeClient(TNodeClients(Collection).GetOwner)
  else
    fParentItem := nil;
  fActionTreeView := GetActionTreeView;
  {$ENDIF}
  SetDefaultValues;
  fActionLink := TNodeClientLink.Create(self);
  ANodeClientsClass := ItemsClass;
  if ANodeClientsClass = nil then ANodeClientsClass := TNodeClients;
  ANodeClientClass := ItemClass;
  if ANodeClientClass = nil then ANodeClientClass := TNodeClient;
  fItems := ANodeClientsClass.Create(Self, ANodeClientClass);
  UpdateNode;
end;

destructor TNodeClient.Destroy;
begin
  FreeAndNil(fActionLink);
  FreeAndNil(fActionNode);
  FreeAndNil(fItems);
  inherited;
end;

procedure TNodeClient.UpdateNode;
var AActionTreeView: TActionTreeView;
  Coll: TCollection;
  Parent: TNodeClient;
  IsNormalOwner: boolean;
begin
  AActionTreeView := ActionTreeView;
  if (AActionTreeView <> nil) and
    (not AActionTreeView.InCreateNode) and
    (Collection is TNodeClients) and
    (TNodeClients(Collection).UpdateCount = 0) then
  begin
    Parent := self;
    while Parent.ParentItem <> nil do
      Parent := Parent.ParentItem;
    Coll := Parent.Collection;
    IsNormalOwner := (Coll is TNodeClients) and (TNodeClients(Coll).GetOwner is TActionTreeView);
    if IsNormalOwner then
      TActionTreeView(TNodeClients(Coll).GetOwner).fInCreateNodeClient := true;
    try
      DoUpdateNode;
    finally
      if IsNormalOwner then
        TActionTreeView(TNodeClients(Coll).GetOwner).fInCreateNodeClient := false;
    end;
  end;
end;

procedure TNodeClient.DoUpdateNode;
var AVisible: boolean;
  Parent: TNodeClient;
  AActionTreeView: TActionTreeView;
  I: integer;
  Node: TTreeNode;
  function ItemsCount: integer;
  var I: integer;
  begin
    result := 0;
    for I := 0 to AActionTreeView.Items.Count - 1 do
      if AActionTreeView.Items[I].Level = 0 then inc(Result);
  end;
  function ItemByIndex(Index: integer): TTreeNode;
  var I, J: integer;
  begin
    result := nil;
    J := 0;
    for I := 0 to AActionTreeView.Items.Count - 1 do
      if AActionTreeView.Items[I].Level = 0 then
      begin
        if J = Index then
        begin
          Result := AActionTreeView.Items[I];
          Exit;
        end;
        inc(J);
      end;
  end;

  function ChildItemByIndex(Parent: TTreeNode; Index: integer): TTreeNode;
  var I, J: integer;
  begin
    result := nil;
    J := 0;
    for I := 0 to Parent.Count - 1 do
    begin
      if J = Index then
      begin
        Result := Parent.Item[I];
        Exit;
      end;
      inc(J);
    end;
  end;


begin
  Parent := ParentItem;
  AActionTreeView := ActionTreeView;
  AVisible := ((Parent = nil) or ((Parent <> nil) and (Parent.ActionNode <> nil))) and
    (AActionTreeView.IsDesigning or RootVisible);
  if AVisible then
  begin
    I := Index;
    if fActionNode = nil then
    begin
      if Parent <> nil then
      begin
        if I >= Parent.ActionNode.Count then
          fActionNode := TActionNode(AActionTreeView.Items.AddChild(Parent.ActionNode, '-'))
        else
          fActionNode := TActionNode(AActionTreeView.Items.Insert(Parent.ActionNode.Item[I], '-'));
      end
      else
      begin
        if I >= ItemsCount then
          fActionNode := TActionNode(AActionTreeView.Items.AddChild(nil, '-'))
        else
          fActionNode := TActionNode(AActionTreeView.Items.Insert(ItemByIndex(I), '-'));
      end;
      fActionNode.fNodeClient := Self;
    end;
    fActionNode.ImageIndex := ImageIndex;
    fActionNode.SelectedIndex := ImageIndex;
    fActionNode.Text := Caption;
    fActionNode.StateIndex := AActionTreeView.StateIndex[Checked];
    fActionNode.Expanded := Expanded;
    if fActionNode.Index <> I then
    begin
      if fActionNode.Parent = nil then
      begin
        Node := ItemByIndex(I);
      end
      else
      begin
        Node := ChildItemByIndex(fActionNode.Parent, I);
      end;
      if Node <> nil then
        fActionNode.MoveTo(Node, naInsert)
    end;
  end
  else
  begin
    FreeAndNil(fActionNode);
  end;
end;

function TNodeClient.RootVisible: boolean;
var Parent: TNodeClient;
begin
  result := Visible;
  Parent := ParentItem;
  while Result and (Parent <> nil) do
  begin
    result := Parent.Visible;
    Parent := Parent.ParentItem;
  end;
end;

function TNodeClient.GetActionTreeView: TActionTreeView;
var Parent: TNodeClient;
begin
  result := nil;
  Parent := self;
  while (Parent <> nil) and (Parent.Collection <> nil) and (result = nil) do
  begin
    if (Parent.Collection is TNodeClients) and
      (TNodeClients(Parent.Collection).GetOwner is TActionTreeView) then
      result := TActionTreeView(TNodeClients(Parent.Collection).GetOwner);
    Parent := Parent.ParentItem;
  end;
end;

{$IFDEF DELPHI_2009_UP}

procedure TNodeClient.SetCollection(Value: TCollection);
begin
  if Value <> Collection then
  begin
    if (Value <> nil) and (not (Value.InheritsFrom(TNodeClients))) then
      AssignError(Value, TNodeClientsClass.ClassName);
    inherited;
    FreeAndNil(fActionNode);
    if (Collection is TNodeClients) and
      (Collection.Owner is TNodeClient) then
      fParentItem := TNodeClient(Collection.Owner)
    else
      fParentItem := nil;
    fActionTreeView := GetActionTreeView;
    UpdateNode;
  end;
end;
{$ENDIF}

function TNodeClient.NeedUpdate: boolean;
begin
  result := (fActionNode <> nil) and
    (Collection is TNodeClients) and
    (TNodeClients(Collection).UpdateCount = 0);
end;

procedure TNodeClient.Invalidate;
begin
  if NeedUpdate then
  begin
    if ActionTreeView <> nil then
      ActionTreeView.Invalidate;
  end;
end;

function TNodeClient.ItemClass: TNodeClientClass;
begin
  result := nil;
end;

function TNodeClient.ItemsClass: TNodeClientsClass;
begin
  result := nil;
end;

function TNodeClient.CaptionStored: boolean;
begin
  result := (Caption <> '') and ((Action = nil) or ((Action <> nil) and (Action.Caption <> Caption)));
end;

function TNodeClient.CheckedStored: Boolean;
begin
  result := (Checked) and ((Action = nil) or ((Action <> nil) and (Action.Checked <> Checked)));
end;

function TNodeClient.EnabledStored: Boolean;
begin
  result := (not Enabled) and ((Action = nil) or ((Action <> nil) and (Action.Enabled <> Enabled)));
end;

function TNodeClient.ImageIndexStored: Boolean;
begin
  result := (ImageIndex >= 0) and ((Action = nil) or ((Action <> nil) and (Action.ImageIndex <> ImageIndex)));
end;

function TNodeClient.VisibleStored: Boolean;
begin
  result := (not Visible) and ((Action = nil) or ((Action <> nil) and (Action.Visible <> Visible)));
end;

function TNodeClient.GetAction: TCustomAction;
begin
  if fActionLink = nil then
    result := nil
  else
    result := TCustomAction(fActionLink.Action);
end;

procedure TNodeClient.SetAction(const Value: TCustomAction);
begin
  if fActionLink.Action <> Value then
  begin
    fActionLink.Action := Value;
    if NeedUpdate then UpdateNode;
  end;
end;

procedure TNodeClient.SetItems(const Value: TNodeClients);
begin
  if Value = nil then
    fItems.Clear
  else
    fItems.Assign(Value);
end;

procedure TNodeClient.SetCaption(const Value: TCaption);
begin
  if fCaption <> Value then
  begin
    fCaption := Value;
    if NeedUpdate then fActionNode.Text := fCaption;
  end;
end;

procedure TNodeClient.SetChecked(const Value: boolean);
begin
  if fChecked <> Value then
  begin
    if (Action is TCustomAction) and
      (TCustomAction(Action).Checked <> Value) then
    begin
      if (ActionTreeView <> nil) and
        (ActionTreeView.IsDesigning) then
      begin
        fChecked := Value;
        TCustomAction(Action).Checked := Value
      end
      else
        Exit;
    end;
    fChecked := Value;
    if NeedUpdate then
    begin
      if (fActionNode <> nil) and (ActionTreeView <> nil) then
        fActionNode.StateIndex := ActionTreeView.StateIndex[fChecked];
    end;
  end;
end;


procedure TNodeClient.SetEnabled(const Value: boolean);
begin
  if fEnabled <> Value then
  begin
    if (Action is TCustomAction) and
      (TCustomAction(Action).Enabled <> Value) then
    begin
      if (ActionTreeView <> nil) and
        (ActionTreeView.IsDesigning) then
      begin
        fEnabled := Value;
        TCustomAction(Action).Enabled := Value;
        fEnabled := TCustomAction(Action).Enabled;
        if NeedUpdate then Invalidate;
      end;
    end
    else
    begin
      fEnabled := Value;
      if NeedUpdate then Invalidate;
    end;
  end;
end;

procedure TNodeClient.SetExpanded(const Value: boolean);
begin
  if fExpanded <> Value then
  begin
    fExpanded := Value;
    if NeedUpdate then
      fActionNode.Expanded := fExpanded;
  end;
end;

procedure TNodeClient.SetImageIndex(const Value: TImageIndex);
begin
  if fImageIndex <> Value then
  begin
    fImageIndex := Value;
    if NeedUpdate then
    begin
      fActionNode.ImageIndex := fImageIndex;
      fActionNode.SelectedIndex := fImageIndex;
      if (fActionNode <> nil) and (ActionTreeView <> nil) then
        fActionNode.StateIndex := ActionTreeView.StateIndex[fChecked];
      Invalidate;
    end;
  end;
end;

procedure TNodeClient.SetVisible(const Value: boolean);
begin
  if fVisible <> Value then
  begin
    if (Action is TCustomAction) and
      (TCustomAction(Action).Visible <> Value) then
    begin
      if (ActionTreeView <> nil) and
        (ActionTreeView.IsDesigning) then
      begin
        fVisible := Value;
        TCustomAction(Action).Visible := Value
      end
      else
        Exit;
    end;
    fVisible := Value;
    Items.BeginUpdate;
    try
      UpdateNode;
    finally
      Items.EndUpdate;
    end;
  end;
end;

function TNodeClient.GetDisplayName: string;
begin
  Result := Trim(Copy(Caption, 1, 32));
  if (Action is TCustomAction) then
  begin
    if Result <> '' then Result := Result + ' ';
    Result := Result + '(' + TCustomAction(Action).Name + ')';
  end;
end;

procedure TNodeClient.AssignValues(Source: TPersistent);
begin
  if Source = nil then
  begin
    SetDefaultValues;
  end
  else
  begin
    if Source is TNodeClient then
    begin
      fVisible := TNodeClient(Source).Visible;
      fEnabled := TNodeClient(Source).Enabled;
      fChecked := TNodeClient(Source).Checked;
      fExpanded := TNodeClient(Source).Expanded;
      fImageIndex := TNodeClient(Source).ImageIndex;
      fCaption := TNodeClient(Source).Caption;
      if fActionLink <> nil then
      begin
        fActionLink.Action := TNodeClient(Source).Action;
      end;
    end
    else
      AssignError(Source, ClassName);
  end;
end;

procedure TNodeClient.Assign(Source: TPersistent);
type TAssignRec = record
    Sourc, Dest: TNodeClient;
    I: integer;
  end;
var ARec: array of TAssignRec;
  N: integer;
  NewItem: TCollectionItem;
  Sourc, Dest: TNodeClient;
  procedure Push(ASourc, ADest: TNodeClient);
  begin
    inc(N);
    if N > Length(ARec) then
    begin
      SetLength(ARec, (N + 3) * 2);
      FillChar(ARec[N - 1], SizeOf(ARec[0]) * (Length(ARec) - (N - 1)), 0);
    end;
    with ARec[N - 1] do
    begin
      Sourc := ASourc;
      Dest := ADest;
      I := 0;
      Dest.Items.BeginUpdate;
    end;
  end;
  function Pop: boolean;
  begin
    result := N > 0;
    if result then
    begin
      Dec(N);
      ARec[N].Dest.Items.EndUpdate;
      FillChar(ARec[N], SizeOf(ARec[0]), 0);
    end;
  end;
begin
  if Source is TCustomAction then
  begin
    if Collection <> nil then Collection.BeginUpdate;
    try
      Action := TAction(Source);
    finally
      if Collection <> nil then Collection.EndUpdate;
    end;
  end
  else
  begin
    N := 0;
    if Collection <> nil then Collection.BeginUpdate;
    try
      AssignValues(Source);
      if Source = nil then
      begin
        Items.Clear;
      end
      else
      begin
        Push(TNodeClient(Source), self);
        while N > 0 do
        begin
          while ARec[N - 1].I < ARec[N - 1].Sourc.Items.Count do
          begin
            if ARec[N - 1].Sourc.Items[ARec[N - 1].I] is TNodeClient then
            begin
              Sourc := ARec[N - 1].Sourc.Items[ARec[N - 1].I];
              NewItem := ARec[N - 1].Dest.Items.Add;
              if NewItem is TNodeClient then
              begin
                Dest := TNodeClient(NewItem);
                Dest.AssignValues(Sourc);
                if Sourc.Items.Count > 0 then
                begin
                  Inc(ARec[N - 1].I);
                  Push(Sourc, Dest);
                  Continue;
                end;
              end;
            end;
            Inc(ARec[N - 1].I);
          end;
          Pop;
        end;
      end;
      UpdateNode;
    finally
      while Pop do
        ;
      if Collection <> nil then Collection.EndUpdate;
    end;
  end;
end;

{ TNodeClients }

procedure TNodeClients.Assign(Source: TPersistent);
var Clients: TNodeClients;
    AActionTreeView: TActionTreeView;
    NewItem: TCollectionItem;
    I: integer;
begin
  AActionTreeView := GetActionTreeView(self);
  try
    if AActionTreeView <> nil then AActionTreeView.BeginUpdate;
    BeginUpdate;
    if Source = nil then
    begin
      Clear;
    end
    else
    begin
      Clients := nil;
      if Source is TNodeClient then
        Clients := TNodeClient(Source).Items
      else if Source is TNodeClients then
        Clients := TNodeClients(Source)
      else if Source is TActionTreeView then
        Clients := TActionTreeView(Source).NodeClients;
      if Clients <> nil then
      begin
        Clear;
        for I := 0 to Clients.Count - 1 do
        begin
          NewItem := Add;
          NewItem.Assign(Clients[I]);
        end;
      end
      else inherited;
    end;
  finally
    EndUpdate;
    if AActionTreeView <> nil then AActionTreeView.EndUpdate;
  end;
end;

constructor TNodeClients.Create(AOwner: TPersistent;
  ItemClass: TNodeClientClass);
begin
  inherited Create(AOwner, ItemClass);
end;

class function TNodeClients.GetActionTreeView(
  NodeClients: TNodeClients): TActionTreeView;
var
  Parent: TPersistent;
begin
  result := nil;
  Parent := NodeClients;
  while (not (Parent is TActionTreeView)) and (Parent <> nil) do
  begin
    if Parent is TNodeClients then
      Parent := TNodeClients(Parent).GetOwner
    else
      Parent := nil;
  end;
  if Parent is TActionTreeView then
    result := TActionTreeView(Parent)
end;

class procedure TNodeClients.IterateClients(NodeClients: TNodeClients;
  ActionProc: TIterateProc);
type
  TNodeClientRec = record
    NodeClient: TNodeClient;
    Executed: boolean;
  end;
var A: array of TNodeClientRec;
  N, I, J: integer;
  Found: Boolean;
  AActionTreeView: TActionTreeView;
  procedure Realloc(N: integer);
  var OldN: integer;
  begin
    OldN := Length(A);
    if OldN <= N then
    begin
      SetLength(A, (N + 3) * 2);
      FillChar(A[OldN], SizeOf(A[0]) * (Length(A) - OldN), 0);
    end;
  end;
begin
  if (NodeClients <> nil) and (NodeClients.Count > 0) then
  begin
    AActionTreeView := GetActionTreeView(NodeClients);
    if AActionTreeView <> nil then
      AActionTreeView.BeginUpdate;
    try
      N := 0;
      for I := 0 to NodeClients.Count - 1 do
        if NodeClients.Items[I] is TNodeClient then
        begin
          inc(N);
          Realloc(N);
          A[N - 1].NodeClient := TNodeClient(NodeClients.Items[I]);
        end;
      Found := True;
      while Found do
      begin
        Found := false;
        for I := 0 to N - 1 do
          if not A[I].Executed then
          begin
            A[I].Executed := True;
            if (A[I].NodeClient <> nil) and
              (A[I].NodeClient.Items <> nil) and
              (A[I].NodeClient.Items.Count > 0) then
              for J := 0 to A[I].NodeClient.Items.Count - 1 do
                if A[I].NodeClient.Items.Items[J] is TNodeClient then
                begin
                  Found := True;
                  inc(N);
                  Realloc(N);
                  A[N - 1].NodeClient := TNodeClient(A[I].NodeClient.Items.Items[J]);
                end;
          end;
      end;
      //for I := N - 1 downto 0 do ActionProc(A[I].NodeClient);
      if (@ActionProc <> nil) then
        for I := 0 to N - 1 do
          ActionProc(A[I].NodeClient);
      for I := 0 to N - 1 do
        if A[I].NodeClient.ActionNode <> nil then
          A[I].NodeClient.ActionNode.Expanded := A[I].NodeClient.Expanded;
      if (NodeClients.GetOwner is TNodeClient) and
        (TNodeClient(NodeClients.GetOwner).ActionNode <> nil) then
        TNodeClient(NodeClients.GetOwner).ActionNode.Expanded := TNodeClient(NodeClients.GetOwner).Expanded;
    finally
      if AActionTreeView <> nil then AActionTreeView.EndUpdate;
    end;
  end;
end;

function TNodeClients.GetItem(Index: Integer): TNodeClient;
begin
  result := TNodeClient(inherited GetItem(Index));
end;

procedure TNodeClients.SetItem(Index: Integer; const Value: TNodeClient);
begin
  inherited SetItem(Index, Value);
end;

procedure TNodeClients.Update(Item: TCollectionItem);
begin
  inherited;
  if Item = nil then
    IterateClients(self, UpdateNode)
  else
  begin
    if Item is TNodeClient then IterateClients(TNodeClient(Item).Items, UpdateNode);
  end;
end;

procedure TNodeClients.UpdateNode(Item: TNodeClient);
begin
  if Item <> nil then Item.UpdateNode;
end;

{ TActionNode }

constructor TActionNode.Create(AOwner: TTreeNodes);
var AActionTreeView: TActionTreeView;
begin
  inherited;
  ImageIndex := -1;
  SelectedIndex := -1;
  StateIndex := 0;
  if (AOwner is TDummyNodes) and (AOwner.Owner is TActionTreeView) then
  begin
    AActionTreeView := TActionTreeView(AOwner.Owner);
    if (AActionTreeView <> nil) and (not AActionTreeView.InCreateNodeClient) then
    begin
      //          TNodeClient
      {$IFDEF DELPHI_2005_UP}
      if (csDesigning in AActionTreeView.ComponentState) or
         (AActionTreeView.IsDesigning) then
        AActionTreeView.fNeedFreeEmptyNodes := true
      else
        {$ENDIF}
        //  ,         TNodeClient
        raise Exception.Create(StrECreateNode);
    end;
  end;
end;

destructor TActionNode.Destroy;
begin
  if fNodeClient <> nil then fNodeClient.fActionNode := nil;
  inherited;
end;

procedure TActionNode.DrawMarked(TreeView: TActionTreeView; R: TRect);
var Y: integer;
    Canvas: TCanvas;
begin
  Canvas := TreeView.Canvas;
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Color := TreeView.GetColorDef(TreeView.fBtnFrameColor, clWindowText);
  case TreeView.MarkedNodeType of
    mntBefore:
      begin
        Canvas.MoveTo(R.left, R.top + 2);
        Canvas.LineTo(R.left + 2, R.top);
        Canvas.LineTo(R.Right - 3, R.top);
        Canvas.LineTo(R.Right, R.top + 3);
      end;
    mntAfter:
      begin
        Canvas.MoveTo(R.left, R.Bottom - 3);
        Canvas.LineTo(R.left + 2, R.Bottom - 1);
        Canvas.LineTo(R.Right - 3, R.Bottom - 1);
        Canvas.LineTo(R.Right, R.Bottom - 4);
      end;
    mntChild:
      begin
        Y := (R.Bottom + R.Top) div (2);
        Canvas.MoveTo(R.left, Y - 2);
        Canvas.LineTo(R.left + 2, Y);
        Canvas.LineTo(R.Right - 2, Y);
        Canvas.LineTo(R.Right, Y - 2);
        Canvas.LineTo(R.Right, Y + 2);
        Canvas.LineTo(R.Right - 2, Y);
        Canvas.MoveTo(R.left + 2, Y);
        Canvas.LineTo(R.left, Y + 2);
        Canvas.LineTo(R.left, Y - 2);
      end;
  end;
end;

procedure TActionNode.DrawImage(TreeView: TActionTreeView;
                                State: TCustomDrawState;
                                Image: TCustomImageList;
                                Index: integer;
                                R: TRect);
var Img: TRichImageList;
begin
  if (Image <> nil) then
  begin
    Img := TreeView.fRichImageList;
    if (Img <> nil) and (Index >= 0) and (Index < Image.Count) then
    begin
      Img.ImageList := Image;
      Img.Enabled := ([cdsGrayed, cdsDisabled] * State = []);
      Img.Hot := (cdsSelected in State) and (Img.Enabled);
      {if not Img.Enabled then
        if (cdsSelected in State) then Img.Blend := 128
                                  else Img.Blend := 64
      else
        Img.Blend := 255;}
      Img.Index := Index;
      Img.Width := Image.Width;
      Img.Height := Image.Height;
      Img.BkColor :=  TreeView.Canvas.Brush.Color;


      TreeView.Canvas.Draw((R.Left + R.Right - Image.Width) div (2),
                           (R.Top + R.Bottom - Image.Height) div (2),
                           Img);
    end;
  end;
end;

procedure TActionNode.DoDraw(Canvas: TCanvas; TreeView: TActionTreeView;
  State: TCustomDrawState; var PaintImages, DefaultDraw: boolean);
var RText, RState, RImage, R: TRect;
  MinTextX: integer;
  NeedImage, NeedState: boolean;
  OldColor: TColor;
begin
  RText := DisplayRect(True);
  if RText.Left < TreeView.ClientWidth then
  begin
    NeedState := False;
    NeedImage := False;
    MinTextX := RText.Left;
    InflateRect(RImage, -1, -1);
    if (TreeView.StateImages <> nil) and (StateIndex > 0) then
    begin
      RState := Rect(RText.Left, RText.Top, RText.Left + TreeView.StateImages.Width, RText.Bottom);
      OffsetRect(RState, -TreeView.StateImages.Width, 0);
      if TreeView.Images <> nil then
        OffsetRect(RState, -TreeView.Images.Width - 3, 0);
      if RState.Left < MinTextX then
        MinTextX := RState.Left;
      NeedState := True;
    end;

    if TreeView.Images <> nil then
    begin
      RImage := Rect(RText.Left - TreeView.Images.Width - 5, RText.Top, RText.Left - 1, RText.Bottom);
      if RImage.Left < MinTextX then
        MinTextX := RImage.Left;
      NeedImage := True;
    end;
    RText.Right := RText.Left + Canvas.TextWidth(Text) + 4;
    if cdsFocused in State then
    begin
      OldColor := Canvas.Font.Color;
      Canvas.Rectangle(RText);
      if (TreeView.StateImages = nil) and (TreeView.Images = nil) and (cdsChecked in state) then
      begin
        SetTextColor(Canvas.Handle, ColorToRGB(Canvas.Pen.Color));
        Canvas.DrawFocusRect(RText);
        SetTextColor(Canvas.Handle, ColorToRGB(OldColor));
      end;
    end
    else
      Canvas.FillRect(RText);
    R := RText;
    OffsetRect(R, 2, 1);
    windows.SetBkMode(Canvas.Handle, TRANSPARENT);
    windows.DrawText(Canvas.Handle, PChar(Text), Length(Text), R, DT_LEFT);

    if NeedState then
    begin
      DrawImage(TreeView, State, TreeView.StateImages, StateIndex, RState);
    end
    else
    begin
      if (cdsChecked in State) then
      begin
        if TreeView.StateImages = nil then
        begin
          State := State + [cdsSelected, cdsFocused];
          TreeView.UpdateCanvasColor(Canvas, State);
          if not NeedImage then
          begin
            RImage := RText;
            InflateRect(RImage, -1, -1);
            Canvas.Brush.Style := bsClear;
          end;

          Canvas.Rectangle(RImage);
        end;
      end;
    end;
    if NeedImage then
      DrawImage(TreeView, State, TreeView.Images, ImageIndex, RImage);

    if (DropTarget) and
      (TreeView.MarkedNodeType <> mntNone) and
      (TreeView.Dragged) then DrawMarked(TreeView, Rect(MinTextX, RText.Top, RText.Right, RText.Bottom));

    if DefaultDraw then
      windows.ExcludeClipRect(Canvas.Handle, MinTextX, RText.Top, TreeView.ClientWidth, RText.Bottom);
  end;
end;

function TActionNode.GetAction: TCustomAction;
begin
  if fNodeClient = nil then
    Result := nil
  else
    Result := fNodeClient.Action;
end;

function TActionNode.RootEnabled: boolean;
var Parent: TTreeNode;
begin
  result := (fNodeClient <> nil) and (fNodeClient.Enabled);
  if Result then
  begin
    Parent := self.Parent;
    while (Parent is TActionNode) and (Result) do
    begin
      Result := (TActionNode(Parent).fNodeClient <> nil)
        and (TActionNode(Parent).fNodeClient.Enabled);
      Parent := Parent.Parent;
    end;
  end;
end;

{ TGliphState }

constructor TGliphState.Create(AOwner: TActionTreeView);
begin
  inherited Create(AOwner);
  fChecked := 2;
  fUnchecked := 1;
end;

procedure TGliphState.Assign(Source: TPersistent);
var SourceName: string;
  isChange: boolean;
begin
  if Source = nil then
  begin
    isChange := (GetChecked <> 2) or
      (GetUnchecked <> 1);
    if isChange then
    begin
      fChecked := 2;
      fUnchecked := 1;
      Changed;
    end;
  end
  else
  begin
    SourceName := Source.ClassType.ClassName;
    if not (Source is TGliphState) then
      raise EConvertError.CreateFmt(SAssignError, [SourceName, ClassType.ClassName]);
    isChange := (GetChecked <> TGliphState(Source).Checked) or
      (GetUnchecked <> TGliphState(Source).Unchecked);
    if isChange then
    begin
      fChecked := TGliphState(Source).Checked;
      fUnchecked := TGliphState(Source).Unchecked;
      Changed;
    end;
  end;
end;

function TGliphState.GetActionTreeView: TActionTreeView;
begin
  if Owner is TActionTreeView then
    Result := TActionTreeView(Owner)
  else
    Result := nil;

end;

function TGliphState.GetHandle: THandle;
begin
  result := 0;
  if ActionTreeView <> nil then
    with ActionTreeView do
    begin
      if ([csLoading, csDestroying, csUpdating] * ComponentState = []) and
        (StateImages <> nil) then
        Result := StateImages.Handle;
    end
  else
    Result := GetHandle;
end;

procedure TGliphState.Changed;
var NeedUpdate: boolean;
begin
  inherited;
  if ActionTreeView <> nil then
    with ActionTreeView do
    begin
      NeedUpdate := (fNodeClients <> nil) and
        ([csLoading, csDestroying, csUpdating] * ComponentState = []) and
        (StateImages <> nil);
      if NeedUpdate then fNodeClients.BeginUpdate;
      try
        fStateIndex[False] := fUnchecked;
        fStateIndex[True] := fChecked;
      finally
        if NeedUpdate then fNodeClients.EndUpdate;
      end;
    end;
end;

function TGliphState.GetChecked: TImageIndex;
begin
  if ActionTreeView <> nil then
    fChecked := ActionTreeView.fStateIndex[True];
  Result := fChecked;
end;

procedure TGliphState.SetChecked(const Value: TImageIndex);
begin
  if GetChecked <> Value then
  begin
    if not CheckImageIndex(Value) then raise EListError.CreateFMT(NoIndexImg, [Value]);
    fChecked := Value;
    fUnchecked := GetUnchecked;
    Changed;
  end;
end;

function TGliphState.GetUnchecked: TImageIndex;
begin
  if ActionTreeView <> nil then
    fUnchecked := ActionTreeView.fStateIndex[False];
  Result := fUnchecked;
end;

procedure TGliphState.SetUnchecked(const Value: TImageIndex);
begin
  if GetUnchecked <> Value then
  begin
    if not CheckImageIndex(Value) then raise EListError.CreateFMT(NoIndexImg, [Value]);
    fChecked := GetChecked;
    fUnchecked := Value;
    Changed;
  end;
end;

{ TActionTreeView }

function TActionTreeView.ActionNodeClass: TActionNodeClass;
begin
  result := nil;
end;

function TActionTreeView.NodeClientsClass: TNodeClientsClass;
begin
  result := nil;
end;

function TActionTreeView.NodeClientsItemClass: TNodeClientClass;
begin
  result := nil;
end;

constructor TActionTreeView.Create(AOwner: TComponent);
var ANodeClientsClass: TNodeClientsClass;
  ANodeClientClass: TNodeClientClass;
begin
  inherited;
  fIsDesigning := csDesigning in ComponentState;
  DoubleBuffered := True;
  //DragMode := dmAutomatic;
  DragKind := dkDrag;
  StateIndex[false] := 1;
  StateIndex[True] := 2;
  fGliphState := TGliphState.Create(Self);
  fBtnFrameColor := $00C66931;
  fSelectedColor := $00EFD3C6;
  fHighlightTextColor := clNone;
  fDisabledFontColor := clGrayText;
  ReadOnly := True;
  ANodeClientsClass := NodeClientsClass;
  if ANodeClientsClass = nil then ANodeClientsClass := TNodeClients;
  ANodeClientClass := NodeClientsItemClass;
  if ANodeClientClass = nil then ANodeClientClass := TNodeClient;
  fNodeClients := ANodeClientsClass.Create(Self, ANodeClientClass);

  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;
  FStateChangeLink := TChangeLink.Create;
  FStateChangeLink.OnChange := ImageListChange;

end;

destructor TActionTreeView.Destroy;
begin
  FreeAndNil(FImageChangeLink);
  FreeAndNil(FStateChangeLink);
  FreeAndNil(fGliphState);
  FreeAndNil(fNodeClients);
  FreeAndNil(fRichImageList);
  inherited;
end;

procedure TActionTreeView.ImageListChange(Sender: TObject);
begin
  ClearHasButtonBitmap;
end;

procedure TActionTreeView.SetImages(Value: TCustomImageList);
begin
  if Value <> Images then
  begin
    if Images <> nil then
      Images.UnRegisterChanges(FImageChangeLink);
    if Value <> nil then
      Value.RegisterChanges(FImageChangeLink);
  end;
  inherited Images := Value;
end;

procedure TActionTreeView.SetStateImages(Value: TCustomImageList);
begin
  if Value <> StateImages then
  begin
    if StateImages <> nil then
      StateImages.UnRegisterChanges(FImageChangeLink);
    if Value <> nil then
      Value.RegisterChanges(FImageChangeLink);
  end;
  inherited StateImages := Value;
end;

function TActionTreeView.GetColorDef(Color, DefaultColor: TColor): TColor;
begin
  if Color = clNone then
    result := self.Color
  else if Color = clDefault then
    result := DefaultColor
  else
    result := Color;
end;

//     

procedure TActionTreeView.UpdateCanvasColor(Canvas: TCanvas;
  State: TCustomDrawState);
begin
  if (cdsSelected in State) then
  begin
    Canvas.Brush.Color := GetColorDef(fSelectedColor, clHighlight);
    if fSelectedColor <> clNone then
      Canvas.Brush.Style := bsSolid
    else
      Canvas.Brush.Style := bsClear;
    Canvas.Pen.Color := GetColorDef(fBtnFrameColor, clHighlight);
    if (fBtnFrameColor <> clNone) and (cdsFocused in State) then
      Canvas.Pen.Style := psSolid
    else
      Canvas.Pen.Style := psClear;
  end
  else
  begin
    Canvas.Brush.Color := Color;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Color := Color;
    Canvas.Pen.Style := psClear;
  end;
end;

function TActionTreeView.CreateRichImageList: TRichImageList;
begin
  Result := TRichImageList.Create;
  Result.Adjusted := True;
  Result.Transparent := True;
  Result.Border := CepNoBorder;
  Result.BorderColor := clNone;
end;

{$IFDEF DELPHI_2005_UP}

procedure TActionTreeView.WMSETREDRAW(var Message: TWMSETREDRAW);
var I: integer;
begin
  if fNeedFreeEmptyNodes and (Message.Redraw <> 0) then
  begin
    fNeedFreeEmptyNodes := false;
    fNodeClients.BeginUpdate;
    try
      for I := Items.Count - 1 downto 0 do
        if (Items[I] is TActionNode) and (TActionNode(Items[I]).fNodeClient = nil) then
          Items[I].Free;
    finally
      fNodeClients.EndUpdate;
    end;
  end;
  inherited;
end;
{$ENDIF}

procedure TActionTreeView.CNNotify(var Message: TWMNotify);
type
  PFontHandles = ^TFontHandles;
  TFontHandles = record
    OurFont,
      StockFont: Integer;
  end;
var
  Node: TTreeNode;
  NodeClient: TNodeClient;
  DefaultDraw, PaintImages: Boolean;
  TmpItem: TTVItem;
  State: TCustomDrawState;
  FontColor: TColor;
  function GetNodeFromItem(const Item: TTVItem): TTreeNode;
  begin
    Result := nil;
    if Items <> nil then
      with Item do
        if (state and TVIF_PARAM) <> 0 then
          Result := Pointer(lParam)
        else
          Result := Items.GetNode(hItem);
  end;
begin
  with Message do
    case NMHdr^.code of
      NM_CUSTOMDRAW:
        with PNMCustomDraw(NMHdr)^ do
        begin
          Result := CDRF_DODEFAULT;
          if (dwDrawStage and CDDS_ITEM) = 0 then
          begin
            Result := Result or CDRF_NOTIFYITEMDRAW;
          end
          else
          begin
            if RC.Right - RC.Left = 0 then
            begin
              Exit;
            end;
            FillChar(TmpItem, SizeOf(TmpItem), 0);
            TmpItem.hItem := HTREEITEM(dwItemSpec);
            Node := GetNodeFromItem(TmpItem);
            if Node is TActionNode then
            begin
              Canvas.Lock;
              try
                Canvas.Handle := hdc;
                Canvas.Font := Font;
                try
                  if fRichImageList = nil then
                    fRichImageList := CreateRichImageList;
                  //    
                  State := TCustomDrawState(Word(uItemState));
                  NodeClient := TActionNode(Node).NodeClient;
                  if Node.Selected then State := State + [cdsSelected];
                  if NodeClient <> nil then
                  begin
                    if (IsDesigning and ((TActionNode(Node).NodeClient <> nil) and
                                         (not TActionNode(Node).NodeClient.Enabled))) or
                       ((not IsDesigning) and (not TActionNode(Node).RootEnabled)) then State := State + [cdsDisabled];
                    if NodeClient.Checked then State := State + [cdsChecked];
                  end
                  else
                    State := State + [cdsDisabled];
                  //   
                  FontColor := Font.Color;
                  if (cdsDisabled in State) and (fDisabledFontColor <> clNone) then
                  begin
                    FontColor := GetColorDef(fDisabledFontColor, clGrayText);
                  end
                  else
                  begin
                    if (cdsSelected in State) and (fHighlightTextColor <> clNone) then
                    begin
                      FontColor := GetColorDef(fHighlightTextColor, clHighlightText);
                    end;
                  end;
                  Canvas.Font.Color := FontColor;
                  //     
                  UpdateCanvasColor(Canvas, State);
                  //   
                  PaintImages := False;
                  DefaultDraw := True;
                  TActionNode(Node).DoDraw(Canvas, Self, State, PaintImages, DefaultDraw);
                  if not PaintImages then
                    Result := Result or TVCDRF_NOIMAGES;
                  if not DefaultDraw then
                    Result := Result or CDRF_SKIPDEFAULT
                  else
                finally
                  Canvas.Handle := 0;
                end;
                if (result and CDRF_SKIPDEFAULT) = 0 then
                  SelectObject(hdc, Font.Handle);
              finally
                Canvas.Unlock;
              end;
            end
          end;
        end;
      TVN_BEGINDRAG:
        begin
          Dragged := True;
          with PNMTreeView(NMHdr)^ do
            Node := GetNodeFromItem(ItemNew);
          if Node is TActionNode then
            FDragNode := TActionNode(Node)
          else
            FDragNode := nil;
          inherited;
        end;
    else
      inherited;
    end;
end;

function TActionTreeView.CreateNode: TTreeNode;
var
  LClass: TActionNodeClass;
begin
  fInCreateNode := True;
  try
    LClass := ActionNodeClass;
    if LClass = nil then
      LClass := TActionNode;
    Result := LClass.Create(Items);
  finally
    fInCreateNode := False;
  end;
end;

{$IFDEF DELPHI_6_UP}

function TActionTreeView.CreateNodes: TTreeNodes;
begin
  result := TDummyNodes.Create(self);
end;
{$ENDIF}

function TActionTreeView.GetStateIndex(Enabled: boolean): TImageIndex;
begin
  Result := fStateIndex[Enabled]
end;

function TActionTreeView.ExecuteSelected: boolean;
begin
  if (Enabled) and
    (not (Selected is TActionNode)) or
    (TActionNode(Selected).Action = nil) or
    (not TActionNode(Selected).RootEnabled) then
  begin
    result := False;
  end
  else
  begin
    result := TActionNode(Selected).Action.Execute;
  end;
end;

procedure TActionTreeView.KeyPress(var Key: Char);
begin
  inherited;
  if (Key = ' ') then
  begin
    if not ExecuteSelected then MessageBeep(0);
    Key := #0;
  end;
end;

procedure TActionTreeView.DblClick;
var P: TPoint;
begin
  inherited;
  P := Mouse.CursorPos;
  P := ScreenToClient(P);
  if GetNodeAt(P.X, P.Y) = Selected then
  begin
    if (not ExecuteSelected) and ((Selected <> nil) and (Selected.Count = 0)) then
      MessageBeep(0)
  end
  else
    MessageBeep(0);
end;



procedure TActionTreeView.SetStateIndex(Enabled: boolean;
  const Value: TImageIndex);
begin
  if fStateIndex[Enabled] <> Value then
  begin
    if fNodeClients <> nil then fNodeClients.BeginUpdate;
    try
      fStateIndex[Enabled] := Value;
    finally
      if fNodeClients <> nil then fNodeClients.EndUpdate;
    end;
  end;
end;


procedure TActionTreeView.SetBtnFrameColor(const Value: TColor);
begin
  if fBtnFrameColor <> Value then
  begin
    fBtnFrameColor := Value;
    invalidate;
  end;
end;

procedure TActionTreeView.SetDisabledFontColor(const Value: TColor);
begin
  if fDisabledFontColor <> Value then
  begin
    fDisabledFontColor := Value;
    invalidate;
  end;
end;

procedure TActionTreeView.SetHighlightTextColor(const Value: TColor);
begin
  if fHighlightTextColor <> Value then
  begin
    fHighlightTextColor := Value;
    invalidate;
  end;
end;

procedure TActionTreeView.SetSelectedColor(const Value: TColor);
begin
  if fSelectedColor <> Value then
  begin
    fSelectedColor := Value;
    invalidate;
  end;
end;

procedure TActionTreeView.SetNodeClients(const Value: TNodeClients);
begin
  if Value = nil then
    fNodeClients.Clear
  else
    fNodeClients := Value;
end;

procedure TActionTreeView.Assign(Source: TPersistent);
begin
  BeginUpdate;
  try
    if Source is TActionTreeView then
    with TActionTreeView(Source) do
    begin
      self.StateImages := StateImages;
      self.Images := Images;
      self.StateIndex[False] := StateIndex[False];
      self.StateIndex[True] := StateIndex[True];
      self.NodeClients.Assign(NodeClients);
    end
    else if Source is TNodeClients then
    begin
      NodeClients.Assign(TNodeClients(Source));
    end
    else
      inherited;
  finally
    EndUpdate;
  end;
end;

procedure TActionTreeView.BeginUpdate;
begin
  inherited Items.BeginUpdate;
  inc(fUpdateCount);
end;

procedure TActionTreeView.EndUpdate;
begin
  if fUpdateCount > 0 then Dec(fUpdateCount);
  inherited Items.EndUpdate;
end;

procedure TActionTreeView.Expand(Node: TTreeNode);
begin
  inherited;
  if (Node is TActionNode) and (TActionNode(Node).NodeClient <> nil) and
    (TNodeClients(TActionNode(Node).NodeClient.Collection).UpdateCount = 0) and
    (UpdateCount = 0) then
    TActionNode(Node).NodeClient.fExpanded := True;
end;

procedure TActionTreeView.Collapse(Node: TTreeNode);
begin
  inherited;
  if (Node is TActionNode) and (TActionNode(Node).NodeClient <> nil) and
    (TNodeClients(TActionNode(Node).NodeClient.Collection).UpdateCount = 0) and
    (UpdateCount = 0) then
    TActionNode(Node).NodeClient.fExpanded := False;
end;

procedure TActionTreeView.CMFocusChanged(var Message: TCMFocusChanged);
begin
  if not Focused then Invalidate;
  inherited;
end;

procedure TActionTreeView.DropSelectedNode(DelChilds: boolean);
var Dest, Sourc: TNodeClient;
    Sources: array of TNodeClient;
    I: integer;
begin
  if (Selected is TActionNode) then
  begin
    Sourc := TActionNode(Selected).NodeClient;
    if (Sourc <> nil) then
    try
      BeginUpdate;
      Sourc.Items.BeginUpdate;
      if (not DelChilds) and (Sourc.Items.Count > 0) then
      begin
        SetLength(Sources, Sourc.Items.Count);
        FillChar(Sources[0], SizeOf(Sources[0]), 0);
        for I := 0 to Sourc.Items.Count - 1 do
          Sources[I] := Sourc.Items[I];
        Dest := Sourc;
        for I := 0 to Length(Sources) - 1 do
          if Sources[I] <> nil then
            Dest := CopyNodeClient(Sources[I], Dest, mntAfter, False);
      end;
      FreeAndNil(Sourc);
    finally
      if Sourc <> nil then Sourc.Items.EndUpdate;
      EndUpdate;
    end;
  end;
end;


// ,   

procedure TActionTreeView.SetDragged(const Value: boolean);
begin
  if fDragged <> Value then
  begin
    fDragged := Value;
    Invalidate;
  end;
end;

procedure TActionTreeView.SetGliphState(const Value: TGliphState);
begin
  fGliphState.Assign(Value);
end;

procedure TActionTreeView.SetMarkedNodeType(const Value: TMarkedNodeType);
begin
  if fMarkedNodeType <> Value then
  begin
    fMarkedNodeType := Value;
    if (Dragged) and
      (HandleAllocated) then
    begin
      fMarkedChange := true;
    end;
  end;
end;

procedure TActionTreeView.SetTargetNode(const Value: TActionNode);
begin
  if fTargetNode <> Value then
  begin
    //if DropTarget <> Value then fMarkedChange := False;
    DropTarget := Value;
    fTargetNode := Value;
  end;
end;

procedure TActionTreeView.DragOver(Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
var Parent: TTreeNode;
  ATargetNode, ADragNode: TActionNode;
  AMarkedNodeType: TMarkedNodeType;
  procedure UpdateTarget;
  var Node: TTreeNode;
    R: TRect;
    H: integer;
  begin
    AMarkedNodeType := mntNone;
    Node := GetNodeAt(X, Y);
    if (Node is TActionNode) and (Node <> fDragNode) then
      ATargetNode := TActionNode(Node)
    else
      ATargetNode := nil;
    if ATargetNode <> nil then
    begin
      R := ATargetNode.DisplayRect(False);
      H := (R.Bottom - R.Top) div (3);
      if (Y >= R.Bottom - H) and (Y <= R.Bottom) then
        AMarkedNodeType := mntAfter
      else if (Y >= R.Top) and (Y <= R.Top + H) then
        AMarkedNodeType := mntBefore
      else if (Y > R.Top + H) and (Y < R.Bottom - H) then
        AMarkedNodeType := mntChild
      else
        ATargetNode := nil;
    end;
  end;
begin
  inherited;
  if (not Accept) and (IsDesigning) then
  begin
    Accept := (Source is TActionTreeView)
      {$IFDEF DELPHI_6_UP} or
    (Source is TActionDragObject) or
      (Source is TActionItemDragObject)
      {$ENDIF};
  end;
  if Accept then
  begin
    case State of
      dsDragEnter: Dragged := True;
      dsDragLeave: Dragged := False;
    end;
    UpdateTarget;
    if Source is TActionTreeView then
    begin
      ADragNode := TActionTreeView(Source).DragNode;
      Accept := (ADragNode <> nil) and
        (((ATargetNode = nil) and (ADragNode.Parent <> nil))
        or ((ATargetNode <> nil)
        and ((AMarkedNodeType <> mntChild)
        or ((ADragNode.Parent <> ATargetNode) and (ADragNode <> ATargetNode)))));
      if Accept then
      begin
        if ATargetNode <> nil then
        begin
          Parent := ATargetNode;
          while (Parent <> nil) and (Accept) do
          begin
            Accept := Parent <> ADragNode;
            Parent := Parent.Parent;
          end;
        end
        else
          AMarkedNodeType := mntRoot;
      end;
    end
      {$IFDEF DELPHI_6_UP}
    else if Source is TActionDragObject then
    begin
      Accept := TActionDragObject(Source).ActionCount > 0;
      if Accept and (ATargetNode = nil) then AMarkedNodeType := mntRoot;
    end
      {$ENDIF};
    TargetNode := ATargetNode;
    if Accept then
      MarkedNodeType := AMarkedNodeType
    else
      MarkedNodeType := mntNone;
  end;
  //Accept := False;
end;

procedure TActionTreeView.DoEndDrag(Target: TObject; X, Y: Integer);
begin
  inherited;
  Dragged := False;
  if not fDragged then fDragNode := nil;
  TargetNode := nil;
  if not fDragged then MarkedNodeType := mntNone;
end;

procedure TActionTreeView.DoStartDrag(var DragObject: TDragObject);
var P: TPoint;
  Node: TTreeNode;
begin
  if fDragNode = nil then
  begin
    GetCursorPos(P);
    with ScreenToClient(P) do
      Node := GetNodeAt(X, Y);
    if Node is TActionNode then
      FDragNode := TActionNode(Node)
    else
      FDragNode := nil;
  end;
  inherited;
  Dragged := True;
end;

procedure TActionTreeView.DoDragOver(Source: TDragObject; X, Y: Integer;
  CanDrop: Boolean);
begin
  if fMarkedChange then
  begin
    TDragObject(Source).HideDragImage;
    Repaint;
    TDragObject(Source).ShowDragImage;
    fMarkedChange := False;
  end;
end;

procedure TActionTreeView.CMDrag(var Message: TCMDrag);
var CanDrop: Boolean;
begin
  inherited;
  with Message, DragRec^ do
    case DragMessage of
      dmDragMove:
        begin
          CanDrop := Message.Result <> 0;
          with ScreenToClient(Pos) do
            DoDragOver(Source, X, Y, CanDrop);
        end;
    end;
end;

function TActionTreeView.CopyNodeClient(Sourc: TPersistent; Dest: TNodeClient; MarkedNodeType: TMarkedNodeType;
  DelSource: boolean): TNodeClient;
var NewItem: TCollectionItem;
  I: integer;
  procedure DoAssign;
  begin
    if NewItem is TNodeClient then Result := TNodeClient(NewItem);
    if result <> nil then
    begin
      Result.Assign(Sourc);
      if DelSource then FreeAndNil(Sourc);
    end;
  end;
begin
  result := nil;
  if (((Dest <> nil) and (Dest.Collection <> nil)) or (MarkedNodeType = mntRoot)) and
    (MarkedNodeType <> mntNone) then
  begin
    case MarkedNodeType of
      mntBefore, mntAfter:
        begin
          I := Dest.Index;
          if MarkedNodeType = mntAfter then inc(I);
          Dest.Collection.BeginUpdate;
          try
            NewItem := Dest.Collection.Insert(I);
            DoAssign;
          finally
            Dest.Collection.EndUpdate;
          end;
        end;
      mntChild:
        begin
          Dest.Items.BeginUpdate;
          try
            NewItem := Dest.Items.Add;
            DoAssign;
          finally
            Dest.Items.EndUpdate;
          end;
        end;
      mntRoot:
        begin
          fNodeClients.BeginUpdate;
          try
            NewItem := fNodeClients.Add;
            DoAssign;
          finally
            fNodeClients.EndUpdate;
          end;
        end;
    end;
  end;
end;

procedure TActionTreeView.DragDrop(Source: TObject; X, Y: Integer);
var Dest, Sourc: TNodeClient;
  Collection: TCollection;
  DelSource: boolean;
  {$IFDEF DELPHI_6_UP}
  I: integer;
  NewItem: TNodeClient;
  {$ELSE}
  KeyState: TKeyBoardState;
  {$ENDIF}
begin
  inherited;
  if MarkedNodeType <> mntNone then
  begin
    if (TargetNode <> nil) then
      Dest := TargetNode.NodeClient
    else
      Dest := nil;
    if Dest <> nil then
      Collection := Dest.Collection
    else
      Collection := NodeClients;

    if (Source is TActionTreeView) and
      (TActionTreeView(Source).DragNode <> nil) and
      (HandleAllocated) then
    begin
      if Collection <> nil then Collection.BeginUpdate;
      try
        Sourc := TActionTreeView(Source).DragNode.NodeClient;
        {$IFDEF DELPHI_6_UP}
        DelSource := not (ssCtrl in KeyboardStateToShiftState);
        {$ELSE}
        GetKeyboardState(KeyState);
        DelSource := not (ssCtrl in KeyboardStateToShiftState(KeyState));
        {$ENDIF}
        CopyNodeClient(Sourc,
          Dest,
          MarkedNodeType,
          DelSource);
      finally
        if Collection <> nil then Collection.EndUpdate;
      end;
    end
      {$IFDEF DELPHI_6_UP}
    else if ((Source is TActionDragObject) or (Source is TCategoryDragObject)) and
      (TActionDragObject(Source).ActionCount > 0) then
    begin
      if Collection <> nil then Collection.BeginUpdate;
      try
        DelSource := False;
        if Source is TCategoryDragObject then
        begin
          Dest := CopyNodeClient(nil,
            Dest,
            MarkedNodeType,
            DelSource);
          if Dest = nil then Exit;
          Dest.Caption := TCategoryDragObject(Source).Category;
          MarkedNodeType := mntChild;
        end;
        NewItem := CopyNodeClient(TActionDragObject(Source).Actions[0],
          Dest,
          MarkedNodeType,
          DelSource);
        for I := 1 to TActionDragObject(Source).ActionCount - 1 do
        begin
          NewItem := CopyNodeClient(TActionDragObject(Source).Actions[I],
            NewItem,
            mntAfter,
            DelSource);
        end;
      finally
        if Collection <> nil then Collection.EndUpdate;
      end;
    end
      {$ENDIF};
  end;
end;

end.

