unit AlignBandsUtl;
(*
 * Module:          
 *              
 * Version:    1.0 (14.11.2007) b
 * Author:      
 * Comments:         .
 *                    
 *              : roschinspb@mail.ru
 *             http://www.delphikingdom.com/asp/users.asp?ID=1271
 *             http://www.roschinspb.boom.ru
 *
 * Copyright:   Cep Po 2007 .
 *)

interface
uses Windows, Types, Math, RTLConsts, SysUtils, Classes, Controls, ActnMan;

resourcestring
  ErrTBandRect = ' TBandRect (%d)';
  ErrInputParam = '  ';
  ErrId = ' %d  ';
  ErrNoId = '  %d';

const PageSize = 16;
      RealignCount = 3;
type
  TBandRect = record
    SourceRect, ResultRect: TRect;
    Row, RowCount: integer;
    MinW, MinH: integer;
    Id: integer;
    Fx, Fy, Vx, Vy, X, Y: Double;
  end;
  PBandRect = ^TBandRect;
  TBandRects = array [0..PageSize - 1] of TBandRect;
  PBandRects = ^TBandRects;

  TEnableReg = record
    Left, Width: integer;
    Index, Count: integer;
  end;
  TEnableRegs = array [0..15] of TEnableReg;
  PEnableRegs = ^TEnableRegs;

  TBandList = class (TObject)
  private
    fArray:PBandRects;
    fRowSize: integer;
    fMinRow: integer;
    fBufCount: integer;
    fCount: integer;
    fRighToLeft: boolean;
    fMaxRow: integer;
    fBoundsRect: TRect;
    fUpdateCount: integer;
    fonAligned: TNotifyEvent;
    fAlligning: boolean;
    procedure SetRowSize(const Value: integer);
    procedure UpdateRect(var BandRect: TBandRect);
    function GetBands(I: integer): TBandRect;
    procedure SetBands(I: integer; const Value: TBandRect);
    procedure CalcMinRow;
    function FirstLastIndex(Row, MinLeft, MaxRight: integer; var First, Last: integer): boolean;
    procedure SetRighToLeft(const Value: boolean);
    procedure UpdateSize(ARect: TRect);
  protected
    procedure InternalAdd(const BandRect: TBandRect;
                          const AArray:PBandRects;
                          var ACount: integer;
                          const ABufCount: integer);
    procedure AdjustRow(Row: integer; Left, Right: integer; MinLeft: integer = Low(Integer); MaxRight: integer = High(Integer));
    procedure Sort;
    procedure AdjustAll(ARect: TRect);
  public
    function CalcEnableRegs(Row: integer; EnableReg: TEnableReg;
      var EnableRegs: PEnableRegs): integer;
    constructor Create;
    destructor Destroy; override;
    property Alligning: boolean read fAlligning;
    property Count: integer read fCount;
    procedure Add(BandRect: TBandRect); overload;
    procedure Add(const ARect: TRect; const MinWidth, MinHeight, Id: integer); overload;
    procedure Delete(Index: integer);
    procedure SetSourceRect(I: integer);
    procedure Reset;
    procedure Init;
    procedure Realign(var ARect: TRect; const N: byte = RealignCount);
    property BoundsRect: TRect read fBoundsRect;
    property MinRow: integer read fMinRow;
    property MaxRow: integer read fMaxRow;
    property RowSize: integer read fRowSize write SetRowSize;
    property Bands[I: integer]: TBandRect read GetBands write SetBands; default;
    property RighToLeft: boolean read fRighToLeft write SetRighToLeft;
    function FindBand(Id: integer): integer;
    function BandById(Id: integer): TBandRect;
    procedure ChangeId(OldId, NewId: integer);
    procedure BeginUpdate;
    procedure EndUpdate;

    property onAligned: TNotifyEvent read fonAligned write fonAligned;
  end;

  TBandItems = class;

  TBandItem = class (TCollectionItem)
  private
    fUpdateCount:integer;
    fChanged:boolean;
    fChangedAll:boolean;
    fVisible: boolean;
    fBoundsRect: TRect;
    fInsets: TRect;
    fSysArea: Byte;
    fControl: TControl;
    fMinHeight: Word;
    fMinWidth: Word;
    fDestroy: boolean;
    procedure SetBoundsRect(const Value: TRect);
    function GetClientRect: TRect;
    procedure SetClientRect(const Value: TRect);
    procedure SetMinHeight(const Value: Word);
    procedure SetMinWidth(const Value: Word);
    procedure SetInsets(const Value: TRect);
    procedure SetSysArea(const Value: Byte);

    procedure SetVisible(const Value: boolean);
    procedure SetControl(const Value: TControl);
    function GetBandItems: TBandItems;
    procedure CalcConstraints(var MinW: Integer; var MinH: Integer);
    function GetSysRect: TRect;
    function GetHandRect: TRect;
  protected
    procedure Changed(AllItems: Boolean);
    procedure SetCollection(Value: TCollection); override;
    procedure ControlChanged; virtual;
  public
    procedure BeginUpdate;
    procedure EndUpdate;

    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    property BandItems: TBandItems read GetBandItems;

    function IsVert: boolean;
    function IsRTL: boolean;
    function IntInsets: TRect;
    property BoundsRect: TRect read fBoundsRect write SetBoundsRect;
    property ClientRect: TRect read GetClientRect write SetClientRect;
    property SysRect: TRect read GetSysRect;
    property HandRect: TRect read GetHandRect;
    property MinWidth:Word read fMinWidth write SetMinWidth;
    property MinHeight:Word read fMinHeight write SetMinHeight;
    property Insets: TRect read fInsets write SetInsets;
    property SysArea: Byte read fSysArea write SetSysArea;
    property Control: TControl read fControl write SetControl;
    property Visible: boolean read fVisible write SetVisible;
  end;

  TBandItems = class (TOwnedCollection)
  private
    fBandList: TBandList;
    fOrientation: TBarOrientation;
    fBoundsRect: TRect;
    procedure SetOrientation(const Value: TBarOrientation);
    procedure Aligned(Sender: TObject);
    function GetRowSize: integer;
    procedure SetRowSize(const Value: integer);
    function GetBoundsRect: TRect;
    procedure SetBoundsRect(const Value: TRect);
    function GetItem(Index: Integer): TBandItem;
    procedure IntAligned;
  protected
    function IsVert: boolean;
    function RevertRect(const R: TRect): TRect;
    procedure DoAligned; virtual;
    procedure DoPainted; virtual;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AOwner: TPersistent; ItemClass: TCollectionItemClass);
    destructor Destroy; override;
    procedure BeginUpdate; override;
    procedure EndUpdate; override;
    function Add(Control: TControl): TBandItem;
    procedure Delete(Index: Integer);
    property BoundsRect: TRect read GetBoundsRect write SetBoundsRect;
    property BandList: TBandList read fBandList;
    function BandByControl(Control: TControl):TBandItem;
    property RowSize: integer read GetRowSize write SetRowSize default 24;
    property Items[Index: Integer]: TBandItem read GetItem; default;
    property Orientation: TBarOrientation read fOrientation write SetOrientation;
  end;

implementation

function CompareBand(BandRect1, BandRect2: TBandRect): Integer;
begin
  if BandRect1.ResultRect.Top < BandRect2.ResultRect.Top then
    result := -2
  else if BandRect1.ResultRect.Top > BandRect2.ResultRect.Top then
    result := 2
  else if BandRect1.ResultRect.Left < BandRect2.ResultRect.Left then
    result := -1
  else if BandRect1.ResultRect.Left > BandRect2.ResultRect.Left then
    result := 1
  else result := 0;
end;

{ BandList }
constructor TBandList.Create;
begin
  inherited;
  fRowSize := 24;
end;

procedure TBandList.Delete(Index: integer);
begin
  if (Index >= 0) and (Index < fCount) then
  begin
    move(fArray^[Index + 1], fArray^[Index], SizeOf(fArray^[0]) * (fCount - Index - 1));
    dec(fCount);
    FillChar(fArray^[fCount], SizeOf(fArray^[0]) * (fBufCount - fCount), 0);
    CalcMinRow;
    Realign(fBoundsRect);
  end else
    raise Exception.CreateFmt(SListIndexError, [Index]);
end;

destructor TBandList.Destroy;
begin
  ReallocMem(fArray, 0);
  inherited;
end;


procedure TBandList.EndUpdate;
begin
  if fUpdateCount > 1 then
  begin
    Dec(fUpdateCount);
  end else
  if fUpdateCount = 1 then
  begin
    fUpdateCount := 0;
    Realign(fBoundsRect);
  end else fUpdateCount := 0;
end;

procedure TBandList.UpdateRect(var BandRect: TBandRect);
var X: integer;
begin
  BandRect.ResultRect := BandRect.SourceRect;
  X := Max(BandRect.MinW, BandRect.ResultRect.Right - BandRect.ResultRect.Left);
  BandRect.ResultRect.Right := BandRect.ResultRect.Left + X;
  X := Max(BandRect.MinH, BandRect.ResultRect.Bottom - BandRect.ResultRect.Top);
  BandRect.ResultRect.Bottom := BandRect.ResultRect.Top + X;

  BandRect.Row := (BandRect.ResultRect.Top + (RowSize div (2))) div (RowSize);
  BandRect.RowCount := Max((BandRect.ResultRect.Bottom - BandRect.ResultRect.Top + (RowSize - 1{ div (2)})) div (RowSize), 1);
  BandRect.ResultRect.Top := BandRect.Row * RowSize;
  BandRect.ResultRect.Bottom := BandRect.ResultRect.Top + BandRect.RowCount * RowSize;
end;

procedure TBandList.CalcMinRow;
var
  I: Integer;
begin
  fMinRow := MaxInt;
  fMaxRow := Low(Integer);
  for I := 0 to fCount - 1 do
  begin
    fMinRow := Min(fArray^[I].Row, fMinRow);
    fMaxRow := Max(fArray^[I].Row, fMaxRow);
  end;
end;

procedure TBandList.ChangeId(OldId, NewId: integer);
var I, J: integer;
begin
  if OldId <> NewId then
  begin
    I := FindBand(OldId);
    if I = -1 then
      raise Exception.CreateFmt(ErrNoId, [OldId]);
    J := FindBand(NewId);
    if J <> -1 then
      raise Exception.CreateFmt(ErrId, [NewId]);
    fArray^[I].Id := NewId;
  end;
end;

procedure TBandList.Init;
var I: integer;
begin
  for I := 0 to Count - 1 do
  begin
    fArray^[I].SourceRect := fArray^[I].ResultRect;
  end;
  Sort;
end;

procedure TBandList.InternalAdd(const BandRect: TBandRect;
                                const AArray:PBandRects;
                                var ACount: integer;
                                const ABufCount: integer);
var I: integer;
begin
  if (ACount >= ABufCount) or (AArray = nil) then
    raise Exception.Create(ErrInputParam);
  I := 0;
  while (I < ACount) and (CompareBand(BandRect, AArray^[I]) >= 0) do inc(I);
  inc(ACount);
  Move(AArray^[I], AArray^[I + 1], (ACount - I - 1) * SizeOf(AArray^[0]));
  Move(BandRect, AArray^[I], SizeOf(AArray^[0]));
end;

procedure TBandList.UpdateSize(ARect: TRect);
var I, J: integer;
    R, RR: TRect;
begin
  for I := 0 to Count - 1 do
  begin
    R := fArray^[I].ResultRect;
    R.Right := R.Left + fArray^[I].SourceRect.Right - fArray^[I].SourceRect.Left;
    for J := 0 to Count - 1 do
    if (I <> J) and (Windows.IntersectRect(RR, R, fArray^[J].ResultRect)) then begin
      R.Right := Min(R.Right, RR.Left);
      R.Right := Max(R.Right, R.Left + fArray^[I].MinW);
      R.Right := Max(Min(R.Right, fArray^[J].ResultRect.Left), fArray^[I].MinW)
    end;
    fArray^[I].ResultRect := R;
  end;
end;

procedure TBandList.Realign(var ARect: TRect; const N: byte = RealignCount);
var I: integer;
  procedure UpdateMinMax(NeedBottom: boolean);
  var I: integer;
  begin
      ARect.Top := High(Integer);
      if NeedBottom then ARect.Bottom := Low(Integer);
      for I := 0 to Count - 1 do
      begin
        ARect.Top := Min(ARect.Top, fArray^[I].ResultRect.Top);
        if NeedBottom then ARect.Bottom := Max(ARect.Bottom, fArray^[I].ResultRect.Bottom);
      end;
  end;
begin
  if fAlligning then Exit;
  fAlligning := true;
  try
    if (Count > 0) and (fUpdateCount = 0) then
    begin
      if N > 0 then
      begin
        Reset;
        AdjustAll(ARect);
      end;
      UpdateMinMax(false);
      for I := 0 to Count - 1 do
        OffsetRect(fArray^[I].ResultRect, 0, -ARect.Top);
      for I := 1 to N - 1 do
      begin
        AdjustAll(ARect);
        if I < N - 1 then
        begin
          Sort;
          UpdateMinMax(false);
          UpdateSize(ARect);
        end;
      end;
      UpdateMinMax(true);
    end;
    fBoundsRect := ARect;
    if @fonAligned <> nil then fonAligned(self);
  finally
    fAlligning := false;
  end;
end;

procedure TBandList.Add(BandRect: TBandRect);
var N, I: integer;
begin
  for I := 0 to Count - 1 do
    if BandRect.Id = fArray^[I].Id then
      raise Exception.CreateFmt(ErrId, [BandRect.Id]);
  UpdateRect(BandRect);
  N := ((fCount + PageSize) div (PageSize)) * PageSize;
  if N > fBufCount then
  begin
    fBufCount := N;
    ReallocMem(fArray, SizeOf(fArray^[0]) * fBufCount);
  end;
  InternalAdd(BandRect, fArray, fCount, fBufCount);
  CalcMinRow;
  Realign(fBoundsRect);
end;

procedure TBandList.Sort;
var I, N, Size: integer;
    P: PBandRects;
begin
  if fCount <= 0 then
  begin
    Exit;
  end;
  N := 0;
  P := nil;
  try
    Size := SizeOf(P^[0]) * fBufCount;
    ReallocMem(P, Size);
    FillChar(P^, Size, 0);
    for I := 0 to fCount - 1 do
      InternalAdd(fArray^[I], P, N, fBufCount);
    Move(P^, fArray^, Size);
  finally
    ReallocMem(P, 0);
  end;
  CalcMinRow;
end;

function TBandList.GetBands(I: integer): TBandRect;
begin
  if (I < 0) or (I >= fCount) then
    raise Exception.CreateFmt(SListIndexError, [I]);
  System.Move(fArray^[I], result, SizeOf(result));
end;

procedure TBandList.SetBands(I: integer; const Value: TBandRect);
var BandRect: TBandRect;
begin
  if (I < 0) or (I >= fCount) then
    raise Exception.CreateFmt(SListIndexError, [I]);
  if (not CompareMem(Addr(Value.SourceRect), Addr(fArray^[I].SourceRect), SizeOf(Value.SourceRect))) or
     (Value.MinW <> fArray^[I].MinW) or
     (Value.MinH <> fArray^[I].MinH) then
  begin
    BandRect := Value;
    UpdateRect(BandRect);
    System.Move(BandRect, fArray^[I], SizeOf(BandRect));
    if fUpdateCount = 0 then
    begin
      Sort;
      Realign(fBoundsRect);
    end;
  end;
end;

procedure TBandList.SetRighToLeft(const Value: boolean);
begin
  if fRighToLeft <> Value then
  begin
    fRighToLeft := Value;
    if fUpdateCount = 0 then Realign(fBoundsRect);
  end;
end;

procedure TBandList.SetRowSize(const Value: integer);
var V, I, H, OldRowCount: integer;
begin
  V := Max(8, Value);
  if (V <> fRowSize) then
  begin
    fRowSize := V;
    for I := 0 to Count - 1 do
    begin
      OldRowCount := fArray^[I].RowCount;
      if OldRowCount > 0 then
      begin
        H := fRowSize * OldRowCount;
        fArray^[I].SourceRect.Bottom := fArray^[I].SourceRect.Top + H;
      end;
      UpdateRect(fArray^[I]);
    end;
    Sort;
    Realign(fBoundsRect);
  end;
end;

procedure TBandList.SetSourceRect(I: integer);
var BandRect: TBandRect;
begin
  BandRect := Bands[I];
  UpdateRect(BandRect);
  fArray^[I] := BandRect;
  Sort;
end;

procedure TBandList.Reset;
var BandRect: TBandRect;
    I: integer;
begin
  for I := 0 to Count - 1 do
  begin
    BandRect := Bands[I];
    UpdateRect(BandRect);
    fArray^[I] := BandRect;
  end;
  Sort;
end;

procedure TBandList.Add(const ARect: TRect; const MinWidth, MinHeight,
  Id: integer);
var BandRect: TBandRect;
begin
  FillChar(BandRect, SizeOf(BandRect), 0);
  BandRect.SourceRect := ARect;
  BandRect.MinW := MinWidth;
  BandRect.MinH := MinHeight;
  BandRect.Id := Id;
  Add(BandRect);
end;

procedure TBandList.AdjustAll(ARect: TRect);
var I, J, N, L, R: integer;
    Reg: TEnableReg;
    Regs: PEnableRegs;
    Need, F: boolean;
    TmpR: TRect;
  function UpRows: boolean;
  var I, J, Row: integer;
      R: TRect;
  begin
    result := false;
    I := 0;
    while I < Count do
    begin
      F := False;
      R := fArray^[I].ResultRect;
      Row := fArray^[I].Row;
      if Row > MinRow then
      begin
        Need := true;
        while Need and (Row > MinRow) do
        begin
          OffsetRect(R, 0, -RowSize);
          Dec(Row);
          for J := 0 to I - 1 do
            if InterSectRect(TmpR, R, fArray^[J].ResultRect) then
            begin
              Need := false;
              break;
            end;
          if Need then
          begin
            result := true;
            F := True;
            fArray^[I].ResultRect := R;
            fArray^[I].Row := Row { (fArray^[I].ResultRect.Top + (RowSize div (2))) div (RowSize)};
          end;
        end;
      end;
      if F then
      begin
        I := 0;
        Sort;
      end else inc(I);
    end;
  end;  
  function DownRows: boolean;
  var I, J, Row: integer;
      R: TRect;
  begin
    result := false;
    I := 0;
    while I < Count do
    begin
      F := False;
      R := fArray^[I].ResultRect;
      Row := fArray^[I].Row;
      if Row > MinRow then
      begin
        Need := True;
        while (Need) and (Row <= MaxRow) do
        begin
          Need := False;
          for J := 0 to I - 1 do
            if InterSectRect(TmpR, R, fArray^[J].ResultRect) then
            begin
              Need := True;
              break;
            end;
          if Need then
          begin
            result := true;
            F := True;
            OffsetRect(R, 0, RowSize);
            Inc(Row);
            fArray^[I].ResultRect := R;
            fArray^[I].Row := Row { (fArray^[I].ResultRect.Top + (RowSize div (2))) div (RowSize)};
          end;
        end;
      end;
      if F then
      begin
        I := 0;
        Sort;
      end else inc(I);
    end;
  end;  
begin
  DownRows;
  UpRows;
  AdjustRow(MinRow, ARect.Left, ARect.Right);
  I := MinRow + 1;
  while I <= MaxRow do
  begin
    Reg.Left := ARect.Left;
    Reg.Width := ARect.Right - ARect.Left;
    Reg.Index := 0;
    Reg.Count := Count;
    Regs := nil;
    try
      N := CalcEnableRegs(I, Reg, Regs);
      for J := 0 to N - 1 do
      begin
          if J > 0 then L := (Regs^[J - 1].Left + Regs^[J - 1].Width + Regs^[J].Left) div (2)
                   else L := Low(Integer);
          if J < (N - 1) then R := (Regs^[J].Left + Regs^[J].Width + Regs^[J + 1].Left) div (2)
                         else R := High(Integer);
          AdjustRow(I, Regs^[J].Left, Regs^[J].Left + Regs^[J].Width, L, R);
      end;
    finally
      ReallocMem(Regs, 0);
    end;
    CalcMinRow;
    inc(I);
  end;
  UpRows;
  DownRows;
end;

function TBandList.FindBand(Id: integer): integer;
begin
  for result := 0 to Count - 1 do
    if Id = fArray^[result].Id then Exit;
  result := -1;
end;

function TBandList.FirstLastIndex(Row, MinLeft, MaxRight: integer; var First, Last: integer): boolean;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and
        ((fArray^[I].Row < Row) or ((fArray^[I].Row = Row) and ((fArray^[I].ResultRect.Right - fArray^[I].ResultRect.Left + 1) div (2) <= MinLeft))) do
    inc(I);
  First := I;
  while (I < Count) and
        (fArray^[I].Row = Row) and
        ((fArray^[I].ResultRect.Right - fArray^[I].ResultRect.Left + 1) div (2) <= MaxRight) do
      inc(I);
  Last := I - 1;
  result := Last >= First;
end;

procedure TBandList.AdjustRow(Row, Left, Right, MinLeft, MaxRight: integer);
var I, First, Last, MinWidth, Width, W, N: integer;
    K: Double;
    OutBands: array of Integer;
  function CalcMinWidth: integer;
  var i: integer;
  begin
    result := 0;
    for I := First to Last do
      inc(result, fArray^[i].MinW);
  end;
  function CalcWidth: integer;
  var i: integer;
  begin
    result := 0;
    for I := First to Last do
      inc(result, fArray^[i].ResultRect.Right - fArray^[i].ResultRect.Left);
  end;
  procedure MoveLeftSingl(Left: integer; I: integer);
  begin
      if (fArray^[I].ResultRect.Left < Left) then
        OffsetRect(fArray^[I].ResultRect, Left - fArray^[I].ResultRect.Left, 0);
  end;
  procedure MoveRightSingl(Right: integer; I: integer);
  begin
      if (fArray^[I].ResultRect.Right > Right) then
        OffsetRect(fArray^[I].ResultRect, Right - fArray^[I].ResultRect.Right, 0);
  end;
  procedure MoveLeft(Left: integer);
  var I: integer;
  begin
    I := First;
    while I <= Last do
    begin
      MoveLeftSingl(Left, I);
      Left := fArray^[I].ResultRect.Right;
      inc(I);
    end;
  end;
  procedure MoveRight(Right: integer);
  var I: integer;
  begin
    I := Last;
    while I >= First do
    begin
      MoveRightSingl(Right, I);
      Right := fArray^[I].ResultRect.Left;
      dec(I);
    end;
  end;
begin
  W := Right - Left;
  if not FirstLastIndex(Row, MinLeft, MaxRight, First, Last) then Exit;
  N := 0;
  Width := CalcWidth;
  MinWidth := CalcMinWidth;
  while (Last > First) and (MinWidth > (Right - Left)) do
  begin
    inc(N);
    SetLength(OutBands, N);
    if RighToLeft then
    begin
      OutBands[N - 1] := First;
      inc(First);
    end
    else begin
      OutBands[N - 1] := Last;
      Dec(Last);
    end;
    MinWidth := CalcMinWidth;
  end;

  if (MinWidth <= W) and (Width > W) then
  begin
    K := (W - MinWidth) / (Width - MinWidth);
    for I := First to Last do
      fArray^[I].ResultRect.Right := fArray^[I].ResultRect.Left + fArray^[I].MinW + Round(K * (fArray^[I].ResultRect.Right - fArray^[I].ResultRect.Left - fArray^[I].MinW));
    if RighToLeft then I := First else I := Last;
    Width := CalcWidth;
    while Width <> W do
    begin
      if Width > W then
      begin
        if (fArray^[I].ResultRect.Right > (fArray^[I].ResultRect.Left + fArray^[I].MinW)) then Dec(fArray^[I].ResultRect.Right);
      end else
        inc(fArray^[I].ResultRect.Right);
      if RighToLeft then inc(I) else Dec(I);
      if I < First then I := Last;
      if I > Last then I := First;
      Width := CalcWidth;
      if (Width = MinWidth) and (Width > W) then break;
    end;
  end else
  begin
    if MinWidth > W then
      for I := First to Last do
        fArray^[I].ResultRect.Right := fArray^[I].ResultRect.Left + fArray^[I].MinW;
  end;
  if RighToLeft then
  begin
    MoveLeft(Left);
    MoveRight(Right);
  end else
  begin
    MoveRight(Right);
    MoveLeft(Left);
  end;
  if Length(OutBands) > 0 then
  begin
    for i := 0 to Length(OutBands) - 1 do
    begin
      OffsetRect(fArray^[OutBands[I]].ResultRect, 0, RowSize);
      if RighToLeft then
      begin
        MoveLeftSingl(Left, OutBands[I]);
        MoveRightSingl(Right, OutBands[I]);
      end else
      begin
        MoveRightSingl(Right, OutBands[I]);
        MoveLeftSingl(Left, OutBands[I]);
      end;
      inc(fArray^[OutBands[I]].Row);
    end;
    Sort;
  end;
end;

function TBandList.BandById(Id: integer): TBandRect;
var I: integer;
begin
  I := FindBand(Id);
  if I <> -1 then
    result := fArray^[I]
  else
    raise Exception.CreateFmt(ErrNoId, [Id]);
end;

procedure TBandList.BeginUpdate;
begin
  Inc(fUpdateCount);
end;

function TBandList.CalcEnableRegs(Row: integer; EnableReg: TEnableReg;
  var EnableRegs: PEnableRegs): integer;
var I, J, K, MaxX: integer;
    Reg, NewReg: TEnableReg;
  function TestReg(var OldReg, NewReg: TEnableReg): boolean;
  var OldRight, NewRight, Right: integer;
  begin
    result := false;
    OldRight := OldReg.Left + OldReg.Width;
    NewRight := OldRight;
    Right := Reg.Left + Reg.Width;
    if (OldRight < Reg.Left) or
       (OldReg.Left > Right) then Exit;
    FillChar(NewReg, SizeOf(NewReg), 0);
    if OldReg.Left < Reg.Left then
    begin
      if NewRight > Right then
      begin //1
        NewReg.Left := Right {+ 1};
        NewReg.Width := NewRight - NewReg.Left;
        result := NewReg.Width > 0;
      end;
      OldRight := Reg.Left {- 1};  //2
    end else
      OldReg.Left := Right {+ 1}; //3
    OldReg.Width := OldRight - OldReg.Left;
  end;
begin
  result := 0;
  I := 0;
  while (I < Count) and (fArray^[I].Row < Row) do inc(I);
  if I <= Count then
  begin
    inc(result);
    ReallocMem(EnableRegs, SizeOf(EnableRegs^[0]) * result);
    Move(EnableReg, EnableRegs^[result - 1], SizeOf(EnableRegs^[0]));
    for J := 0 to I - 1  do
      if fArray^[J].Row + fArray^[J].RowCount > Row then
      begin
        K := 0;
        Reg.Left := fArray^[J].ResultRect.Left;
        Reg.Width := fArray^[J].ResultRect.Right - fArray^[J].ResultRect.Left;
        while K < result do
        begin
          if TestReg(EnableRegs^[K], NewReg) then
          begin
            inc(result);
            ReallocMem(EnableRegs, SizeOf(EnableRegs^[0]) * result);
            Move(NewReg, EnableRegs^[result - 1], SizeOf(EnableRegs^[0]));
          end;
          if EnableRegs^[K].Width <= 0 then
          begin
            Dec(result);
            Move(EnableRegs^[K + 1], EnableRegs^[K], SizeOf(EnableRegs^[0]) * (result - K));
            ReallocMem(EnableRegs, SizeOf(EnableRegs^[0]) * result);
          end else
            inc(K);
        end;
      end;
    if Result > 0 then
    begin
      K := I;
      for J := 0 to result - 1 do
      begin
        if J = result - 1 then MaxX := High(Integer)
                          else MaxX := (EnableRegs^[J].Left + EnableRegs^[J].Width + EnableRegs^[J + 1].Left) div 2;
        EnableRegs^[J].Index := K;
        EnableRegs^[J].Count := 0;
        while (K < fCount) and (fArray^[K].Row = Row) and
              {((BandRects[K].R.Left + BandRects[K].R.Right + 1) div (2) < MaxX)}
              (fArray^[K].ResultRect.Left div 2 < MaxX) do
        begin
          inc(K);
          inc(EnableRegs^[J].Count);
        end;
      end;
    end;
  end;
end;

{ TBandItem }

procedure TBandItem.Changed(AllItems: Boolean);
begin
  if AllItems then fChangedAll := true
              else fChanged := true;
  if fUpdateCount = 0 then inherited;
end;

procedure TBandItem.BeginUpdate;
begin
  inc(fUpdateCount);
end;

procedure TBandItem.EndUpdate;
var
  I: Integer;
begin
  dec(fUpdateCount);
  if fUpdateCount = 0 then
  begin
    if Collection <> nil then
    begin
      for I := 0 to Collection.Count - 1 do
        if TBandItem(Collection.Items[I]).fChangedAll then
        begin
          Changed(true);
          Exit;
        end;
      for I := 0 to Collection.Count - 1 do
        if TBandItem(Collection.Items[I]).fChanged then
        begin
          TBandItem(Collection.Items[I]).Changed(false);
        end;
    end;
  end;
  if fUpdateCount < 0 then fUpdateCount := 0;
end;

function TBandItem.GetBandItems: TBandItems;
begin
  result := TBandItems(Collection);
end;

function TBandItem.IsVert: boolean;
begin
  result := BandItems.IsVert;
end;

function TBandItem.IsRTL: boolean;
begin
  result := BandItems.Orientation = boRightToLeft;
end;

function TBandItem.IntInsets: TRect;
begin
  if IsVert then
  begin
    result.Top := Insets.Left;
    result.Bottom := Insets.Right + SysArea;
    result.Left := Insets.Top;
    result.Right := Insets.Bottom;
  end else
  begin
    if IsRTL then
    begin
      result.Left := Insets.Right + SysArea;
      result.Right := Insets.Left;
    end else
    begin
      result.Right := Insets.Right + SysArea;
      result.Left := Insets.Left;
    end;
    result.Top := Insets.Top;
    result.Bottom := Insets.Bottom;
  end;
end;

procedure TBandItem.CalcConstraints(var MinW: Integer; var MinH: Integer);
var
  IR: TRect;
begin
  IR := IntInsets;
  MinW := IR.Left + IR.Right + fMinWidth;
  MinH := IR.Top + IR.Bottom + fMinHeight;
end;

procedure TBandItem.SetBoundsRect(const Value: TRect);
var MinW, MinH: integer;
    R: TRect;
begin
  R := Value;
  CalcConstraints(MinW, MinH);
  if R.Bottom - R.Top < MinH then R.Bottom := R.Top + MinH;
  if R.Right - R.Left < MinW then
    if IsRTL then R.Left := R.Right - MinW
             else R.Right := R.Left + MinW;
  if not CompareMem(@fBoundsRect, @R, SizeOf(R)) then
  begin
    fBoundsRect := R;
    if Visible then Changed(true);
  end;
end;

procedure TBandItem.SetMinHeight(const Value: Word);
begin
  fMinHeight := Value;
  SetBoundsRect(fBoundsRect);
end;

procedure TBandItem.SetMinWidth(const Value: Word);
begin
  fMinWidth := Value;
  SetBoundsRect(fBoundsRect);
end;

procedure TBandItem.SetClientRect(const Value: TRect);
var R, IR: TRect;
begin
  IR := IntInsets;
  R.Left := Value.Left - IR.Left;
  R.Right := Value.Right + IR.Right;
  R.Top := Value.Top - IR.Top;
  R.Bottom := Value.Bottom + IR.Bottom;
  SetBoundsRect(R);
end;

function TBandItem.GetClientRect: TRect;
var IR: TRect;
begin
  IR := IntInsets;
  result.Left := BoundsRect.Left + IR.Left;
  result.Right := BoundsRect.Right - IR.Right;
  result.Top := BoundsRect.Top + IR.Top;
  result.Bottom := BoundsRect.Bottom - IR.Bottom;
end;

function TBandItem.GetHandRect: TRect;
begin
  result := BoundsRect;
  if IsVert then
  begin
    Result.Bottom := Result.Top + Insets.Left - Insets.Right;
    inc(Result.Top, Insets.Right);
  end else
  begin
    if IsRTL then
    begin
      result.Left := result.Right - Insets.Left + Insets.Right;
      Dec(result.Right, Insets.Right);
    end else
    begin
      result.Right := result.Left + Insets.Left - Insets.Right;
      inc(Result.Left, Insets.Right);
    end;
  end;
end;

function TBandItem.GetSysRect: TRect;
begin
  result := ClientRect;
  if IsVert then
  begin
    result.Top := result.Bottom;
    result.Bottom := result.Top + fSysArea;
  end else
  begin
    if IsRTL then
    begin
      result.Right := result.Left;
      result.Left := result.Right - fSysArea;
    end else
    begin
      result.Left := result.Right;
      result.Right := result.Left + fSysArea;
    end;
  end;
end;

procedure TBandItem.SetInsets(const Value: TRect);
var R: TRect;
begin
  R.Left := Max(0, Value.Right);
  R.Right := Max(0, Value.Right);
  R.Top := Max(0, Value.Top);
  R.Bottom := Max(0, Value.Bottom);
  if not CompareMem(@fInsets, @R, SizeOf(R)) then
  begin
    fInsets := R;
    SetBoundsRect(fBoundsRect);
  end;
end;

procedure TBandItem.SetSysArea(const Value: Byte);
var R: TRect;
    D, I: integer;
    BandRect: TBandRect;
  procedure UpdateR(var R: TRect; Reverted: boolean);
  begin
    if not Reverted then R := BandItems.RevertRect(R);
    if IsVert then
      inc(R.Right, D)
    else
      if IsRTL then dec(R.Left, D)
               else inc(R.Right, D);
    if not Reverted then R := BandItems.RevertRect(R);
  end;
begin
  if fSysArea <> Value then
  begin
    D := Value - fSysArea;
    fSysArea := Value;
    BandItems.BeginUpdate;
    try
      if BandItems.BandList.FindBand(Id) >= 0 then
      begin
        BandRect := BandItems.BandList.BandById(Id);
        UpdateR(BandRect.SourceRect, true);
        I := BandItems.BandList.FindBand(Id);
        BandItems.BandList[I] := BandRect;
        R := BoundsRect;
        UpdateR(R, false);
        SetBoundsRect(R);
      end;
    finally
      BandItems.EndUpdate;
    end;
  end;
end;


constructor TBandItem.Create(Collection: TCollection);
begin
  if not (Collection is TBandItems) then
    raise Exception.Create(ErrInputParam);
  inherited;
  fInsets := Rect(11, 2, 2, 2);
  fVisible := true;
  fChangedAll := true;
end;

destructor TBandItem.Destroy;
begin
  fDestroy := true;
  inherited;
end;

procedure TBandItem.ControlChanged;
begin
  if Control <> nil then
  begin
    Visible := Control.Visible;
    MinWidth := Control.Constraints.MinWidth;
    MinHeight := Control.Constraints.MaxHeight;
    ClientRect := Control.BoundsRect;
  end
  else
  begin
    Visible := false;
    MinWidth := 0;
    MinHeight := 0;
  end;
end;

procedure TBandItem.SetCollection(Value: TCollection);
var Ok: boolean;
begin
  Ok := Value is TBandItems;
  if Not Ok then
    Ok :=(Value = nil) and
         ((Collection is TOwnedCollection) and
          (TOwnedCollection(Collection).Owner is TControl) and
          ((csDestroying in TControl(TOwnedCollection(Collection).Owner).ComponentState)
            or fDestroy));
  if not Ok then raise Exception.Create(ErrInputParam);
  if Value = nil then inherited
  else
  begin
    BeginUpdate;
    try
      inherited;
      if Visible then Changed(true);
    finally
      EndUpdate;
    end;
  end;
end;

procedure TBandItem.SetControl(const Value: TControl);
begin
  if fControl <> Value then
  begin
    BeginUpdate;
    try
      fControl := Value;
      fControl.Align := alNone;
      ControlChanged;
    finally
      EndUpdate;
    end;
  end;
end;

procedure TBandItem.SetVisible(const Value: boolean);
begin
  if fVisible <> Value then
  begin
    fVisible := Value;
    if Control <> nil then Control.Visible := fVisible;
    Changed(true);
  end;
end;



{ TBandItems }

function TBandItems.Add(Control: TControl): TBandItem;
begin
  BeginUpdate;
  try
    result := TBandItem(inherited Add);
    result.Control := Control;
    if GetOwner is TWinControl then
    begin
      Control.Parent := TWinControl(GetOwner);
    end;
  finally
    EndUpdate;
  end;
end;

procedure TBandItems.Delete(Index: Integer);
begin
  BeginUpdate;
  try
    inherited Delete(Index);
  finally
    EndUpdate;
  end;
end;

procedure TBandItems.Aligned(Sender: TObject);
begin
  IntAligned;
  DoAligned;
end;

function TBandItems.BandByControl(Control: TControl): TBandItem;
var I: integer;
begin
  result := nil;
  for I := 0 to Count - 1 do
    if Items[I].Control = Control then
    begin
      result := Items[I];
      break;
    end;
end;

procedure TBandItems.IntAligned;
var
  I: Integer;
  R: TRect;
  J: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    J := fBandList.FindBand(TBandItem(Items[I]).ID);
    if J >= 0 then
    begin
      R := RevertRect(fBandList.Bands[J].ResultRect);
      TBandItem(Items[I]).fBoundsRect := R;
    end;
  end;
end;

procedure TBandItems.BeginUpdate;
begin
  inherited;
  if fBandList <> nil then fBandList.BeginUpdate;
end;

procedure TBandItems.EndUpdate;
begin
  inherited;
  if fBandList <> nil then fBandList.EndUpdate;
end;

constructor TBandItems.Create(AOwner: TPersistent;
  ItemClass: TCollectionItemClass);
begin
  inherited;
  fBandList := TBandList.Create;
  fBandList.RowSize := 24;
  fBandList.onAligned := Aligned;
end;

destructor TBandItems.Destroy;
begin
  FreeAndNil(fBandList);
  inherited;
end;

procedure TBandItems.DoAligned;
var I: integer;
    R1, R2: TRect;
begin
  if GetOwner is TWinControl then
  begin
    TWinControl(GetOwner).DisableAlign;
    try
      for I := 0 to Count - 1 do
      if Items[I].Control <> nil then
      begin
        R1 := Items[I].Control.BoundsRect;
        R2 := Items[I].ClientRect;
        if not CompareMem(@R1, @R2, SizeOf(R1)) then
          Items[I].Control.BoundsRect := R2;
        Items[I].Control.Visible := Items[I].Visible;
        if Items[I].Control.Parent <> GetOwner then
          Items[I].Control.Parent := TWinControl(GetOwner);
      end;
    finally
      TWinControl(GetOwner).EnableAlign;
    end;
  end;
end;

procedure TBandItems.DoPainted;
begin
  if (GetOwner is TWinControl) and (TWinControl(GetOwner).HandleAllocated) then
  begin
    InvalidateRect(TWinControl(GetOwner).Handle, nil, true);
  end;
end;

function TBandItems.GetItem(Index: Integer): TBandItem;
begin
  result := TBandItem(inherited GetItem(Index));
end;

function TBandItems.GetRowSize: integer;
begin
  result := fBandList.RowSize;
end;

procedure TBandItems.SetRowSize(const Value: integer);
begin
  fBandList.RowSize := Value;
end;

function TBandItems.IsVert: boolean;
begin
  result := Orientation in [boTopToBottom, boBottomToTop];
end;

  {procedure XCHG(var X1, X2: integer);
  assembler; asm
    push  eBx
    push  eCx
    mov   eBx, [X1]
    mov   eCx, [X2]
    XCHG  eBx, eCx
    mov   [X1], eBx
    mov   [X2], eCx
    pop   eCx
    pop   eBx
  end;}
  procedure XCHG(var X1, X2: integer);
  var X: integer;
  begin
    X := X1;
    X1 := X2;
    X2 := X;
  end;

function TBandItems.RevertRect(const R: TRect): TRect;
begin
  result := R;
  if IsVert then
  begin
    XCHG(result.Left, result.Top);
    XCHG(result.Right, result.Bottom);
  end;
end;

function TBandItems.GetBoundsRect: TRect;
begin
  result := RevertRect(fBoundsRect);
end;

procedure TBandItems.SetBoundsRect(const Value: TRect);
var R: TRect;
begin
  R := RevertRect(Value);
  if ((R.Right - R.Left) <> (fBoundsRect.Right - fBoundsRect.Left)) or
     ((R.Bottom - R.Top) <> (fBoundsRect.Bottom - fBoundsRect.Top)) then
  begin
    fBoundsRect := R;
    fBandList.Realign(R);
  end;
end;

procedure TBandItems.SetOrientation(const Value: TBarOrientation);
var I, J: integer;
    OldBoundsRect: TRect;
    BandRect: TBandRect;
begin
  if fOrientation <> Value then
  begin
    OldBoundsRect := BoundsRect;
    fOrientation := Value;
    BeginUpdate;
    try
      fBandList.Reset;
      fBandList.RighToLeft := fOrientation in [boRightToLeft, boBottomToTop];
      for I := 0 to Count - 1 do
      begin
        TBandItem(Items[I]).fChangedAll := true;
        J := fBandList.FindBand(TBandItem(Items[I]).ID);
        if J >= 0 then
        begin
          BandRect := fBandList.Bands[J];
          TBandItem(Items[I]).CalcConstraints(BandRect.MinW, BandRect.MinH);
          if IsVert then XCHG(BandRect.MinW, BandRect.MinH);
          fBandList.Bands[J] := BandRect;
        end;
      end;
      BoundsRect := OldBoundsRect;
    finally
      EndUpdate;
    end;
  end;
end;

procedure TBandItems.Update(Item: TCollectionItem);
var I, J: integer;
    BandRect: TBandRect;
    All, Need: boolean;
    BandItem: TBandItem;
begin
  All := false;
  Need := false;
  inherited;
  fBandList.BeginUpdate;
  try
    // 
    J := 0;
    while J < fBandList.Count do
    begin
      BandRect := fBandList.Bands[J];
      BandItem := TBandItem(FindItemID(BandRect.Id));
      if (BandItem = nil) or (not BandItem.Visible) then
      begin
        fBandList.Delete(J);
        Need := true;
        J := 0;
      end else inc(J);
    end;
    //   
    for I := 0 to Count - 1 do
      if TBandItem(Items[I]).fChanged then
      begin
        TBandItem(Items[I]).fChanged := false;
        Need := true;
      end;
    //      fBandList
    for I := 0 to Count - 1 do
    if (TBandItem(Items[I]).fChangedAll) and
       (TBandItem(Items[I]).Visible) then
    begin
      J := fBandList.FindBand(TBandItem(Items[I]).ID);
      if J < 0 then
      begin
        BandRect.SourceRect := RevertRect(TBandItem(Items[I]).BoundsRect);
        //SA := TBandItem(Items[I]).fSysArea;
        //TBandItem(Items[I]).fSysArea := 0;
        if fBandList.RighToLeft then
          Dec(BandRect.SourceRect.Left, TBandItem(Items[I]).SysArea)
        else
          Inc(BandRect.SourceRect.Right, TBandItem(Items[I]).SysArea);
        TBandItem(Items[I]).CalcConstraints(BandRect.MinW, BandRect.MinH);
        if IsVert then XCHG(BandRect.MinW, BandRect.MinH);
        BandRect.Id := Items[I].ID;
        fBandList.Add(BandRect);
        J := fBandList.FindBand(TBandItem(Items[I]).ID);
        //TBandItem(Items[I]).fSysArea := SA;
      end;
      if J >= 0 then
      begin
        All := true;
        TBandItem(Items[I]).fChangedAll := false;
        BandRect := fBandList.Bands[J];
        TBandItem(Items[I]).CalcConstraints(BandRect.MinW, BandRect.MinH);
        if IsVert then XCHG(BandRect.MinW, BandRect.MinH);
        BandRect.SourceRect := RevertRect(TBandItem(Items[I]).BoundsRect);
        fBandList.Bands[J] := BandRect;
      end;
    end;
  finally
    fBandList.EndUpdate;
  end;
  if Need or All then DoPainted;
end;

end.
