unit ImageIndex;

(*
  * Module:           TImageIndexProperty
  * Version:    2.9 (26.06.2010)
  * Author:      
  * Comments:         .
  *                    
  *              : roschinspb@mail.ru
  *             http://www.delphikingdom.com/asp/users.asp?ID=1271
  *             http://www.roschinspb.narod.ru
  *
  * Copyright:   2006  2010
  *
  * Version:    2.9 (22.06.2010)
  * Realized:           
  *                     
  *
  * Version:    1.8.0.0 (20.01.2010)
  * Realized:        TCustomGliphs
  *)
interface

{$I '..\Directives.inc'}

uses Windows,
  SysUtils,
  Classes,
  CommCtrl,
{$IFDEF DELPHI_2005_UP}
  GraphUtil,
{$ELSE}
  GraphUtilEx,
{$ENDIF}
{$IFDEF DELPHI_6_UP}
  DesignIntf,
  DesignEditors,
  VCLEditors,
{$ELSE}
  DsgnIntf,
{$ENDIF}
  TypInfo,
  Graphics,
  Controls,
  ImgList,
  SpeedButtonCep,
  IniFiles,
  uBaseForm,
  ufImageIndex;

resourcestring
  //    
  NameGroupProps = ' ';
  //      
  NameItemIsDialog = ' ';

const //  ,     
  GroupProps: array [0 .. 1] of string = ('Gliphs', 'Blend');
  StrDialog = 'Dialog';

type
  //   ImgList
  TImageIndexProperty = class(TEnumProperty {$IFDEF VER130}{$ELSE},
{$IFDEF DELPHI_2006_UP}
    ICustomPropertyDrawing80,
{$ENDIF}
    ICustomPropertyDrawing, ICustomPropertyListDrawing {$ENDIF})
  private
    fCountImages: integer;
    fHandle: THandle;
    fW, fH, fDefaultHeight: integer;
    procedure UpdateInfo;
    function GetIndexImage(value: string): integer;
    procedure SaveIsDialog;
    procedure LoadIsDialog;
  public
    procedure GetValues(Proc: TGetStrProc); override;
    function GetValue: string; override;
    procedure SetValue(const value: string); override;
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
    procedure ListMeasureHeight(const value: string; ACanvas: TCanvas; var AHeight: integer); {$IFDEF VER130} override;
{$ELSE}{$ENDIF}
    procedure ListMeasureWidth(const value: string; ACanvas: TCanvas; var AWidth: integer); {$IFDEF VER130} override;
{$ELSE}{$ENDIF}
    procedure ListDrawValue(const value: string; ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean);
{$IFDEF VER130} override; {$ELSE}{$ENDIF}
    procedure PropDrawName(ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean); {$IFDEF VER130} override;
{$ELSE}{$ENDIF}
    procedure DrawScaledImage(const ARect: TRect; ACanvas: TCanvas);
    procedure PropDrawValue(ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean); {$IFDEF VER130} override;
{$ELSE}{$ENDIF}
{$IFDEF DELPHI_2006_UP}
    function PropDrawNameRect(const ARect: TRect): TRect;
    function PropDrawValueRect(const ARect: TRect): TRect;
{$ENDIF}
  end;

procedure Register;

implementation

uses StdCtrls;

var
  IsDialog: Boolean;
  LastIsDialogLoad: TDateTime;

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TImageIndex), TCustomGliphs, '', TImageIndexProperty);
{$IFDEF DELPHI_2005_UP}
  RegisterPropertiesInCategory(NameGroupProps, TSpeedButtonCep, GroupProps);
{$ENDIF}
end;

{ TImageIndexProperty }
//  ,    

procedure TImageIndexProperty.UpdateInfo;
var
  Component: TPersistent;
begin
  try
    fHandle := 0;
    fCountImages := 0;
    fW := 0;
    fH := 0;
    if PropCount > 0 then
    begin
      Component := GetComponent(0);
      if Component is TCustomGliphs then
      begin
        fHandle := TCustomGliphs(Component).Handle;
      end
      else
      begin
        if Component is TControl then
        begin
          fHandle := TControl(Component).Perform(TCM_GETIMAGELIST, 0, 0);
        end;
      end;
      if fHandle <> 0 then
      begin
        ImageList_GetIconSize(fHandle, fW, fH);
        fCountImages := ImageList_GetImageCount(fHandle);
      end;
    end;
  except
    on E: Exception do
    begin
      raise Exception.Create('UpdateInfo: ' + #13#10 + E.ClassName + #13#10 + E.Message);
    end;
  end;
end;

//   

function TImageIndexProperty.GetIndexImage(value: string): integer;
var
  i, Code: integer;
begin
  if value = NameItemIsDialog then
    Result := -2
  else
  begin
    Result := -1;
    Val(value, i, Code);
    if (Code = 0) and (i >= 0) then
      Result := i;
  end;
end;

function TImageIndexProperty.GetAttributes: TPropertyAttributes;
begin
  LoadIsDialog;
  Result := [paMultiSelect, paRevertable];
  if IsDialog then
    Result := Result + [paDialog]
  else
    Result := Result + [paValueList];
end;

procedure TImageIndexProperty.GetValues(Proc: TGetStrProc);
var
  i: integer;
begin
  // inherited;
  UpdateInfo;
  for i := -1 to fCountImages - 1 do
    Proc(inttostr(i));
  Proc(NameItemIsDialog);
end;

function TImageIndexProperty.GetValue: string;
begin
  Result := inherited GetValue;
end;

procedure TImageIndexProperty.SetValue(const value: string);
var
  index, Code: integer;
begin
  if value = NameItemIsDialog then
  begin
    IsDialog := True;
    SaveIsDialog;
  end
  else
  begin
    if value <> '' then
    begin
      Val(trim(value), index, Code);
      if (Code <> 0) or (index <= -1) then
        inherited SetValue('-1')
      else
      begin
        inherited SetValue(inttostr(index));
      end;
    end
    else
      inherited;
  end;
end;

//  

procedure TImageIndexProperty.ListDrawValue(const value: string; ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean);
var
  i, X, Y, Dx, Dy: integer;
  Bk: TColor;
  Style: Cardinal;
  R: TRect;
  NewW: integer;
  S: string;
begin
  inherited;
  try
    i := GetIndexImage(value);
    if i = -2 then
      S := NameItemIsDialog
    else
    begin
      S := inttostr(i);
      if i >= 0 then
        S := ' ' + S;
    end;
    R := ARect;
    ACanvas.FillRect(R);
    SetBkMode(ACanvas.Handle, TRANSPARENT);
    UpdateInfo;
    NewW := ARect.Bottom - ARect.Top;
    if i >= 0 then
    begin
      if (i < fCountImages) then
      begin
        Dy := ARect.Bottom - ARect.Top;
        if Dy < fH then
        begin
          Y := ARect.Top;
        end
        else
        begin
          Y := (Dy - fH - 1) div (2) + ARect.Top;
          Dy := fH;
        end;
        Dx := ARect.Right - ARect.Left;
        if Dx < fW then
          X := ARect.Left
        else
        begin
          X := ARect.Left;
          Dx := fW;
        end;
        Bk := ColorToRGB(ACanvas.Brush.Color);
        if ASelected then
          Style := ILD_SELECTED
        else
          Style := ILD_Normal;
{$IFDEF DELPHI_2006_UP}
        Inc(X);
{$ELSE}
        Inc(X, 2);
{$ENDIF};
        ImageList_DrawEx(self.fHandle, i, ACanvas.Handle, X, Y, Dx, Dy, Bk, CLR_DEFAULT, Style);
      end;
    end;
    if self.fHandle = 0 then
    begin
      ACanvas.Pen.Color := clGrayText;
{$IFDEF DELPHI_2006_UP}
      X := 1;
{$ELSE}
      X := 2;
{$ENDIF};
      ACanvas.Rectangle(ARect.Left + X + 1, ARect.Top + 1, ARect.Left + NewW + X - 1, ARect.Top + NewW - 1);
    end;
    if NewW < fW then
      NewW := fW;
    if NewW > 32 then
      NewW := 32;
    Inc(R.Left, NewW + 4);
    Windows.DrawText(ACanvas.Handle, PChar(S), Length(S), R, DT_VCenter);
  except
    on E: Exception do
    begin
      raise Exception.Create('ListDrawValue: ' + #13#10 + E.ClassName + #13#10 + E.Message);
    end;
  end;
end;

procedure TImageIndexProperty.ListMeasureHeight(const value: string; ACanvas: TCanvas; var AHeight: integer);
var
  i: integer;
begin
  inherited;
  fDefaultHeight := AHeight;
  i := GetIndexImage(value);
  if i >= 0 then
  begin
    // UpdateInfo;
    if (i < fCountImages) and (AHeight < fH) then
      AHeight := fH;
    if AHeight > 26 then
      AHeight := 26;
  end;
end;

procedure TImageIndexProperty.ListMeasureWidth(const value: string; ACanvas: TCanvas; var AWidth: integer);
var
  i: integer;
begin
  inherited;
  i := GetIndexImage(value);
  if i >= 0 then
  begin
    // UpdateInfo;
    if (i < fCountImages) and (AWidth < fW) then
      AWidth := fW;
    if AWidth > 32 then
      AWidth := 32;
  end;
end;

procedure TImageIndexProperty.LoadIsDialog;
var
  Ini: TIniFile;
begin
  if (Now - LastIsDialogLoad) > (10 / 24 / 60 / 60) then
  begin
    LastIsDialogLoad := Now;
    Ini := TIniFile.Create(DefaultIniFileName(''));
    try
      IsDialog := Ini.ReadBool(ClassName, StrDialog, false);
    finally
      FreeAndNil(Ini);
    end;
  end;
end;

procedure TImageIndexProperty.SaveIsDialog;
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(DefaultIniFileName(''));
  try
    Ini.WriteBool(ClassName, StrDialog, IsDialog);
  finally
    FreeAndNil(Ini);
  end;
  LastIsDialogLoad := Now;
end;

procedure TImageIndexProperty.DrawScaledImage(const ARect: TRect; ACanvas: TCanvas);
var
  NewW: integer;
  X: integer;
  i: integer;
  Scale1: Double;
  Scale2: Double;
  NewB: TBitmap;
  NewH: integer;
  B: TBitmap;
  Y: integer;
  OldPenColor, OldBrushColor: TColor;
  OldBrushStyle: TBrushStyle;
  S: string;
{$IFNDEF DELPHI_2006_UP}
  R: TRect;
{$ENDIF}
begin
  S := GetVisualValue;
  if S <> '' then
    i := GetIndexImage(S)
  else
    i := -1;
  NewH := ARect.Bottom - ARect.Top;
  NewW := NewH;
{$IFNDEF DELPHI_2006_UP}
  S := inttostr(i);
  if i >= 0 then
    S := ' ' + S;
  ACanvas.FillRect(ARect);
  R := ARect;
  Inc(R.Left, NewW + 4);
  Windows.DrawText(ACanvas.Handle, PChar(S), Length(S), R, DT_VCenter);
{$ENDIF}
  if i >= 0 then
    UpdateInfo;
  if (i >= 0) and (i < fCountImages) then
  begin
    B := TBitmap.Create;
    try
{$IFDEF DELPHI_2005_UP}
      B.SetSize(fW, fH);
{$ELSE}
      B.Width := fW;
      B.Height := fH;
{$ENDIF}
      B.Canvas.Brush.Color := ACanvas.Brush.Color;
      B.Canvas.FillRect(Rect(0, 0, fW, fH));
      ImageList_DrawEx(self.fHandle, i, B.Canvas.Handle, 0, 0, fW, fH, ColorToRGB(ACanvas.Brush.Color), CLR_DEFAULT,
        ILD_Normal);
      Scale1 := NewW;
      Scale1 := Scale1 / fW;
      Scale2 := NewH;
      Scale2 := Scale2 / fH;
      if Scale1 > Scale2 then
        Scale1 := Scale2;
      NewB := TBitmap.Create;
      try
        ScaleImage(B, NewB, Scale1);
        X := (NewW - NewB.Width) div (2);
        Y := (NewH - NewB.Height) div (2);
        ACanvas.Draw(ARect.Left + X + 1, ARect.Top + Y, NewB);
      finally
        FreeAndNil(NewB);
      end;
    finally
      FreeAndNil(B);
    end;
  end
  else
  begin
    OldBrushStyle := ACanvas.Brush.Style;
    OldBrushColor := ACanvas.Brush.Color;
    OldPenColor := ACanvas.Pen.Color;
    try
      ACanvas.Brush.Style := bsClear;
      ACanvas.Pen.Color := clGrayText;
      ACanvas.Rectangle(ARect.Left + 2, ARect.Top + 1, ARect.Left + NewW, ARect.Top + NewH - 1);
    finally
      ACanvas.Brush.Style := OldBrushStyle;
      ACanvas.Brush.Color := OldBrushColor;
      ACanvas.Pen.Color := OldPenColor;
    end;
  end;
end;

procedure TImageIndexProperty.PropDrawValue(ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean);
begin
  if GetVisualValue <> '' then
  begin
    DrawScaledImage(ARect, ACanvas);
  end
  else
  begin
{$IFDEF DELPHI_6_UP}
    DefaultPropertyDrawValue(self, ACanvas, ARect);
{$ELSE}
    inherited;
{$ENDIF}
  end;
end;

procedure TImageIndexProperty.PropDrawName(ACanvas: TCanvas; const ARect: TRect; ASelected: Boolean);
begin
{$IFDEF DELPHI_6_UP}
  DefaultPropertyDrawName(self, ACanvas, ARect);
{$ELSE}
  inherited;
{$ENDIF}
end;
{$IFDEF DELPHI_2006_UP}

function TImageIndexProperty.PropDrawNameRect(const ARect: TRect): TRect;
begin
  Result := ARect;
end;

function TImageIndexProperty.PropDrawValueRect(const ARect: TRect): TRect;
begin
  Result := ARect;
  dec(Result.Right, (Result.Right - Result.Left) - (Result.Bottom - Result.Top + 1));
end;
{$ENDIF}

procedure TImageIndexProperty.Edit;
var
  F: TfrmImageIndex;
  CheckBoxIsDialog: TCheckBox;
begin
  F := TfrmImageIndex.Create(nil);
  try
    UpdateInfo;
    if GetVisualValue <> '' then
      F.ImageIndex := GetIndexImage(GetVisualValue)
    else
      F.ImageIndex := -1;
    F.ImageListHandle := fHandle;
    F.Caption := GetName;
    CheckBoxIsDialog := TCheckBox.Create(F);
    CheckBoxIsDialog.Checked := IsDialog;
    CheckBoxIsDialog.Parent := F;
    CheckBoxIsDialog.Caption := NameItemIsDialog;
    CheckBoxIsDialog.Left := F.IndexEdit.Left + F.IndexEdit.Width + 2 * F.Margin;
    CheckBoxIsDialog.Width := 132;
    CheckBoxIsDialog.Top := F.ClientHeight - CheckBoxIsDialog.Height - F.Margin -
      ((F.IndexEdit.Height - CheckBoxIsDialog.Height) div (2)) - 1;
    CheckBoxIsDialog.Parent := F;
    CheckBoxIsDialog.Anchors := [akLeft, akBottom];
    if F.ShowModal = mrOk then
    begin
      SetOrdValue(F.ImageIndex);
      if IsDialog <> CheckBoxIsDialog.Checked then
      begin
        IsDialog := CheckBoxIsDialog.Checked;
        SaveIsDialog;
      end;
    end;
  finally
    FreeAndNil(F);
  end;
end;

initialization

IsDialog := false;
LastIsDialogLoad := 0;

end.
