unit RegisterIdeMenu;
(*
 * Module:           
 *             
 * Version:    002.00 (29.08.2008)
 * Author:      
 *
 * Copyright:   2008 .
 *)
interface
{$I Directives.inc}
uses
  Windows, SysUtils, Classes, ToolsAPI, Menus, ActnList, Graphics, CepUtils,
  {$IFDEF VER130}Consts, {$ELSE}GraphUtil, {$ENDIF}GraphUtilEx;

const
 {$IFDEF DELPHI_2005_UP}
   DelayValue = 500;
 {$ELSE}
   DelayValue = 200;
 {$ENDIF}
ResourceString
  StrNoParent = '   "%s"  ';
  StrErrorParams = '    "%s"   ' +
                    '';
  StrErrorRegister = '  "%s"  ';
  StrELen = ' "%s"    %d ';
type
  TAWideChar = array [1..128] of WideChar;
  TMenuItemRec = packed Record
    ParentMenuName: TAWideChar;
    BeforeMenuName: TAWideChar;
    MenuName: TAWideChar;
    Category: TAWideChar;
    PrefixActionName: TAWideChar;
    Caption: TAWideChar;
    ResName: TAWideChar;
    Checked: boolean;
    ShortCut: TShortCut;
    Handler: TNotifyEvent;
    Updater: TNotifyEvent;
    ToolBarName: TAWideChar;
    ToolBarCaption: TAWideChar;
  end;

function MenuItemRec(ParentMenuName,
                     BeforeMenuName,
                     MenuName,
                     Category,
                     PrefixActionName,
                     Caption: string;
                     ResName: string;
                     Checked: boolean;
                     Key: TShortCut;
                     ToolBarName: string = '';
                     ToolBarCaption: string = '';
                     Handler: TNotifyEvent = nil;
                     Updater: TNotifyEvent = nil): TMenuItemRec; overload;

function MenuItemRec(ParentMenuName,
                     BeforeMenuName,
                     MenuName,
                     Category,
                     PrefixActionName,
                     Caption: string;
                     ResName: string;
                     Checked: boolean;
                     Key: Char;
                     Shift: TShiftState;
                     ToolBarName: string = '';
                     ToolBarCaption: string = '';
                     Handler: TNotifyEvent = nil;
                     Updater: TNotifyEvent = nil): TMenuItemRec; overload;


function RegisterMenuItem(MenuItemRec: TMenuItemRec): Integer;
function UnregisterMenuItem(MenuName: String): Boolean;
function RegisterImage(ResourceName: string; Ident: string): integer;
function IdeMainMenu: TMenu;
function IDEActionList: TCustomActionList;
function IntFindMenuByName(Item: TComponent; Name: TComponentName): TMenuItem;
procedure ShowSubItemNames(Item: TMenuItem);

var ShowMessages: boolean;

implementation

uses
  InternalTimer, CSplashWindow, Forms,
  {$IFDEF DELPHI_2006_UP}ComCtrls, Controls, ActnMan,{$ENDIF}
  TextUtlCep;

type
  TMenuItemRecEx = record
    Rec: TMenuItemRec;
    RegisteredFlag: boolean;
    Action: TAction;
    MenuItem: TMenuItem;
    ImageIndex: integer;
    ToolBar: {$IFDEF DELPHI_2006_UP}TToolBar;{$ELSE}TObject;{$ENDIF}
  end;

  TRegisteredMenuList = class (TComponent)
  private
    fRegisteredItems: array of TMenuItemRecEx;
    fTimer: TInternalTimer;
    fResourceBitmapList: TStringList;
    fModuleList: TStringList;
    fToolBarNotifierIndex: integer;
    procedure RemoveMenuItems;
    procedure AddMenuItems(Sender: TObject);
    function AddMenuItem(var MenuItemRec: TMenuItemRecEx): TMenuItem;
    function FindMenuByName(Item: TComponent; Name: TComponentName): TMenuItem;
    function DeleteRegisteredItem(I: integer): boolean;
    function CreateBitmap(ResourceName: string): TBitmap;
    function AddButton(var MenuItemRec: TMenuItemRecEx): boolean;
  protected
    procedure UpdateResourceList;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    procedure AddAction(Owner: TComponent; var MenuItemRec: TMenuItemRecEx);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function RegisterMenuItem(MenuItemRec: TMenuItemRec): Integer;
    function UnregisterMenuItem(MenuName: String): Boolean;
    function RegisterImage(ResourceName: string; Ident: string): integer;
  end;

var
  vRegisteredMenuList: TRegisteredMenuList;

function AtoStr(const A: TAWideChar): string;
{$IFNDEF DELPHI_2009_UP}
var I, Len, J: integer;
{$ENDIF}
begin
  {$IFDEF DELPHI_2009_UP}
     result := PChar(@A);
  {$ELSE}
     result := '';
     Len := Low(A);
     while (Len <= High(A)) and (A[Len] <> #0) do inc(Len);
     if Len - Low(A) > 0 then
     begin
       SetLength(result, Len - Low(A));
       J := 1;
       for i := Low(A) to Len do
       begin
         WideCharToMultiByte(1251, Word(A[I]), result[J]);
         inc(J);
       end;
     end;
  {$ENDIF}
end;

function MenuItemRec(ParentMenuName,
                     BeforeMenuName,
                     MenuName,
                     Category,
                     PrefixActionName,
                     Caption: string;
                     ResName: string;
                     Checked: boolean;
                     Key: TShortCut;
                     ToolBarName: string = '';
                     ToolBarCaption: string = '';
                     Handler: TNotifyEvent = nil;
                     Updater: TNotifyEvent = nil): TMenuItemRec;
  function StrToA(const S: string): TAWideChar;
  var TmpS: WideString;
  begin
    FillChar(Result, SizeOf(Result), 0);
    if Length(S) > Length(result) then
      raise EConvertError.CreateFMT(StrELen, [S, Length(result)]);
    TmpS := S;
    Move(TmpS[1], result, Length(TmpS) * SizeOf(TmpS[1]));
  end;
begin
  FillChar(Result, SizeOf(Result), 0);
  Result.ParentMenuName := StrToA(ParentMenuName);
  Result.BeforeMenuName := StrToA(BeforeMenuName);
  Result.MenuName := StrToA(MenuName);
  Result.Category := StrToA(Category);
  Result.PrefixActionName := StrToA(PrefixActionName);
  Result.Caption := StrToA(Caption);
  Result.ResName := StrToA(ResName);
  Result.Checked := Checked;
  Result.ShortCut := Key;
  Result.Handler := Handler;
  Result.Updater := Updater;
  Result.ToolBarName := StrToA(ToolBarName);
  Result.ToolBarCaption := StrToA(ToolBarCaption);
end;

function MenuItemRec(ParentMenuName,
                     BeforeMenuName,
                     MenuName,
                     Category,
                     PrefixActionName,
                     Caption: string;
                     ResName: string;
                     Checked: boolean;
                     Key: Char;
                     Shift: TShiftState;
                     ToolBarName: string = '';
                     ToolBarCaption: string = '';
                     Handler: TNotifyEvent = nil;
                     Updater: TNotifyEvent = nil): TMenuItemRec;
var AKey: TShortCut;
begin
  if Key <> #0 then
    AKey := ShortCut(Ord(Key), Shift)
  else
    AKey := 0;
  result := MenuItemRec(ParentMenuName,
                        BeforeMenuName,
                        MenuName,
                        Category,
                        PrefixActionName,
                        Caption,
                        ResName,
                        Checked,
                        AKey,
                        ToolBarName,
                        ToolBarCaption,
                        Handler,
                        Updater);
end;


function RegisteredMenuList: TRegisteredMenuList;
begin
  if vRegisteredMenuList = nil then
    vRegisteredMenuList := TRegisteredMenuList.Create(nil);
  result := vRegisteredMenuList;
end;

function RegisterMenuItem(MenuItemRec: TMenuItemRec): Integer;
begin
  result := -1;
  if RegisteredMenuList <> nil then
    result := RegisteredMenuList.RegisterMenuItem(MenuItemRec);
end;

function UnregisterMenuItem(MenuName: String): Boolean;
begin
  result := False;
  if vRegisteredMenuList <> nil then
    result := vRegisteredMenuList.UnregisterMenuItem(MenuName);
end;

function RegisterImage(ResourceName: string; Ident: string): integer;
begin
  result := -1;
  if RegisteredMenuList <> nil then
    result := RegisteredMenuList.RegisterImage(ResourceName, Ident);
end;

function IdeMainMenu: TMenu;
var
  hRes: HResult;
  MenuServices: INTAServices40;
begin
  Result := nil;
  hRes := BorlandIDEServices.QueryInterface(INTAServices40, MenuServices);
  if hRes <> S_OK then exit;
  if MenuServices = nil then exit;
  Result := MenuServices.MainMenu;
end;

function IDEActionList: TCustomActionList;
var
  NTAServices: INTAServices;
begin
  NTAServices := BorlandIDEServices as INTAServices;
  Result := NTAServices.ActionList;
end;

{ TRegisteredMenuList }

procedure TRegisteredMenuList.AddMenuItems(Sender: TObject);
var I: integer;
begin
  for I := 0 to Length(fRegisteredItems) - 1 do
  begin
    if (fRegisteredItems[I].RegisteredFlag) then
      Continue;
    try
      if AddMenuItem(fRegisteredItems[I]) <> nil then
      begin
        fRegisteredItems[I].RegisteredFlag := True;
        //AddButton(fRegisteredItems[I]);
      end;
    except
      Continue;
    end;
  end;
end;

constructor TRegisteredMenuList.Create(AOwner: TComponent);
begin
  inherited;
  fToolBarNotifierIndex := -1;
end;

procedure TRegisteredMenuList.RemoveMenuItems;
begin
  while (Length(fRegisteredItems) > 0) and DeleteRegisteredItem(Length(fRegisteredItems) - 1) do;
end;

destructor TRegisteredMenuList.Destroy;
  procedure UnregisterNotyfier;
  {$IFDEF DELPHI_2006_UP}
  var NTAServices: INTAServices90;
  {$ENDIF}
  begin
    {$IFDEF DELPHI_2006_UP}
    NTAServices := BorlandIDEServices as INTAServices90;
    NTAServices.UnregisterToolbarNotifier(fToolBarNotifierIndex);
    {$ENDIF}
    fToolBarNotifierIndex := -1;
  end;
begin
  FreeAndNil(fTimer);
  RemoveMenuItems;
  FreeAndNil(fResourceBitmapList);
  FreeAndNil(fModuleList);
  if fToolBarNotifierIndex >= 0 then UnregisterNotyfier;
  inherited;
end;

function IntFindMenuByName(Item: TComponent; Name: TComponentName): TMenuItem;
var
  I: Integer;
begin
  result := nil;
  if Item = nil then Exit;
  Name := Trim(UpperCase(Name));
  if Name = '' then Exit;
  if Item is TMenu then
  begin
    for I := 0 to TMenu(Item).Items.Count - 1 do
    if Trim(UpperCase(TMenu(Item).Items[I].Name)) = Name then
    begin
      result := TMenu(Item).Items[I];
      Exit;
    end;
    for I := 0 to TMenu(Item).Items.Count - 1 do
    if (TMenu(Item).Items[I].Count > 0) then
    begin
      result := IntFindMenuByName(TMenu(Item).Items[I], Name);
      if result <> nil then Exit;
    end;
  end
  else
  if Item is TMenuItem then
  begin
    for I := 0 to TMenuItem(Item).Count - 1 do
    if Trim(UpperCase(TMenuItem(Item).Items[I].Name)) = Name then
    begin
      result := TMenuItem(Item).Items[I];
      Exit;
    end;
    for I := 0 to TMenuItem(Item).Count - 1 do
    if (TMenuItem(Item).Items[I].Count > 0) then
    begin
      result := IntFindMenuByName(TMenuItem(Item).Items[I], Name);
      if result <> nil then Exit;
    end;
  end;
end;

function TRegisteredMenuList.FindMenuByName(Item: TComponent; Name: TComponentName): TMenuItem;
begin
  result := IntFindMenuByName(Item, Name);
end;

procedure TRegisteredMenuList.Notification(AComponent: TComponent; Operation: TOperation);
var I: integer;
begin
  inherited;
  if Operation = opRemove then
  begin
    for I := Length(fRegisteredItems) - 1 downto 0 do
      if fRegisteredItems[I].Action = AComponent then
      begin
        fRegisteredItems[I].Action := nil;
        UnregisterMenuItem(AtoStr(fRegisteredItems[I].Rec.MenuName));
      end;
  end;
end;

procedure ShowSubItemNames(Item: TMenuItem);
var S: string;
    I: integer;
begin
  if Item <> nil then
  begin
    S := '';
    for I := 0 to Item.Count - 1 do
    begin
      if S <> '' then S := S + #13#10;
      S := S + Format('%d  ''%s'' '#9' (%s)', [I, Item.Items[I].name, Item.Items[i].Caption]);
    end;
    MessageBox(0, PChar(S), PChar(Item.Name), MB_ICONINFORMATION);
  end;
end;

function TRegisteredMenuList.CreateBitmap(ResourceName: string): TBitmap;
var TmpBitmap: TBitmap;
    NTAServices: INTAServices;
    X, Y, W, H: integer;
  function LoadFromResource(ResourceName: string): boolean;
  var i, LastI: integer;
      Name, ResName: string;
      Size, LastSize: word;
    procedure NameToSize(FullName: string; var Name: string; var Size: Word);
    var i: integer;
    begin
      i := Length(FullName);
      Name := FullName;
      Size := 0;
      while (i > 0) and
            (i > Length(FullName) - 4) and
            (Ord(FullName[i]) >= Ord('0')) and
            (Ord(FullName[i]) <= Ord('9')) do dec(i);
      if (i > 0) and (FullName[i] = '_') and (i < Length(FullName)) then
      begin
        Name := Copy(FullName, 1, i - 1);
        Size := StrToInt(Copy(FullName, i + 1, Length(FullName)));
      end;
    end;
  begin
    result := False;
    NameToSize(ResourceName, Name, Size);
    LastSize := 0;
    LastI := -1;
    for i := fResourceBitmapList.Count - 1 downto 0 do
    begin
      NameToSize(fResourceBitmapList[i], ResName, Size);
      if ResName = Name then
      begin
        if Size = W then
        begin
          LastI := I;
          Break;
        end;
        if (Size = 0) and (LastI = -1) then
        begin
          LastI := I;
        end;
        if (Size > W) and (Size < LastSize) and (Size > W) then
        begin
          LastI := I;
          LastSize := Size;
        end;
        if (Size < W) and ((Size > LastSize) or (LastSize > W) or (LastSize = 0)) then
        begin
          LastI := I;
          LastSize := Size;
        end;
      end;
    end;
    if LastI >= 0 then
    begin
      try
        TmpBitmap.LoadFromResourceName(Cardinal(fResourceBitmapList.Objects[LastI]),
                                       fResourceBitmapList[LastI]);
        Result := True;
      except
        on E: Exception do
        begin
          if ShowMessages then
            MessageBox(0, PChar(E.Message), nil, MB_ICONEXCLAMATION or MB_TOPMOST);
          Result := False;
        end;
      end;
    end;
  end;
begin
  result := nil;
  TmpBitmap := nil;
  ResourceName := Trim(AnsiUpperCase(ResourceName));
  if ResourceName = '' then Exit;
  TmpBitmap := TBitmap.Create;
  try
    UpdateResourceList;
    NTAServices := BorlandIDEServices as INTAServices;
    W := NTAServices.ImageList.Width;
    H := NTAServices.ImageList.Height;
    if LoadFromResource(ResourceName) then
    begin
      TmpBitmap.Transparent := False;
      Result := TBitmap.Create;
      try
        Result.PixelFormat := pf32Bit;
        Result.Transparent := True;
        Result.TransparentMode := tmFixed;
        Result.TransparentColor := TmpBitmap.Canvas.Pixels[0, TmpBitmap.Height - 1];
        Result.Width := W;
        Result.Height := H;
        Result.Canvas.Brush.Color := Result.TransparentColor;
        Result.Canvas.FillRect(Rect(0, 0, Result.Width, Result.Height));
        X := (Result.Width - TmpBitmap.Width) div (2);
        Y := (Result.Height - TmpBitmap.Height) div (2);
        Result.Canvas.Draw(X, Y, TmpBitmap);
      except
        FreeAndNil(Result);
      end;
    end;
  finally
    FreeAndNil(TmpBitmap);
  end;
end;

type
  TCustomActionListEx = class (TCustomActionList);

procedure TRegisteredMenuList.AddAction(Owner: TComponent; var MenuItemRec: TMenuItemRecEx);
var ActionList: TCustomActionListEx;
    ActName: string;
    I: integer;
    NewAction: TAction;
  procedure UpdateProps;
  var ResName, Cap, Hint: string;
      I: integer;
  begin
    NewAction.Name := ActName;
    NewAction.Category := StripHotKey(AtoStr(MenuItemRec.Rec.Category));
    Cap := AtoStr(MenuItemRec.Rec.Caption);
    Hint := Cap;
    I := Pos('|', Cap);
    if I > 0 then
    begin
      Cap := Copy(Cap, 1, I - 1);
      Hint := Copy(Hint, I + 1, Length(Hint));
    end;
    NewAction.Caption := Cap;
    NewAction.Hint := Hint;
    NewAction.OnExecute := MenuItemRec.Rec.Handler;
    NewAction.OnUpdate := MenuItemRec.Rec.Updater;
    NewAction.Checked := MenuItemRec.Rec.Checked;
    NewAction.ActionList := IDEactionList;
    NewAction.Visible := True;
    if MenuItemRec.Rec.ShortCut <> 0 then
      NewAction.ShortCut := MenuItemRec.Rec.ShortCut;
    ResName := AtoStr(MenuItemRec.Rec.ResName);
    if MenuItemRec.ImageIndex <> 0 then
      NewAction.ImageIndex := MenuItemRec.ImageIndex
    else
    begin
      if ResName <> '' then
      begin
        MenuItemRec.ImageIndex := RegisterImage(ResName, ActName);
        NewAction.ImageIndex := MenuItemRec.ImageIndex;
      end;
    end;
  end;
begin
  NewAction := nil;
  ActionList := TCustomActionListEx(IDEActionList);
  if ActionList = nil then Exit;
  if AtoStr(MenuItemRec.Rec.MenuName) <> '' then
    ActName := AtoStr(MenuItemRec.Rec.PrefixActionName) + AtoStr(MenuItemRec.Rec.MenuName)
  else
    ActName := AtoStr(MenuItemRec.Rec.PrefixActionName) + AtoStr(MenuItemRec.Rec.Category);
  for I := 0 to ActionList.ActionCount - 1 do
    if (ActionList[I] is TAction) and (AnsiUpperCase(ActName) = AnsiUpperCase(ActionList[I].Name)) then
    begin
      NewAction := TAction(ActionList[I]);
      Break;
    end;
  if NewAction = nil then
  begin
    NewAction := TAction.Create(Owner);
    try
      NewAction.FreeNotification(self);
      UpdateProps;
    except
      FreeAndNil(NewAction);
      MenuItemRec.Action := nil;
      raise;
    end;
  end
  else
  begin
    UpdateProps;
  end;
  MenuItemRec.Action := NewAction;
end;

function TRegisteredMenuList.AddMenuItem(var MenuItemRec: TMenuItemRecEx): TMenuItem;
var NewItem, ParentItem, BeforeItem: TMenuItem;
    MainMenu: TMenu;
begin
  result := nil;
  if (AtoStr(MenuItemRec.Rec.Caption) = '') or
     ((@MenuItemRec.Rec.Handler = nil) and (AtoStr(MenuItemRec.Rec.Caption) <> '-')) or
     (AtoStr(MenuItemRec.Rec.PrefixActionName) = '') then Exit;

  ParentItem := nil;
  NewItem := nil;
  MainMenu := IdeMainMenu;
  if AtoStr(MenuItemRec.Rec.ParentMenuName) <> '' then
  begin
    ParentItem := FindMenuByName(MainMenu, AtoStr(MenuItemRec.Rec.ParentMenuName));
    //ShowSubItemNames(ParentItem);
    if ParentItem = nil then
    begin
      if ShowMessages then
        MessageBox(0, PChar(Format(StrNoParent, [AtoStr(MenuItemRec.Rec.ParentMenuName)])), nil, MB_ICONEXCLAMATION or MB_TOPMOST);
      Exit;
    end;
  end;
  try
    if ParentItem = nil then
    begin
      NewItem := TMenuItem.Create(MainMenu);
      BeforeItem := FindMenuByName(MainMenu, AtoStr(MenuItemRec.Rec.BeforeMenuName));
    end
    else
    begin
      NewItem := TMenuItem.Create(ParentItem);
      BeforeItem := FindMenuByName(ParentItem, AtoStr(MenuItemRec.Rec.BeforeMenuName));
    end;
    if (BeforeItem <> nil) and (BeforeItem.Parent <> ParentItem) then
      BeforeItem := nil;

    NewItem.Name := AtoStr(MenuItemRec.Rec.MenuName);
    if AtoStr(MenuItemRec.Rec.Caption) = '-' then
    begin
      NewItem.Caption := '-';
      NewItem.OnClick := MenuItemRec.Rec.Handler;
    end
    else
    begin
      AddAction(IDEActionList, MenuItemRec);
      NewItem.Action := MenuItemRec.Action;
      if ParentItem = nil then
      begin
        if BeforeItem <> nil then
          MainMenu.Items.Insert(MainMenu.Items.IndexOf(BeforeItem), NewItem)
        else
          MainMenu.Items.Add(NewItem);
      end
      else
      begin
        if BeforeItem <> nil then
          ParentItem.Insert(ParentItem.IndexOf(BeforeItem), NewItem)
        else
          ParentItem.Add(NewItem);
      end;
      MenuItemRec.MenuItem := NewItem;
      if NewItem.Action <> nil then
      begin
        if NewItem.Action is TAction then
        begin
          TAction(NewItem.Action).Enabled := True;
          NewItem.Action.Update;
          NewItem.Enabled := TAction(NewItem.Action).Enabled;
        end;
        {$IFDEF Debug}
         // if TAction(NewItem.Action).Enabled then
         //   MessageBox(0, PChar(NewItem.Action.Name + ' enabled'), 'Test', MB_ICONINFORMATION)
         // else
         //   MessageBox(0, PChar(NewItem.Action.Name + ' disabled'), 'Test', MB_ICONINFORMATION);
        {$ENDIF}
      end;
      result := NewItem;
    end;
  Except
    FreeAndNil(MenuItemRec.Action);
    FreeAndNil(NewItem);
    raise;
  end;
end;

{$IFDEF DELPHI_2006_UP}
type

TToolBarNotifier = class(TNotifierObject, INTACustomizeToolbarNotifier)
public
  procedure CreateButton(AOwner: TComponent; var Button: TControl; Action: TBasicAction);
  procedure FilterAction(Action: TBasicAction; ViewingAllCommands: Boolean; var DisplayName: string; var Display: Boolean; var Handled: Boolean);
  procedure FilterCategory(var Category: string; var Display: Boolean; var Handled: Boolean);
  procedure ResetToolbar(var Toolbar: TWinControl);
  procedure ShowToolbar(Toolbar: TWinControl; Show: Boolean);
  procedure ToolbarModified(Toolbar: TWinControl);
  procedure Destroyed;
end;

var ToolBarNotifier: TToolBarNotifier;

{ TIdeNotifier }

procedure TToolBarNotifier.CreateButton(AOwner: TComponent; var Button: TControl; Action: TBasicAction);
begin

end;

procedure TToolBarNotifier.Destroyed;
begin
  //   - 
end;

procedure TToolBarNotifier.FilterCategory(var Category: string; var Display, Handled: Boolean);
begin

end;

procedure TToolBarNotifier.ResetToolbar(var Toolbar: TWinControl);
begin

end;

procedure TToolBarNotifier.ShowToolbar(Toolbar: TWinControl; Show: Boolean);
begin

end;

procedure TToolBarNotifier.ToolbarModified(Toolbar: TWinControl);
begin

end;

procedure TToolBarNotifier.FilterAction(Action: TBasicAction; ViewingAllCommands: Boolean; var DisplayName: string;
  var Display, Handled: Boolean);
begin

end;
{$ENDIF}


function TRegisteredMenuList.AddButton(var MenuItemRec: TMenuItemRecEx): boolean;
{$IFDEF DELPHI_2006_UP}
var ToolBarName, ToolBarCaption, btnName: string;
    hRes: HResult;
    NTAServices: INTAServices90;
    ToolBar: TToolBar;
    Button: TToolButton;
    Control: TControl;
    I: integer;
    IsDivider: boolean;
{$ENDIF}
begin
  result := False;
  {$IFDEF DELPHI_2006_UP}
  ToolBarName := AtoStr(MenuItemRec.Rec.ToolBarName);
  if (MenuItemRec.Action <> nil) and (ToolBarName <> '') then
  begin
    hRes := BorlandIDEServices.QueryInterface(INTAServices90, NTAServices);
    if hRes <> S_OK then exit;
    if NTAServices = nil then exit;
    ToolBarCaption := AtoStr(MenuItemRec.Rec.ToolBarCaption);
    ToolBar := NTAServices.ToolBar[ToolBarName];
    if ToolBar = nil then
    begin
      //  
      //Owners:  UsefulLotion(TDockToolBar) AppBuilder(TAppBuilder) (TApplication)
      //Parents: UsefulLotion(TDockToolBar) ControlBar1(TControlBar) AppBuilder(TAppBuilder)
      {AppBuilder := Application.FindComponent('AppBuilder');
      if AppBuilder is TWinControl then
      begin
        ControlBar1 := TWinControl(AppBuilder).FindChildControl('ControlBar1');
        if ControlBar1 is TWinControl then
        begin
          NTAServices.ReadToolbar(AppBuilder, TWinControl(ControlBar1),
                                  ToolBarName,
                                  TWinControl(ToolBar));
        end;
      end;  }
      if ToolBar = nil then
      begin
        ToolBar := NTAServices.NewToolbar(ToolBarName, ToolBarCaption);
        if ToolBar <> nil then
        begin
          //ToolBar.Visible := True;
          ToolBar.CustomizeKeyName := ToolBarName;
          ToolBar.CustomizeValueName := 'tb' + ToolBarName;
        end;
      end;
      if ToolBar <> nil then
        ToolBar.Customizable := True;
    end
    else
    if ToolBarCaption <> '' then
      ToolBar.Caption := ToolBarCaption;
    if ToolBar <> nil then
    begin
      if ToolBarNotifier = nil then
      begin
        ToolBarNotifier := TToolBarNotifier.Create;
        fToolBarNotifierIndex := NTAServices.RegisterToolbarNotifier(ToolBarNotifier as INTACustomizeToolbarNotifier);
      end;
      MenuItemRec.ToolBar := ToolBar;
      btnName := 'btn' + AtoStr(MenuItemRec.Rec.MenuName);
      Button := nil;
      for I := 0 to ToolBar.ButtonCount - 1 do
      if UpperCase(ToolBar.Buttons[I].Name) = UpperCase(btnName) then
      begin
        Button := ToolBar.Buttons[I];
        Break;
      end;
      if (Button = nil) and (ToolBar.FindChildControl(btnName) <> nil) then Exit;
      if Button = nil then
      begin
        if MenuItemRec.Action.ImageIndex = -1 then
          MenuItemRec.Action.Update;
        IsDivider := AtoStr(MenuItemRec.Rec.Caption) = '-';
        Control := NTAServices.AddToolButton(ToolBarName, btnName, MenuItemRec.Action, IsDivider);
        if Control is TToolButton then
          Button := TToolButton(Control);
      end
      else
      begin
        Button.Action := MenuItemRec.Action;
      end;
      Result := Button <> nil;
    end;
  end;
  {$ENDIF}
end;

function TRegisteredMenuList.RegisterMenuItem(MenuItemRec: TMenuItemRec): Integer;
  function CheckRegistered(MenuName: string): boolean;
  var
    I: Integer;
  begin
    result := false;
    MenuName := Trim(UpperCase(MenuName));
    if MenuName = '' then Exit;
    for I := 0 to Length(fRegisteredItems) - 1 do
    begin
      if UpperCase(AtoStr(fRegisteredItems[I].Rec.MenuName)) = UpperCase(MenuName) then
      begin
        result := True;
        Exit;
      end;
    end;
    result := FindMenuByName(IdeMainMenu, MenuName) <> nil;
  end;
begin
  result := -1;
  if (AtoStr(MenuItemRec.Caption) = '') or
    ((@MenuItemRec.Handler = nil) and (AtoStr(MenuItemRec.Caption) <> '-')) or
     (AtoStr(MenuItemRec.PrefixActionName) = '') then
  begin
    if ShowMessages then
      MessageBox(0, PChar(Format(StrErrorParams, [AtoStr(MenuItemRec.Caption)])), nil, MB_ICONEXCLAMATION or MB_TOPMOST);
    Exit;
  end;

  if (MenuItemRec.MenuName <> '') and (CheckRegistered(AtoStr(MenuItemRec.MenuName))) then
  begin
    if ShowMessages then
      MessageBox(0, PChar(Format(StrErrorRegister, [AtoStr(MenuItemRec.Caption)])), nil, MB_ICONEXCLAMATION or MB_TOPMOST);
    Exit;
  end;

  if (AtoStr(MenuItemRec.ParentMenuName) <> '') and (not CheckRegistered(AtoStr(MenuItemRec.ParentMenuName))) then
  begin
    if ShowMessages then
      MessageBox(0, PChar(Format(StrNoParent, [AtoStr(MenuItemRec.ParentMenuName)])), nil, MB_ICONEXCLAMATION or MB_TOPMOST);
    Exit;
  end;

  Result := Length(fRegisteredItems);
  SetLength(fRegisteredItems, Result + 1);
  FillChar(fRegisteredItems[Result], SizeOf(fRegisteredItems[Result]), 0);
  fRegisteredItems[Result].Rec := MenuItemRec;
  AddAction(IDEActionList, fRegisteredItems[Result]);
  AddButton(fRegisteredItems[Result]);
  //      
  if fTimer = nil then
    fTimer := TInternalTimer.Create(nil);
  fTimer.ExecuteDelay(DelayValue, AddMenuItems);
end;

function TRegisteredMenuList.DeleteRegisteredItem(I: integer): boolean;
var Action: TAction;
    MenuItem: TMenuItem;
begin
  result := (I >= 0) and (I < Length(fRegisteredItems));
  if Result then
  begin
    Action := fRegisteredItems[I].Action;
    if Action <> nil then
    begin
      Action.OnExecute := nil;
      Action.OnUpdate := nil;
      Action.OnHint := nil;
    end;
    fRegisteredItems[I].Action := nil;
    //Action.Free;   
    MenuItem := fRegisteredItems[I].MenuItem;
    fRegisteredItems[I].MenuItem := nil;
    MenuItem.Free;
    fRegisteredItems[I].RegisteredFlag := False;
    Move(fRegisteredItems[I + 1],
         fRegisteredItems[I],
         SizeOf(fRegisteredItems[I]) * (Length(fRegisteredItems) - 1 - I));
    SetLength(fRegisteredItems, Length(fRegisteredItems) - 1);
  end;
end;

function TRegisteredMenuList.UnregisterMenuItem(MenuName: String): boolean;
var DelItemList: TStringList;
    Found: boolean;
    ParentName: string;
    I, J, Count: integer;
  function IndexMenuItemRec(MenuName: string): integer;
  var I: integer;
  begin
    Result := -1;
    for I := 0 to Length(fRegisteredItems) - 1 do
    if MenuName = AnsiUpperCase(AtoStr(fRegisteredItems[i].Rec.MenuName)) then
    begin
      result := I;
      Break;
    end;
  end;
begin
  Result := False;
  DelItemList := TStringList.Create;
  try
    MenuName := AnsiUpperCase(MenuName);
    DelItemList.Add(MenuName);
    repeat
      Found := false;
      Count := DelItemList.Count;
      for I := 0 to Count - 1 do
      if Integer(DelItemList.Objects[I]) = 0 then
      begin
        //     ,   
        for J := 0 to Length(fRegisteredItems) - 1 do
        begin
          ParentName := AnsiUpperCase(AtoStr(fRegisteredItems[J].Rec.ParentMenuName));
          if DelItemList[I] = ParentName then
          begin
            DelItemList.Add(AnsiUpperCase(AtoStr(fRegisteredItems[J].Rec.MenuName)));
            Found := True;
          end;
        end;
        DelItemList.Objects[I] := Pointer(1);
      end;
    until Found = False;
    for I := DelItemList.Count - 1 downto 0 do
    begin
      //     
      DeleteRegisteredItem(IndexMenuItemRec(DelItemList[I]));
    end;
  finally
    FreeAndNil(DelItemList);
  end;
end;

procedure TRegisteredMenuList.UpdateResourceList;
var ModuleList: TStringList;
    I, J: integer;
    Found: boolean;
begin
  if fResourceBitmapList = nil then
  begin
    fResourceBitmapList := TStringList.Create;
  end;
  if fModuleList = nil then
  begin
    fModuleList := TStringList.Create;
    fModuleList.Sorted := True;
    fModuleList.Duplicates := dupIgnore;
  end;
  ModuleList := TStringList.Create;
  try
    EnumDLLs(ModuleList);
    for I := ModuleList.Count - 1 downto 0 do
    begin
      Found := fModuleList.Find(ModuleList[I], J);
      if not Found then
      begin
        fModuleList.AddObject(ModuleList[I], ModuleList.Objects[I]);
        EnumResources(Cardinal(ModuleList.Objects[I]), fResourceBitmapList, RT_BITMAP, '');
      end;
    end;
  finally
    ModuleList.Free;
  end;
end;

function TRegisteredMenuList.RegisterImage(ResourceName, Ident: string): integer;
var Bitmap: TBitmap;
    NTAServices: INTAServices;
begin
  result := -1;
  UpdateResourceList;
  Bitmap := CreateBitmap(ResourceName);
  try
    if Bitmap <> nil then
    begin
      NTAServices := BorlandIDEServices as INTAServices;
      {$IFDEF DELPHI_7_UP}
      Result := NTAServices.AddMasked(Bitmap, Bitmap.TransparentColor, Ident);
      {$ELSE}
      Result := NTAServices.AddMasked(Bitmap, Bitmap.TransparentColor);
      {$ENDIF}
    end;
  finally
    FreeAndNil(Bitmap);
  end;
end;

initialization
  ShowMessages := True;
  vRegisteredMenuList := nil;
  {$IFDEF DELPI_2006_UP}
  ToolBarNotifier := nil;
  {$ENDIF}

finalization
  FreeAndNil(vRegisteredMenuList);
  {$IFDEF DELPI_2006_UP}
  ToolBarNotifier := nil;
  {$ENDIF}
end.






