unit TextUtlCep;

(*
 * Module:         
 * Version:    1.1.0.0 (24.09.2008)
 * Author:      
 * Comments:     ()  
 *             http://www.delphikingdom.com/asp/viewitem.asp?catalogid=722
 *             ,     
 *
 * Copyright:   2008 Cep Po
 *
 * Version:    1.1.0.0 (24.09.2008)
 * Realized:     Delphi 2009
 *                
 *)
interface
{$I Directives.inc}
uses Windows, Messages, Consts, SysUtils, Classes, Graphics, ComCtrls, RichEdit, ComStrs
{$IFDEF DELPHI_6_UP}, RTLConsts{$ELSE}, GraphUtilEx{$ENDIF}
{$IFDEF DELPHI_2009_UP}, Character{$ENDIF};

type
  TTokenRec = record
    StartWord, EndWord: WideString;
  end;
{$Z4}
  TCharsCollection = set of AnsiChar;

const PageSize = 10240 div 2;
  CommentToken1: TTokenRec = (StartWord: '{'; EndWord: '}');
  CommentToken2: TTokenRec = (StartWord: '(*'; EndWord: '*)');
  CommentToken3: TTokenRec = (StartWord: '//'; EndWord: #13#10);
  CharBackSpaces: TCharsCollection = [' ', #0, #9, #10, #13, #160];

  RTFformat: WideString = 'Rich Text Format';
  UM_PAINTLINE = WM_USER + 13;


resourcestring
  ResInvalidValue = '   ''%s''';

type
  TClipboardFormats = array[0..2] of integer;
  TRichEdit = class(ComCtrls.TRichEdit)
  private
    fRightLine: integer;
    fFirstLeftLine: integer;
    fLeftLine: integer;
    procedure SaveToStream(Stream: TStream; TextType: Integer);
    procedure LoadFromStream(Stream: TStream; TextType: Integer);
    function GetTextW: WideString;
    procedure SetTextW(const Value: WideString);
    procedure WMERASEBKGND(var Message: TWMERASEBKGND); message WM_ERASEBKGND;
    procedure UMPAINTLINE(var Message: TMessage); message UM_PAINTLINE;
    procedure SetRightLine(const Value: integer);
    procedure DrawLines(DC: HDC);
    procedure SetFirstLeftLine(const Value: integer);
    procedure SetLeftLine(const Value: integer);
  public
    function IsEnablePaste(var Formats: TClipboardFormats): integer;
    procedure CopyAll;
    procedure PasteAll;
    property TextW: WideString read GetTextW write SetTextW;
    property RightLine: integer read fRightLine write SetRightLine;
    property LeftLine: integer read fLeftLine write SetLeftLine;
    property FirstLeftLine: integer read fFirstLeftLine write SetFirstLeftLine;
  end;



  TTokenPositions = record
    Offset, StartPos, EndPos: LongInt;
  end;

  TFindedText = class
  private
    fWideText: PWideChar;
    fLen: LongInt;
    fTokenPositions: array of TTokenPositions;
    fMaxBufferSize: LongInt;
    fRightMargin: Word;
    fCodePage: LongInt;
    function InternalFind(const Offset: Integer; var StartPos, EndPos: LongInt;
      const Next: boolean): boolean;
  protected
    procedure SetWideText(const Value: PWideChar); virtual;
    //       
    //Offset   ( )   ,     
    //   0.      
    //  
    //       , 
    //   ,  -1.
    function DoFind(var Offset: LongInt): LongInt; virtual; abstract;
    function ScanWhileBS(var X: integer): boolean;
  public
    constructor Create(WideText: PWideChar); virtual;
    procedure Clear; virtual;
    property Len: LongInt read fLen;
    property WideText: PWideChar read fWideText write SetWideText;
    property RightMargin: Word read fRightMargin write fRightMargin default 80;
    property CodePage: LongInt read fCodePage write fCodePage default 1251;
    property MaxBufferSize: LongInt read fMaxBufferSize;
    //     .
    //     ,   
    //c     
    //Offset -  ( )   ,     
    //   0
    function FindNext(var Offset: LongInt): LongInt; virtual;
    // ,       
    // Offset  ,   True,  StartPos, EndPos 
    //     ,    
    //False,  StartPos  EndPos   .
    function InToken(const Offset: LongInt; var StartPos, EndPos: LongInt): boolean; virtual;
    //    
    //Offset - .    ,     
    //StartPos, EndPos -      
    //S -       
    //   True   .  False.    , 
    //tartPos, EndPos, S   .
    function GetStripText(const Offset: LongInt; var StartPos, EndPos: LongInt; var S: WideString): boolean; virtual; abstract;
    //    
    function GetTokenText(const StartPos, EndPos: LongInt; const S: WideString): WideString; virtual; abstract;
  end;

  TQuotedText = class(TFindedText)
  private
    fOwner: TQuotedText;
    fQuote: WideChar;
    fCommentTokens: array of TTokenRec;
    fEnableFastFind: boolean;
    fPrefix: WideChar;
    fHexPrefix: WideChar;
    procedure SetQuote(const Value: WideChar);
    function GetCommentTokens(Index: integer): TTokenRec;
    procedure SetCommentTokens(Index: integer; const Value: TTokenRec);
    function GetCountCommentTokens: integer;
    procedure UpdateEnableFastFind;
    procedure SetPrefix(const Value: WideChar);
    procedure SetHexPrefix(const Value: WideChar);
  protected
    function ExtractChar(var Offset: Integer; var W: Word): boolean; virtual;
    function CheckQuotedText(var Offset: Integer): boolean; virtual;
    function FindFirstToken(var X1, X2: Integer): Integer; virtual;
    function DoFind(var Offset: LongInt): LongInt; override;
    property Prefix: WideChar read fPrefix write SetPrefix;
    property HexPrefix: WideChar read fHexPrefix write SetHexPrefix;
    function IsSimbol(C: WideChar): boolean;
    function IsLetterOrDigit(C: WideChar): boolean;
    procedure NextCommentOrLiteral(var StartPos, EndPos: Integer);
    function NextSimbol(var StartPos: Integer; First: boolean = true): boolean;
    function NextWord(var StartPos, EndPos: Integer; First: boolean = true): boolean;
    function NextWordFromList(const WordList: TStringList; var StartPos, EndPos: Integer; First: boolean = true): Integer;
    function NextBrackets(const Open, Close: WideChar; var StartPos, EndPos: Integer; First: boolean = true): boolean;
  public
    constructor Create(WideText: PWideChar); overload; override;
    destructor Destroy; override;
    procedure Assign(Source: TQuotedText); virtual;
    constructor Create(AOwner: TQuotedText); reintroduce; overload; virtual;
    property Quote: WideChar read fQuote write SetQuote;
    property CommentTokens[Index: integer]: TTokenRec read GetCommentTokens write SetCommentTokens;
    procedure AddCommentTokens(Value: TTokenRec);
    property CountCommentTokens: integer read GetCountCommentTokens;
    function GetStripText(const Offset: LongInt; var StartPos, EndPos: LongInt; var S: WideString): boolean; override;
    function GetTokenText(const StartPos, EndPos: LongInt; const S: WideString): WideString; override;
  end;

  TCommentedText = class(TQuotedText)
  protected
    function DoFind(var Offset: LongInt): LongInt; override;
  public
    function GetStripText(const Offset: LongInt; var StartPos, EndPos: LongInt; var S: WideString): boolean; override;
    function GetTokenText(const StartPos, EndPos: LongInt; const S: WideString): WideString; override;
  end;

  TLiteralText = class(TQuotedText)
  private
    fConcateChar: WideString;
    fDefaultHex: boolean;
    fCommentList: TObject;
    fColumn: integer;
    procedure SetConcateChar(const Value: WideString);
    function ScanWhileBackSpace(var X: Integer;
      AcceptConcate: boolean): boolean;
  protected
    function DoFind(var Offset: LongInt): LongInt; override;
  public
    constructor Create(WideText: PWideChar); override;
    destructor Destroy; override;
    property Prefix default '#';
    property HexPrefix default '$';
    property DefaultHex: boolean read fDefaultHex write fDefaultHex;
    property ConcateChar: WideString read fConcateChar write SetConcateChar;
    function GetStripText(const Offset: LongInt; var StartPos, EndPos: LongInt; var S: WideString): boolean; override;
    function GetTokenText(const StartPos, EndPos: LongInt; const S: WideString): WideString; override;
    procedure UpdateComments(StartPos, EndPos: LongInt);
    property Column: integer read fColumn write fColumn;
    function MakeText(StripText: WideString;
      const Hex: boolean = false;
      const FirstLeft: Longint = 0;
      Left: Longint = 0;
      Right: Longint = 80;
      const CodePage: LongInt = 1251;
      const NoComments: boolean = true): WideString;
    function GetMargins(const StartPos, EndPos: Integer;
      var FirstLeft, Left, Right: integer): boolean;
    procedure TruncRight(var Result: WideString);
  end;

  TFunctionText = class(TQuotedText)
  private
    fList: TStringList;
    fObjName: WideString;
    fObjClass: WideString;
    fObjReturn: WideString;
    fClassEndPos: integer;
    fParams: TStringList;
    fParamTypes: TStringList;
    fObjKind: Integer;
  protected
    function DoFind(var Offset: LongInt): LongInt; override;
  public
    constructor Create(WideText: PWideChar); override;
    destructor Destroy; override;

    property ObjClass: WideString read fObjClass;
    property ObjName: WideString read fObjName;
    property ObjReturn: WideString read fObjReturn;
    property ObjKind: Integer read fObjKind;
    property ClassEndPos: integer read fClassEndPos;
    property Params: TStringList read fParams;
    property List: TStringList read fList;
    property ParamTypes: TStringList read fParamTypes;

  end;


function UpCaseW(Ch: WideChar): WideChar;
function HexToInt(s: AnsiString): Longword;

function FindFirstWord(const WText: PWideChar; var Ofs: LongInt; Quote: WideChar; Token0, Token1, Token2, Token3: DWORD; CharCount: LongInt): integer; overload;
function FindFirstWord(const WText: PWideChar; var Ofs: LongInt; Quote: WideChar; Token0, Token1, Token2, Token3: DWORD): integer; overload;

function FindEOS(const WText: PWideChar; var Ofs: LongInt; CharCount: LongInt): integer; overload;
function FindEOS(const WText: PWideChar; var Ofs: LongInt): integer; overload;

function FindQuotedText(Ch: WideChar; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt; overload;
function FindQuotedText(Ch: WideChar; const S: PWideChar; var Offset: LongInt): LongInt; overload;

function PosStrEx(const SubStr: WideString; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt; overload;
function PosStrEx(const SubStr: WideString; const S: PWideChar; var Offset: LongInt): LongInt; overload;
function PosStrIns(const SubStr: WideString; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;

function PosWord(const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;


function WideCharToMultiByte(CodePage: LongInt; WC: Word; var C: AnsiChar): boolean;
//        .
//   ,    
function StringToList(const S: WideString; const Separator: WideChar; List: TStringList):
  integer;
function ListToString(List: TStringList; const Separator: WideChar): WideString;

procedure PosToColRow(const WText: PWideChar; const Ofs: LongInt; var Col, Row: integer; Len: LongInt = -1);

implementation
const
  ReadError = $0001;
  WriteError = $0002;
  NoError = $0000;


function UpCaseW(Ch: WideChar): WideChar;
asm
        CMP     AX,'a'
        JB      @@exit
        CMP     AX,'z'
        JA      @@exit
        SUB     AX,'a' - 'A'
@@exit:
end;

function HexToInt(s: AnsiString): Longword;
var
  b: Byte;
  c: AnsiChar;
begin
  Result := 0;
  for b := 1 to Length(s) do
  begin
    Result := Result * 16;
    c := s[b];
    case UpCase(c) of
      '0'..'9': Inc(Result, Ord(c) - Ord('0'));
      'A'..'F': Inc(Result, Ord(c) - Ord('A') + 10);
    else
      raise EConvertError.Create('No Hex-Number');
    end;
  end;
end;



{ TRichEdit }

function StreamSave(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  Buffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  try
    pcb := 0;
    if StreamInfo^.Converter <> nil then
      {$IFDEF DELPHI_2009_UP}
    begin
      SetLength(Buffer, cb);
      Move(pbBuff^, Buffer[0], cb);
      pcb := StreamInfo^.Converter.ConvertWriteStream(StreamInfo^.Stream, Buffer, cb);
    end;
    {$ELSE}
      pcb := StreamInfo^.Converter.ConvertWriteStream(StreamInfo^.Stream, PChar(pbBuff), cb);
    {$ENDIF}
  except
    Result := WriteError;
  end;
end;

//     ComCtrls

function AdjustLineBreaks(Dest, Source: PAnsiChar): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ESI,EDX
        MOV     EDX,EAX
        CLD
@@1:    LODSB
@@2:    OR      AL,AL
        JE      @@4
        CMP     AL,0AH
        JE      @@3
        STOSB
        CMP     AL,0DH
        JNE     @@1
        MOV     AL,0AH
        STOSB
        LODSB
        CMP     AL,0AH
        JE      @@1
        JMP     @@2
@@3:    MOV     EAX,0A0DH
        STOSW
        JMP     @@1
@@4:    STOSB
        LEA     EAX,[EDI-1]
        SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;

//  ,      PWideChar

function AdjustLineBreaksW(Dest, Source: PWideChar): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ESI,EDX
        MOV     EDX,EAX
        CLD
@@1:    LODSW
@@2:    OR      AX,AX
        JE      @@4
        CMP     AX,000AH
        JE      @@3
        STOSW
        CMP     AX,000DH
        JNE     @@1
        MOV     AX,000AH
        STOSW
        LODSW
        CMP     AX,000AH
        JE      @@1
        JMP     @@2
@@3:    MOV     EAX,000A000DH
        STOSD
        JMP     @@1
@@4:    STOSW
        LEA     EAX,[EDI-2]
        SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;


//     

function StreamLoad(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  Buffer, pBuff: PAnsiChar;
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  TmpBuffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  {$IFDEF DELPHI_2009_UP}
  Buffer := AnsiStrAlloc(cb + 1);
  {$ELSE}
  Buffer := StrAlloc(cb + 1);
  {$ENDIF}
  try
    cb := cb div 2;
    pcb := 0;
    pBuff := Buffer + cb;
    try
      if StreamInfo^.Converter <> nil then
        {$IFDEF DELPHI_2009_UP}
      begin
        SetLength(TmpBuffer, cb);
        FillChar(pBuff^, cb, 0);
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, TmpBuffer, cb);
        Move(TmpBuffer[0], pBuff^, pcb);
      end;
      {$ELSE}
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, pBuff, cb);
      {$ENDIF}
      if pcb > 0 then
      begin
        pBuff[pcb] := #0;
        if pBuff[pcb - 1] = #13 then
          pBuff[pcb - 1] := #0;
        pcb := AdjustLineBreaks(Buffer, pBuff);
        Move(Buffer^, pbBuff^, pcb);
      end;
    except
      Result := ReadError;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

//  ,   WideString

function StreamLoadW(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  Buffer, pBuff: PAnsiChar;
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  TmpBuffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  {$IFDEF DELPHI_2009_UP}
  Buffer := AnsiStrAlloc((cb + 1) * 2);
  {$ELSE}
  Buffer := StrAlloc((cb + 1) * 2);
  {$ENDIF}
  try
    cb := cb;
    pcb := 0;
    pBuff := Buffer + cb;
    try
      if StreamInfo^.Converter <> nil then
        {$IFDEF DELPHI_2009_UP}
      begin
        SetLength(TmpBuffer, cb);
        FillChar(pBuff^, cb, 0);
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, TmpBuffer, cb);
        Move(TmpBuffer[0], pBuff^, pcb);
      end;
      {$ELSE}
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, pBuff, cb);
      {$ENDIF}
      if pcb > 0 then
      begin
        PWideChar(pBuff)[(pcb) div (2)] := #0;
        if PWideChar(pBuff)[(pcb div 2) - 1] = #13 then
          PWideChar(pBuff)[(pcb div 2) - 1] := #0;
        pcb := AdjustLineBreaksW(PWideChar(Buffer), PWideChar(pBuff));
        Move(Buffer^, pbBuff^, pcb);
      end;
    except
      Result := ReadError;
    end;
  finally
    StrDispose(Buffer);
  end;
end;
//    

procedure TRichEdit.LoadFromStream(Stream: TStream;
  TextType: Integer);
var
  EditStream: TEditStream;
  StreamInfo: TRichEditStreamInfo;
  Converter: TConversion;
begin
  Converter := DefaultConverter.Create;
  StreamInfo.Stream := Stream;
  StreamInfo.Converter := Converter;
  try
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      if (TextType and SF_UNICODE) = SF_UNICODE then
        pfnCallBack := @StreamLoadW
      else
        pfnCallBack := @StreamLoad;
      dwError := 0;
    end;
    SendMessage(Handle, EM_STREAMIN, TextType, Longint(@EditStream));
    if EditStream.dwError <> 0 then
      raise EOutOfResources.Create(sRichEditLoadFail);
  finally
    Converter.Free;
  end;
end;

//        

procedure TRichEdit.SaveToStream(Stream: TStream; TextType: Integer);
var
  EditStream: TEditStream;
  StreamInfo: TRichEditStreamInfo;
  Converter: TConversion;
begin
  Converter := DefaultConverter.Create;
  try
    StreamInfo.Stream := Stream;
    StreamInfo.Converter := Converter;
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      pfnCallBack := @StreamSave;
      dwError := 0;
    end;
    SendMessage(Handle, EM_STREAMOUT, TextType, LongInt(@EditStream));
    if EditStream.dwError <> 0 then
      raise EOutOfResources.Create(sRichEditSaveFail);
  finally
    Converter.Free;
  end;
end;

//     

procedure TRichEdit.CopyAll;
var
  uFormat: integer; //  RTF
  Stream: TMemoryStream;
  hMem: THandle;
  P: pointer;
  procedure CopyByFormat(TextType, cbFormat: Longint);
  begin
    if (cbFormat = 0) or (cbFormat = 0) then
      Exit;
    Stream.Clear;
    SaveToStream(Stream, TextType);
    Stream.Position := 0;
    hMem := GlobalAlloc(GMEM_DDESHARE or GMEM_MOVEABLE, Stream.Size + 2);
    try
      P := GlobalLock(hMem);
      FillChar(P^, Stream.Size + 2, 0);
      CopyMemory(P, Stream.Memory, Stream.Size);
    finally
      GlobalUnlock(hMem);
    end;
    SetClipboardData(cbFormat, hMem);
  end;
begin
  uFormat := RegisterClipboardFormatW(PWideChar(RTFformat));
  if OpenClipBoard(Handle) then
  begin
    try
      EmptyClipboard;
      Stream := TMemoryStream.Create;
      try
        CopyByFormat(SF_TEXT or SF_UNICODE, CF_UNICODETEXT);
        CopyByFormat(SF_RTFNOOBJS, uFormat);
        CopyByFormat(SF_TEXT, CF_TEXT);
      finally
        FreeAndNil(Stream);
      end;
    finally
      CloseClipboard;
    end;
  end
  else
    RaiseLastOSError;
end;

//   

procedure TRichEdit.PasteAll;
var
  Formats: TClipboardFormats;
  cbFormat: integer;
  Stream: TMemoryStream;
  hMem: THandle;
  P: pointer;
  Size: DWORD;
  procedure CopyByFormat(TextType, cbFormat: Longint);
  begin
    if (cbFormat = 0) or (cbFormat = 0) then
      Exit;
    Stream.Clear;
    hMem := GetClipboardData(cbFormat);
    if hMem <> 0 then
    begin
      P := GlobalLock(hMem);
      try
        Size := GlobalSize(hMem);
        Stream.WriteBuffer(P^, Size);
        Stream.Position := 0;
      finally
        GlobalUnlock(hMem);
      end;
      LoadFromStream(Stream, TextType);
    end;
  end;
begin
  if OpenClipBoard(Handle) then
  begin
    try
      cbFormat := IsEnablePaste(Formats);
      if cbFormat <> 0 then
      begin
        Stream := TMemoryStream.Create;
        try
          case cbFormat of
            CF_UNICODETEXT: CopyByFormat(SF_TEXT or SF_UNICODE, cbFormat);
            CF_TEXT: CopyByFormat(SF_TEXT, cbFormat);
          else
            CopyByFormat(SF_RTF, cbFormat);
          end;
        finally
          FreeAndNil(Stream);
        end;
      end;
    finally
      CloseClipboard;
    end;
  end
  else
    RaiseLastOSError;
end;

//     
// Formats       
// (UNICODE, RTF, TEXT)

function TRichEdit.IsEnablePaste(var Formats: TClipboardFormats): integer;
var
  F, I: integer;
  Buf: array[0..255] of WideChar;
  wS: WideString;
begin
  result := 0;
  FillChar(Formats, SizeOf(Formats), 0);
  F := EnumClipboardFormats(0);
  while F <> 0 do
  begin
    if F = CF_TEXT then
      Formats[2] := F;
    if F = CF_UNICODETEXT then
      Formats[0] := F;
    if (GetClipboardFormatNameW(F, PWideChar(@Buf), Length(Buf)) > 0) then
    begin
      wS := WideString(PWideChar(@Buf));
      if wS = RTFformat then
        Formats[1] := F;
    end;
    F := EnumClipboardFormats(F);
  end;
  for I := Low(Formats) to High(Formats) do
    if Formats[I] <> 0 then
    begin
      result := Formats[I];
      break;
    end;
end;

function TRichEdit.GetTextW: WideString;
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    SaveToStream(Stream, SF_TEXT or SF_UNICODE);
    result := PWideChar(Stream.Memory);
    SetLength(result, (Stream.Size) div (SizeOf(Result[1])));
    Stream.Clear;
  finally
    FreeAndNil(Stream);
  end;
end;

procedure TRichEdit.SetFirstLeftLine(const Value: integer);
begin
  if fFirstLeftLine <> Value then
  begin
    fFirstLeftLine := Value;
    if HandleAllocated then
      InvalidateRect(Handle, nil, true);
  end;
end;

procedure TRichEdit.SetLeftLine(const Value: integer);
begin
  if fLeftLine <> Value then
  begin
    fLeftLine := Value;
    if HandleAllocated then
      InvalidateRect(Handle, nil, true);
  end;
end;

procedure TRichEdit.SetRightLine(const Value: integer);
begin
  if fRightLine <> Value then
  begin
    fRightLine := Value;
    if HandleAllocated then
      InvalidateRect(Handle, nil, true);
  end;
end;

procedure TRichEdit.SetTextW(const Value: WideString);
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    Stream.WriteBuffer(Value[1], (Length(Value) + 1) * SizeOf(Value[1]));
    Stream.Position := 0;
    LoadFromStream(Stream, SF_TEXT or SF_UNICODE);
    Stream.Clear;
  finally
    FreeAndNil(Stream);
  end;
end;

procedure TRichEdit.UMPAINTLINE(var Message: TMessage);
var
  DC: HDC;
begin
  if HandleAllocated and
    ((fRightLine > 0) or (fFirstLeftLine > 0) or (LeftLine > 0)) then
  begin
    DC := GetWindowDC(Handle);
    try
      DrawLines(DC);
    finally
      ReleaseDC(Handle, DC);
    end;
  end;
end;

procedure TRichEdit.DrawLines(DC: HDC);
var
  R: TRect;
  OldFont: HFont;
  Pen, OldPen: HPen;
  procedure DoDrawLine(X: integer);
  var
    P: tagPoint;
  begin
    if X > 0 then
    begin
      X := (R.Right * X) div 2 + 1;
      MoveToEx(DC, X, 0, @P);
      LineTo(DC, X, Height);
    end;
  end;
begin
  if DC <> 0 then
  begin
    R := Rect(0, 0, 1000, 1000);
    OldFont := SelectObject(DC, Font.Handle);
    try
      DrawText(DC, '00', 2, R, DT_CALCRECT);
      SetBkColor(DC, ColorToRGB(Color));
      Pen := CreatePen(PS_DOT, 1, clGray);
      OldPen := SelectObject(DC, Pen);
      try
        DoDrawLine(fRightLine);
        DoDrawLine(fLeftLine);
        DoDrawLine(fFirstLeftLine);
      finally
        SelectObject(DC, OldPen);
        DeleteObject(Pen);
      end;
    finally
      SelectObject(DC, OldFont);
    end;
  end;
end;

procedure TRichEdit.WMERASEBKGND(var Message: TWMERASEBKGND);
begin
  inherited;
  if HandleAllocated then
    PostMessage(Handle, UM_PAINTLINE, 0, 0);
end;






function FindFirstWord(const WText: PWideChar; var Ofs: LongInt; Quote: WideChar; Token0, Token1, Token2, Token3: DWORD; CharCount: LongInt): integer;
  assembler;
asm
  push   eBx
  push   eCx
  push   eDx
  push   eDi
  push   eSi

  mov    eSi, eAx
  mov    eDi, eAx
  mov    eAx, DS: eDx

  SUB    CharCount, eAx

  shl    eAx, 1
  add    eSi, eAx
  @Cikl:
    Sub   CharCount, 1
    jl    @NoFind
    mov    eAx, Ds: [eSi]
    cmp    Ax, 0
    je     @NoFind
    // Quote
    Cmp    Ax, Cx
    je     @_Q
    // Token0
    @N0:
    mov    eBx, dword ptr [Token0]
    Cmp    eAx, eBx                  //   
    je     @T0                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N1                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T0                       //   
    // Token1
    @N1:
    mov    eBx, dword ptr [Token1]
    Cmp    eAx, eBx                  //   
    je     @T1                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N2                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T1                       //   
    // Token2
    @N2:
    mov    eBx, dword ptr [Token2]
    Cmp    eAx, eBx                  //   
    je     @T2                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N3                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T2                       //   
    // Token3
    @N3:
    mov    eBx, dword ptr [Token3]
    Cmp    eAx, eBx                  //   
    je     @T3                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N4                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T3                       //   
    @N4:
    add    eSi, 2
    jmp    @Cikl
  @NoFind:
  mov    eAx, $FFffFFff
  jmp    @ENDS
  @_Q:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 4
  jmp    @ENDS
  @T0:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 0
  jmp    @ENDS
  @T1:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 1
  jmp    @ENDS
  @T2:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 2
  jmp    @ENDS
  @T3:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 3
  jmp    @ENDS



 @ENDS:
  pop    eSi
  pop    eDi
  pop    eDx
  pop    eCx
  pop    eBx
end;

function FindFirstWord(const WText: PWideChar; var Ofs: LongInt; Quote: WideChar; Token0, Token1, Token2, Token3: DWORD): integer;
  assembler;
asm
  push   eBx
  push   eCx
  push   eDx
  push   eDi
  push   eSi

  mov    eSi, eAx
  mov    eDi, eAx
  mov    eAx, DS: eDx
  shl    eAx, 1
  add    eSi, eAx
  @Cikl:
    mov    eAx, Ds: [eSi]
    cmp    Ax, 0
    je     @NoFind
    // Quote
    Cmp    Ax, Cx
    je     @_Q
    // Token0
    @N0:
    mov    eBx, dword ptr [Token0]
    Cmp    eAx, eBx                  //   
    je     @T0                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N1                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T0                       //   
    // Token1
    @N1:
    mov    eBx, dword ptr [Token1]
    Cmp    eAx, eBx                  //   
    je     @T1                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N2                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T1                       //   
    // Token2
    @N2:
    mov    eBx, dword ptr [Token2]
    Cmp    eAx, eBx                  //   
    je     @T2                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N3                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T2                       //   
    // Token3
    @N3:
    mov    eBx, dword ptr [Token3]
    Cmp    eAx, eBx                  //   
    je     @T3                       //   
    cmp    eBx, $FFff                // ,      0
    jA     @N4                       // c  ,  
    cmp    Ax, Bx                    //    
    je     @T3                       //   
    @N4:
    add    eSi, 2
    jmp    @Cikl
  @NoFind:
  mov    eAx, $FFffFFff
  jmp    @ENDS
  @_Q:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 4
  jmp    @ENDS
  @T0:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 0
  jmp    @ENDS
  @T1:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 1
  jmp    @ENDS
  @T2:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 2
  jmp    @ENDS
  @T3:
  mov    eAx, eSi
  sub    eAx, eDi
  shr    eAx, 1
  mov    eDx, SS: eSp + 8
  mov    DS: eDx, eAx
  mov    eAx, 3
  jmp    @ENDS

 @ENDS:
  pop    eSi
  pop    eDi
  pop    eDx
  pop    eCx
  pop    eBx
end;

function FindEOS(const WText: PWideChar; var Ofs: LongInt; CharCount: LongInt): integer;
  assembler;
asm
  push   eBx
  push   eCx
  push   eDx
  push   eDi
  push   eSi

  mov    eSi, eAx
  mov    eDi, eAx
  mov    eAx, DS: eDx

  SUB    eCx, eAx
  push   eCx

  shl    eAx, 1
  add    eSi, eAx
  @Cikl:
    Sub    DWORD PTR SS: eSp, 1
    jl    @NoFind
    mov    eAx, Ds: [eSi]
    cmp    Ax, 0
    je     @NoFind

    Cmp    eAx, $000A000D
    je     @Finded2
    Cmp    eAx, $000D000A
    je     @Finded2

    Cmp    Ax, $000D
    je     @Finded1
    Cmp    Ax, $000A
    je     @Finded1

    add    eSi, 2
    jmp    @Cikl
  @NoFind:
    mov    eAx, $FFffFFff
    jmp    @ENDS
  @Finded1:
    ADD    eSi, 2
    mov    eAx, eSi
    sub    eAx, eDi
    shr    eAx, 1
    mov    eDx, SS: eSp + 12
    mov    DS: eDx, eAx
    SUB    eAx, 1
  jmp    @ENDS
  @Finded2:
    ADD    eSi, 4
    mov    eAx, eSi
    sub    eAx, eDi
    shr    eAx, 1
    mov    eDx, SS: eSp + 12
    mov    DS: eDx, eAx
    SUB    eAx, 2
  jmp    @ENDS
 @ENDS:

  pop    eCx
  pop    eSi
  pop    eDi
  pop    eDx
  pop    eCx
  pop    eBx
end;


function FindEOS(const WText: PWideChar; var Ofs: LongInt): integer;
  assembler;
asm
  push   eBx
  push   eCx
  push   eDx
  push   eDi
  push   eSi

  mov    eSi, eAx
  mov    eDi, eAx
  mov    eAx, DS: eDx
  shl    eAx, 1
  add    eSi, eAx
  @Cikl:
    mov    eAx, Ds: [eSi]
    cmp    Ax, 0
    je     @NoFind

    Cmp    eAx, $000A000D
    je     @Finded2
    Cmp    eAx, $000D000A
    je     @Finded2

    Cmp    Ax, $000D
    je     @Finded1
    Cmp    Ax, $000A
    je     @Finded1

    add    eSi, 2
    jmp    @Cikl
  @NoFind:
    mov    eAx, $FFffFFff
    jmp    @ENDS
  @Finded1:
    ADD    eSi, 2
    mov    eAx, eSi
    sub    eAx, eDi
    shr    eAx, 1
    mov    eDx, SS: eSp + 8
    mov    DS: eDx, eAx
    SUB    eAx, 1
  jmp    @ENDS
  @Finded2:
    ADD    eSi, 4
    mov    eAx, eSi
    sub    eAx, eDi
    shr    eAx, 1
    mov    eDx, SS: eSp + 8
    mov    DS: eDx, eAx
    SUB    eAx, 2
  jmp    @ENDS
 @ENDS:
  pop    eSi
  pop    eDi
  pop    eDx
  pop    eCx
  pop    eBx
end;

function FindQuotedText(Ch: WideChar; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;
  assembler;
asm
  Push  eBx
  Push  eCx
  Push  eDx
  Push  eSi
  Push  eDi

  Mov   eSi, eDx
  Mov   eDi, eCx
  Mov   eBx, DWORD PTR DS:eDi

  Sub   CharCount, eBx

  Shl   eBx, 1
  Add   eSi, eBx
  Mov   eBx, eAx
  Xor   eCx, eCx

  @CiklOutQuote:
    Sub   CharCount, 1
    jl    @NoFind
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, 0
    je   @NoFind
    CMP  Ax, Bx
    je   @EndCiklOutQuote
    add  eSi, 2
    jmp  @CiklOutQuote
  @EndCiklOutQuote:
  Mov   eCx, eSi
  Sub   eCx, eDx
  add   eSi, 2
  Shr   eCx, 1
  @CiklInQuote:
    Sub   CharCount, 1
    jl    @NoFind
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, 0
    je   @NoFind
    CMP  Ax, Bx
    je   @M1
    add  eSi, 2
    jmp  @CiklInQuote
  @M1:
    add  eSi, 2
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, Bx
    jne  @EndCiklInQuote
    add  eSi, 2
    jmp  @CiklInQuote
  @EndCiklInQuote:
  Mov   eAx, eSi
  Sub   eAx, eDx
  Shr   eAx, 1
  Mov   DWORD PTR DS:eDi, eAx
  Mov   eAx, eCx
  jmp   @ENDS

  @NoFind:
  Mov   eAx, -1
  jmp   @ENDS

  @ENDS:
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eCx
  Pop   eBx
end;

function FindQuotedText(Ch: WideChar; const S: PWideChar; var Offset: LongInt): LongInt;
  assembler;
asm
  Push  eBx
  Push  eCx
  Push  eDx
  Push  eSi
  Push  eDi

  Mov   eSi, eDx
  Mov   eDi, eCx
  Mov   eBx, DWORD PTR DS:eDi

  Shl   eBx, 1
  Add   eSi, eBx
  Mov   eBx, eAx
  Xor   eCx, eCx

  @CiklOutQuote:
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, 0
    je   @NoFind
    CMP  Ax, Bx
    je   @EndCiklOutQuote
    add  eSi, 2
    jmp  @CiklOutQuote
  @EndCiklOutQuote:
  Mov   eCx, eSi
  Sub   eCx, eDx
  add   eSi, 2
  Shr   eCx, 1
  @CiklInQuote:
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, 0
    je   @NoFind
    CMP  Ax, Bx
    je   @M1
    add  eSi, 2
    jmp  @CiklInQuote
  @M1:
    add  eSi, 2
    Mov  Ax, WORD PTR DS:eSi
    CMP  Ax, Bx
    jne  @EndCiklInQuote
    add  eSi, 2
    jmp  @CiklInQuote
  @EndCiklInQuote:
  Mov   eAx, eSi
  Sub   eAx, eDx
  Shr   eAx, 1
  Mov   DWORD PTR DS:eDi, eAx
  Mov   eAx, eCx
  jmp   @ENDS

  @NoFind:
  Mov   eAx, -1
  jmp   @ENDS

  @ENDS:
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eCx
  Pop   eBx
end;

function PosStrEx(const SubStr: WideString; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;
  assembler;
asm
  CMP   eAx, 0
  JE   @Rets
  CMP   eDx, 0
  JE   @Rets

  Push  eBx
  Push  eDx
  Push  eSi
  Push  eDi
  Push  eCx

  Mov   eSi, eDx                 // eSi = S^
  Mov   eDi, eCx                 // eDi = Offset^
  Mov   eBx, DWORD PTR DS:eDi    // eBx = Offset

  Sub   CharCount, eBx

  Shl   eBx, 1
  Add   eSi, eBx                 // eSi = S[Offset]^
  Mov   eCx, DS: [eAx - 4]       // eCx = Length(SubStr)
  Shr   eCx, 1
  Cmp   eCx, 0
  Jle   @NoFind1
  Mov   eDi, eAx                 // eDi = SubStr^
  Push  eCx
  Push  eDi

  @Cikl:
    Mov   eDi, SS: [eSp]
    Mov   eCx, SS: [eSp + 4]
    Mov   Bx, WORD PTR DS:eDi
    @CiklFirst:
       Mov  Ax, DS: eSi
       Add  eSi, 2
       Sub   CharCount, 1
       jl    @NoFind
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @CiklFirst

    @CiklOk:
       Mov  Ax, DS: eSi
       Add  eSi, 2
       Sub   CharCount, 1
       jl    @NoFind
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @Cikl

    @Ok:
       Add  eDi, 2
       Sub  eCx, 1
       JBE  @Found
       Mov   Bx, WORD PTR DS:eDi
    Jmp   @CiklOk

  @Found:
  Pop   eDi
  Pop   eCx
  Mov   eDi, SS: eSp
  Mov   eAx, eSi
  Sub   eAx, SS: [eSp + 12]
  Shr   eAx, 1
  //Inc   eAx
  Mov   Ds: eDi, eAx
  Sub   eAx, eCx
  jmp   @ENDS

  @NoFind:
  Pop   eDi
  Pop   eCx
  @NoFind1:
  Mov   eAx, -1
  jmp   @ENDS
  @Rets:
  Mov   eAx, -1
  Ret

  @ENDS:
  Pop   eCx
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eBx
end;

function PosStrIns(const SubStr: WideString; const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;
  assembler;
asm
  CMP   eAx, 0
  JE   @Rets
  CMP   eDx, 0
  JE   @Rets

  Push  eBx
  Push  eDx
  Push  eSi
  Push  eDi
  Push  eCx

  Mov   eSi, eDx                 // eSi = S^
  Mov   eDi, eCx                 // eDi = Offset^
  Mov   eBx, DWORD PTR DS:eDi    // eBx = Offset

  Sub   CharCount, eBx

  Shl   eBx, 1
  Add   eSi, eBx                 // eSi = S[Offset]^
  Mov   eCx, DS: [eAx - 4]       // eCx = Length(SubStr)
  Shr   eCx, 1
  Cmp   eCx, 0
  Jle   @NoFind1
  Mov   eDi, eAx                 // eDi = SubStr^
  Push  eCx
  Push  eDi

  @Cikl:
    Mov   eDi, SS: [eSp]
    Mov   eCx, SS: [eSp + 4]
    Mov   Bx, WORD PTR DS:eDi
    CMP     BX, 'a'
    JB      @CiklFirst
    CMP     BX,'z'
    JA      @CiklFirst
    SUB     BX,'a' - 'A'
    @CiklFirst:
       Mov  Ax, DS: eSi
       CMP     AX,'a'
       JB      @M1
       CMP     AX,'z'
       JA      @M1
       SUB     AX,'a' - 'A'
       @M1:
       Add  eSi, 2
       Sub   CharCount, 1
       jl    @NoFind
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @CiklFirst

    @CiklOk:
       Mov  Ax, DS: eSi
       CMP     AX,'a'
       JB      @M2
       CMP     AX,'z'
       JA      @M2
       SUB     AX,'a' - 'A'
       @M2:
       Add  eSi, 2
       Sub   CharCount, 1
       jl    @NoFind
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @Cikl

    @Ok:
       Add  eDi, 2
       Sub  eCx, 1
       JBE  @Found
       Mov  Bx, WORD PTR DS:eDi
       CMP     BX,'a'
       JB      @CiklOk
       CMP     BX,'z'
       JA      @CiklOk
       SUB     BX,'a' - 'A'
    Jmp   @CiklOk

  @Found:
  Pop   eDi
  Pop   eCx
  Mov   eDi, SS: eSp
  Mov   eAx, eSi
  Sub   eAx, SS: [eSp + 12]
  Shr   eAx, 1
  //Inc   eAx
  Mov   Ds: eDi, eAx
  Sub   eAx, eCx
  jmp   @ENDS

  @NoFind:
  Pop   eDi
  Pop   eCx
  @NoFind1:
  Mov   eAx, -1
  jmp   @ENDS
  @Rets:
  Mov   eAx, -1
  Ret

  @ENDS:
  Pop   eCx
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eBx
end;

function PosStrEx(const SubStr: WideString; const S: PWideChar; var Offset: LongInt): LongInt;
  assembler;
asm
  CMP   eAx, 0
  JE   @Rets
  CMP   eDx, 0
  JE   @Rets

  Push  eBx
  Push  eDx
  Push  eSi
  Push  eDi
  Push  eCx

  Mov   eSi, eDx                 // eSi = S^
  Mov   eDi, eCx                 // eDi = Offset^
  Mov   eBx, DWORD PTR DS:eDi    // eBx = Offset
  Shl   eBx, 1
  Add   eSi, eBx                 // eSi = S[Offset]^
  Mov   eCx, DS: [eAx - 4]       // eCx = Length(SubStr)
  Shr   eCx, 1
  Cmp   eCx, 0
  Jle   @NoFind1
  Mov   eDi, eAx                 // eDi = SubStr^
  Push  eCx
  Push  eDi

  @Cikl:
    Mov   eDi, SS: [eSp]
    Mov   eCx, SS: [eSp + 4]
    Mov   Bx, WORD PTR DS:eDi
    @CiklFirst:
       Mov  Ax, DS: eSi
       Add  eSi, 2
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @CiklFirst

    @CiklOk:
       Mov  Ax, DS: eSi
       Add  eSi, 2
       CMP  Ax, Bx
       Je   @Ok
       CMP  Ax, 0
       Je   @NoFind
    Jmp   @Cikl

    @Ok:
       Add  eDi, 2
       Sub  eCx, 1
       JBE  @Found
       Mov   Bx, WORD PTR DS:eDi
    Jmp   @CiklOk

  @Found:
  Pop   eDi
  Pop   eCx
  Mov   eDi, SS: eSp
  Mov   eAx, eSi
  Sub   eAx, SS: [eSp + 12]
  Shr   eAx, 1
  //Inc   eAx
  Mov   Ds: eDi, eAx
  Sub   eAx, eCx
  jmp   @ENDS

  @NoFind:
  Pop   eDi
  Pop   eCx
  @NoFind1:
  Mov   eAx, -1
  jmp   @ENDS
  @Rets:
  Mov   eAx, -1
  Ret

  @ENDS:
  Pop   eCx
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eBx
end;

function PosWord(const S: PWideChar; var Offset: LongInt; CharCount: LongInt): LongInt;
  assembler;
asm
  CMP   eAx, 0
  JE   @Rets
  CMP   eDx, 0
  JL   @Rets

  Push  eAx
  Push  eBx
  Push  eDx
  Push  eSi
  Push  eDi
  Push  eCx

  mov   eSi, eAx
  mov   eDi, eDx                 // eDi = Offset^
  mov   eBx, DWORD PTR DS:eDi    // eBx = Offset

  Sub   eCx, eBx

  Shl   eBx, 1
  Add   eSi, eBx                 // eSi = S[Offset]^

  @CiklNoChar:
    Sub   eCx, 1
    jl    @NoFind

    Mov   Ax, WORD PTR DS:eSi
    CMP   Ax, 0                  // #0
    je    @NoFind
    Add   eSi, 2
    CMP   Ax, 36                 // '$'
    je    @Found
    CMP   AX, 95                 // '_'
    je    @Found
    CMP   Ax, 48                 // '0'
    jb    @CiklNoChar
    CMP   Ax, 57                 // '9'
    jbe   @Found
    CMP   Ax, 65                 // 'A'
    jb    @CiklNoChar
    CMP   Ax, 90                 // 'Z'
    jbe   @Found
    CMP   Ax, 97                 // 'a'
    jb    @CiklNoChar
    CMP   Ax, 122                // 'z'

    jbe   @Found
  jmp    @CiklNoChar
  @Found:
  Mov   eAx, eSi
  Sub   eAx, SS: [eSp + 20]
  Shr   eAx, 1
  Dec   eAx
  Mov   eBx, eAx
  @CiklChar:
    Mov   Ax, WORD PTR DS:eSi
    Add   eSi, 2
    CMP   Ax, 0                  // #0

    Sub   eCx, 1
    jl    @OkFind
    je    @OkFind
    CMP   Ax, 36                 // '$'
    je    @CiklChar
    CMP   AX, 95                 // '_'
    je    @CiklChar
    CMP   Ax, 48                 // '0'
    jb    @OkFind
    CMP   Ax, 57                 // '9'
    jbe   @CiklChar
    CMP   Ax, 65                 // 'A'
    jb    @OkFind
    CMP   Ax, 90                 // 'Z'
    jbe   @CiklChar
    CMP   Ax, 97                 // 'a'
    jb    @OkFind
    CMP   Ax, 122                // 'z'
    jbe   @CiklChar
  @OkFind:
  Mov   eAx, eSi
  Sub   eAx, SS: [eSp + 20]
  Shr   eAx, 1
  Dec   eAx
  Mov   DWORD PTR Ds:eDi, eAx
  Mov   eAx, eBx
  jmp   @ENDS

  @NoFind:
  Mov   eAx, -1
  jmp   @ENDS
  @Rets:
  Mov   eAx, -1
  Ret

  @ENDS:
  Pop   eCx
  Pop   eDi
  Pop   eSi
  Pop   eDx
  Pop   eBx
  Add   eSp, 4
end;

const CodePages: array[1250..1258, 128..255] of Word =
  (($20AC, $0000, $201A, $0000, $201E, $2026, $2020, $2021, $0000, $2030, $0160, $2039, $015A, $0164, $017D, $0179, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $0000, $2122, $0161, $203A, $015B, $0165, $017E, $017A, $00A0, $02C7, $02D8, $0141, $00A4, $0104, $00A6, $00A7, $00A8, $00A9, $015E, $00AB, $00AC, $00AD, $00AE, $017B, $00B0, $00B1, $02DB, $0142, $00B4, $00B5, $00B6, $00B7, $00B8, $0105, $015F, $00BB, $013D, $02DD, $013E, $017C, $0154, $00C1, $00C2, $0102, $00C4, $0139, $0106, $00C7, $010C, $00C9, $0118, $00CB, $011A, $00CD, $00CE, $010E, $0110, $0143, $0147, $00D3, $00D4, $0150, $00D6, $00D7, $0158, $016E, $00DA, $0170, $00DC, $00DD, $0162, $00DF, $0155, $00E1, $00E2, $0103, $00E4, $013A, $0107, $00E7, $010D, $00E9, $0119, $00EB, $011B, $00ED, $00EE, $010F, $0111, $0144, $0148, $00F3, $00F4, $0151, $00F6, $00F7, $0159, $016F, $00FA, $0171, $00FC, $00FD, $0163, $02D9),
    ($0402, $0403, $201A, $0453, $201E, $2026, $2020, $2021, $20AC, $2030, $0409, $2039, $040A, $040C, $040B, $040F, $0452, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $0000, $2122, $0459, $203A, $045A, $045C, $045B, $045F, $00A0, $040E, $045E, $0408, $00A4, $0490, $00A6, $00A7, $0401, $00A9, $0404, $00AB, $00AC, $00AD, $00AE, $0407, $00B0, $00B1, $0406, $0456, $0491, $00B5, $00B6, $00B7, $0451, $2116, $0454, $00BB, $0458, $0405, $0455, $0457, $0410, $0411, $0412, $0413, $0414, $0415, $0416, $0417, $0418, $0419, $041A, $041B, $041C, $041D, $041E, $041F, $0420, $0421, $0422, $0423, $0424, $0425, $0426, $0427, $0428, $0429, $042A, $042B, $042C, $042D, $042E, $042F, $0430, $0431, $0432, $0433, $0434, $0435, $0436, $0437, $0438, $0439, $043A, $043B, $043C, $043D, $043E, $043F, $0440, $0441, $0442, $0443, $0444, $0445, $0446, $0447, $0448, $0449, $044A, $044B, $044C, $044D, $044E, $044F),
    ($20AC, $0000, $201A, $0192, $201E, $2026, $2020, $2021, $02C6, $2030, $0160, $2039, $0152, $0000, $017D, $0000, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $02DC, $2122, $0161, $203A, $0153, $0000, $017E, $0178, $00A0, $00A1, $00A2, $00A3, $00A4, $00A5, $00A6, $00A7, $00A8, $00A9, $00AA, $00AB, $00AC, $00AD, $00AE, $00AF, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00B8, $00B9, $00BA, $00BB, $00BC, $00BD, $00BE, $00BF, $00C0, $00C1, $00C2, $00C3, $00C4, $00C5, $00C6, $00C7, $00C8, $00C9, $00CA, $00CB, $00CC, $00CD, $00CE, $00CF, $00D0, $00D1, $00D2, $00D3, $00D4, $00D5, $00D6, $00D7, $00D8, $00D9, $00DA, $00DB, $00DC, $00DD, $00DE, $00DF, $00E0, $00E1, $00E2, $00E3, $00E4, $00E5, $00E6, $00E7, $00E8, $00E9, $00EA, $00EB, $00EC, $00ED, $00EE, $00EF, $00F0, $00F1, $00F2, $00F3, $00F4, $00F5, $00F6, $00F7, $00F8, $00F9, $00FA, $00FB, $00FC, $00FD, $00FE, $00FF),
    ($20AC, $0000, $201A, $0192, $201E, $2026, $2020, $2021, $0000, $2030, $0000, $2039, $0000, $0000, $0000, $0000, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $0000, $2122, $0000, $203A, $0000, $0000, $0000, $0000, $00A0, $0385, $0386, $00A3, $00A4, $00A5, $00A6, $00A7, $00A8, $00A9, $0000, $00AB, $00AC, $00AD, $00AE, $2015, $00B0, $00B1, $00B2, $00B3, $0384, $00B5, $00B6, $00B7, $0388, $0389, $038A, $00BB, $038C, $00BD, $038E, $038F, $0390, $0391, $0392, $0393, $0394, $0395, $0396, $0397, $0398, $0399, $039A, $039B, $039C, $039D, $039E, $039F, $03A0, $03A1, $0000, $03A3, $03A4, $03A5, $03A6, $03A7, $03A8, $03A9, $03AA, $03AB, $03AC, $03AD, $03AE, $03AF, $03B0, $03B1, $03B2, $03B3, $03B4, $03B5, $03B6, $03B7, $03B8, $03B9, $03BA, $03BB, $03BC, $03BD, $03BE, $03BF, $03C0, $03C1, $03C2, $03C3, $03C4, $03C5, $03C6, $03C7, $03C8, $03C9, $03CA, $03CB, $03CC, $03CD, $03CE, $0000),
    ($20AC, $0000, $201A, $0192, $201E, $2026, $2020, $2021, $02C6, $2030, $0160, $2039, $0152, $0000, $0000, $0000, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $02DC, $2122, $0161, $203A, $0153, $0000, $0000, $0178, $00A0, $00A1, $00A2, $00A3, $00A4, $00A5, $00A6, $00A7, $00A8, $00A9, $00AA, $00AB, $00AC, $00AD, $00AE, $00AF, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00B8, $00B9, $00BA, $00BB, $00BC, $00BD, $00BE, $00BF, $00C0, $00C1, $00C2, $00C3, $00C4, $00C5, $00C6, $00C7, $00C8, $00C9, $00CA, $00CB, $00CC, $00CD, $00CE, $00CF, $011E, $00D1, $00D2, $00D3, $00D4, $00D5, $00D6, $00D7, $00D8, $00D9, $00DA, $00DB, $00DC, $0130, $015E, $00DF, $00E0, $00E1, $00E2, $00E3, $00E4, $00E5, $00E6, $00E7, $00E8, $00E9, $00EA, $00EB, $00EC, $00ED, $00EE, $00EF, $011F, $00F1, $00F2, $00F3, $00F4, $00F5, $00F6, $00F7, $00F8, $00F9, $00FA, $00FB, $00FC, $0131, $015F, $00FF),
    ($20AC, $0000, $201A, $0192, $201E, $2026, $2020, $2021, $02C6, $2030, $0000, $2039, $0000, $0000, $0000, $0000, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $02DC, $2122, $0000, $203A, $0000, $0000, $0000, $0000, $00A0, $00A1, $00A2, $00A3, $20AA, $00A5, $00A6, $00A7, $00A8, $00A9, $00D7, $00AB, $00AC, $00AD, $00AE, $00AF, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00B8, $00B9, $00F7, $00BB, $00BC, $00BD, $00BE, $00BF, $05B0, $05B1, $05B2, $05B3, $05B4, $05B5, $05B6, $05B7, $05B8, $05B9, $0000, $05BB, $05BC, $05BD, $05BE, $05BF, $05C0, $05C1, $05C2, $05C3, $05F0, $05F1, $05F2, $05F3, $05F4, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $05D0, $05D1, $05D2, $05D3, $05D4, $05D5, $05D6, $05D7, $05D8, $05D9, $05DA, $05DB, $05DC, $05DD, $05DE, $05DF, $05E0, $05E1, $05E2, $05E3, $05E4, $05E5, $05E6, $05E7, $05E8, $05E9, $05EA, $0000, $0000, $200E, $200F, $0000),
    ($20AC, $067E, $201A, $0192, $201E, $2026, $2020, $2021, $02C6, $2030, $0679, $2039, $0152, $0686, $0698, $0688, $06AF, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $06A9, $2122, $0691, $203A, $0153, $200C, $200D, $06BA, $00A0, $060C, $00A2, $00A3, $00A4, $00A5, $00A6, $00A7, $00A8, $00A9, $06BE, $00AB, $00AC, $00AD, $00AE, $00AF, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00B8, $00B9, $061B, $00BB, $00BC, $00BD, $00BE, $061F, $06C1, $0621, $0622, $0623, $0624, $0625, $0626, $0627, $0628, $0629, $062A, $062B, $062C, $062D, $062E, $062F, $0630, $0631, $0632, $0633, $0634, $0635, $0636, $00D7, $0637, $0638, $0639, $063A, $0640, $0641, $0642, $0643, $00E0, $0644, $00E2, $0645, $0646, $0647, $0648, $00E7, $00E8, $00E9, $00EA, $00EB, $0649, $064A, $00EE, $00EF, $064B, $064C, $064D, $064E, $00F4, $064F, $0650, $00F7, $0651, $00F9, $0652, $00FB, $00FC, $200E, $200F, $06D2),
    ($20AC, $0000, $201A, $0000, $201E, $2026, $2020, $2021, $0000, $2030, $0000, $2039, $0000, $00A8, $02C7, $00B8, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $0000, $2122, $0000, $203A, $0000, $00AF, $02DB, $0000, $00A0, $0000, $00A2, $00A3, $00A4, $0000, $00A6, $00A7, $00D8, $00A9, $0156, $00AB, $00AC, $00AD, $00AE, $00C6, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00F8, $00B9, $0157, $00BB, $00BC, $00BD, $00BE, $00E6, $0104, $012E, $0100, $0106, $00C4, $00C5, $0118, $0112, $010C, $00C9, $0179, $0116, $0122, $0136, $012A, $013B, $0160, $0143, $0145, $00D3, $014C, $00D5, $00D6, $00D7, $0172, $0141, $015A, $016A, $00DC, $017B, $017D, $00DF, $0105, $012F, $0101, $0107, $00E4, $00E5, $0119, $0113, $010D, $00E9, $017A, $0117, $0123, $0137, $012B, $013C, $0161, $0144, $0146, $00F3, $014D, $00F5, $00F6, $00F7, $0173, $0142, $015B, $016B, $00FC, $017C, $017E, $02D9),
    ($20AC, $0000, $201A, $0192, $201E, $2026, $2020, $2021, $02C6, $2030, $0000, $2039, $0152, $0000, $0000, $0000, $0000, $2018, $2019, $201C, $201D, $2022, $2013, $2014, $02DC, $2122, $0000, $203A, $0153, $0000, $0000, $0178, $00A0, $00A1, $00A2, $00A3, $00A4, $00A5, $00A6, $00A7, $00A8, $00A9, $00AA, $00AB, $00AC, $00AD, $00AE, $00AF, $00B0, $00B1, $00B2, $00B3, $00B4, $00B5, $00B6, $00B7, $00B8, $00B9, $00BA, $00BB, $00BC, $00BD, $00BE, $00BF, $00C0, $00C1, $00C2, $0102, $00C4, $00C5, $00C6, $00C7, $00C8, $00C9, $00CA, $00CB, $0300, $00CD, $00CE, $00CF, $0110, $00D1, $0309, $00D3, $00D4, $01A0, $00D6, $00D7, $00D8, $00D9, $00DA, $00DB, $00DC, $01AF, $0303, $00DF, $00E0, $00E1, $00E2, $0103, $00E4, $00E5, $00E6, $00E7, $00E8, $00E9, $00EA, $00EB, $0301, $00ED, $00EE, $00EF, $0111, $00F1, $0323, $00F3, $00F4, $01A1, $00F6, $00F7, $00F8, $00F9, $00FA, $00FB, $00FC, $01B0, $20AB, $00FF));


function WideCharToMultiByte(CodePage: LongInt; WC: Word; var C: AnsiChar): boolean;
var I: integer;
begin
  result := false;
  if CodePage = 0 then
  begin
    result := (WC < 128);
    if Result then Byte(C) := Word(WC);
  end
  else
  begin
    if (WC < 128) then
    begin
      result := True;
      Byte(C) := Word(WC)
    end
    else
    begin
      for I := 128 to 255 do
        if CodePages[CodePage, I] = WC then
        begin
          result := true;
          Byte(C) := I;
          break;
        end;
    end;
  end;
end;

function PosChar(Ch: WideChar; const S: WideString; DupIgnore: Boolean; Offset: Integer
  = 1): Integer; {$IFDEF DELPHI_2006_UP}inline; {$ENDIF}
var
  X: Integer;
begin
  Result := 0;
  repeat
    X := PosStrEx(Ch, PWideChar(S), Offset);
    inc(X);
    if X > 0 then
    begin
      if DupIgnore and (X < Length(S)) and (S[X + 1] = Ch) then
        Offset := X + 2
      else
      begin
        Result := X;
        Exit;
      end;
    end;
  until X = 0;
end;

function StringToList(const S: WideString; const Separator: WideChar; List: TStringList):
  integer;
var
  X1, X2: integer;
  CurrS: WideString;
  I: integer;
begin
  result := 0;
  List.Clear;
  X1 := 1;
  X2 := 1;
  while X2 > 0 do
  begin
    X2 := PosChar(Separator, S, True, X2);
    if X2 > 0 then
      CurrS := Copy(S, X1, X2 - X1)
    else
      CurrS := Copy(S, X1, Length(S) - X1 + 1);
    I := 1;
    while I <= Length(CurrS) do
    begin
      if CurrS[I] = Separator then Delete(CurrS, I, 1);
      inc(I);
    end;
    List.Add(CurrS);
    inc(Result);
    if X2 > 0 then
    begin
      inc(X2);
      X1 := X2;
    end;
  end;
end;

function ListToString(List: TStringList; const Separator: WideChar): WideString;
var
  I, J: integer;
  CurrS: WideString;
begin
  result := '';
  for I := 0 to List.Count - 1 do
  begin
    if result <> '' then result := result + Separator;
    CurrS := List[I];
    for J := List.Count downto 1 do
      if CurrS[J] = Separator then Insert(Separator, CurrS, J + 1);
    result := result + CurrS;
  end;
end;

procedure PosToColRow(const WText: PWideChar; const Ofs: LongInt; var Col, Row: integer; Len: LongInt = -1);
var X1, X2, X3: LongInt;
begin
  X2 := 0;
  X1 := 0;
  X3 := 0;
  Col := -1;
  Row := -1;
  while (X1 >= 0) and (X2 < Ofs) do
  begin
    if X1 >= 0 then X3 := X2;
    if Len = -1 then
      X1 := FindEOS(WText, X2)
    else
      X1 := FindEOS(WText, X2, Len);
    inc(Row);
  end;
  if Row >= 0 then Col := Ofs - X3;
end;

function IsCharBS(const Ch: WideChar): boolean; {$IFDEF DELPHI_2006_UP}inline; {$ENDIF}
begin
  result := (Word(Ch) <= 255) and (AnsiChar(Ch) in CharBackSpaces);
end;


//------------------------------------------------------------------------------
// ()  
//MaxMatching -    ( 3-4)
//strInputMatching -  
//strInputStandart - -

//    
// if IndistinctMatching(4, " ", "   - ") > 40 then ...
type
  TRetCount = packed record
    lngSubRows: Word;
    lngCountLike: Word;
  end;

//------------------------------------------------------------------------------

function Matching(StrInputA: WideString;
  StrInputB: WideString;
  lngLen: Integer): TRetCount;
var
  TempRet: TRetCount;
  PosStrB: Integer;
  PosStrA: Integer;
  StrA: WideString;
  StrB: WideString;
  StrTempA: WideString;
  StrTempB: WideString;
begin
  StrA := string(StrInputA);
  StrB := string(StrInputB);

  for PosStrA := 1 to Length(strA) - lngLen + 1 do
  begin
    StrTempA := System.Copy(strA, PosStrA, lngLen);

    for PosStrB := 1 to Length(strB) - lngLen + 1 do
    begin
      StrTempB := System.Copy(strB, PosStrB, lngLen);
      if SysUtils.AnsiCompareText(StrTempA, StrTempB) = 0 then
      begin
        Inc(TempRet.lngCountLike);
        break;
      end;
    end;

    Inc(TempRet.lngSubRows);
  end; // PosStrA

  Matching.lngCountLike := TempRet.lngCountLike;
  Matching.lngSubRows := TempRet.lngSubRows;
end; { function }

//------------------------------------------------------------------------------

function IndistinctMatching(MaxMatching: Integer;
  strInputMatching: WideString;
  strInputStandart: WideString): Integer;
var
  gret: TRetCount;
  tret: TRetCount;
  lngCurLen: Integer; //  
begin
    //   - ,  
  if (MaxMatching = 0) or (Length(strInputMatching) = 0) or
    (Length(strInputStandart) = 0) then
  begin
    IndistinctMatching := 0;
    exit;
  end;

  gret.lngCountLike := 0;
  gret.lngSubRows := 0;
    //      
  for lngCurLen := 1 to MaxMatching do
  begin
        //  A   B
    tret := Matching(strInputMatching, strInputStandart, lngCurLen);
    gret.lngCountLike := gret.lngCountLike + tret.lngCountLike;
    gret.lngSubRows := gret.lngSubRows + tret.lngSubRows;
        //  B   A
    tret := Matching(strInputStandart, strInputMatching, lngCurLen);
    gret.lngCountLike := gret.lngCountLike + tret.lngCountLike;
    gret.lngSubRows := gret.lngSubRows + tret.lngSubRows;
  end;

  if gret.lngSubRows = 0 then
  begin
    IndistinctMatching := 0;
    exit;
  end;

  IndistinctMatching := Trunc((gret.lngCountLike / gret.lngSubRows) * 100);
end;


type
  TPosComment = (pcSingle, pcLeft, pcRight, pcCenter);
  TCommentRec = record
    Comment: WideString;
    IsBefore: boolean;
    PosComment: TPosComment;
    StripText: WideString;
    Col, Row, Pos, D, IndexMaxRel: LongInt;
  end;

  TCommentList = class
  private
    fOwner: TLiteralText;
    fComments: array of TCommentRec;
    fCount: integer;
  public
    constructor Create(AOwner: TLiteralText; const StartPos, EndPos: LongInt);
    destructor Destroy; override;
    function MakeText(const TokenText: WideString): WideString;
  end;

{ TCommentList }

constructor TCommentList.Create(AOwner: TLiteralText; const StartPos,
  EndPos: Integer);
var CT: TCommentedText;
  QT: TQuotedText;
  WS: WideString;
  Ofs, X, I, N, DBefore, DAfter, D: integer;
  TmpX1, TmpX2: LongInt;
  IsBefore: boolean;

begin
  fOwner := AOwner;
  if (fOwner = nil) or (fOwner.WideText = nil) or
    ((EndPos - StartPos) < 4) then Exit;
  WS := Copy(fOwner.WideText, StartPos + 1, EndPos - StartPos);
  QT := nil;
  CT := nil;
  try
    QT := TQuotedText.Create(PWideChar(WS));
    CT := TCommentedText.Create(PWideChar(WS));
    for I := 0 to fOwner.CountCommentTokens - 1 do
    begin
      CT.AddCommentTokens(fOwner.CommentTokens[I]);
      QT.AddCommentTokens(fOwner.CommentTokens[I]);
    end;
    Ofs := 0;
    X := CT.FindNext(Ofs);
    //   
    while (Ofs >= 0) and (X >= 0) do
    begin
      N := fCount;
      inc(fCount);
      if fCount > Length(fComments) then
        SetLength(fComments, fCount * 2);
      FillChar(fComments[N], SizeOf(fComments[0]), 0);
      while Word(CT.WideText[Ofs - 1]) in [13, 10] do Dec(Ofs);
      fComments[N].Comment := Copy(CT.WideText, X + 1, Ofs - X);
      PosToColRow(CT.WideText, X, fComments[N].Col, fComments[N].Row, CT.Len);
      fComments[N].Pos := X;
      while (X >= 0) and (not (Word(CT.WideText[X]) in [13, 10])) do
      begin
        Dec(X);
        if not IsCharBS(CT.WideText[X]) then
        begin
          fComments[N].PosComment := pcRight;
          break;
        end;
      end;
      X := Ofs;
      while (not (Word(CT.WideText[X]) in [0, 13, 10])) do
      begin
        inc(X);
        if not IsCharBS(CT.WideText[X]) then
        begin
          if fComments[N].PosComment = pcRight then
            fComments[N].PosComment := pcCenter
          else
            fComments[N].PosComment := pcLeft;
          break;
        end;
      end;
      X := CT.FindNext(Ofs);
    end;
    //   ,    
    for I := 0 to fCount - 1 do fComments[I].D := -1;
    if fCount = 0 then Exit;
    Ofs := 0;
    X := QT.FindNext(Ofs);
    while (Ofs >= 0) and (X >= 0) do
    begin
      for I := 0 to fCount - 1 do
      begin
        D := -1;
        IsBefore := false;
        DBefore := X - (fComments[I].Pos + Length(fComments[I].Comment));
        DAfter := fComments[I].Pos - Ofs;
        case fComments[I].PosComment of
          pcLeft: begin
              D := DBefore;
              IsBefore := True;
            end;
          pcRight: begin
              D := DAfter;
              IsBefore := False;
            end;
          pcSingle, pcCenter: begin
              if (DBefore >= 0) and (DAfter >= 0) and (DBefore < DAfter) then
              begin
                D := DBefore;
                IsBefore := True;
              end
              else
              begin
                D := DAfter;
                IsBefore := False;
              end;
            end;
        end;
        if (D >= 0) and ((fComments[I].D = -1) or (D < fComments[I].D)) then
        begin
          fComments[I].D := D;
          fComments[I].IsBefore := IsBefore;
          QT.GetStripText(X, TmpX1, TmpX2, fComments[I].StripText);
        end;
      end;
      X := QT.FindNext(Ofs);
    end;
  finally
    FreeAndNil(QT);
    FreeAndNil(CT);
  end;
end;

destructor TCommentList.Destroy;
begin

  inherited;
end;

function TCommentList.MakeText(const TokenText: WideString): WideString;
var StripStrings: array of WideString;
  TokenStrings: array of WideString;
  WS: WideString;
  Ofs, X1, X2, I, J, K, MaxRel, Rel, IndexMaxRel: integer;
  QT: TQuotedText;
  procedure DoStripWS;
  var X, Ofs, I: integer;
  begin
    if QT = nil then
    begin
      QT := TQuotedText.Create(PWideChar(WS));
      for I := 0 to fOwner.CountCommentTokens - 1 do
        QT.AddCommentTokens(fOwner.CommentTokens[I]);
    end
    else
      QT.WideText := PWideChar(WS);
    Ofs := 0;
    X := QT.FindNext(Ofs);
    if not QT.GetStripText(X, X, Ofs, WS) then
    begin
      WS := '';
      Exit;
    end;
  end;
begin
  result := TokenText;
  QT := nil;
  if fCount > 0 then
  begin
    try
      X1 := 0;
      X2 := 0;
      Ofs := 0;
      while X2 < Length(TokenText) do
      begin
        X2 := FindEOS(PWideChar(TokenText), Ofs);
        if X2 = -1 then X2 := Length(TokenText);
        SetLength(StripStrings, Length(StripStrings) + 1);
        SetLength(TokenStrings, Length(TokenStrings) + 1);
        WS := Copy(TokenText, X1 + 1, X2 - X1);
        TokenStrings[Length(TokenStrings) - 1] := WS;
        DoStripWS;
        StripStrings[Length(StripStrings) - 1] := WS;
        X1 := Ofs;
      end;
      for I := 0 to fCount - 1 do
      begin
        IndexMaxRel := -1;
        MaxRel := 0;
        for J := 0 to Length(StripStrings) - 1 do
        begin
          Rel := IndistinctMatching(4, StripStrings[J], fComments[I].StripText);
          if Rel > MaxRel then
          begin
            MaxRel := Rel;
            IndexMaxRel := J;
          end;
        end;
        fComments[I].IndexMaxRel := IndexMaxRel;
      end;
      for I := 0 to fCount - 1 do
        if fComments[I].IndexMaxRel >= 0 then
        begin
          J := fComments[I].IndexMaxRel;
          case fComments[I].PosComment of
            pcSingle: begin {   } end;
            pcLeft: begin
                X1 := 1;
                while (X1 <= fComments[I].Col) and (TokenStrings[J][X1] = ' ') do inc(X1);
                X2 := 0;
                for K := 1 to Length(fComments[I].Comment) do
                  if TokenStrings[J][X1 + X2] = ' ' then
                    inc(X2)
                  else
                    break;
                Delete(TokenStrings[J], X1, X2);
                Insert(fComments[I].Comment, TokenStrings[J], X1);
              end;
            pcRight: begin
                while Length(TokenStrings[J]) < fComments[I].Col do
                  TokenStrings[J] := TokenStrings[J] + ' ';
                TokenStrings[J] := TokenStrings[J] + fComments[I].Comment;
              end;
            pcCenter: begin {   } end;
          end;
        end;
      result := '';
      for J := 0 to Length(TokenStrings) - 1 do
      begin
        if result <> '' then result := result + #13#10;
        result := result + TokenStrings[J];
      end;
    finally
      FreeAndNil(QT);
    end;
  end;
end;


{ TFindedText }

procedure TFindedText.Clear;
var I: integer;
begin
  SetLength(fTokenPositions, fMaxBufferSize);
  for I := 0 to fMaxBufferSize - 1 do
  begin
    //FillChar(fTokenPositions[I], SizeOf(fTokenPositions[I]), 0);
    fTokenPositions[I].Offset := -1;
    fTokenPositions[I].StartPos := -1;
    fTokenPositions[I].EndPos := -1;
  end;
end;

constructor TFindedText.Create(WideText: PWideChar);
begin
  if WideText = nil then
    raise Exception.CreateFMT(ResInvalidValue, ['nil']);
  inherited Create;
  fRightMargin := 80;
  fCodePage := 1251;
  SetWideText(WideText);
end;

function TFindedText.InternalFind(const Offset: LongInt;
  var StartPos, EndPos: LongInt;
  const Next: boolean): boolean;
var I: integer;
  OldOffset, NewOffset, FindedStartPos: LongInt;
  //   
  function IsFinded(const I: integer): boolean;
  begin
    if Next then
      Result := (I >= 0) and
        (Offset < fTokenPositions[I].StartPos) and
        (Offset >= fTokenPositions[I].Offset)
    else
      Result := (I >= 0) and
        (Offset <= fTokenPositions[I].EndPos) and
        (Offset >= fTokenPositions[I].Offset);
  end;
  //   
  procedure UpdateResult(I: Integer);
  begin
    if Next then
      Result := (Offset <= fTokenPositions[I].EndPos)
    else
      Result := (Offset <= fTokenPositions[I].EndPos) and (Offset >= fTokenPositions[I].StartPos);
    if Result then
    begin
      StartPos := fTokenPositions[I].StartPos;
      EndPos := fTokenPositions[I].EndPos;
    end;
  end;
  //   
  procedure UpdateTokenPos;
  var I: integer;
  begin
    result := false;
    I := OldOffset div PageSize;
    if (FindedStartPos >= 0) then
    begin
      fTokenPositions[I].Offset := OldOffset;
      fTokenPositions[I].StartPos := FindedStartPos;
      fTokenPositions[I].EndPos := NewOffset;
      UpdateResult(I);
    end
    else
    begin
      fTokenPositions[I].Offset := OldOffset;
      fTokenPositions[I].StartPos := -1;
      fTokenPositions[I].EndPos := -1;
    end;
  end;
begin
  Result := false;
  if Offset < 0 then Exit;
  FindedStartPos := -1;
  I := fMaxBufferSize - 1;
  while (I >= 0) and (not IsFinded(I)) do Dec(I);
  if IsFinded(I) then
    UpdateResult(I) // ,    ,    
  else
  begin
    NewOffset := 0;
    for I := 0 to fMaxBufferSize - 1 do
      if (fTokenPositions[I].Offset >= 0) then
      begin
        if (Offset < fTokenPositions[I].Offset) then
          break
        else
          NewOffset := fTokenPositions[I].EndPos;
      end;
    repeat
      OldOffset := NewOffset;
      FindedStartPos := DoFind(NewOffset);
      UpdateTokenPos;
    until (FindedStartPos < 0) or (NewOffset >= Offset);
  end;
end;

function TFindedText.FindNext(var Offset: Integer): LongInt;
var StartPos, EndPos: LongInt;
begin
  StartPos := Offset;
  if InternalFind(Offset, StartPos, EndPos, True) then
  begin
    Offset := EndPos;
    Result := StartPos;
  end
  else
    Result := -1;
end;

function TFindedText.InToken(const Offset: Integer; var StartPos,
  EndPos: Integer): boolean;
begin
  Result := InternalFind(Offset, StartPos, EndPos, False);
end;

function TFindedText.ScanWhileBS(var X: integer): boolean;
begin
  while (X < fLen) and IsCharBS(fWideText[X]) do inc(X);
  result := X < fLen;
end;

procedure TFindedText.SetWideText(const Value: PWideChar);
begin
  fWideText := Value;
  fLen := Length(fWideText);
  fMaxBufferSize := (fLen + PageSize - 1) div (PageSize);
  Clear;
end;

{ TQuotedText }

procedure TQuotedText.UpdateEnableFastFind;
var
  I: Integer;
begin
  if CountCommentTokens = 3 then
  begin
    fEnableFastFind := true;
    for I := 0 to CountCommentTokens - 1 do
      if (Length(CommentTokens[I].StartWord) < 1) or
        (Length(CommentTokens[I].StartWord) > 2) then
        fEnableFastFind := false;
  end
  else
    fEnableFastFind := false;
end;

procedure TQuotedText.AddCommentTokens(Value: TTokenRec);
const EmptyToken: TTokenRec = (StartWord: ''; EndWord: '');
begin
  SetLength(fCommentTokens, Length(fCommentTokens) + 1);
  fCommentTokens[Length(fCommentTokens) - 1] := EmptyToken;
  CommentTokens[Length(fCommentTokens) - 1] := Value;
  UpdateEnableFastFind;
end;

constructor TQuotedText.Create(WideText: PWideChar);
begin
  inherited;
  fQuote := '''';
end;

//       #123.
//  Offset   '#',       

function TQuotedText.ExtractChar(var Offset: LongInt; var W: Word): boolean;
var Ofs, I, N, Code: LongInt;
  Hex: Boolean;
  S: AnsiString;
  function IsCifra(const Ch: WideChar; IsHex: boolean): boolean; {$IFDEF DELPHI_2006_UP}inline; {$ENDIF}
  var C: WideChar;
  begin
    C := UpCaseW(Ch);
    result := ((C >= '0') and (C <= '9')) or
      (IsHex and (C >= 'A') and (C <= 'F'));
  end;
begin
  Ofs := Offset;
  result := False;
  if (Prefix <> #0) then
    if fWideText[Ofs] = Prefix then inc(Ofs)
    else Exit;
  //      
  Hex := (fHexPrefix <> #0) and
    (Ofs < fLen) and
    (fWideText[Ofs] = fHexPrefix);
  if Hex then inc(Ofs);

  //   
  if (Ofs < fLen) and (IsCifra(fWideText[Ofs], Hex)) then
  begin
    I := 0;
    while (Ofs < fLen) and (fWideText[Ofs] = '0') do Inc(Ofs); //  
    S := '';
    while (Ofs + I < fLen) and
      (I < 6) and
      (IsCifra(fWideText[Ofs + I], Hex)) do
    begin
      S := S + AnsiChar(fWideText[Ofs + I]);
      Inc(I);
    end;
    if (I > 6) or (I < 1) then Exit;
    if S = '' then S := '0';
    //   
    try
      if Hex then
        N := HexToInt(S)
      else
      begin
        Val(string(S), N, Code);
        if Code > 0 then Exit;
      end;
      if (N > 65535) then Exit;
    except
      Exit; //       .
    end;
    W := N;
    result := True;
    Offset := Ofs + I;
  end;
end;


function TQuotedText.CheckQuotedText(var Offset: LongInt): boolean;
var OldOffset: LongInt;
  W: Word;
begin
  result := false;
  if Offset >= 0 then
  begin
    OldOffset := Offset;
    if (fQuote <> #0) and (fWideText[Offset] = fQuote) then
      result := OldOffset = FindQuotedText(fQuote, fWideText, Offset, fLen)
    else if (fPrefix <> #0) and (fWideText[Offset] = fPrefix) then
      result := ExtractChar(Offset, W);
    if not result then Offset := OldOffset;
  end;
end;

procedure TQuotedText.Assign(Source: TQuotedText);
var I: integer;
begin
  if Source <> nil then
  begin
    WideText := Source.WideText;
    Quote := Source.Quote;
    SetLength(fCommentTokens, 0);
    for I := 0 to Source.CountCommentTokens - 1 do
      AddCommentTokens(Source.CommentTokens[I]);
    CodePage := Source.CodePage;
    Prefix := Source.Prefix;
    HexPrefix := Source.HexPrefix;
  end
  else
    raise EConvertError.CreateFMT(SAssignError, ['nil', ClassType.ClassName]);
end;

function TQuotedText.NextBrackets(const Open, Close: WideChar; var StartPos,
  EndPos: Integer; First: boolean): boolean;
var N, X1, X2: integer;
begin
  Result := False;
  X1 := StartPos;
  N := 0;
  while NextSimbol(X1, First) do
  begin
    First := False;
    if WideText[X1] = Open then
    begin
      inc(N);
      break;
    end;
    inc(X1);
  end;
  if N > 0 then
  begin
    X2 := X1 + 1;
    while (N > 0) and (NextSimbol(X2, False)) do
    begin
      if WideText[X2] = Open then Inc(N);
      if WideText[X2] = Close then Dec(N);
      inc(X2);
    end;
    if N = 0 then
    begin
      StartPos := X1;
      EndPos := X2;
      Result := true;
    end;
  end;
end;

constructor TQuotedText.Create(AOwner: TQuotedText);
var W: PWideChar;
begin
  if AOwner <> nil then
  begin
    Create(AOwner.WideText);
    Assign(AOwner);
    fOwner := AOwner;
  end
  else
  begin
    W := nil;
    Create(W);
  end;
end;

function TQuotedText.FindFirstToken(var X1, X2: LongInt): Integer;
var I: integer;
  OldOffset, Ofs, MinX1, MinX2: LongInt;
  PrefixStr: WideString;
  W: Word;
begin
  MinX1 := -1;
  MinX2 := -1;
  result := -1;
  OldOffset := X1;
  if fEnableFastFind then
  begin
    PrefixStr := Prefix;
    Ofs := OldOffset;
    repeat
      result := FindFirstWord(fWideText, Ofs, fQuote,
        PDWORD(@CommentTokens[0].StartWord[1])^,
        PDWORD(@CommentTokens[1].StartWord[1])^,
        PDWORD(@CommentTokens[2].StartWord[1])^,
        PDWORD(@PrefixStr[1])^,
        fLen);
      if result = -1 then break;
      MinX1 := Ofs;
      if Result in [3, 4] then
      begin
        MinX2 := Ofs;
        if not CheckQuotedText(MinX2) then result := -1
        else result := 3;
        inc(Ofs);
      end
      else
      begin
        inc(Ofs, Length(CommentTokens[Result].StartWord));
        MinX2 := Ofs;
        if PosStrEx(CommentTokens[Result].EndWord, fWideText, MinX2, fLen) < 0 then result := -1;
      end;
    until result >= 0;
  end
  else
  begin
    for I := 0 to CountCommentTokens - 1 do
    begin
      Ofs := OldOffset;
      X1 := PosStrEx(CommentTokens[I].StartWord, fWideText, Ofs, fLen);
      if X1 >= 0 then
      begin
        X2 := PosStrEx(CommentTokens[I].EndWord, fWideText, Ofs, fLen);
        if (X2 >= 0) and ((X1 < MinX1) or (MinX1 < 0)) then
        begin
          MinX1 := X1;
          MinX2 := Ofs;
          result := I;
        end;
      end;
    end;
    Ofs := OldOffset;
    X1 := FindQuotedText(fQuote, fWideText, Ofs, fLen);
    if (X1 >= 0) and ((X1 < MinX1) or (MinX1 < 0)) then
    begin
      MinX1 := X1;
      MinX2 := Ofs;
      result := CountCommentTokens;
    end;
    if Prefix <> #0 then
    begin
      Ofs := OldOffset;
      X1 := PosStrEx(Prefix, fWideText, Ofs, fLen);
      if (X1 >= 0) and ((X1 < MinX1) or (MinX1 < 0)) then
      begin
        Dec(Ofs);
        if ExtractChar(Ofs, w) then
        begin
          MinX1 := X1;
          MinX2 := Ofs;
          result := CountCommentTokens;
        end;
      end;
    end;
  end;
  if result >= 0 then
  begin
    X1 := MinX1;
    X2 := MinX2;
  end
  else
  begin
    X1 := -1;
    X2 := -1;
  end;
end;

destructor TQuotedText.Destroy;
begin
  inherited;
end;

function TQuotedText.DoFind(var Offset: Integer): LongInt;
var X1, X2, I: LongInt;
begin
  Result := -1;
  X2 := Offset;
  repeat
    X1 := X2;
    I := FindFirstToken(X1, X2);
  until (I < 0) or (I = CountCommentTokens);
  if (I >= 0) then
  begin
    Result := X1;
    Offset := X2;
  end;
end;

function TQuotedText.GetCommentTokens(Index: integer): TTokenRec;
begin
  if (Index >= 0) and (Index < Length(fCommentTokens)) then
    result := fCommentTokens[Index]
  else
    EListError.CreateFmt(SListIndexError, [Index]);
end;

function TQuotedText.GetCountCommentTokens: integer;
begin
  result := Length(fCommentTokens);
end;

procedure TQuotedText.SetCommentTokens(Index: integer; const Value: TTokenRec);
begin
  if (Index >= 0) and (Index < Length(fCommentTokens)) then
  begin
    if (Value.StartWord <> fCommentTokens[Index].StartWord) or
      (Value.EndWord <> fCommentTokens[Index].EndWord) then
    begin
      fCommentTokens[Index] := value;
      Clear;
    end;
  end
  else
    EListError.CreateFmt(SListIndexError, [Index]);
end;

procedure TQuotedText.SetHexPrefix(const Value: WideChar);
begin
  if fHexPrefix <> Value then
  begin
    fHexPrefix := Value;
    Clear;
  end;
end;

procedure TQuotedText.SetPrefix(const Value: WideChar);
begin
  if fPrefix <> Value then
  begin
    fPrefix := Value;
    Clear;
  end;
end;

procedure TQuotedText.SetQuote(const Value: WideChar);
begin
  if fQuote <> Value then
  begin
    fQuote := Value;
    Clear;
  end;
end;

function TQuotedText.GetStripText(const Offset: Integer; var StartPos,
  EndPos: Integer; var S: WideString): boolean;
var Ofs, N: LongInt;
begin
  result := False;
  Ofs := Offset;
  if Ofs >= fLen then Exit;
  if ScanWhileBS(Ofs) then
  begin
    Result := InToken(Ofs, StartPos, EndPos);
    if Result then
    begin
      N := EndPos - StartPos - 2;
      SetLength(S, N);
      Move(fWideText[StartPos + 1], S[1], SizeOf(S[1]) * N);
    end;
  end;
end;

function TQuotedText.GetTokenText(const StartPos, EndPos: Integer;
  const S: WideString): WideString;
begin
  Result := '';
  if fQuote <> #0 then Result := Result + fQuote;
  Result := Result + S;
  if fQuote <> #0 then Result := Result + fQuote;
end;


function TQuotedText.IsLetterOrDigit(C: WideChar): boolean;
begin
  {$IFDEF DELPHI_2009_UP}
    result := ((C > ' ') and Character.IsLetterOrDigit(C)) or (C = '_');
  {$ELSE}
    result := ((C >= 'A') and (C <= 'Z')) or
              ((C >= 'z') and (C <= 'z')) or
              ((C >= '0') and (C <= '9'));
  {$ENDIF}
end;

function TQuotedText.IsSimbol(C: WideChar): boolean;
begin
  result := (Ord(C) > 32) and (Ord(C) < 128) and
            (not IsLetterOrDigit(C)) and
            (not (AnsiChar(C) in CharBackSpaces));
end;

procedure TQuotedText.NextCommentOrLiteral(var StartPos, EndPos: Integer);
var
  I, J, X1: Integer;
  PrefixStr: WideString;
begin
  PrefixStr := Prefix;
  X1 := StartPos;
  if fEnableFastFind then
  begin
    I := FindFirstWord(fWideText, X1, fQuote,
        PDWORD(@CommentTokens[0].StartWord[1])^,
        PDWORD(@CommentTokens[1].StartWord[1])^,
        PDWORD(@CommentTokens[2].StartWord[1])^,
        PDWORD(@PrefixStr[1])^,
        fLen);
  end
  else
    I := -1;
  if I = -1 then
  begin
    StartPos := fLen;
    EndPos := StartPos;
  end
  else
  begin
    //  ,  
    if I in [3, 4] then
    begin
      StartPos := X1;
      CheckQuotedText(X1);
      if X1 = StartPos then
        X1 := fLen;
      EndPos := X1;
    end
    else
    begin
      StartPos := X1;
      inc(X1, Length(CommentTokens[I].StartWord));
      J := PosStrEx(CommentTokens[I].EndWord, fWideText, X1, fLen);
      if J < 0 then
        EndPos := fLen
      else
        EndPos := X1;
    end;
  end;
end;

function TQuotedText.NextSimbol(var StartPos: Integer; First: boolean): boolean;
var X, MaxX, NextX: integer;
begin
  result := False;
  if (StartPos >= fLen) or (StartPos < 0) then Exit;
  //       ,  
  if First then
  begin
    NextX := -1;
    repeat
      X := NextX + 1;
      NextCommentOrLiteral(X, NextX);
    until (NextX >= StartPos) or (NextX >= fLen);
    if X < StartPos then X := NextX
                    else X := StartPos;
  end
  else
    X := StartPos;
  while not Result do
  begin
    MaxX := X;
    NextCommentOrLiteral(MaxX, NextX);
    //if MaxX = NextX then Inc(NextX);
    while (X < MaxX) and (not IsSimbol(fWideText[X])) do inc(X);
    if X >= fLen then Break;
    if (X < fLen) and (X = MaxX) then
    begin
      //  ,  
      X := NextX;
    end
    else
    begin
      StartPos := X;
      Result := X < fLen;
    end;
  end;
end;

function TQuotedText.NextWord(var StartPos, EndPos: Integer;
  First: boolean): boolean;
var X, Start, MaxX, NextX: integer;
    PrefixStr: WideString;
begin
  result := False;
  if (StartPos >= fLen) or (StartPos < 0) then Exit;
  //       ,  
  if First then
  begin
    NextX := -1;
    repeat
      X := NextX + 1;
      NextCommentOrLiteral(X, NextX);
    until (NextX >= StartPos) or (NextX >= fLen);
    if X < StartPos then X := NextX
                    else X := StartPos;
  end
  else
    X := StartPos;
  if (X < fLen) and IsLetterOrDigit(fWideText[X]) then
  begin
    //   
    while (X >= 0) and (IsLetterOrDigit(fWideText[X])) do Dec(X);
    Start := StartPos;
    StartPos := X + 1;
    X := Start + 1;
    while (X < fLen) and (IsLetterOrDigit(fWideText[X])) do inc(X);
    EndPos := X;
    result := true;
  end
  else
  begin
    PrefixStr := Prefix;
    while not Result do
    begin
      MaxX := X;
      NextCommentOrLiteral(MaxX, NextX);
      while (X < MaxX) and (not IsLetterOrDigit(fWideText[X])) do inc(X);
      if X >= fLen then Break;
      if (X < fLen) and (X = MaxX) then
      begin
        //  ,  
        X := NextX;
        Continue;
      end
      else
      begin
        StartPos := X;
      end;
      if X < fLen then
      begin
        StartPos := X;
        while (X < fLen) and (IsLetterOrDigit(fWideText[X])) do inc(X);
        EndPos := X;
        result := true;
      end;
    end;
  end;
end;

function TQuotedText.NextWordFromList(const WordList: TStringList; var StartPos,
  EndPos: Integer; First: boolean): Integer;
var
  I, X1, X2, J: Integer;
  Eq: boolean;
begin
  result := -1;
  X1 := StartPos;
  X2 := EndPos;
  while Result = -1 do
  begin
    if NextWord(X1, X2, First) then
    begin
      First := false;
      for I := 0 to WordList.Count - 1 do
      begin
        Eq := X2 - X1 = Length(WordList[I]);
        if Eq then
        for J := 1 to X2 - X1 do
          if UpCaseW(fWideText[X1 - 1 + J]) <> UpCaseW(WideChar(WordList[I][J])) then
          begin
            Eq := False;
            break;
          end;
        if Eq then
        begin
          result := I;
          StartPos := X1;
          EndPos := X2;
          Exit;
        end;
      end;
      X1 := X2 + 1;
    end
    else
      Exit;
  end;
end;

{ TCommentedText }

function TCommentedText.DoFind(var Offset: Integer): LongInt;
var X1, X2, I: LongInt;
begin
  Result := -1;
  X2 := Offset;
  repeat
    X1 := X2;
    I := FindFirstToken(X1, X2);
  until (I < CountCommentTokens);
  if (I >= 0) then
  begin
    Result := X1;
    Offset := X2;
  end;
end;

function TCommentedText.GetStripText(const Offset: Integer; var StartPos,
  EndPos: Integer; var S: WideString): boolean;
var Ofs, N, R: LongInt;
begin
  result := False;
  Ofs := Offset;
  if Ofs >= fLen then Exit;
  if ScanWhileBS(Ofs) then
  begin
    Result := InToken(Ofs, StartPos, EndPos);
    if Result then
    begin
      R := FindFirstToken(StartPos, EndPos);
      N := EndPos - StartPos - Length(CommentTokens[R].StartWord) - Length(CommentTokens[R].EndWord);
      SetLength(S, N);
      Move(fWideText[StartPos + Length(CommentTokens[R].StartWord)], S[1], SizeOf(S[1]) * N);
    end;
  end;
end;

function TCommentedText.GetTokenText(const StartPos, EndPos: Integer;
  const S: WideString): WideString;
var Ofs, X1, X2, R: LongInt;
begin
  Ofs := StartPos;
  Result := '';
  R := -1;
  if ScanWhileBS(Ofs) then
  begin
    if InToken(Ofs, X1, X2) then R := FindFirstToken(X1, X2);
  end;
  if R = -1 then R := 0;
  Result := Result + CommentTokens[R].StartWord;
  Result := Result + S;
  Result := Result + CommentTokens[R].EndWord;
end;

{ TLiteralText }

constructor TLiteralText.Create(WideText: PWideChar);
begin
  inherited;
  fPrefix := '#';
  fHexPrefix := '$';
  fConcateChar := '+';
  AddCommentTokens(CommentToken1);
  AddCommentTokens(CommentToken2);
  AddCommentTokens(CommentToken3);
end;

procedure TLiteralText.SetConcateChar(const Value: WideString);
begin
  if fConcateChar <> Value then
  begin
    fConcateChar := Value;
    Clear;
  end;
end;

//  X,  fWideText[X]  ,  
//  AcceptConcate,      '+'

function TLiteralText.ScanWhileBackSpace(var X: Integer; AcceptConcate: boolean): boolean;
var I, Ofs: integer;
  Found: boolean;
  TMPS: WideString;
begin
  Found := True;
  while (X < fLen) and Found do
  begin
    Found := ScanWhileBS(X);
    if Found then
    begin
      if AcceptConcate and (ConcateChar <> '') then
      begin
        TMPS := Copy(fWideText, X + 1, Length(ConcateChar));
        Found := TMPS = ConcateChar;
        if Found then Inc(X, Length(ConcateChar));
      end
      else
        Found := False;
      if not Found then
        for I := 0 to CountCommentTokens - 1 do
        begin
          Ofs := X;
          if (PosStrEx(CommentTokens[I].StartWord, fWideText, Ofs, fLen) = X) and
            (PosStrEx(CommentTokens[I].EndWord, fWideText, Ofs, fLen) >= 0) then
          begin
            X := Ofs;
            Found := True;
            break;
          end;
        end;
    end;
  end;
  Result := X < fLen;
end;

destructor TLiteralText.Destroy;
begin
  FreeAndNil(fCommentList);
  inherited;
end;

function TLiteralText.DoFind(var Offset: Integer): LongInt;
var X1, X2, R: integer;
  W: Word;
begin
  result := -1;
  X2 := Offset;
  repeat
    X1 := X2;
    R := FindFirstToken(X1, X2);
  until (R < 0) or (R = CountCommentTokens);
  if R = CountCommentTokens then
  begin
    Result := X1;
    Offset := X2;
    while ScanWhileBackSpace(X2, True) do
    begin
      if (fQuote <> #0) and (fWideText[X2] = fQuote) then
      begin
        if inherited DoFind(X2) = -1 then Exit;
        Offset := X2;
      end
      else
        if (fPrefix <> #0) and (fWideText[X2] = fPrefix) then
        begin
          if not ExtractChar(X2, W) then Exit;
          Offset := X2;
        end
        else
          Exit;
    end;
  end;
end;


function TLiteralText.GetStripText(const Offset: Integer; var StartPos,
  EndPos: Integer; var S: WideString): boolean;
var Ofs, X, X1, X2, N: LongInt;
  TmpS: WideString;
  W: Word;
  I: Integer;
begin
  result := False;
  S := '';
  Ofs := Offset;
  if Ofs >= fLen then Exit;
  if ScanWhileBS(Ofs) then
  begin
    Result := InToken(Ofs, StartPos, EndPos);
    if Result then
    begin
      X := StartPos;
      while ScanWhileBackSpace(X, True) do
      begin
        if (fQuote <> #0) and (fWideText[X] = fQuote) then
        begin
          X2 := X;
          X1 := FindQuotedText(fQuote, fWideText, X2, fLen);
          if X1 < 0 then Exit;
          Result := True;
          N := X2 - X1 - 2;
          SetLength(TmpS, N);
          Move(fWideText[X1 + 1], TmpS[1], SizeOf(TmpS[1]) * N);
          X := X2;
          S := S + TmpS;
        end
        else
          if (fPrefix <> #0) and (fWideText[X] = fPrefix) then
          begin
            if not ExtractChar(X, W) then break;
            S := S + ' ';
            Word(S[Length(S)]) := W;
          end
          else
            break;
      end;
      if fQuote <> #0 then
      begin
        I := 1;
        while I < Length(S) do
        begin
          if (S[I] = fQuote) and (S[I + 1] = fQuote) then Delete(S, I, 1);
          inc(I);
        end;
      end;
      UpdateComments(StartPos, EndPos);
    end
    else
      FreeAndNil(fCommentList);
  end;
end;

procedure TLiteralText.TruncRight(var Result: WideString);
var X, L: integer;
begin
  while Length(Result) > 0 do
  begin
    X := Length(Result) - Length(ConcateChar) + 1;
    L := Length(ConcateChar);
    if (X > 0) and (Copy(Result, X, L) = ConcateChar) then
      Delete(Result, X, L)
    else if (Word(Result[Length(Result)]) in [13, 10, 32]) then
      Delete(Result, Length(Result), 1)
    else
      break;
  end;
end;

procedure TLiteralText.UpdateComments(StartPos, EndPos: Integer);
begin
  FreeAndNil(fCommentList);
  fCommentList := TCommentList.Create(Self, StartPos, EndPos);
end;

function TLiteralText.MakeText(StripText: WideString;
  const Hex: boolean = false;
  const FirstLeft: Longint = 0;
  Left: Longint = 0;
  Right: Longint = 80;
  const CodePage: LongInt = 1251;
  const NoComments: boolean = true): WideString;
var I, I1, IndexChar, CountChar: integer;
  Text: WideString;
  Br: WideString;
  InQuoted: boolean;
  NoRepresented, ConcateAdded: boolean;
  ACh: AnsiChar;
  function IsConcate(const S: WideString; const I: integer): boolean;
  begin
    result := (I > 0) and (Copy(S, I, Length(fConcateChar)) = fConcateChar);
  end;
  //  ,        
  function ScanInQuoted(const S: WideString; I: integer): Integer;
  begin
    Result := -1;
    //      ,     
    if (S[I] = ' ') and
      ((I < Length(S)) and (S[I + 1] = Quote) and (S[I + 2] <> Quote)) then
      Dec(I);
    while (I > 0) and (S[I] <> ' ') do
    begin
      if S[I] = Quote then
      begin
        if I > 1 then
          if S[I - 1] = Quote then Dec(I)
          else Exit
        else
          Exit;
      end;
      Dec(I);
    end;
    if I > 0 then
      result := I + 1;
  end;
  //  ,       
  function ScanOutQuoted(const S: WideString; I: integer): Integer;
  begin
    Result := -1;
    while (I > 0) and
      (not IsConcate(S, I - Length(ConcateChar) + 1)) and
      (not (Ord(S[I]) in [Ord(Prefix), Ord(Quote), 32, 13, 10, 9, 0])) do
      Dec(I);
    if I > 0 then
    begin
      if (Ord(S[I]) in [Ord(Prefix), 32]) then
      begin
        if (I > 1) and (S[I - 1] = ' ') then Dec(I);
        result := I
      end
      else if S[I] = Quote then
        result := I + 1
      else if (IsConcate(S, I - Length(ConcateChar) + 1)) then
        result := I + Length(ConcateChar);
    end;
  end;
  //   
  procedure AddBreak(I: integer; Last: boolean);
  var L, R, X, BreakPos: integer;
    S: WideString;
  begin
    if I1 = 0 then L := FirstLeft
    else L := Left;
    R := Right;
    if (R - L < 15) then R := L + 15;
    if (not Last) and (not ConcateAdded) then
      Dec(R, Length(ConcateChar));
    if InQuoted and (CountChar > 3) then
    begin
      L := (R + L) div (2);
      Dec(R, 2);
      if I - I1 > R then X := I1 + R
      else X := I;
      BreakPos := ScanInQuoted(Text, X);
      if BreakPos < L then BreakPos := X;
      S := Quote;
      if not ConcateAdded then
      begin
        S := S + ' ' + ConcateChar;
        ConcateAdded := True;
      end;
      I1 := BreakPos + Length(S) + 1;
      S := S + Br + Quote;
      Insert(S, Text, BreakPos);
    end
    else
    begin
      if InQuoted then
        I := Length(Text) - CountChar - 1;
      if (not ConcateAdded) and (Text[I1 + R] <> Prefix) then Dec(R);
      if I - I1 > R then X := I1 + R
      else X := I;
      if X > Length(Text) then
        BreakPos := Length(Text) + 1
      else
      begin
        BreakPos := ScanOutQuoted(Text, X);
        if BreakPos < L then BreakPos := X;
      end;
      //   
      while (BreakPos < Length(Text)) and (Text[BreakPos] = ' ') do
        Delete(Text, BreakPos, 1);
      while (BreakPos > 1) and (Text[BreakPos - 1] = ' ') do
      begin
        Dec(BreakPos);
        Delete(Text, BreakPos, 1);
      end;
      ConcateAdded := Text[BreakPos] <> Prefix;
      //    "+"
      if ((BreakPos <= Length(Text) - Length(ConcateChar)) and IsConcate(Text, BreakPos)) or
        IsConcate(Text, BreakPos - Length(ConcateChar)) then
        S := ''
      else
        S := ' ' + ConcateChar;
      I1 := BreakPos + Length(S) + 1;
      S := S + Br;
      Insert(S, Text, BreakPos);
    end;
  end;
  //  
  procedure AddToResult(S: WideString; Last: boolean; var NewStr: boolean);
  var I, X, R, L: integer;
  begin
    I := Length(Text);
    X := Length(S);
    if not InQuoted then
      while (X > 0) and (S[X] = ' ') do Dec(X);
    X := I - I1 + X;
    R := Right;
    if InQuoted then Dec(R);
    if Last then Dec(R);
    if I1 = 0 then L := FirstLeft
    else L := Left;
    if (R - L < 15) then R := L + 15;
    //    
    NewStr := (X > R);
    Text := Text + S;
    if NewStr then
    begin
      AddBreak(I1 + R, Last);
    end;
  end;
  //     ,     
  procedure SetInQuoted(Value, BeginStr, EndStr: boolean);
  var S: WideString;
    NewStr: boolean;
  begin
    if (fQuote <> #0) and (Value <> InQuoted) then
    begin
      CountChar := 0;
      InQuoted := Value;
      if InQuoted then
      begin
        if BeginStr or ConcateAdded then
          S := ''
        else
          S := ' ' + ConcateChar;
        if (not EndStr) and (not BeginStr) and (not ConcateAdded) then
          S := S + ' ';
        S := S + Quote;
        AddToResult(S, EndStr, NewStr);
        ConcateAdded := False;
      end
      else
      begin
        S := Quote;
        if (not EndStr) and (not ConcateAdded) then
        begin
          S := S + ' ' + ConcateChar;
          ConcateAdded := true;
        end;
        if (not BeginStr) and (not EndStr) then
          S := S + ' ';
        AddToResult(S, EndStr, NewStr);
      end;
    end;
  end;
  //   
  procedure AddNormaChar(Ch: WideChar; BeginStr, EndStr: boolean);
  var S: WideString;
    NewStr: boolean;
  begin
    SetInQuoted(True, BeginStr, false);
    S := Ch;
    if Ch = Quote then S := S + S;
    inc(CountChar, Length(S));
    AddToResult(S, EndStr, NewStr);
    ConcateAdded := False;
  end;
  //   
  procedure AddSingleChar(Ch: WideChar; BeginStr, EndStr: boolean);
  var S: WideString;
    NewStr: boolean;
  begin
    SetInQuoted(false, BeginStr, false);
    S := Prefix;
    if Hex and (HexPrefix <> #0) then
    begin
      S := S + HexPrefix;
      if Word(Ch) <= 15 then
        S := S + IntToHex(Word(Ch), 1)
      else if Word(Ch) <= 255 then
        S := S + IntToHex(Word(Ch), 2)
      else
        S := S + IntToHex(Word(Ch), 4);
    end
    else
      S := S + InttoStr(Word(Ch));
    inc(CountChar);
    AddToResult(S, EndStr, NewStr);
    ConcateAdded := False;
    if Ord(Ch) = 10 then
    begin
      //    
      if (not EndStr) {and (not NewStr)} then AddBreak(Length(Text) + 1, False);
    end;
  end;
begin
  //  
  result := '';
  ConcateAdded := false;
  CountChar := 0;
  for I := 0 to FirstLeft - 1 do result := result + ' ';
  if StripText = '' then
  begin
    result := result + Quote + Quote;
    Exit;
  end;
  Text := Result;
  //        30 
  if Right - Left < 30 then Left := Right - 30;
  if Left < 0 then
  begin
    inc(Right, -Left);
    Left := 0;
  end;
  //  
  Br := #13#10;
  for I := 0 to Left - 1 do Br := Br + ' ';
  //  
  InQuoted := false;
  I1 := 0;
  for IndexChar := 1 to Length(StripText) do
  begin
    if (StripText[IndexChar] < ' ') or (StripText[IndexChar] = #160) then
      AddSingleChar(StripText[IndexChar], IndexChar = 1, IndexChar = Length(StripText))
    else
    begin
      if CodePage = 1200 then
        AddNormaChar(StripText[IndexChar], IndexChar = 1, IndexChar = Length(StripText))
      else
      begin
        NoRepresented := not WideCharToMultiByte(CodePage, Word(StripText[IndexChar]), ACh);
        //NoRepresented := true;
        if NoRepresented then
          AddSingleChar(StripText[IndexChar], IndexChar = 1, IndexChar = Length(StripText))
        else
          AddNormaChar(StripText[IndexChar], IndexChar = 1, IndexChar = Length(StripText));
      end;
    end;
  end;
  SetInQuoted(False, False, True);
  Result := Text;
  if (not NoComments) and (fCommentList is TCommentList) then
    Result := TCommentList(fCommentList).MakeText(Result);
end;

function TLiteralText.GetMargins(const StartPos, EndPos: Integer;
  var FirstLeft, Left, Right: integer): boolean;
var Ofs, X1, X2: LongInt;
  NoQuote: boolean;
begin
  result := false;
  Left := 0;
  FirstLeft := 0;
  Right := RightMargin;
  Ofs := StartPos;
  if Ofs >= fLen then Exit;
  if ScanWhileBS(Ofs) then
  begin
    NoQuote := not InToken(Ofs, X1, X2);
    if NoQuote then
    begin
      X1 := StartPos - 1;
      Left := fColumn;
      FirstLeft := fColumn;
    end
    else
    begin
      while (X1 - Left >= 0) and (not (Word(fWideText[X1 - Left]) in [13, 10])) do Inc(Left);
      FirstLeft := Left;
      Result := True;
    end;
  end;
  if RightMargin - Left < 40 then
    Left := RightMargin - 40;
  if Left < 0 then Left := 0;
  Left := (Left div 2) * 2;
  FirstLeft := (FirstLeft div 2) * 2;
end;

function TLiteralText.GetTokenText(const StartPos, EndPos: LongInt;
  const S: WideString): WideString;
var Left, Right, FirstLeft: LongInt;
begin
  result := '';
  GetMargins(StartPos, EndPos, FirstLeft, Left, Right);
  result := MakeText(S, DefaultHex, FirstLeft, Left, Right, CodePage);
end;

{ TFunctionText }

constructor TFunctionText.Create(WideText: PWideChar);
begin
  inherited;
  fList := TStringList.Create;
  fParams := TStringList.Create;
  fParamTypes := TStringList.Create;
  StringToList(UpperCase('procedure|function|property|class'), '|', fList);
end;

destructor TFunctionText.Destroy;
begin
  FreeAndNil(fList);
  FreeAndNil(fParams);
  FreeAndNil(fParamTypes);
  inherited;
end;

function TFunctionText.DoFind(var Offset: Integer): LongInt;
var X1, X2, NextX, I, StartPos, EndPos: Integer;
    TmpList: TStringList;
  procedure CheckEnd;
  begin
    if WideText[X1] = ';' then
    begin
      Result := StartPos;
      Offset := X1 + 1;
    end;
  end;
  procedure FillParams(StartPos, EndPos: integer);
  var OldLen: integer;
      X1, X2, FirstIndex, I: integer;
      WS, WT: WideString;
      Found: boolean;
  begin
    OldLen := fLen;
    try
      fLen := EndPos;
      FirstIndex := 0;
      X1 := StartPos;
      while NextWord(X1, X2, False) do
      begin
        WT := '';
        WS := Copy(WideText, X1 + 1, X2 - X1);
        X1 := X2;
        if ScanWhileBS(X1) then
        begin
          if (WideText[X1] = ',') or (WideText[X1] = ';') then
          begin
            fParams.add(WS);
            if WideText[X1] = ';' then
              for I := FirstIndex to fParams.Count - 1 do
                fParamTypes.Add('');
          end;
          if (WideText[X1] = ':') then
          begin
            inc(X1);
            fParams.add(WS);
            if NextWord(X1, X2, False) then
            begin
              Found := False;
              while (not Found) and NextSimbol(X2, false) do
              begin
                Found := WideText[X2] = ';';
                inc(X2);
              end;
              if not Found then X2 := fLen else Dec(X2);
              WT := Copy(WideText, X1 + 1, X2 - X1);
              for I := FirstIndex to fParams.Count - 1 do
                fParamTypes.Add(WT);
              FirstIndex := fParams.Count;
              X1 := X2;
            end;
          end;
        end
        else
          fParams.add(WS);
      end;
      for I := FirstIndex to fParams.Count - 1 do
        fParamTypes.Add('');
    finally
      fLen := OldLen;
    end;
  end;
begin
  X1 := Offset;
  Result := -1;
  repeat
    I := NextWordFromList(fList, X1, X2, false);
    fObjKind := I;
    StartPos := X1;
    fObjClass := '';
    fObjName := '';
    fObjReturn := '';
    fParams.Clear;
    fParamTypes.Clear;
    case I of
      0, 1: begin
           X1 := X2 + 1;
           if NextWord(X1, X2, False) then
           begin
             fObjName := Copy(WideText, X1 + 1, X2 - X1);
             X1 := X2;
             if NextSimbol(X1, False) then
             begin
               if WideText[X1] = '.' then
               begin
                 fObjClass := fObjName;
                 if NextWord(X1, X2, False) then
                   fObjName := Copy(WideText, X1 + 1, X2 - X1)
                 else
                   Continue;
                 X1 := X2;
                 if not NextSimbol(X1, False) then Continue;
               end;
               CheckEnd;
               if (Result = -1) and (WideText[X1] = '(') then
               begin
                 if NextBrackets('(', ')', X1, X2, False) then
                 begin
                   FillParams(X1 + 1, X2 - 1);
                   X1 := X2;
                   CheckEnd;
                   if (Result = -1) and (WideText[X1] = ':') then
                   begin
                     if NextWord(X1, X2, False) then
                       fObjReturn := Copy(WideText, X1 + 1, X2 - X1)
                     else
                       Continue;
                     X1 := X2;
                     CheckEnd;
                   end
                   else
                     X1 := X2;
                 end;
               end;
             end;
           end;
         end;
      3: begin
           fClassEndPos := -1;
           StartPos := X1;
           repeat
             while (X1 > 0) and (fWideText[X1] <> '=') do
               Dec(x1);
             if X1 <= 0 then break;
             Dec(X1);
             NextX := X1;
             NextCommentOrLiteral(NextX, EndPos);
           until (X1 < NextX) or (NextX = -1) or (X1 <= 0);
           if (NextX = -1) or (X1 <= 0) then
           begin
             Offset := X2;
             Exit;
           end;
           repeat
             while (X1 > 0) and (not IsLetterOrDigit(fWideText[X1])) do Dec(x1);
             while (X1 > 0) and (IsLetterOrDigit(fWideText[X1])) do Dec(x1);
             if X1 <= 0 then break;
             NextX := X1;
           until (NextWord(NextX, EndPos) and (NextX = X1 + 1)) or (X1 <= 0);
           if (NextX = -1) or (X1 <= 0) then
           begin
             Offset := X2;
             Exit;
           end;
           fObjName := Copy(WideText, NextX + 1, EndPos - NextX);
           X1 := NextX;
           if NextSimbol(X1, False) then
           begin
             if (WideText[X1] = '=') then
             begin
               TmpList := TStringList.Create;
               try
                 TmpList.Add(UpperCase('class'));
                 if NextWordFromList(TmpList, X1, X2, False) = 0 then
                 begin
                   StartPos := X1;
                   if NextSimbol(x1, false) then
                   begin
                     CheckEnd;
                     TmpList.Clear;
                     TmpList.Add(UpperCase('end'));
                     if (Result = -1) and (WideText[X1] = '(') then
                     begin
                       if not NextBrackets('(', ')', X1, X2, False) then
                         exit;
                     end;
                     fObjClass := Copy(WideText, StartPos + 1, X2 - StartPos);
                     EndPos := X2;
                     if (NextWordFromList(TmpList, X1, X2, False) = 0) and NextSimbol(x2, false) then
                     begin
                       if WideText[X2] = ';' then
                       begin
                         Result := NextX;
                         Offset := EndPos;
                         fClassEndPos := X2;
                       end;
                     end;
                   end;
                 end;
               finally
                 TmpList.Free;
               end;
             end;
           end;
         end;
      else
      begin
        X1 := X2;
      end;
    end;
  until (I = -1) or (Result <> -1);
end;

end.


