unit UMainExpertBuild;
{$I Directives.inc}
interface
uses Windows, SysUtils, Classes, ToolsAPI, Menus, IniFiles,
  RegisterIdeMenu, uIDEutl, TextUtlCep, uBaseForm, UnitFormBuild, ActnList,
  CepUtils, PCustFunc, WindowThread, uWixGenerator;

const
  ExceptExeName = 'ExpertCepBuild.BPL; ControlsCepEx.BPL; ExspertCep.BPL; EditorsCepEx.BPL';
resourcestring
  Res_BuildMenuName = '   ';    // Alt + B
  ErrExecute = '  ';
  OkExecute = '';
  StrExecute = ' ' + #13#10 + '%s';
  StrBuild = ' ' + #13#10 + '%s';
  StrSave = ' ' + #13#10 + '%s';
  StrNoBuild = '     ' + #13#10 + '''%s''';
type
  TExpertBuild = class(TObject)
  private
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure DoBuild(Sender: TObject);
    procedure DoUpdateBuild(Sender: TObject);
    procedure onBuild(Sender: TObject);
  end;

procedure LogIDEMessage(const psFile, psMessage: string; const piY, piX: integer);
function GetCurrentProject: IOTAProject;
{$IFDEF DELPHI_2005_UP}
procedure Register;
{$ENDIF}


implementation
uses UMainExpertCep{$IFDEF DELPHI_2005_UP}, DesignIntf{$ENDIF};
const
  MAIN_MENU = 'ProjectMenu'; //    Progect
  BUILD_MENU = 'ProjectInformationItem'; //   information for
  CategoryIDE = 'ExpertCep';
  PrefixActionName = 'actCep';
  BuildMain = 'BuildCurrentProjectCep';
  ResourceBuild = 'BUILDCEP';

  RUN_MENU = 'RunMenu';
  MENU_SUBRUN = 'RunRunItem';
var ExpertBuild: TExpertBuild;

{$IFDEF DELPHI_2005_UP}
procedure Register;
begin
  ForceDemandLoadState(dlDisable);
end;
{$ENDIF}

procedure LogIDEMessage(const psFile, psMessage: string; const piY, piX: integer);
var
  lciMessages: IOTAMessageServices40;
  hRes: HResult;
begin
  { no empty lines in this log }
  if psMessage = '' then
    exit;

  hRes := BorlandIDEServices.QueryInterface(IOTAMessageServices40, lciMessages);
  if hRes <> S_OK then
    exit;
  if lciMessages = nil then
    exit;

  if (piY >= 0) and (piX >= 0) then
    lciMessages.AddToolMessage(psFile, psMessage, 'CEP', piY, piX)
  else
    lciMessages.AddTitleMessage('CEP: ' + psFile + ' ' + psMessage);
end;

{ the function GetCurrentProject is from Erik's OpenTools API FAQ and Resources
  http://www.gexperts.org/opentools/
}
// Modified from code posted by Ray Lischner (www.tempest-sw.com)
function GetCurrentProject: IOTAProject;
var
  Services: IOTAModuleServices;
  Module: IOTAModule;
  Project: IOTAProject;
  ProjectGroup: IOTAProjectGroup;
  MultipleProjects: boolean;
  I: integer;
begin
  Result   := nil;
  MultipleProjects := False;
  Services := BorlandIDEServices as IOTAModuleServices;
  for I := 0 to Services.ModuleCount - 1 do
  begin
    Module := Services.Modules[I];
    if Module.QueryInterface(IOTAProjectGroup, ProjectGroup) = S_OK then
    begin
      Result := ProjectGroup.ActiveProject;
      Exit;
    end
    else if Module.QueryInterface(IOTAProject, Project) = S_OK then
    begin
      if Result = nil then
        Result := Project
      else
        MultipleProjects := True;
    end;
  end;
  if MultipleProjects then
    Result := nil;
end;


{ TExpertBuild }

constructor TExpertBuild.Create;
begin
  inherited;
  RegisterMenuItem(MenuItemRec(MAIN_MENU,
                               BUILD_MENU,
                               BuildMain,
                               CategoryIDE,
                               PrefixActionName,
                               Res_BuildMenuName,
                               ResourceBuild,
                               False,
                               'B',
                               [ssAlt],
                               DefaultToolBarName,
                               '',
                               ExpertBuild.DoBuild,
                               ExpertBuild.DoUpdateBuild));
end;

destructor TExpertBuild.Destroy;
begin
  UnregisterMenuItem(BuildMain);
  inherited;
end;

procedure TExpertBuild.DoBuild(Sender: TObject);
var F: TFormBuild;
    NeedRun: boolean;
    miRun, Item: TMenuItem;
    MainMenu: TMenu;
    Proj: IOTAProject;
    Target: string;
begin
  Proj := GetCurrentProject;
  if Proj <> nil then
  begin
    Item := nil;
    MainMenu := IdeMainMenu;
    miRun := IntFindMenuByName(MainMenu,RUN_MENU);
    if miRun <> nil then
    begin
      {$IFDEF DEBUG} ShowSubItemNames(miRun); {$ENDIF}
      Item := IntFindMenuByName(miRun, MENU_SUBRUN);
    end;
    F := TFormBuild.Create(nil);
    try
      Target := GetTargetFileName(Proj);
      F.Caption := Target;
      F.ExeName := Target;
      F.cbRun.Enabled := Item <> nil;
      if Item = nil then F.cbRun.Checked := False;
      F.onDefaultAction := onBuild;
      NeedRun := (F.ShowModal = IdOk) and (Item <> nil) and (F.cbRun.Checked);
    finally
      FreeAndNil(F);
    end;
    if NeedRun then
    begin
      if (Item.Action is TCustomAction) and (TCustomAction(Item.Action).Enabled) then
      begin
        Item.Action.Execute;
      end;
    end;
  end;
end;

procedure TExpertBuild.onBuild(Sender: TObject);
var F: TThreadWindow;
    TextBefore, TextAfter: string;
    OutStream: TMemoryStream;
    Proj: IOTAProject;
    StepCount, Step: integer;
  function DoExecute(Text: string): boolean;
  begin
    Text := Trim(Text);
    result := Text <> '';
    if Result then
    begin
      Text := StringReplace(Text,
                            Macros[0],
                            Proj.FileName,
                            [rfReplaceAll, rfIgnoreCase]);
      Text := StringReplace(Text,
                            Macros[1],
                            GetTargetFileName(Proj),
                            [rfReplaceAll, rfIgnoreCase]);
      Text := StringReplace(Text,
                            Macros[2],
                            Proj.ProjectOptions.Values['HostApplication'],
                            [rfReplaceAll, rfIgnoreCase]);
      Text := StringReplace(Text,
                            Macros[3],
                            Proj.ProjectOptions.Values['RunParams'],
                            [rfReplaceAll, rfIgnoreCase]);
      result := ExecuteProcess(Text, '', ExtractFileDir(Proj.FileName), '', OutStream);
      if not Result then
      begin
        LogIDEMessage('', ErrExecute + ': ' + Text, -1, -1);
        //MessageBox(0, PChar(ErrExecute + ': ' + #13#10 + Text), nil, MB_ICONEXCLAMATION OR MB_TOPMOST);
      end
      else
        LogIDEMessage('', OkExecute + ': ' + Text, -1, -1);
    end;
  end;
  procedure SetWindowThreadText(const S: string; const Args: array of const);
  begin
    F.BoundsRect := Rect(8, 8, 400, 66);
    F.Text := Format(S, Args);
    F.Percent := (Step * 1000) div (StepCount);
    inc(Step);
  end;
begin
  Proj := GetCurrentProject;
  if (Sender is TFormBuild) and (Proj <> nil) then
  with TFormBuild(Sender) do
  begin
    if Pos(UpperCase(ExtractFileName(ExeName)), UpperCase(ExceptExeName)) > 0 then
    begin
      MessageBox(0, PChar(Format(StrNoBuild, [ExeName])), PChar(ExtractFileName(ExeName)), MB_ICONEXCLAMATION or MB_TOPMOST);
      Exit;
    end;
    Step := 0;
    StepCount := 1;
    if (TextBefore <> '') and (actRunBefore.Checked) then Inc(StepCount);
    if (cbSave.Checked) or ((cbMSI.Enabled) and (cbMSI.Checked)) then Inc(StepCount);
    if (actMSI.Enabled) and (actMSI.Checked) then Inc(StepCount);
    if (TextAfter <> '') and (actRunAfter.Checked) then Inc(StepCount);
    TextBefore := Trim(eBefore.Text);
    TextAfter := Trim(eAfter.Text);
    F := TThreadWindow.Show;
    try
      F.Enabled := False;
      F.Interval := 100;
      F.Percent := 0;
      F.Font := Font.Handle;
      F.BoundsRect := Rect(8, 8, 400, 66);
      ClearText;
      OutStream := TMemoryStream.Create;
      try
        Stream := OutStream;
        if (TextBefore <> '') and (actRunBefore.Checked) then
        begin
          SetWindowThreadText(StrExecute, [TextBefore]);
          DoExecute(TextBefore);
          CopyTextFromStream;
        end;
        SetWindowThreadText(StrBuild, [ExeName]);
        if cbRun.Checked then Close;
        if not Proj.ProjectBuilder.BuildProject(cmOTABuild, false, true) then
        begin
          Exit;
        end
        else
        begin
          if cbRun.Checked then ModalResult := IdOk;
        end;
        if (cbSave.Checked) or ((cbMSI.Enabled) and (cbMSI.Checked)) then
        begin
          SetWindowThreadText(StrSave, [Proj.FileName]);
          Proj.Save(False, True);
        end;
        if (actMSI.Enabled) and (actMSI.Checked) then
        begin
          SetWindowThreadText(StrBuild, [ChangeFileExt(ExeName, '.msi')]);
          //frmWixGenerator.LoadFromIni;
          MakeMSI;
          CopyTextFromStream;
        end;
        if (TextAfter <> '') and (actRunAfter.Checked) then
        begin
          SetWindowThreadText(StrExecute, [TextAfter]);
          DoExecute(TextAfter);
          CopyTextFromStream;
        end;
      finally
        FreeAndNil(OutStream);
        Stream := nil;
      end;
    finally
      FreeAndNil(F);
    end;
  end;
end;


procedure TExpertBuild.DoUpdateBuild(Sender: TObject);
begin
  if Sender is TAction then
  begin
    TAction(Sender).Enabled := GetCurrentProject <> nil;
  end;
end;

initialization
  ExpertBuild := TExpertBuild.Create;

finalization
  FreeAndNil(ExpertBuild);

end.
