unit UMultiThread;
(*
 * Module:     TMultiThreadFor      
 *               .
 * Version:    2.5 (14.11.2008)
 * Author:      
 * Comments:      :
 *             .   for Z := StartValueZ to FinishValueZ do
 *             .     for Y := StartValueY to FinishValueY do
 *             .       for X := StartValueX to FinishValueZ do
 *             .         Execute(X, Y, Z)
 *                 ().
 *             :
 *             1.      onExecute   ,
 *                   ,
 *               .
 *             2.  onStartThread, onFinishThread, onExecute 
 *               ,       
 *                 .
 *             3.  onExecute,       
 *             ,   TMultiThreadFor     
 *             ,    .
 *                   .
 *                    
 *              : roschinspb@mail.ru
 *             http://www.delphikingdom.com/asp/users.asp?ID=1271
 *             http://www.roschinspb.narod.ru
 *
 * Copyright:   2008
 *
 * Version:    2.5 (14.11.2008)
 * Realized:   
 *)
interface
{$I Directives.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls,
  InternalTimer;

resourcestring
  StrEStarted = '  ,     %s';
  StrENoStarted = '  ';
  StrEParam = '  (%s)  %s';

const
  OrderNames: array[0..2] of string = ('X', 'Y', 'Z');
  MaxPercent = 1000;             //   onPercent   0..MaxPercent
  DelayTerminate = 1000;         //       Stop
  DelayInTimer = 10;             //     
  TimerInterval = 40;            //    onPercent
  IndexNull = $FFFFFFFF;

type
  //  TMultiThreadFor
  TMultiThreadStatus = (mtsNull, //   
    mtsStarted,                  //  
    mtsTerminated,               //  
    mtsError,                    //     
    mtsSucceeds);                //   
  //   
  TOrderRec = record
    StartValue: Extended;        //  
    FinishValue: Extended;       //  
    Count: DWORD;                //  
  end;
  //      
  TErrorStack = record
    Exception: ExceptClass;
    Message: string;
    Index: DWORD;
    VectorIndex: array of DWORD;
    Vector: array of Extended;
  end;

  TMultiThreadFor = class;

  //    TMultiThreadFor
  EMultiThread = class(Exception);

  //  
  TMultiThreadForEvent = procedure(Sender: TMultiThreadFor; ThreadIndex: Word) of object;
  TMultiThreadForExecute = procedure(Sender: TMultiThreadFor;
                                     ThreadIndex: Word;
                                     const Index: DWORD;
                                     const VectorIndex: array of DWORD;
                                     const Vector: array of Extended) of object;
  TBeforeStart = procedure(Sender: TMultiThreadFor) of object;
  TAfterFinish = procedure(Sender: TMultiThreadFor; const ErrorStack: array of TErrorStack) of object;
  TonPercent = procedure(Sender: TMultiThreadFor; const Percent: integer) of object;

  //        
  TOrderItem = class(TCollectionItem)
  private
    fItem: TOrderRec;
    fDisplayName: string;
    function GetCount: DWORD;
    function GetFinishValue: Extended;
    function GetStartValue: Extended;
    procedure SetCount(const Value: DWORD);
    procedure SetFinishValue(const Value: Extended);
    procedure SetStartValue(const Value: Extended);
    procedure SetItem(const Value: TOrderRec);
    procedure CheckCollection(Collection: TCollection);
  protected
    {$IFDEF DELPHI_6_UP}
    procedure SetCollection(Value: TCollection); override;
    {$ENDIF}
    procedure SetIndex(Value: Integer); override;
    function GetDisplayName: string; override;
    procedure SetDisplayName(const Value: string); override;
  public
    constructor Create(Collection: TCollection); override;
    property Item: TOrderRec read fItem write SetItem;
    procedure Assign(Source: TPersistent); override;
    procedure LoadFromStream(S: TStream);
    procedure SaveToStream(S: TStream);
  published
    property StartValue: Extended read GetStartValue write SetStartValue;
    property FinishValue: Extended read GetFinishValue write SetFinishValue;
    property Count: DWORD read GetCount write SetCount default 1;
    property DisplayName;
  end;

  //    
  TOrderItems = class(TOwnedCollection)
  private
    function GetItem(Index: Integer): TOrderItem;
    procedure SetItem(Index: Integer; const Value: TOrderItem);
  protected
    procedure Update(Item: TCollectionItem); override;
  published
  public
    constructor Create(AOwner: TPersistent);
    property Items[Index: Integer]: TOrderItem read GetItem write SetItem; default;
  end;

  //  
  TMultiThreadFor = class(TComponent)
  private
    fForParams: array of TOrderRec;
    fDelta: array of Extended;
    fThreads: array of TThread;
    fIndexes: array of DWORD;
    fExecuteCount: DWORD;
    FLock: TRTLCriticalSection;
    fStarted: boolean;
    fDefThreadCount: Word;
    fThreadCount: Word;
    fonStartThread: TMultiThreadForEvent;
    fonFinishThread: TMultiThreadForEvent;
    fonExecute: TMultiThreadForExecute;
    fPartCount: DWORD;
    fOrders: TOrderItems;
    fErrorStack: array of TErrorStack;
    fonBeforeStart: TBeforeStart;
    fonAfterFinish: TAfterFinish;
    fonPercent: TonPercent;
    fLastPercent: integer;
    fTimer: TInternalTimer;
    fStatus: TMultiThreadStatus;
    procedure CheckStarted;
    procedure SetDefThreadCount(const Value: Word);
    procedure SetonStartThread(const Value: TMultiThreadForEvent);
    procedure SetonFinishThread(const Value: TMultiThreadForEvent);
    procedure SetonExecute(const Value: TMultiThreadForExecute);
    procedure SetPartCount(const Value: DWORD);
    procedure SetOrderItems(const Value: TOrderItems);
    function GetThreads(Index: integer): TThread;
    procedure DelCriticalSection;
    procedure SetonBeforeStart(const Value: TBeforeStart);
    procedure SetonAfterFinish(const Value: TAfterFinish);
    procedure DoTimer(Sender: TObject);
  protected
    procedure IndexToVector(Index: DWORD;
                        var VectorIndex: array of DWORD;
                        var Vector: array of Extended);
    procedure DoStartThread(ThreadIndex: Word); virtual;
    procedure DoFinishThread(ThreadIndex: Word); virtual;
    procedure DoExecute(const ThreadIndex: Word;
                        const Index: DWORD;
                        const VectorIndex: array of DWORD;
                        const Vector: array of Extended); virtual;
    procedure DoBeforeStart; virtual;
    procedure DoAfterFinish(const ErrorStack: array of TErrorStack); virtual;
    procedure DoPercent(const Percent: integer); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // .        
    procedure Start;
    //    .
    procedure Stop;
    //   ,    Delay .
    function Wait(Delay: DWORD): DWORD;
    // Threads    ()
    property Threads[Index: integer]: TThread read GetThreads;
    // ThreadCount    
    property ThreadCount: Word read fThreadCount;
    //  Started = true,    
    property Started: boolean read fStarted;
    // Status     TMultiThreadFor
    property Status: TMultiThreadStatus read fStatus;
    //   
    procedure Lock;
    //   
    procedure UnLock;
  published
    //  DefThreadCount   ,   .
    //  0,    ,    .
    property DefThreadCount: Word read fDefThreadCount write SetDefThreadCount default 0;
    //  PartCount   ,  
    //   
    property PartCount: DWORD read fPartCount write SetPartCount default 1000;
    //   
    property Orders: TOrderItems read fOrders write SetOrderItems;

    //       
    //  onStartThread     
    property onStartThread: TMultiThreadForEvent read fonStartThread write SetonStartThread;
    //  onFinishThread     ,  
    //    ,    
    property onFinishThread: TMultiThreadForEvent read fonFinishThread write SetonFinishThread;
    //   onExecute      
    //   
    property onExecute: TMultiThreadForExecute read fonExecute write SetonExecute;

    //       .
    //   ,      .
    //  onBeforeStart     
    property onBeforeStart: TBeforeStart read fonBeforeStart write SetonBeforeStart;
    //  onAfterFinish   / 
    property onAfterFinish: TAfterFinish read fonAfterFinish write SetonAfterFinish;
    //  onPercent     
    //    .
    property onPercent: TonPercent read fonPercent write fonPercent;
  end;

procedure Register;

implementation
uses {$IFDEF DELPHI_6_UP}RTLConsts;
     {$ELSE}Consts;
     {$ENDIF}

procedure Register;
begin
  RegisterComponents('Cep', [TMultiThreadFor]);
end;

type

  TThreadItem = class(TThread)
    fOwner: TMultiThreadFor;
    fThreadIndex: Word;
    fException: ExceptClass;
    fMessage: string;
  public
    constructor Create(AOwner: TMultiThreadFor; AThreadIndex: Word);
    procedure Execute; override;
  end;

{ TThreadItem }

constructor TThreadItem.Create(AOwner: TMultiThreadFor; AThreadIndex: Word);
begin
  if AOwner = nil then
    raise EMultiThread.CreateFmt(StrEParam, ['nil', 'AOwner']);
  inherited Create(True);
  fOwner := AOwner;
  fThreadIndex := AThreadIndex;
end;

procedure TThreadItem.Execute;
var CurrIndex: DWORD;
  I: integer;
  VectorIndex: array of DWORD;
  Vector: array of Extended;
  function GetCurrIndex: DWORD;
  var I: integer;
  begin
    fOwner.Lock;
    try
      Result := IndexNull;
      // ,        
      for I := 0 to Length(fOwner.fThreads) - 1 do
        if (fOwner.fThreads[I] <> nil) and
          (TThreadItem(fOwner.fThreads[I]).fException <> nil) then Exit;
      //    
      for I := 0 to Length(fOwner.fIndexes) - 1 do
        if //(I <> fThreadIndex) and
          (fOwner.fIndexes[I] <> IndexNull) and
          ((fOwner.fIndexes[I] > Result) or (Result = IndexNull)) then
          Result := fOwner.fIndexes[I];
      if Result = IndexNull then
        Result := 0
      else
        Result := Result + fOwner.fPartCount;
      fOwner.fIndexes[fThreadIndex] := Result;
    finally
      fOwner.UnLock;
    end;
  end;
begin
  try
    try
      fOwner.DoStartThread(fThreadIndex);
      SetLength(VectorIndex, Length(fOwner.fForParams));
      SetLength(Vector, Length(fOwner.fForParams));
      while not Terminated do
      begin
        CurrIndex := GetCurrIndex;
        if CurrIndex = IndexNull then Exit;
        for I := 0 to fOwner.fPartCount - 1 do
        begin
          if Terminated then Exit;
          //     ,    
          if (CurrIndex >= fOwner.fExecuteCount) then
          begin
            fOwner.Lock;
            try
              fOwner.fIndexes[fThreadIndex] := fOwner.fExecuteCount;
              Exit;
            finally
              fOwner.UnLock;
            end;
          end;
          fOwner.IndexToVector(CurrIndex, VectorIndex, Vector);
          fOwner.DoExecute(fThreadIndex, CurrIndex, VectorIndex, Vector);
          inc(CurrIndex);
        end;
      end;
    except
      on E: Exception do
      begin
        fException := ExceptClass(E.ClassType);
        fMessage := E.Message;
      end;
    end;
  finally
    fOwner.DoFinishThread(fThreadIndex);
  end;
end;

type
  TOrderItemWrapper = class(TComponent)
  private
    FOrderItem: TOrderItem;
  published
    property OrderItem: TOrderItem read FOrderItem write FOrderItem;
  end;


{ TMultiThreadFor }

procedure TMultiThreadFor.CheckStarted;
begin
  if fStarted then
    raise EMultiThread.CreateFMT(StrEStarted, [ClassType.ClassName]);
end;

constructor TMultiThreadFor.Create(AOwner: TComponent);
begin
  inherited;
  fOrders := TOrderItems.Create(self);
  fPartCount := 1000;
end;

destructor TMultiThreadFor.Destroy;
var I: integer;
begin
  FreeAndNil(fTimer);
  Stop;
  for I := 0 to Length(fThreads) - 1 do
    FreeAndNil(fThreads[I]);
  DelCriticalSection;
  FreeAndNil(fOrders);
  inherited;
end;

procedure TMultiThreadFor.DoFinishThread(ThreadIndex: Word);
begin
  if Assigned(fonFinishThread) then fonFinishThread(self, ThreadIndex);
end;

procedure TMultiThreadFor.DoExecute(const ThreadIndex: Word;
                                    const Index: DWORD;
                                    const VectorIndex: array of DWORD;
                                    const Vector: array of Extended);
begin
  if Assigned(fonExecute) then fonExecute(self,
                                          ThreadIndex,
                                          Index,
                                          VectorIndex,
                                          Vector);
end;

procedure TMultiThreadFor.DoStartThread(ThreadIndex: Word);
begin
  if Assigned(fonStartThread) then fonStartThread(self, ThreadIndex);
end;

procedure TMultiThreadFor.DoAfterFinish(const ErrorStack: array of TErrorStack);
begin
  if ([csDestroying, csDesigning] * ComponentState) = [] then
    if Assigned(fonAfterFinish) then fonAfterFinish(self, ErrorStack);
end;

procedure TMultiThreadFor.DoBeforeStart;
begin
  if ([csDestroying, csDesigning] * ComponentState) = [] then
    if Assigned(fonBeforeStart) then fonBeforeStart(self);
end;

procedure TMultiThreadFor.DoPercent(const Percent: integer);
begin
  if ([csDestroying, csDesigning] * ComponentState) = [] then
    if Assigned(fonPercent) then fonPercent(self, Percent);
end;

procedure TMultiThreadFor.DoTimer(Sender: TObject);
var I, NewPercent, N: integer;
  Percent: Double;
  ExecutedCount: DWORD;
  NeedStop: boolean;
begin
  if fStarted and (fExecuteCount > 0) then
  begin
    ExecutedCount := 0;
    NeedStop := True;
    // ,     
    for I := 0 to Length(fThreads) - 1 do
      if (fThreads[I] = nil) or
        (TThreadItem(fThreads[I]).fException <> nil) then
      begin
        //   , 
        //     .    ,
        //   
        if Wait(DelayInTimer) <> WAIT_TIMEOUT then Stop;
        Exit;
      end
      else
      begin
        if fIndexes[I] <> fExecuteCount then
          NeedStop := False; //     
      end;
    //    ,  100%    
    if NeedStop then
    begin
      try
        DoPercent(MaxPercent);
        Exit;
      finally
        Stop;
      end;
    end
    else
    begin
      N := 0;
      for I := 0 to Length(fThreads) - 1 do
        if (fThreads[I] <> nil) and (fIndexes[I] <> IndexNull) then
        begin
          inc(ExecutedCount, fIndexes[I]);
          inc(N);
        end;
      if N > 0 then
        Percent := ExecutedCount / fExecuteCount / N
      else
        Percent := 0;
      NewPercent := Round(Percent * MaxPercent);
      if NewPercent >= MaxPercent then
        NewPercent := MaxPercent - 1;
      if NewPercent <> fLastPercent then
      begin
        DoPercent(NewPercent);
        fLastPercent := NewPercent;
      end;
    end;
  end
  else
  begin
    if fTimer <> nil then fTimer.Stop;
    //DoPercent(-1);
  end;
end;

function TMultiThreadFor.GetThreads(Index: integer): TThread;
begin
  result := fThreads[Index];
end;

procedure TMultiThreadFor.IndexToVector(Index: DWORD;
                                    var VectorIndex: array of DWORD;
                                    var Vector: array of Extended);
var M, X: DWORD;
    I: integer;
begin
  M := fExecuteCount;
  for I := 0 to High(fForParams) do
  begin
    M := M div fForParams[High(fForParams) - I].Count;
    if M > 1 then X := Index mod M else X := Index;
    Index := Index div fForParams[I].Count;
    VectorIndex[I] := X;
    Vector[I] := fForParams[I].StartValue + X * fDelta[I];
  end;
end;

procedure TMultiThreadFor.SetDefThreadCount(const Value: Word);
begin
  if fDefThreadCount <> Value then
  begin
    CheckStarted;
    fDefThreadCount := Value;
  end;
end;

procedure TMultiThreadFor.SetonExecute(const Value: TMultiThreadForExecute);
begin
  if @fonExecute <> @Value then
  begin
    CheckStarted;
    fonExecute := Value;
  end;
end;

procedure TMultiThreadFor.SetonFinishThread(const Value: TMultiThreadForEvent);
begin
  if @fonFinishThread <> @Value then
  begin
    CheckStarted;
    fonFinishThread := Value;
  end;
end;

procedure TMultiThreadFor.SetonStartThread(const Value: TMultiThreadForEvent);
begin
  if @fonStartThread <> @Value then
  begin
    CheckStarted;
    fonStartThread := Value;
  end;
end;

procedure TMultiThreadFor.SetonBeforeStart(const Value: TBeforeStart);
begin
  if @fonBeforeStart <> @Value then
  begin
    CheckStarted;
    fonBeforeStart := Value;
  end;
end;

procedure TMultiThreadFor.SetonAfterFinish(const Value: TAfterFinish);
begin
  if @fonAfterFinish <> @Value then
  begin
    CheckStarted;
    fonAfterFinish := Value;
  end;
end;

procedure TMultiThreadFor.SetOrderItems(const Value: TOrderItems);
begin
  fOrders.Assign(Value);
end;

procedure CheckCount(Value: DWORD; ValueName: string);
begin
  if (Value < 1) or (Value = IndexNull) then
    raise EMultiThread.CreateFmt(StrEParam, [inttostr(Value), ValueName]);
end;

procedure TMultiThreadFor.SetPartCount(const Value: DWORD);
begin
  if fPartCount <> Value then
  begin
    CheckStarted;
    CheckCount(Value, 'PartCount');
    fPartCount := Value;
  end;
end;

procedure TMultiThreadFor.Start;
var Info: TSystemInfo;
  I: integer;
begin
  if not Started then
  begin
    fStatus := mtsNull;
    DelCriticalSection;
    DoBeforeStart;
    fLastPercent := -1;
    SetLength(fErrorStack, 0);
    if fDefThreadCount = 0 then
    begin
      GetSystemInfo(Info);
      fThreadCount := Info.dwNumberOfProcessors;
    end
    else
      fThreadCount := fDefThreadCount;
    SetLength(fDelta, Length(fForParams));
    fExecuteCount := 1;
    for I := 0 to High(fForParams) do
      if fForParams[I].Count > 1 then
      begin
        fDelta[I] := (fForParams[I].FinishValue - fForParams[I].StartValue) / (fForParams[I].Count - 1);
        fExecuteCount := fExecuteCount * fForParams[I].Count;
      end
      else
        fDelta[I] := 0;
    SetLength(fIndexes, fThreadCount);
    SetLength(fThreads, fThreadCount);
    for I := 0 to fThreadCount - 1 do
    begin
      fIndexes[I] := IndexNull;
      fThreads[I] := nil;
    end;
    try
      for I := 0 to fThreadCount - 1 do
      begin
        fThreads[I] := TThreadItem.Create(Self, I);
      end;
      FillChar(FLock, SizeOf(FLock), 0);
      InitializeCriticalSection(FLock);
      if fTimer = nil then
        fTimer := TInternalTimer.Create(self);
      fTimer.ExecuteInterval(TimerInterval, DoTimer);
      fStarted := true;
      fStatus := mtsStarted;
    except
      for I := 0 to fThreadCount - 1 do
      begin
        FreeAndNil(fThreads[I]);
      end;
      raise;
    end;
    for I := 0 to fThreadCount - 1 do
      fThreads[I].Resume;
  end;
end;

procedure TMultiThreadFor.Lock;
begin
  if not fStarted then
    raise EMultiThread.Create(StrENoStarted);
  EnterCriticalSection(FLock);
end;

procedure TMultiThreadFor.UnLock;
begin
  if not fStarted then
    raise EMultiThread.Create(StrENoStarted);
  LeaveCriticalSection(FLock);
end;

function TMultiThreadFor.Wait(Delay: DWORD): DWORD;
var I, J, MaxJ, Index: integer;
  N: DWORD;
  Handles: TWOHandleArray;
begin
  result := 0;
  if not fStarted then Exit;
  MaxJ := fThreadCount div MAXIMUM_WAIT_OBJECTS;
  for J := 0 to MaxJ do
  begin
    Lock;
    try
      N := fThreadCount - (J * MAXIMUM_WAIT_OBJECTS);
      if N > MAXIMUM_WAIT_OBJECTS then N := MAXIMUM_WAIT_OBJECTS;
      for I := 0 to N - 1 do
      begin
        Index := I + (J * MAXIMUM_WAIT_OBJECTS);
        if fThreads[Index] <> nil then
          Handles[I] := fThreads[Index].Handle
        else
          Handles[I] := 0;
      end;
    finally
      UnLock;
    end;
    result := WaitForMultipleObjects(N, @Handles, True, Delay);
    //   
    if Result = WAIT_TIMEOUT then Exit;
    //    
    if (Result < WAIT_OBJECT_0 + N) then
      Result := WAIT_OBJECT_0;
    //   
    if (Result >= WAIT_ABANDONED_0) and (Result < WAIT_ABANDONED_0 + N) then
      Result := WAIT_ABANDONED_0;
  end;
end;

procedure TMultiThreadFor.DelCriticalSection;
begin
  DeleteCriticalSection(FLock);
  FillChar(FLock, SizeOf(FLock), 0);
end;

procedure TMultiThreadFor.Stop;
var I: integer;
begin
  if Started then
  begin
    //  
    for I := 0 to fThreadCount - 1 do
    begin
      if (fThreads[I] <> nil) and
        (fIndexes[I] <> fExecuteCount) and
        (TThreadItem(fThreads[I]).fException = nil) then
      begin
        fThreads[I].Terminate;
        fStatus := mtsTerminated;
      end;
    end;
    try
      //   ,      
      Wait(DelayTerminate);
      //   
      SetLength(fErrorStack, fThreadCount);
      for I := 0 to fThreadCount - 1 do
      begin
        SetLength(fErrorStack[I].Vector, Length(fForParams));
        SetLength(fErrorStack[I].VectorIndex, Length(fForParams));
        fErrorStack[I].Index := fIndexes[I];
        IndexToVector(fIndexes[I],
                      fErrorStack[I].VectorIndex,
                      fErrorStack[I].Vector);
        if fThreads[I] <> nil then
        begin
          fErrorStack[I].Exception := TThreadItem(fThreads[I]).fException;
          fErrorStack[I].Message := TThreadItem(fThreads[I]).fMessage;
          if fErrorStack[I].Exception <> nil then
          begin
            //  
            if fErrorStack[I].Exception.InheritsFrom(EAbort) then
            begin
              //   EAbort   mtsTerminated
              if fStatus = mtsStarted then fStatus := mtsTerminated;
            end
            else
              fStatus := mtsError;
          end;
        end
        else
        begin
          fErrorStack[I].Exception := nil;
          fErrorStack[I].Message := 'nil';
        end;
      end;
    finally
      //   
      for I := 0 to fThreadCount - 1 do
        FreeAndNil(fThreads[I]);
      fThreadCount := 0;
      SetLength(fThreads, fThreadCount);
    end;
    //   
    DelCriticalSection;
    fStarted := false;
    if fStatus = mtsStarted then
      fStatus := mtsSucceeds;
    DoAfterFinish(fErrorStack);
  end;
end;

{ TOrderItem }

constructor TOrderItem.Create(Collection: TCollection);
begin
  inherited;
  fItem.Count := 1;
end;

function TOrderItem.GetCount: DWORD;
begin
  result := fItem.Count;
end;

function TOrderItem.GetDisplayName: string;
begin
  result := fDisplayName;
end;

function TOrderItem.GetFinishValue: Extended;
begin
  result := fItem.FinishValue;
end;

function TOrderItem.GetStartValue: Extended;
begin
  result := fItem.StartValue;
end;

procedure TOrderItem.LoadFromStream(S: TStream);
var
  Wrapper: TOrderItemWrapper;
begin
  CheckCollection(Collection);
  Wrapper := TOrderItemWrapper.Create(nil);
  try
    Wrapper.OrderItem := TOrderItem.Create(nil);
    S.ReadComponent(Wrapper);
    Assign(Wrapper.OrderItem);
  finally
    Wrapper.OrderItem.Free;
    Wrapper.Free;
  end;
end;

procedure TOrderItem.SaveToStream(S: TStream);
var
  Wrapper: TOrderItemWrapper;
begin
  Wrapper := TOrderItemWrapper.Create(nil);
  try
    Wrapper.OrderItem := Self;
    S.WriteComponent(Wrapper);
  finally
    Wrapper.Free;
  end;
end;

procedure TOrderItem.Assign(Source: TPersistent);
var Tmp: TOrderRec;
  procedure AssignError(Source: TPersistent);
  var
    SourceName: string;
  begin
    if Source <> nil then
      SourceName := Source.ClassName else
      SourceName := 'nil';
    raise EConvertError.CreateResFmt(@SAssignError, [SourceName, ClassName]);
  end;
begin
  if Source = nil then
  begin
    Tmp.Count := 1;
    Tmp.StartValue := 0;
    Tmp.FinishValue := 0;
  end
  else
  begin
    if Source is TOrderItem then
      Tmp := TOrderItem(Source).Item
    else
      AssignError(Source);
  end;
  Item := TMP;
end;

procedure TOrderItem.CheckCollection(Collection: TCollection);
begin
  if (Collection <> nil) and
     (Collection is TOrderItems) and
     (TOrderItems(Collection).GetOwner is TMultiThreadFor) then
    TMultiThreadFor(TOrderItems(Collection).GetOwner).CheckStarted;
end;

procedure TOrderItem.SetItem(const Value: TOrderRec);
begin
  if not CompareMem(@Value, @fItem, SizeOf(fItem)) then
  begin
    CheckCollection(Collection);
    CheckCount(Value.Count, 'Count');
    fItem := Value;
    Changed(false);
  end;
end;

{$IFDEF DELPHI_6_UP}
procedure TOrderItem.SetCollection(Value: TCollection);
begin
  if Value <> Collection then
  begin
    CheckCollection(Value);
    CheckCollection(Collection);
    inherited;
    if (fDisplayName = '') and
      (Collection <> nil) and
      (Collection.Count > 0) and
      (Collection.Count <= Length(OrderNames)) then
      fDisplayName := OrderNames[Collection.Count - 1];
  end;
end;
{$ENDIF}

procedure TOrderItem.SetCount(const Value: DWORD);
begin
  if Value <> Count then
  begin
    CheckCollection(Collection);
    CheckCount(Value, 'Count');
    fItem.Count := Value;
    Changed(false);
  end;
end;

procedure TOrderItem.SetDisplayName(const Value: string);
begin
  fDisplayName := Value;
end;

procedure TOrderItem.SetFinishValue(const Value: Extended);
begin
  if Value <> FinishValue then
  begin
    CheckCollection(Collection);
    fItem.FinishValue := Value;
    Changed(false);
  end;
end;

procedure TOrderItem.SetStartValue(const Value: Extended);
begin
  if Value <> StartValue then
  begin
    CheckCollection(Collection);
    fItem.StartValue := Value;
    Changed(false);
  end;
end;

procedure TOrderItem.SetIndex(Value: Integer);
begin
  if Index <> Value then
  begin
    CheckCollection(Collection);
    inherited;
  end;
end;

{ TOrderItems }

constructor TOrderItems.Create(AOwner: TPersistent);
var S: string;
begin
  if not (AOwner is TMultiThreadFor) then
  begin
    if AOwner = nil then
      S := 'nil'
    else
      S := AOwner.ClassName;
    raise EMultiThread.CreateFmt(StrEParam, [S, 'AOwner']);
  end;
  inherited Create(AOwner, TOrderItem);
end;

function TOrderItems.GetItem(Index: Integer): TOrderItem;
begin
  result := TOrderItem(inherited GetItem(Index));
end;

procedure TOrderItems.SetItem(Index: Integer; const Value: TOrderItem);
begin
  inherited SetItem(Index, TCollectionItem(Value));
end;

procedure TOrderItems.Update(Item: TCollectionItem);
var I: integer;
begin
  if (GetOwner is TMultiThreadFor) then
  begin
    TMultiThreadFor(GetOwner).CheckStarted;
    SetLength(TMultiThreadFor(GetOwner).fForParams, Count);
    SetLength(TMultiThreadFor(GetOwner).fDelta, Count);
    for I := 0 to Count - 1 do
    begin
      TMultiThreadFor(GetOwner).fForParams[I] := TOrderItem(Items[I]).Item;
    end;
  end;
  inherited;
end;

end.
