unit UnitFormBuild;

interface
{$I Directives.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ActnList, ExtCtrls, IniFiles, uBaseForm,
  ImgList, Math, ImageListCep, SpeedButtonCep, InternalTimer, ufWixGenerator, uWixGenerator,
  Menus;

const _Project = '$PROJECT';
      _Target = '$EXENAME';
      _Host = '$HOSTNAME';
      _Params = '$PARAMS';

      Macros: array [0..3] of string = (_Project, _Target, _Host, _Params);


type

  TFormBuild = class(TBaseForm)
    cbAfter: TCheckBox;
    cbBefore: TCheckBox;
    eAfter: TComboBox;
    actRunAfter: TAction;
    actInsert: TAction;
    eBefore: TComboBox;
    actRunBefore: TAction;
    cbMSI: TCheckBox;
    actMSI: TAction;
    actfrmWixGenerator: TAction;
    sbcMSI: TSpeedButtonCep;
    PopupItems: TPopupMenu;
    cbRun: TCheckBox;
    actCopy: TAction;
    sbCopy: TSpeedButtonCep;
    cbSave: TCheckBox;
    actSave: TAction;
    procedure actRunAfterExecute(Sender: TObject);
    procedure actInsertExecute(Sender: TObject);
    procedure ListBox1DblClick(Sender: TObject);
    procedure actInsertUpdate(Sender: TObject);
    procedure actRunBeforeExecute(Sender: TObject);
    procedure EditEnter(Sender: TObject);
    procedure actDefaultExecute(Sender: TObject);
    procedure actCancelExecute(Sender: TObject);
    procedure actfrmWixGeneratorExecute(Sender: TObject);
    procedure actfrmWixGeneratorUpdate(Sender: TObject);
    procedure actMSIExecute(Sender: TObject);
    procedure PopupItemsPopup(Sender: TObject);
    procedure actCopyUpdate(Sender: TObject);
    procedure actCopyExecute(Sender: TObject);
    procedure actMSIUpdate(Sender: TObject);
    procedure actSaveUpdate(Sender: TObject);
    procedure actSaveExecute(Sender: TObject);
  private
    { Private declarations }
    fNormalHeight: integer;
    fInserting: boolean;
    fLastFocusEdit: TComboBox;
    fLastSelStart: integer;
    fLastSelLen: integer;
    ffrmWixGenerator: TfrmWixGenerator;
    fWixGenerator: TCustomWixGenerator;
    fonDefaultAction: TNotifyEvent;
    fMResult: TRichEdit;
    fTimer: TInternalTimer;
    function GetExeName: string;
    procedure SetExeName(const Value: string);
    procedure DoSelecteItem(Edit: TComboBox);
    procedure DoItemClick(Sender: TObject);
    function GetStream: TStream;
    procedure SetStream(const Value: TStream);
    procedure UpdateEnableEdit;
  protected
    procedure DoLoadFromIni(Ini:TIniFile); override;
    procedure DoSaveToIni(Ini:TIniFile); override;
    procedure onAfterExecute(Sender: TObject);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure DoShow; override;
    procedure MakeMSI;
    property ExeName: string read GetExeName write SetExeName;
    procedure CopyTextFromStream;
    procedure ClearText;
    property Stream: TStream read GetStream write SetStream;
    property frmWixGenerator: TfrmWixGenerator read ffrmWixGenerator;
    property onDefaultAction: TNotifyEvent read fonDefaultAction write fonDefaultAction;
  end;

implementation

{$R *.dfm}

procedure TFormBuild.DoSelecteItem(Edit: TComboBox);
var X, X1, X2, I: integer;
    M, SS: string;
begin
  fLastFocusEdit := Edit;
  X := Edit.SelStart;
  X1 := X + 1;
  X2 := X1 + Edit.SelLength;
  SS := UpperCase(Edit.Text);
  while (X1 > 0) and (X1 <= Length(SS)) and (SS[X1] <> ' ') do dec(X1);
  while (X2 > 0) and (X2 <= Length(SS)) and (SS[X2] <> ' ') do inc(X2);
  if (X2 > 0) and (X2 <= Length(SS)) and (SS[X2] = ' ') then dec(X2);
  fLastSelStart := X1 - 1;
  fLastSelLen := X2 - X1;
  M := Copy(SS, X1, X2 - 1);
  fLastSelStart := Edit.SelStart;
  fLastSelLen := Edit.SelLength;
  for I := 0 to PopupItems.Items.Count - 1 do
  begin
    if Trim(M) = Macros[PopupItems.Items[I].Tag] then
    begin
      fLastSelStart := X1 - 1;
      fLastSelLen := X2 - X1;
    end;
  end;
end;

procedure TFormBuild.actInsertExecute(Sender: TObject);
var P: TPoint;
begin
  if fLastFocusEdit <> nil then
  begin
    P.x := 0;
    P.y := fLastFocusEdit.Height;
    P := fLastFocusEdit.ClientToScreen(P);
    PopupItems.Popup(P.x, P.y);
  end;
end;

procedure TFormBuild.PopupItemsPopup(Sender: TObject);
var I: integer;
begin
  fLastFocusEdit := nil;
  if (actRunAfter.Checked) and ((eAfter.Focused)) then
  begin
    DoSelecteItem(eAfter);
  end
  else
  if (actRunBefore.Checked) and ((eBefore.Focused)) then
  begin
    DoSelecteItem(eBefore);
  end;
  for I := 0 to PopupItems.Items.Count - 1 do
  begin
    PopupItems.Items[I].Enabled := fLastFocusEdit <> nil;
  end;
  inherited;
end;

procedure TFormBuild.DoItemClick(Sender: TObject);
var S: string;
begin
  if (Sender is TMenuItem) and (fLastFocusEdit <> nil) then
  begin
    S := fLastFocusEdit.Text;
    Delete(S, fLastSelStart + 1, fLastSelLen);
    Insert(' ' + Macros[TMenuItem(Sender).Tag], S, fLastSelStart + 1);
    fLastFocusEdit.Text := S;
    fLastFocusEdit.SelStart := fLastSelStart;
    fLastFocusEdit.SelLength := Length(Macros[TMenuItem(Sender).Tag]) + 1;
  end;
end;

procedure TFormBuild.actInsertUpdate(Sender: TObject);
begin
  actInsert.Enabled := actRunAfter.Checked or actRunBefore.Checked;
  actInsert.Visible := actRunAfter.Checked or actRunBefore.Checked;
end;

constructor TFormBuild.Create(AOwner: TComponent);
var I: integer;
    Item: TMenuItem;
begin
  inherited;
  DoubleBuffered := True;
  fTimer := TInternalTimer.Create(self);
  fWixGenerator := TCustomWixGenerator.Create(self);
  ffrmWixGenerator := TfrmWixGenerator.Create(self);
  ffrmWixGenerator.WixGenerator := fWixGenerator;
  ffrmWixGenerator.LoadFromIni;
  fNormalHeight := Height;
  PopupItems.Items.Clear;
  for I := Low(Macros) to High(Macros) do
  begin
    Item := TMenuItem.Create(self);
    Item.Caption := Macros[I];
    Item.Tag := I;
    Item.OnClick := DoItemClick;
    PopupItems.Items.Add(Item);
  end;
  fMResult := TRichEdit.Create(self);
  fMResult.ReadOnly := True;
  fMResult.Font.Name := 'Courier new';
  fMResult.Font.Size := 8;
  fMResult.Font.Color := clLime;
  fmResult.PlainText := True;
  fmResult.ScrollBars := ssVertical;
  fMResult.Parent := self;
  fMResult.Anchors := [akTop, akLeft, akRight, akBottom];
  fMResult.Color := clBlack;
end;

procedure TFormBuild.ListBox1DblClick(Sender: TObject);
begin
  actInsert.Execute;
end;

procedure TFormBuild.UpdateEnableEdit;
begin
  eAfter.Enabled := actRunAfter.Checked;
  eBefore.Enabled := actRunBefore.Checked;
end;

procedure TFormBuild.DoLoadFromIni(Ini: TIniFile);
  procedure LoadCB(Edit: TComboBox);
  begin
    Edit.Text := Ini.ReadString(Name, Edit.Name, Edit.Text);
    LoadItems(Ini, Name, Edit.Name, Edit.Items);
  end;
begin
  inherited;
  actRunAfter.Checked := Ini.ReadBool(Name, 'Run', false);
  actRunBefore.Checked := Ini.ReadBool(Name, 'RunBefore', false);
  actMSI.Checked := Ini.ReadBool(Name, 'MSI', false);
  if not actMSI.Checked then
    actSave.Checked := Ini.ReadBool(Name, 'SAVE', false);
  cbRun.Checked := Ini.ReadBool(Name, cbRun.Name, cbRun.Checked);
  LoadCB(eAfter);
  LoadCB(eBefore);
  UpdateEnableEdit;
end;

procedure TFormBuild.DoSaveToIni(Ini: TIniFile);
  procedure SaveCB(Edit: TComboBox);
  var i: integer;
      S: string;
      Found: boolean;
  begin
    S := Trim(Edit.Text);
    Found := False;
    if S <> '' then
      for I := 0 to Edit.Items.Count - 1 do
        if Trim(Edit.Items[I]) = S then
        begin
          Edit.Items.Delete(I);
          Edit.Items.Insert(0, S);
          Edit.ItemIndex := 0;
          Found := True;
          Break;
        end;
    if not Found then
    begin
      Edit.Items.Insert(0, S);
      Edit.ItemIndex := 0;
    end;
    Ini.WriteString(Name, Edit.Name, S);
    SaveItems(Ini, Name, Edit.Name, Edit.Items);
  end;
begin
  inherited;
  Ini.WriteBool(Name, 'Run', actRunAfter.Checked);
  Ini.WriteBool(Name, 'RunBefore', actRunBefore.Checked);
  Ini.WriteBool(Name, 'MSI', actMSI.Checked);
  if not actMSI.Checked then
    Ini.WriteBool(Name, 'SAVE', actSave.Checked);
  Ini.WriteBool(Name, cbRun.Name, cbRun.Checked);
  SaveCB(eAfter);
  SaveCB(eBefore);
end;

procedure TFormBuild.EditEnter(Sender: TObject);
begin
  if (Sender is TComboBox) and (not fInserting) then
  begin
    fLastFocusEdit := TComboBox(Sender);
    fLastSelStart := TComboBox(Sender).SelStart;
    fLastSelLen := TComboBox(Sender).SelLength;
  end;
end;

procedure TFormBuild.actRunAfterExecute(Sender: TObject);
begin
  actRunAfter.Checked := not actRunAfter.Checked;
  UpdateEnableEdit;
end;

procedure TFormBuild.actRunBeforeExecute(Sender: TObject);
begin
  actRunBefore.Checked := not actRunBefore.Checked;
  UpdateEnableEdit;
end;


procedure TFormBuild.actDefaultExecute(Sender: TObject);
begin
  inherited;
  if fMResult.Enabled then
    fMResult.SetFocus;
  if @fonDefaultAction <> nil then
    fonDefaultAction(self);
  if cbRun.Checked then
  begin
    Self.Close;
    ModalResult := IdOk;
  end
  else
  begin
    fTimer.ExecuteDelay(1000, onAfterExecute);
  end;
end;

procedure TFormBuild.actCancelExecute(Sender: TObject);
begin
  inherited;
  close;
  ModalResult := IdCancel;
end;

procedure TFormBuild.actfrmWixGeneratorExecute(Sender: TObject);
begin
  inherited;
  ffrmWixGenerator.Visible := not ffrmWixGenerator.Visible;
end;

procedure TFormBuild.actfrmWixGeneratorUpdate(Sender: TObject);
begin
  inherited;
  actfrmWixGenerator.Checked := (ffrmWixGenerator <> nil) and (ffrmWixGenerator.Visible);
end;

procedure TFormBuild.DoShow;
var Top: integer;
begin
  inherited;
  Top := cbRun.Top + cbRun.Height + 4;
  eBefore.Width := SpeedButton1.Left - BorderWidth - eBefore.Left;
  eAfter.Width := SpeedButton1.Left - BorderWidth - eAfter.Left;
  fMResult.SetBounds(eAfter.Left, Top, eAfter.Width, ClientHeight - Top);
  sbcMSI.Left := ClientWidth - sbcMSI.Width;
  sbcMSI.Top := SpeedButton2.Top + SpeedButton2.Height + Max(BorderWidth, 4);
  sbCopy.Top := ClientHeight - sbCopy.Height;
  sbCopy.Left := ClientWidth - sbCopy.Width;
end;

procedure TFormBuild.actMSIExecute(Sender: TObject);
begin
  inherited;
  actMSI.Checked := not actMSI.Checked;
end;

function TFormBuild.GetExeName: string;
begin
  result := fWixGenerator.ExeName;
end;

procedure TFormBuild.SetExeName(const Value: string);
begin
  if fWixGenerator.ExeName <> Value then
  begin
    fWixGenerator.ExeName := Value;
    if FileExists(ChangeFileExt(Value, '.' + DefaultFileExt)) then
    begin
      fWixGenerator.RefreshByWix(True);
      ffrmWixGenerator.UpdateProps;
    end
    else
    begin
      ffrmWixGenerator.WixGenerator := nil;
      ffrmWixGenerator.UpdateProps;
      ffrmWixGenerator.WixGenerator := fWixGenerator;
    end;
  end;
end;

procedure TFormBuild.MakeMSI;
begin
  if (actMSI.Enabled) and (actMSI.Checked) then
  begin
    //WixGenerator.Strartup := cbStrartup.Checked;
    //WixGenerator.Midas := cbStrartup.Checked;
    ffrmWixGenerator.UpdateProps;
    ffrmWixGenerator.actMake.Update;
    ffrmWixGenerator.actMake.Execute;
  end;
end;


procedure TFormBuild.CopyTextFromStream;
var OldPos: int64;
    WS: WideString;
begin
  if Stream <> nil then
  begin
    OldPos := Stream.Position;
    try
      Stream.Position := 0;
      SetLength(WS, Stream.Size div SizeOf(WS[1]));
      Stream.Position := 0;
      Stream.Read(WS[1], Stream.Size);
      if (WS <> '') and (Word(WS[1]) = 65279) then
        Delete(WS, 1, 1);
      fmResult.TextW := WS;
      fmResult.Repaint;
      invalidate;
      if Application <> nil then
        Application.ProcessMessages;
      fmResult.SelStart := Length(fmResult.TextW) - 2;
      fmResult.SelLength := 0;
    finally
      Stream.Position := OldPos;
    end;
  end;
end;

procedure TFormBuild.ClearText;
begin
  fmResult.TextW := '';
end;

function TFormBuild.GetStream: TStream;
begin
  result := ffrmWixGenerator.Stream;
end;

procedure TFormBuild.SetStream(const Value: TStream);
begin
  ffrmWixGenerator.Stream := Value;
end;

procedure TFormBuild.onAfterExecute(Sender: TObject);
begin
  if Visible then
  begin
    SendToBack;
    BringToFront;
  end;
end;

procedure TFormBuild.actCopyUpdate(Sender: TObject);
begin
  inherited;
  actCopy.Enabled := fMResult.TextW <> '';
end;

procedure TFormBuild.actCopyExecute(Sender: TObject);
begin
  inherited;
  fMResult.CopyAll;
end;

procedure TFormBuild.actMSIUpdate(Sender: TObject);
begin
  inherited;
  actMSI.Enabled := ffrmWixGenerator.WixSource <> '';
end;

procedure TFormBuild.actSaveUpdate(Sender: TObject);
begin
  inherited;
  actSave.Enabled := (not actMSI.Enabled) or (not actMSI.Checked);
  if not actSave.Enabled then actSave.Checked := True;
end;

procedure TFormBuild.actSaveExecute(Sender: TObject);
begin
  inherited;
  Invalidate;
end;

end.
