unit uBaseForm;

interface

{$I Directives.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, ExtCtrls, ComCtrls, RichEdit, ComStrs,
  {$IFDEF DELPHI_9_UP}{$ELSE}GraphUtilEx, {$ENDIF}
  IniFiles, ActnList, ImgList, ImageListCep, SpeedButtonCep;
const
  SettingsFileName = 'CepSettings.ini';
  cLeft = 'Left';
  cTop = 'Top';
  cWidth = 'Width';
  cHeight = 'Height';
  cVisible = 'Visible';

  RTFformat: WideString = 'Rich Text Format';

type
  TClipboardFormats = array[0..2] of integer;

  TRichEdit = class(ComCtrls.TRichEdit)
  private
    procedure SaveToStream(Stream: TStream; TextType: Integer);
    procedure LoadFromStream(Stream: TStream; TextType: Integer);
    function GetTextW: WideString;
    procedure SetTextW(const Value: WideString);
  public
    function IsEnablePaste(var Formats: TClipboardFormats): integer;
    procedure CopyAll;
    procedure PasteAll;
    property TextW: WideString read GetTextW write SetTextW;
  end;


  TBaseForm = class(TForm)
    ActionList1: TActionList;
    actDefault: TAction;
    actCancel: TAction;
    Timer1: TTimer;
    SpeedButton2: TSpeedButtonCep;
    SpeedButton1: TSpeedButtonCep;
    ImageList1: TImageListCep;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure actCancelExecute(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure actCancelUpdate(Sender: TObject);
    procedure actDefaultUpdate(Sender: TObject);
  private
    fIniFileName: string;
    fWait: TObject;
  protected
    procedure SetIniFileName(const Value: string); virtual;
    procedure DoSaveToIni(Ini: TIniFile); virtual;
    procedure DoLoadFromIni(Ini: TIniFile); virtual;
    procedure LoadItems(Ini: TIniFile; SectionName, ValuesName: string; Items: TStrings);
    procedure SaveItems(Ini: TIniFile; SectionName, ValuesName: string; Items: TStrings);
    { Private declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property IniFileName: string read fIniFileName write SetIniFileName;

    procedure WriteRect(Ini: TIniFile; Form: TForm);
    function ReadRect(Ini: TIniFile; Form: TForm): TRect;
    procedure BeginWait;
    procedure StopWait;

    procedure SaveToIni;
    procedure LoadFromIni;

    procedure DoShow; override;

  end;

function DefaultIniFileName(Name: string): string;

implementation
uses SHFolder, Math, WindowThread, CWindowThread;

{$R *.dfm}

const
  ReadError = $0001;
  WriteError = $0002;
  NoError = $0000;

constructor TBaseForm.Create(AOwner: TComponent);
begin
  inherited;
  fIniFileName := DefaultIniFileName('');
  ImageList1.UpdateImages;
end;

destructor TBaseForm.Destroy;
begin
  FreeAndNil(fWait);
  inherited;
end;

procedure TBaseForm.SaveItems(Ini: TIniFile; SectionName, ValuesName: string; Items: TStrings);
var i: integer;
    S: string;
begin
  for i := 0 to 100 do
  begin
    S := inttostr(I);
    while Length(S) < 3 do S := '0' + S;
    S := AnsiLowerCase(ValuesName + S);
    Ini.DeleteKey(SectionName, S);
  end;
  i := 0;
  while (i < Items.Count) and (i < 99) do
  begin
    inc(i);
    S := inttostr(I);
    while Length(S) < 3 do S := '0' + S;
    S := AnsiLowerCase(ValuesName + S);
    Ini.WriteString(SectionName, S, Items[I - 1]);
  end;
end;

procedure TBaseForm.LoadItems(Ini: TIniFile; SectionName, ValuesName: string; Items: TStrings);
var List: TStringList;
    i, j: integer;
    S: string;
begin
  Items.Clear;
  List := TStringList.Create;
  try
    Ini.ReadSection(SectionName, List);
    for i := 1 to 99 do
    begin
      S := inttostr(I);
      while Length(S) < 3 do S := '0' + S;
      S := AnsiLowerCase(ValuesName + S);
      for j := 0 to List.Count - 1 do
        if AnsiLowerCase(List[j]) = S then
        begin
          S := Ini.ReadString(SectionName, List[j], '');
          if S <> '' then
          begin
            Items.Add(S);
            Break;
          end;
        end;
    end;
  finally
    List.Free;
  end;
end;

procedure TBaseForm.DoLoadFromIni(Ini: TIniFile);
begin

end;

procedure TBaseForm.DoSaveToIni(Ini: TIniFile);
begin

end;

procedure TBaseForm.DoShow;
begin
  inherited;
  LoadFromIni;
  SpeedButton1.Left := ClientWidth - SpeedButton1.Width;
  SpeedButton2.Left := ClientWidth - SpeedButton2.Width;
  SpeedButton2.Top := SpeedButton1.Top + SpeedButton1.Height + Max(BorderWidth, 4);
end;

procedure TBaseForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if (self.Visible) and (fsModal in FormState) then
  begin
    SaveToIni;
  end;
end;

function DefaultIniFileName(Name: string): string;
  function DefaultSaveLocation: string;
  var
    P: PChar;
    Size: integer;
  begin
    P := nil;
    try
      Size := MAX_PATH * SizeOf(P[0]);
      P := AllocMem(Size);
      FillChar(P^, Size, 0);
      if SHGetFolderPath(0 {Handle}, CSIDL_COMMON_APPDATA or CSIDL_FLAG_CREATE, 0, 0, P) = S_OK then
        Result := P
      else
        Result := ExtractFilePath(ParamStr(0));
    finally
      FreeMem(P);
    end;
  end;
begin
  Result := Name;
  if Result = '' then Result := SettingsFileName;
  {$IFDEF DELPHI_6_UP}
  Result := IncludeTrailingPathDelimiter(DefaultSaveLocation) + Result;
  {$ELSE}
  if (DefaultSaveLocation <> '') and
    (DefaultSaveLocation[Length(DefaultSaveLocation)] <> '\') then
    Result := '\' + Result;
  Result := DefaultSaveLocation + Result;
  {$ENDIF}
  Result := ChangeFileExt(Result, '.ini');
end;


procedure TBaseForm.BeginWait;
const
  Width = 250;
  Height = 48;
var
  R: TRect;
  FontInfo: tagLOGFONT;
  HFont: THandle;
begin
  if fWait = nil then
  begin
    fWait := TThreadWindow.Show('');
    with TThreadWindow(fWait) do
    begin
      AlphaShow := true;
      FontInfo := GetFontIndirect(0, dStatusFont);
      HFont := CreateFontIndirect(FontInfo);
      try
        R := BoundsRect;
        R.Left := (Screen.Width - Width) div (2);
        R.Top := (Screen.Height - Height) * 2 div (5);
        R.Bottom := R.Top + Height;
        R.Right := R.Left + Width;
        Font := HFont;
        BoundsRect := R;
      finally
        DeleteObject(HFont);
      end;
    end;
  end;
end;

procedure TBaseForm.StopWait;
begin
  FreeAndNil(fWait);
end;

procedure TBaseForm.WriteRect(Ini: TIniFile; Form: TForm);
var
  R: TRect;
begin
  with Form do
  begin
    R := BoundsRect;
    Ini.WriteInteger(Name, cLeft, R.Left);
    Ini.WriteInteger(Name, cTop, R.Top);
    Ini.WriteInteger(Name, cWidth, R.Right - R.Left);
    Ini.WriteInteger(Name, cHeight, R.Bottom - R.Top);
  end;
end;

function TBaseForm.ReadRect(Ini: TIniFile; Form: TForm): TRect;
var
  W, H: integer;
  R: TRect;
begin
  with Form do
  begin
    R.Left := Ini.ReadInteger(Name, cLeft, -1);
    R.Top := Ini.ReadInteger(Name, cTop, -1);
    W := Max(Ini.ReadInteger(Name, cWidth, Width), Constraints.MinWidth);
    H := Max(Ini.ReadInteger(Name, cHeight, Height), Constraints.MinHeight);
    if (R.Left = -1) or (R.Top = -1) then
    begin
      R.Left := (Screen.Width - W) div (2);
      R.Top := (Screen.Height - H) div (2);
    end;
    R.Right := R.Left + W;
    R.Bottom := R.Top + H;
    BoundsRect := R;
    Result := R;
  end;
end;


procedure TBaseForm.LoadFromIni;
var
  Ini: TIniFile;
begin
  BeginWait;
  try
    try
      Ini := TIniFile.Create(fIniFileName);
      try
        ReadRect(Ini, Self);
        DoLoadFromIni(Ini);
      finally
        FreeAndNil(Ini);
      end;
    except
      //    ,     
    end;
  finally
    StopWait;
  end;
end;

procedure TBaseForm.SaveToIni;
var
  Ini: TIniFile;
begin
  BeginWait;
  try
    try
      Ini := TIniFile.Create(fIniFileName);
      try
        WriteRect(Ini, Self);
        DoSaveToIni(Ini);
      finally
        FreeAndNil(Ini);
      end;
    except
      //    ,     
    end;
  finally
    StopWait;
  end;
end;

procedure TBaseForm.SetIniFileName(const Value: string);
begin
  fIniFileName := Value;
end;

{ TRichEdit }

function StreamSave(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  Buffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  try
    pcb := 0;
    if StreamInfo^.Converter <> nil then
      {$IFDEF DELPHI_2009_UP}
    begin
      SetLength(Buffer, cb);
      Move(pbBuff^, Buffer[0], cb);
      pcb := StreamInfo^.Converter.ConvertWriteStream(StreamInfo^.Stream, Buffer, cb);
    end;
    {$ELSE}
      pcb := StreamInfo^.Converter.ConvertWriteStream(StreamInfo^.Stream, PChar(pbBuff), cb);
    {$ENDIF}
  except
    Result := WriteError;
  end;
end;

//     ComCtrls

function AdjustLineBreaks(Dest, Source: PAnsiChar): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ESI,EDX
        MOV     EDX,EAX
        CLD
@@1:    LODSB
@@2:    OR      AL,AL
        JE      @@4
        CMP     AL,0AH
        JE      @@3
        STOSB
        CMP     AL,0DH
        JNE     @@1
        MOV     AL,0AH
        STOSB
        LODSB
        CMP     AL,0AH
        JE      @@1
        JMP     @@2
@@3:    MOV     EAX,0A0DH
        STOSW
        JMP     @@1
@@4:    STOSB
        LEA     EAX,[EDI-1]
        SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;

//  ,      PWideChar

function AdjustLineBreaksW(Dest, Source: PWideChar): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ESI,EDX
        MOV     EDX,EAX
        CLD
@@1:    LODSW
@@2:    OR      AX,AX
        JE      @@4
        CMP     AX,000AH
        JE      @@3
        STOSW
        CMP     AX,000DH
        JNE     @@1
        MOV     AX,000AH
        STOSW
        LODSW
        CMP     AX,000AH
        JE      @@1
        JMP     @@2
@@3:    MOV     EAX,000A000DH
        STOSD
        JMP     @@1
@@4:    STOSW
        LEA     EAX,[EDI-2]
        SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;


//     

function StreamLoad(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  Buffer, pBuff: PAnsiChar;
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  TmpBuffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  {$IFDEF DELPHI_2009_UP}
  Buffer := AnsiStrAlloc(cb + 1);
  {$ELSE}
  Buffer := StrAlloc(cb + 1);
  {$ENDIF}
  try
    cb := cb div 2;
    pcb := 0;
    pBuff := Buffer + cb;
    try
      if StreamInfo^.Converter <> nil then
        {$IFDEF DELPHI_2009_UP}
      begin
        SetLength(TmpBuffer, cb);
        FillChar(pBuff^, cb, 0);
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, TmpBuffer, cb);
        Move(TmpBuffer[0], pBuff^, pcb);
      end;
      {$ELSE}
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, pBuff, cb);
      {$ENDIF}
      if pcb > 0 then
      begin
        pBuff[pcb] := #0;
        if pBuff[pcb - 1] = #13 then
          pBuff[pcb - 1] := #0;
        pcb := AdjustLineBreaks(Buffer, pBuff);
        Move(Buffer^, pbBuff^, pcb);
      end;
    except
      Result := ReadError;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

//  ,   WideString

function StreamLoadW(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  Buffer, pBuff: PAnsiChar;
  StreamInfo: PRichEditStreamInfo;
  {$IFDEF DELPHI_2009_UP}
  TmpBuffer: TConversionBuffer;
  {$ENDIF}
begin
  Result := NoError;
  StreamInfo := PRichEditStreamInfo(Pointer(dwCookie));
  {$IFDEF DELPHI_2009_UP}
  Buffer := AnsiStrAlloc((cb + 1) * 2);
  {$ELSE}
  Buffer := StrAlloc((cb + 1) * 2);
  {$ENDIF}
  try
    cb := cb;
    pcb := 0;
    pBuff := Buffer + cb;
    try
      if StreamInfo^.Converter <> nil then
        {$IFDEF DELPHI_2009_UP}
      begin
        SetLength(TmpBuffer, cb);
        FillChar(pBuff^, cb, 0);
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, TmpBuffer, cb);
        Move(TmpBuffer[0], pBuff^, pcb);
      end;
      {$ELSE}
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, pBuff, cb);
      {$ENDIF}
      if pcb > 0 then
      begin
        PWideChar(pBuff)[(pcb) div (2)] := #0;
        if PWideChar(pBuff)[(pcb div 2) - 1] = #13 then
          PWideChar(pBuff)[(pcb div 2) - 1] := #0;
        pcb := AdjustLineBreaksW(PWideChar(Buffer), PWideChar(pBuff));
        Move(Buffer^, pbBuff^, pcb);
      end;
    except
      Result := ReadError;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

procedure TRichEdit.CopyAll;
var
  uFormat: integer; //  RTF
  Stream: TMemoryStream;
  hMem: THandle;
  P: pointer;
  procedure CopyByFormat(TextType, cbFormat: Longint);
  begin
    if (cbFormat = 0) or (cbFormat = 0) then
      Exit;
    Stream.Clear;
    SaveToStream(Stream, TextType);
    Stream.Position := 0;
    hMem := GlobalAlloc(GMEM_DDESHARE or GMEM_MOVEABLE, Stream.Size + 2);
    try
      P := GlobalLock(hMem);
      FillChar(P^, Stream.Size + 2, 0);
      CopyMemory(P, Stream.Memory, Stream.Size);
    finally
      GlobalUnlock(hMem);
    end;
    SetClipboardData(cbFormat, hMem);
  end;
begin
  uFormat := RegisterClipboardFormatW(PWideChar(RTFformat));
  if OpenClipBoard(Handle) then
  begin
    try
      EmptyClipboard;
      Stream := TMemoryStream.Create;
      try
        CopyByFormat(SF_TEXT or SF_UNICODE, CF_UNICODETEXT);
        CopyByFormat(SF_RTFNOOBJS, uFormat);
        CopyByFormat(SF_TEXT, CF_TEXT);
      finally
        FreeAndNil(Stream);
      end;
    finally
      CloseClipboard;
    end;
  end
  else
    RaiseLastOSError;
end;

function TRichEdit.GetTextW: WideString;
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    SaveToStream(Stream, SF_TEXT or SF_UNICODE);
    result := PWideChar(Stream.Memory);
    SetLength(result, (Stream.Size) div (SizeOf(Result[1])));
    Stream.Clear;
  finally
    FreeAndNil(Stream);
  end;
end;

function TRichEdit.IsEnablePaste(var Formats: TClipboardFormats): integer;
var
  F, I: integer;
  Buf: array[0..255] of WideChar;
  wS: WideString;
begin
  result := 0;
  FillChar(Formats, SizeOf(Formats), 0);
  F := EnumClipboardFormats(0);
  while F <> 0 do
  begin
    if F = CF_TEXT then
      Formats[2] := F;
    if F = CF_UNICODETEXT then
      Formats[0] := F;
    if (GetClipboardFormatNameW(F, PWideChar(@Buf), Length(Buf)) > 0) then
    begin
      wS := WideString(PWideChar(@Buf));
      if wS = RTFformat then
        Formats[1] := F;
    end;
    F := EnumClipboardFormats(F);
  end;
  for I := Low(Formats) to High(Formats) do
    if Formats[I] <> 0 then
    begin
      result := Formats[I];
      break;
    end;
end;

procedure TRichEdit.LoadFromStream(Stream: TStream; TextType: Integer);
var
  EditStream: TEditStream;
  StreamInfo: TRichEditStreamInfo;
  Converter: TConversion;
begin
  Converter := DefaultConverter.Create;
  StreamInfo.Stream := Stream;
  StreamInfo.Converter := Converter;
  try
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      if (TextType and SF_UNICODE) = SF_UNICODE then
        pfnCallBack := @StreamLoadW
      else
        pfnCallBack := @StreamLoad;
      dwError := 0;
    end;
    SendMessage(Handle, EM_STREAMIN, TextType, Longint(@EditStream));
    if EditStream.dwError <> 0 then
      raise EOutOfResources.Create(sRichEditLoadFail);
  finally
    Converter.Free;
  end;
end;

procedure TRichEdit.PasteAll;
var
  Formats: TClipboardFormats;
  cbFormat: integer;
  Stream: TMemoryStream;
  hMem: THandle;
  P: pointer;
  Size: DWORD;
  procedure CopyByFormat(TextType, cbFormat: Longint);
  begin
    if (cbFormat = 0) or (cbFormat = 0) then
      Exit;
    Stream.Clear;
    hMem := GetClipboardData(cbFormat);
    if hMem <> 0 then
    begin
      P := GlobalLock(hMem);
      try
        Size := GlobalSize(hMem);
        Stream.WriteBuffer(P^, Size);
        Stream.Position := 0;
      finally
        GlobalUnlock(hMem);
      end;
      LoadFromStream(Stream, TextType);
    end;
  end;
begin
  if OpenClipBoard(Handle) then
  begin
    try
      cbFormat := IsEnablePaste(Formats);
      if cbFormat <> 0 then
      begin
        Stream := TMemoryStream.Create;
        try
          case cbFormat of
            CF_UNICODETEXT: CopyByFormat(SF_TEXT or SF_UNICODE, cbFormat);
            CF_TEXT: CopyByFormat(SF_TEXT, cbFormat);
          else
            CopyByFormat(SF_RTF, cbFormat);
          end;
        finally
          FreeAndNil(Stream);
        end;
      end;
    finally
      CloseClipboard;
    end;
  end
  else
    RaiseLastOSError;
end;

procedure TRichEdit.SaveToStream(Stream: TStream; TextType: Integer);
var
  EditStream: TEditStream;
  StreamInfo: TRichEditStreamInfo;
  Converter: TConversion;
begin
  Converter := DefaultConverter.Create;
  try
    StreamInfo.Stream := Stream;
    StreamInfo.Converter := Converter;
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      pfnCallBack := @StreamSave;
      dwError := 0;
    end;
    SendMessage(Handle, EM_STREAMOUT, TextType, LongInt(@EditStream));
    if EditStream.dwError <> 0 then
      raise EOutOfResources.Create(sRichEditSaveFail);
  finally
    Converter.Free;
  end;
end;

procedure TRichEdit.SetTextW(const Value: WideString);
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    Stream.WriteBuffer(Value[1], (Length(Value) + 1) * SizeOf(Value[1]));
    Stream.Position := 0;
    LoadFromStream(Stream, SF_TEXT or SF_UNICODE);
    Stream.Clear;
  finally
    FreeAndNil(Stream);
  end;
end;

procedure TBaseForm.FormActivate(Sender: TObject);
begin
  actDefault.Update;
  actCancel.Update;
end;

procedure TBaseForm.actCancelExecute(Sender: TObject);
begin
  if fsModal in FormState then Close;
  ModalResult := mrCancel;
end;

procedure TBaseForm.actCancelUpdate(Sender: TObject);
begin
  if (Sender is TAction) then
    TAction(Sender).Visible := fsModal in FormState;
end;

procedure TBaseForm.actDefaultUpdate(Sender: TObject);
begin
  if (Sender is TAction) then
    TAction(Sender).Visible := fsModal in FormState;
end;

end.
