unit uEditActionTreeView;

interface
{$I Directives.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ActionTreeView, ExtCtrls,
  ActnList,
  {$IFDEF DELPHI_6_UP}
  ActnMan, GraphUtil,
  {$ENDIF}
  ImgList,
  ImageListCep, InternalTimer, ToolWin, SpeedButtonCep, Mask, Menus;
const
  CenterWidth = 8;
type
  TSourceActionTreeView = class (TActionTreeView)
  protected
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean); override;
    function NodeClientsItemClass: TNodeClientClass; override;
    function ActionNodeClass: TActionNodeClass; override;
    procedure KeyPress(var Key: Char); override;
    procedure DblClick; override;
  end;

  TSourceNodeClient = class (TNodeClient)
  private
    fStateIndex: integer;
    fIsFound: boolean;
  protected
    function ItemClass: TNodeClientClass; override;
  public
    property StateIndex: integer read fStateIndex write fStateIndex;
    property IsFound: boolean read fIsFound write fIsFound;
  end;

  TSourceActionNode = class (TActionNode)
  private
  public
    procedure DrawImage(TreeView: TActionTreeView; State: TCustomDrawState;
      Image: TCustomImageList; Index: integer; R: TRect); override;
  end;

  TResultActionTreeView = class (TActionTreeView)
  private
    fIntTimer: TInternalTimer;
    fOnClientsChanged: TNotifyEvent;
    procedure DoDelayChanged;
  protected
    procedure IntTimerProc(Sender: TObject);
    function NodeClientsItemClass: TNodeClientClass; override;
    function ActionNodeClass: TActionNodeClass; override;
    procedure KeyPress(var Key: Char); override;
    procedure DblClick; override;
  public
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property OnClientsChanged: TNotifyEvent read fOnClientsChanged write fOnClientsChanged;
  end;

  TResultNodeClient = class (TNodeClient)
  private
    fStateIndex: integer;
  protected
    procedure AssignValues(Source: TPersistent); override;
    function ItemClass: TNodeClientClass; override;
  public
    property StateIndex: integer read fStateIndex write fStateIndex;
  end;

  TResultActionNode = class (TActionNode)
  private
  public
    procedure DrawImage(TreeView: TActionTreeView; State: TCustomDrawState;
      Image: TCustomImageList; Index: integer; R: TRect); override;
  end;

  TImageNumber = record
    Images: TCustomImageList;
    CurrIndex, AllIndex: integer;
    Added: boolean;
    Action: TCustomAction;
  end;

  TEditActionTreeView = class(TForm)
    pTop: TPanel;
    pBottom: TPanel;
    StateImages: TImageListCep;
    ButtonImages: TImageListCep;
    ActionList1: TActionList;
    actAdd: TAction;
    actAddChild: TAction;
    actDelSingle: TAction;
    actDelAll: TAction;
    actEnabled: TAction;
    actHide: TAction;
    actExpanded: TAction;
    actChecked: TAction;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    meCaption: TMaskEdit;
    pToolBar: TPanel;
    actCopyFromAction: TAction;
    ToolButton10: TToolButton;
    actCopyToAction: TAction;
    ToolButton11: TToolButton;
    ToolButton12: TToolButton;
    Button1: TButton;
    Button2: TButton;
    cbImages: TComboBox;
    cbImageIndex: TComboBox;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure actAddExecute(Sender: TObject);
    procedure actAddChildExecute(Sender: TObject);
    procedure actDelSingleExecute(Sender: TObject);
    procedure actDelAllExecute(Sender: TObject);
    procedure actEnabledExecute(Sender: TObject);
    procedure actHideExecute(Sender: TObject);
    procedure actExpandedExecute(Sender: TObject);
    procedure actAddUpdate(Sender: TObject);
    procedure actDestNeed(Sender: TObject);
    procedure actEnabledUpdate(Sender: TObject);
    procedure actHideUpdate(Sender: TObject);
    procedure actExpandedUpdate(Sender: TObject);
    procedure actCheckedExecute(Sender: TObject);
    procedure actCheckedUpdate(Sender: TObject);
    procedure meCaptionChange(Sender: TObject);
    procedure actCopyFromActionExecute(Sender: TObject);
    procedure actCopyToActionExecute(Sender: TObject);
    procedure actCopyFromActionUpdate(Sender: TObject);
    procedure actCopyToActionUpdate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure cbImagesChange(Sender: TObject);
    procedure cbImageIndexMeasureItem(Control: TWinControl; Index: Integer;
      var Height: Integer);
    procedure cbImageIndexDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure cbImageIndexChange(Sender: TObject);
  private
    { Private declarations }
    fTreeSource: TSourceActionTreeView;
    fTreeResult: TResultActionTreeView;
    fEnableActionEdit: boolean;
    fAllResultImages: TImageListCep;
    fSourceImageNumbers: array of TImageNumber;
    procedure AllignTop;
    procedure SetActionTreeView(const Value: TActionTreeView);
    function GetTreeResult: TActionTreeView;
    function DestNodeClient: TNodeClient;
    procedure SetEnableActionEdit(const Value: boolean);
    function DestNodeAction: TCustomAction;
    procedure UpdateEditEnabled;
    procedure UpdateSourceImages;
    function ImagesEqual: boolean;
    procedure ResultImagesChange(Sender: TObject);
  protected
    procedure Resize; override;
    procedure CreateWnd; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddSourceActions(List: TStringList);
    procedure AddImages(List: TStringList);
    procedure UpdateSourceNodes(Sender: TObject);
    procedure ResultChange(Sender: TObject; Node: TTreeNode);
    property ActionTreeView: TActionTreeView read GetTreeResult write SetActionTreeView;
    property EnableActionEdit: boolean read fEnableActionEdit write SetEnableActionEdit;
  end;

var EditActionTreeView: TEditActionTreeView;

implementation

{$R *.dfm}

{ TEditActionTreeView }

procedure TEditActionTreeView.cbImageIndexChange(Sender: TObject);
var I, Code: integer;
  procedure SetImageIndex(I: integer);
  begin
    if DestNodeClient.ImageIndex <> I then
    begin
      DestNodeClient.ImageIndex := I;
      fTreeResult.DoDelayChanged;
    end;
  end;
begin
  if (DestNodeClient <> nil) then
  begin
    if (cbImageIndex.ItemIndex >= 0) then
      SetImageIndex(cbImageIndex.ItemIndex - 1)
    else
    begin
      val(cbImageIndex.Text, I, Code);
      if Code = 0 then
        SetImageIndex(I);
    end;
  end;
end;

procedure TEditActionTreeView.cbImageIndexDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);
var S: string;
    R: TRect;
begin
  if Control is TComboBox then
  begin
    if Index >= 0 then S := TComboBox(Control).Items[Index]
                  else S := TComboBox(Control).Text;
    with TComboBox(Control).Canvas do
    begin
      if odSelected in State then
      begin
        Font.Color := TComboBox(Control).Font.Color;
        Brush.Color := fTreeResult.SelectedColor;
        Pen.Color := fTreeResult.BtnFrameColor;
      end
      else
      begin
        Brush.Color := TComboBox(Control).Color;
        Pen.Color := TComboBox(Control).Color;
        Font.Color := TComboBox(Control).Font.Color;
      end;
      Rectangle(Rect);
      Brush.Style := bsClear;

      if Index > 0 then
        fAllResultImages.Draw(TComboBox(Control).Canvas,
                       Rect.Left + 1,
                       Rect.Top,
                       Index - 1);
      R := Rect;
      Dec(R.Right, 3);
      inc(R.Top, (Rect.Bottom - Rect.Top - TextHeight('0')) div 2);
      {$IFDEF DELPHI_2005_UP}
      TextRect(R, S, [tfRight]);
      {$ELSE}
      TextRect(R, R.Right - Canvas.TextWidth(S), R.Top, S);
      {$ENDIF}
      ExcludeClipRect(Handle, Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
    end;
  end;
end;

procedure TEditActionTreeView.cbImageIndexMeasureItem(Control: TWinControl;
  Index: Integer; var Height: Integer);
begin
  if Control is TComboBox then
  begin
    if Index = 0 then
      Height :=  TComboBox(Control).Canvas.TextHeight('0')
    else
      Height := fAllResultImages.Height + 1;
  end;
end;

constructor TEditActionTreeView.Create(AOwner: TComponent);
begin
  inherited;
  fAllResultImages := TImageListCep.Create(self);
  fAllResultImages.Width := 16;
  fAllResultImages.Height := 16;
  fAllResultImages.onChange := ResultImagesChange;
  fTreeSource := TSourceActionTreeView.Create(pTop);
  fTreeSource.IsDesigning := True;
  fTreeSource.ParentColor := True;
  fTreeSource.DragMode := dmAutomatic;
  fTreeResult := TResultActionTreeView.Create(pTop);
  fTreeResult.IsDesigning := True;
  fTreeResult.DragMode := dmAutomatic;
  fTreeResult.OnClientsChanged := UpdateSourceNodes;
  fTreeResult.OnChange := ResultChange;
  meCaption.Text := '';
  StateImages.UpdateImages;
  fTreeSource.StateImages := StateImages;
  AllignTop;
  fTreeSource.Parent := pTop;
  fTreeResult.Parent := pTop;
end;

procedure TEditActionTreeView.CreateWnd;
begin
  {$IFDEF DELPHI_2005_UP}
  ToolBar1.DrawingStyle := ComCtrls.dsGradient;
  ToolBar1.GradientEndColor := Color;
  {$ENDIF}
  meCaption.DoubleBuffered := True;
  inherited;
end;

destructor TEditActionTreeView.Destroy;
var NeedClear: boolean;
begin
  NeedClear := self = EditActionTreeView;
  inherited;
  if NeedClear then
  begin
    EditActionTreeView := nil;
  end;
end;

procedure TEditActionTreeView.FormActivate(Sender: TObject);
  procedure SelectSurrentImages;
  var I: integer;
  begin
    for I := 0 to cbImages.Items.Count - 1 do
    begin
      if cbImages.Items.Objects[I] = fTreeResult.Images then
      begin
        cbImages.ItemIndex := I;
        Exit;
      end;
    end;
    if fTreeResult.Images <> nil then
    begin
      cbImages.Items.AddObject(fTreeResult.Images.Name, fTreeResult.Images);
      cbImages.ItemIndex := cbImages.Items.Count - 1;
    end
    else
      cbImages.ItemIndex := -1;
  end;
begin
  SelectSurrentImages;
  cbImagesChange(nil);
  UpdateEditEnabled;
end;

procedure TEditActionTreeView.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  Action := caFree;
end;


procedure TEditActionTreeView.Resize;
begin
  inherited;
  AllignTop;
end;

procedure TEditActionTreeView.UpdateEditEnabled;
var EditEnabled: boolean;
begin
  EditEnabled := fTreeResult.Enabled and
               ((fTreeResult.Selected) is TResultActionNode) and
                (TResultActionNode(fTreeResult.Selected).NodeClient <> nil)
          { and ((EnableActionEdit) or (TResultActionNode(fTreeResult.Selected).Action = nil))};
  meCaption.ReadOnly := not EditEnabled;
  cbImageIndex.Enabled := EditEnabled;

  cbImages.Visible := cbImages.Items.Count > 0;
  cbImages.Enabled := (cbImages.Items.Count > 1) and fTreeResult.Enabled;

  if EditEnabled then
  begin
    meCaption.Color := clWindow;
    cbImageIndex.Color := clWindow;
    cbImages.Color := clWindow;;
  end
  else
  begin
    meCaption.ParentColor := True;
    cbImageIndex.ParentColor := True;
    cbImages.ParentColor := True;
  end;
end;

procedure TEditActionTreeView.ResultChange(Sender: TObject; Node: TTreeNode);
var I: integer;
begin
  I := -1;
  UpdateEditEnabled;
  if Node is TResultActionNode then
  begin
    meCaption.Text := TResultActionNode(Node).Text;
    if TResultActionNode(Node).NodeClient <> nil then
      I := TResultActionNode(Node).NodeClient.ImageIndex;
  end
  else
  begin
    meCaption.Text := '';
  end;
  inc(I);
  if (I >= 0) and (I < cbImageIndex.Items.Count) then
    cbImageIndex.ItemIndex := I
  else
    cbImageIndex.Text := Inttostr(I);
end;

procedure TEditActionTreeView.ResultImagesChange(Sender: TObject);
begin
  cbImageIndex.Invalidate;
end;

procedure TEditActionTreeView.UpdateSourceImages;
var
  I: Integer;
  J: Integer;
  NeedAdd: Boolean;
  Node: TSourceActionNode;
  B: TBitmap;
  function AddSourceImage(Images: TCustomImageList; Index: Integer): Integer;
  var I: integer;
  begin
    result := StateImages.CustomImg.Count;
    for I := 0 to StateImages.ResourceNames.Count - 1 do
      Result := Result + StateImages.ResourceNames.Items[I].GliphCount;
    if B = nil then
    begin
      B := TBitmap.Create;
      B.Canvas.Brush.Color := self.Color;
    end;
    B.Width := Images.Width;
    B.Height := Images.Height;
    B.Canvas.FillRect(Rect(0, 0, B.Width, B.Height));
    Images.Draw(B.Canvas, 0, 0, Index);
    StateImages.CustomImg.InsertBitmap(StateImages.CustomImg.Count, B);
  end;
begin
  B := nil;
  try
  for I := 0 to fTreeSource.Items.Count - 1 do
    if (fTreeSource.Items[I] is TSourceActionNode) and (TSourceActionNode(fTreeSource.Items[I]).Action <> nil) then
    begin
      NeedAdd := True;
      Node := TSourceActionNode(fTreeSource.Items[I]);
      for J := Length(fSourceImageNumbers) - 1 downto 0 do
      begin
        NeedAdd := fSourceImageNumbers[J].Action <> Node.Action;
        if not NeedAdd then
        begin
          if Node.Action.ImageIndex <> fSourceImageNumbers[J].CurrIndex then
          begin
            NeedAdd := True;
            Move(fSourceImageNumbers[J + 1],
                 fSourceImageNumbers[J],
                 (Length(fSourceImageNumbers) - J - 1) * SizeOf(fSourceImageNumbers[0]));
            SetLength(fSourceImageNumbers, Length(fSourceImageNumbers) -1);
          end
          else
          begin
            TSourceNodeClient(Node.NodeClient).StateIndex := fSourceImageNumbers[J].AllIndex;
            Break;
          end;
        end;
      end;
      if NeedAdd then
      begin
        SetLength(fSourceImageNumbers, Length(fSourceImageNumbers) + 1);
        with fSourceImageNumbers[Length(fSourceImageNumbers) - 1] do
        begin
          Action := Node.Action;
          if (Action.ActionList <> nil) and (Action.ActionList.Images <> nil) then
            Images := Action.ActionList.Images
          else
            Images := nil;
          CurrIndex := Action.ImageIndex;
          AllIndex := -1;
          Added := False;
          if (CurrIndex >= 0) and (Images <> nil) then
          begin
            for J := 0 to Length(fSourceImageNumbers) - 2 do
            if (Images = fSourceImageNumbers[J].Images) and (CurrIndex = fSourceImageNumbers[J].CurrIndex) then
            begin
              AllIndex := fSourceImageNumbers[J].AllIndex;
              Added := fSourceImageNumbers[J].Added;
              Break;
            end;
            if AllIndex = -1 then
            begin
              AllIndex := AddSourceImage(Images, CurrIndex);
              Added := AllIndex >= 0;
            end;
          end;
          TSourceNodeClient(Node.NodeClient).StateIndex := AllIndex;
        end;
      end;
    end;
    //StateImages.UpdateImages;
  finally
    B.Free;
  end;
end;

function TEditActionTreeView.GetTreeResult: TActionTreeView;
begin
  Result := fTreeResult;
end;

procedure TEditActionTreeView.meCaptionChange(Sender: TObject);
begin
  if DestNodeClient <> nil then
  begin
    DestNodeClient.Caption := meCaption.Text;
  end;

end;

procedure TEditActionTreeView.SetActionTreeView(const Value: TActionTreeView);
begin
  fTreeResult.Assign(Value);
end;

procedure TEditActionTreeView.SetEnableActionEdit(const Value: boolean);
begin
  fEnableActionEdit := Value;
end;

procedure TEditActionTreeView.UpdateSourceNodes(Sender: TObject);
var I, J: integer;
    SourceNodeClient, ResultNodeClient: TNodeClient;
    SourceAction, ResultAction: TCustomAction;
    Found: boolean;
begin
  UpdateSourceImages;
  for I := 0 to fTreeSource.Items.Count - 1 do
  begin
    SourceNodeClient := TActionNode(fTreeSource.Items[I]).NodeClient;
    if SourceNodeClient <> nil then
    begin
      SourceAction := SourceNodeClient.Action;
      Found := False;
      if SourceAction <> nil then
      begin
        for J := 0 to fTreeResult.Items.Count - 1 do
        begin
          ResultNodeClient := TActionNode(fTreeResult.Items[J]).NodeClient;
          ResultAction := ResultNodeClient.Action;
          Found := ResultAction = SourceAction;
          if Found then Break;
        end;
        TSourceNodeClient(SourceNodeClient).IsFound := Found;
        if SourceAction <> nil then
        begin
          TSourceNodeClient(SourceNodeClient).Caption := '  ' +
            SourceAction.Name + ' (' + Copy(SourceAction.Caption, 1, 32) + ')';
        end;
      end;
    end;
  end;
  fTreeSource.Invalidate;
end;

function TEditActionTreeView.DestNodeClient: TNodeClient;
begin
  Result := nil;
  if fTreeResult.Selected <> nil then
    Result := TActionNode(fTreeResult.Selected).NodeClient;
end;

function TEditActionTreeView.DestNodeAction: TCustomAction;
begin
  Result := nil;
  if (fTreeResult.Selected <> nil) and
     (TActionNode(fTreeResult.Selected).NodeClient <> nil) then
    Result := TActionNode(fTreeResult.Selected).NodeClient.Action;
end;


procedure TEditActionTreeView.actAddChildExecute(Sender: TObject);
var
  Dest: TNodeClient;
begin
  if DestNodeClient <> nil then
  begin
    Dest := TNodeClient(DestNodeClient.Items.Add);
    if Dest.ActionNode <> nil then
    begin
      Dest.ActionNode.Selected := True;
      meCaption.SetFocus;
    end;
  end;
end;

procedure TEditActionTreeView.actDestNeed(Sender: TObject);
begin
  if Sender is TAction then
    TAction(Sender).Enabled := fTreeResult.Enabled and (DestNodeClient <> nil);
end;

procedure TEditActionTreeView.actAddExecute(Sender: TObject);
var Dest: TNodeClient;
begin
  if DestNodeClient <> nil then
  begin
    Dest := TNodeClient(DestNodeClient.Collection.Insert(DestNodeClient.Index + 1));
  end
  else
  begin
    Dest := TNodeClient(fTreeResult.NodeClients.Add);
  end;
  if Dest.ActionNode <> nil then
  begin
    Dest.ActionNode.Selected := True;
    meCaption.SetFocus;
  end;
end;

procedure TEditActionTreeView.actAddUpdate(Sender: TObject);
begin
  if Sender is TAction then
    TAction(Sender).Enabled := fTreeResult.Enabled;
end;

procedure TEditActionTreeView.actCheckedExecute(Sender: TObject);
begin
  if DestNodeClient <> nil then
    DestNodeClient.Checked := not DestNodeClient.Checked;
end;

procedure TEditActionTreeView.actCheckedUpdate(Sender: TObject);
begin
  if Sender is TAction then
  begin
    TAction(Sender).Enabled := fTreeResult.Enabled and (DestNodeClient <> nil) and
                               (fEnableActionEdit or (DestNodeAction = nil));
    if (DestNodeClient <> nil) and DestNodeClient.Checked then
      TAction(Sender).ImageIndex := 11
    else
      TAction(Sender).ImageIndex := 10;
  end;
end;

procedure TEditActionTreeView.actCopyFromActionExecute(Sender: TObject);
begin
  DestNodeClient.Caption := DestNodeAction.Caption;
  meCaption.Text := DestNodeClient.Caption;
end;

procedure TEditActionTreeView.actCopyFromActionUpdate(Sender: TObject);
begin
  if Sender is TAction then
    TAction(Sender).Enabled := (fTreeResult.Enabled) and (DestNodeAction <> nil) and (DestNodeClient <> nil) and
                               (DestNodeAction.Caption <> DestNodeClient.Caption);
end;

function TEditActionTreeView.ImagesEqual: boolean;
var Action: TCustomAction;
    Images: TCustomImageList;
begin
  result := false;
  Action := DestNodeAction;
  if Action <> nil then
  begin
    Images := nil;
    if Action.ActionList <> nil then
      Images := DestNodeAction.ActionList.Images;
    result := (Images = nil) or (Images = fTreeResult.Images);
  end;
end;

procedure TEditActionTreeView.actCopyToActionExecute(Sender: TObject);
begin
  if DestNodeAction <> nil then
  begin
    DestNodeAction.Caption := DestNodeClient.Caption;
    if ImagesEqual then
      DestNodeAction.ImageIndex := DestNodeClient.ImageIndex;
    fTreeResult.DoDelayChanged;
  end;
end;

procedure TEditActionTreeView.actCopyToActionUpdate(Sender: TObject);
begin
  if Sender is TAction then
    TAction(Sender).Enabled := (fTreeResult.Enabled) and (DestNodeAction <> nil) and (DestNodeClient <> nil) and
                               (fEnableActionEdit) and
                               ((DestNodeAction.Caption <> DestNodeClient.Caption) or
                                ((DestNodeAction.ImageIndex <> DestNodeClient.ImageIndex) and ImagesEqual));
  TAction(Sender).Visible := fEnableActionEdit;
end;

procedure TEditActionTreeView.actDelAllExecute(Sender: TObject);
begin
  fTreeResult.DropSelectedNode(True);
  fTreeResult.DoDelayChanged;
end;

procedure TEditActionTreeView.actDelSingleExecute(Sender: TObject);
begin
  fTreeResult.DropSelectedNode(false);
  fTreeResult.DoDelayChanged;
end;

procedure TEditActionTreeView.actEnabledExecute(Sender: TObject);
begin
  if DestNodeClient <> nil then
    DestNodeClient.Enabled := not DestNodeClient.Enabled;
end;

procedure TEditActionTreeView.actEnabledUpdate(Sender: TObject);
begin
  if Sender is TAction then
  begin
    TAction(Sender).Enabled := fTreeResult.Enabled and (DestNodeClient <> nil) and
                               (fEnableActionEdit or (DestNodeAction = nil));
    if (DestNodeClient <> nil) and DestNodeClient.Enabled then
      TAction(Sender).ImageIndex := 5
    else
      TAction(Sender).ImageIndex := 4;
  end;
end;

procedure TEditActionTreeView.actExpandedExecute(Sender: TObject);
begin
  if DestNodeClient <> nil then
    DestNodeClient.Expanded := not DestNodeClient.Expanded;
end;

procedure TEditActionTreeView.actExpandedUpdate(Sender: TObject);
begin
  if Sender is TAction then
  begin
    TAction(Sender).Enabled := fTreeResult.Enabled and (DestNodeClient <> nil);
    if (DestNodeClient <> nil) and DestNodeClient.Expanded then
      TAction(Sender).ImageIndex := 9
    else
      TAction(Sender).ImageIndex := 8;
  end;
end;

procedure TEditActionTreeView.actHideExecute(Sender: TObject);
begin
  if DestNodeClient <> nil then
    DestNodeClient.Visible := not DestNodeClient.Visible;
end;

procedure TEditActionTreeView.actHideUpdate(Sender: TObject);
begin
  if Sender is TAction then
  begin
    TAction(Sender).Enabled := fTreeResult.Enabled and (DestNodeClient <> nil) and
                               (fEnableActionEdit or (DestNodeAction = nil));
    if (DestNodeClient = nil) or DestNodeClient.Visible then
      TAction(Sender).ImageIndex := 7
    else
      TAction(Sender).ImageIndex := 6;
  end;
end;

procedure TEditActionTreeView.AddImages(List: TStringList);
var I: integer;
begin
  for I := 0 to List.Count - 1 do
  begin
    if List.Objects[I] is TCustomImageList then
    begin
      cbImages.Items.AddObject(List[I], List.Objects[I]);
    end;
  end;
end;

procedure TEditActionTreeView.AddSourceActions(List: TStringList);
var
  I, J: Integer;
  Name, FullName: string;
  Items: TNodeClients;
  Item: TNodeClient;
  function CreateItemByName(Name: string): boolean;
  var I: integer;
  begin
    Result := True;
    for I := Items.Count - 1 downto 0 do
    if Items[I].Caption = Name then
    begin
      Item := Items[I];
      Items := Item.Items;
      Result := False;
      Break;
    end;
    if Result then
    begin
      Item := TNodeClient(Items.Add);
      Items := Item.Items;
      Item.Caption := Name;
    end;
  end;
begin
  if List.Count > 0 then
  begin
    fTreeSource.BeginUpdate;
    try
      fTreeSource.NodeClients.BeginUpdate;
      try
        for I := 0 to List.Count - 1 do
        if List.Objects[I] is TCustomAction then
        begin
          Items := fTreeSource.NodeClients;
          FullName := List[I];
          repeat
            J := Pos('.', FullName);
            if J > 0 then
            begin
              Name := Copy(FullName, 1, J - 1);
              Delete(FullName, 1, J);
              if (CreateItemByName(Name)) and (Item is TSourceNodeClient) then
              begin
                TSourceNodeClient(Item).StateIndex := 3;
                TSourceNodeClient(Item).Expanded := False;
              end;
            end;
          until J = 0;
          if TCustomAction(List.Objects[I]).ActionList <> nil then
          begin
            Name := TCustomAction(List.Objects[I]).ActionList.Name;
            if CreateItemByName(Name) and (Item is TSourceNodeClient) then
            begin
              TSourceNodeClient(Item).StateIndex := 5;
              {$IFDEF DELPHI_6_UP}
              if TCustomAction(List.Objects[I]).ActionList is TActionManager then
                TSourceNodeClient(Item).StateIndex := 4;
              {$ENDIF}  
            end;
          end;
          if TCustomAction(List.Objects[I]).Category <> '' then
          begin
            if CreateItemByName(TCustomAction(List.Objects[I]).Category) and
               (Item is TSourceNodeClient) then
              TSourceNodeClient(Item).StateIndex := 6;
            Item.Expanded := False;
          end;
          if TCustomAction(List.Objects[I]).Caption <> '' then
            FullName := '  ' + FullName + ' (' + Copy(TCustomAction(List.Objects[I]).Caption, 1, 32) + ')';
          if CreateItemByName(FullName) and
             (Item is TSourceNodeClient) then
            TSourceNodeClient(Item).StateIndex := 7;
          Item.Assign(TCustomAction(List.Objects[I]));
        end;
      finally
        fTreeSource.NodeClients.EndUpdate;
      end;
      UpdateSourceImages;
    finally
      fTreeSource.EndUpdate;
    end;
  end;
end;

procedure TEditActionTreeView.AllignTop;
var
  W, H: Integer;
  R: TRect;
begin
  W := (pTop.ClientWidth - CenterWidth) div (2);
  fTreeSource.SetBounds(pTop.ClientWidth - W, 0, W, pTop.ClientHeight);
  fTreeResult.SetBounds(0, 0, W, pTop.ClientHeight);


  W := ToolBar1.Width - 330;
  if W < pToolBar.Constraints.MaxWidth then W := pToolBar.Constraints.MaxWidth;
  pToolBar.Width := W;

  W := pToolBar.ClientWidth;
  H := pToolBar.ClientHeight;
  R.Left := 6;
  R.Top := (H - meCaption.Height + 1) div 2;
  R.Right := W - cbImageIndex.Width - 6;
  R.Bottom := R.Top + meCaption.Height;
  meCaption.BoundsRect := R;
  R.Left := W - cbImageIndex.Width - 4;
  R.Top := (H - cbImageIndex.Height + 1) div 2;
  R.Right := R.Left + cbImageIndex.Width;
  R.Bottom := R.Top + cbImageIndex.Height;
  cbImageIndex.BoundsRect := R;
end;

procedure TEditActionTreeView.cbImagesChange(Sender: TObject);
var I: integer;
    B: TBitmap;
begin
  cbImageIndex.Items.Clear;
  cbImageIndex.Items.Add('-1');

  if (cbImages.ItemIndex >= 0) and
     (cbImages.Items.Objects[cbImages.ItemIndex] is TCustomImageList) then
  begin
    fTreeResult.Images := TCustomImageList(cbImages.Items.Objects[cbImages.ItemIndex]);
  end;

  fAllResultImages.CustomImg.Clear;
  if fTreeResult.Images <> nil then
  begin
    for I := 0 to fTreeResult.Images.Count - 1 do
      cbImageIndex.Items.Add(Inttostr(I));
    B := TBitmap.Create;
    try
      B.Width := fTreeResult.Images.Width;
      B.Height := fTreeResult.Images.Height;
      B.Canvas.Brush.Color := Color;
      for I := 0 to fTreeResult.Images.Count - 1 do
      begin
        B.Canvas.FillRect(Rect(0, 0, B.Width, B.Height));
        fTreeResult.Images.Draw(B.Canvas, 0, 0, I);
        fAllResultImages.CustomImg.InsertBitmap(I, B);
      end;
    finally
      B.Free;
    end;
  end;
  if fTreeResult.Selected <> nil then
    ResultChange(fTreeResult, fTreeResult.Selected);
end;

{ TSourceActionTreeView }

function TSourceActionTreeView.ActionNodeClass: TActionNodeClass;
begin
  Result := TSourceActionNode;
end;

function TSourceActionTreeView.NodeClientsItemClass: TNodeClientClass;
begin
  Result := TSourceNodeClient;
end;

procedure TSourceActionTreeView.DblClick;
begin
  //inherited;
end;

procedure TSourceActionTreeView.KeyPress(var Key: Char);
begin
  //inherited;
end;

procedure TSourceActionTreeView.DragOver(Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
begin
  inherited;
  Accept := False;
  MarkedNodeType := mntNone;
end;

{ TSourceNodeClient }

function TSourceNodeClient.ItemClass: TNodeClientClass;
begin
  Result := TSourceNodeClient;
end;

{ TSourceActionNode }

procedure TSourceActionNode.DrawImage(TreeView: TActionTreeView;
  State: TCustomDrawState; Image: TCustomImageList; Index: integer; R: TRect);
begin
  if (NodeClient is TSourceNodeClient) then
  begin
    if (Image = TreeView.StateImages) then
    begin
      State := State - [cdsDisabled];
      inherited DrawImage(TreeView, State, Image, TSourceNodeClient(NodeClient).StateIndex, R);
      if TSourceNodeClient(NodeClient).IsFound then
      begin
        State := State - [cdsChecked];
        OffsetRect(R, 8, 0);
        inherited DrawImage(TreeView, State, Image, 0, R);
      end;
    end
    else
    begin
      inherited DrawImage(TreeView, State, Image, Index, R);
    end;
  end;
end;

{ TResultActionTreeView }

constructor TResultActionTreeView.Create(AOwner: TComponent);
begin
  inherited;
  fIntTimer := TInternalTimer.Create(self);
  fIntTimer.ExecuteDelay(100, IntTimerProc);
end;

destructor TResultActionTreeView.Destroy;
begin
  FreeAndNil(fIntTimer);
  inherited;
end;

procedure TResultActionTreeView.DoDelayChanged;
begin
  fIntTimer.ExecuteDelay(100, IntTimerProc);
end;

function TResultActionTreeView.ActionNodeClass: TActionNodeClass;
begin
  Result := TResultActionNode;
end;

function TResultActionTreeView.NodeClientsItemClass: TNodeClientClass;
begin
  Result := TResultNodeClient;
end;

procedure TResultActionTreeView.DragDrop(Source: TObject; X, Y: Integer);
var Dest, Sourc: TNodeClient;
  Collection: TCollection;
  DelSource: boolean;
begin
  if (MarkedNodeType <> mntNone) and (Source is TSourceActionTreeView) then
  begin
    if (TargetNode <> nil) then
      Dest := TargetNode.NodeClient
    else
      Dest := nil;
    if Dest <> nil then
      Collection := Dest.Collection
    else
      Collection := NodeClients;

    if (TActionTreeView(Source).DragNode <> nil) and
       (HandleAllocated) then
    begin
      if Collection <> nil then Collection.BeginUpdate;
      try
        Sourc := TActionTreeView(Source).DragNode.NodeClient;
        DelSource := False;
        Dest := CopyNodeClient(Sourc,
                               Dest,
                               MarkedNodeType,
                               DelSource);
        if Dest <> nil then
          DoDelayChanged;
      finally
        if Collection <> nil then Collection.EndUpdate;
      end;
    end;
  end
  else
    inherited;
end;

procedure TResultActionTreeView.IntTimerProc(Sender: TObject);
begin
  if @fOnClientsChanged <> nil then fOnClientsChanged(self);
end;

procedure TResultActionTreeView.DblClick;
begin
  //inherited;
end;

procedure TResultActionTreeView.KeyPress(var Key: Char);
begin
  //inherited;
end;

{ TResultNodeClient }

procedure TResultNodeClient.AssignValues(Source: TPersistent);
begin
  inherited;
  if Action <> nil then
  begin
    if (Source is TSourceNodeClient) and
       (TSourceNodeClient(Source).Action <> nil) then
      Caption := TSourceNodeClient(Source).Action.Caption;
  end;
end;

function TResultNodeClient.ItemClass: TNodeClientClass;
begin
  Result := TResultNodeClient;
end;

{ TResultActionNode }

procedure TResultActionNode.DrawImage(TreeView: TActionTreeView;
  State: TCustomDrawState; Image: TCustomImageList; Index: integer; R: TRect);
begin
  inherited;

end;

end.


