unit ufSettings;
(*
 * Module:      ""    
 * Version:    1.1.0.0 (24.09.2008)
 * Author:     Cep Po
 *
 * Copyright:   2008 Cep Po
 *
 * Version:    1.1.0.0 (24.09.2008)
 * Realized:      
 *)
interface
{$I Directives.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Mask, TextUtlCep;

resourcestring
  CodePages = '1251| windows-1251| ANSI Cyrillic; Cyrillic (Windows)|' +
    '0|ANSI| All simbol no represented|' +
    '1200| Unicode| All simbol represented|' +
    '1250| windows-1250| ANSI Central European; Central European (Windows)|' +
    '1252| windows-1252| ANSI Latin 1; Western European (Windows)|' +
    '1253| windows-1253| ANSI Greek; Greek (Windows)|' +
    '1254| windows-1254| ANSI Turkish; Turkish (Windows)|' +
    '1255| windows-1255| ANSI Hebrew; Hebrew (Windows)|' +
    '1256| windows-1256| ANSI Arabic; Arabic (Windows)|' +
    '1257| windows-1257| ANSI Baltic; Baltic (Windows)|' +
    '1258| windows-1258| ANSI/OEM Vietnamese; Vietnamese (Windows)';
type
  TFormSettings = class(TForm)
    cbCodePage: TComboBox;
    Label1: TLabel;
    cbHex: TCheckBox;
    eFirstLeft: TMaskEdit;
    Label2: TLabel;
    GroupBox1: TGroupBox;
    Label3: TLabel;
    Label4: TLabel;
    eLeft: TMaskEdit;
    eRight: TMaskEdit;
    GroupBox2: TGroupBox;
    cbWantReturns: TCheckBox;
    cbWantTabs: TCheckBox;
    cbNoComments: TCheckBox;
    procedure cbCodePageChange(Sender: TObject);
    procedure cbHexClick(Sender: TObject);
    procedure eKeyPress(Sender: TObject; var Key: Char);
    procedure eFirstLeftChange(Sender: TObject);
    procedure eLeftChange(Sender: TObject);
    procedure eRightChange(Sender: TObject);
    procedure eExit(Sender: TObject);
    procedure cbWantClick(Sender: TObject);
    procedure cbNoCommentsClick(Sender: TObject);
  private
    { Private declarations }
    fListCodePages: TStringList;
    fonChanged: TNotifyevent;
    fFirstTextLeft: integer;
    fTextLeft: integer;
    fTextRight: integer;
    fUpdateCount: integer;
    function GetCodePage: Cardinal;
    procedure UpdateText;
    procedure SetFirstTextLeft(const Value: integer);
    procedure SetTextLeft(const Value: integer);
    procedure SetTextRight(const Value: integer);
    function SenderValue(Sender: TObject): integer;
    function GetWantTabs: boolean;
    procedure SetWantTabs(const Value: boolean);
    procedure SetCheckValue(ACheckBox: TCheckBox; const Value: Boolean);
    function GetWantReturns: boolean;
    procedure SetWantReturns(const Value: boolean);
    function GetNoComments: boolean;
    procedure SetNoComments(const Value: boolean);
    procedure DoBeep;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoChanged;
    procedure BeginUpdate;
    procedure EndUpdate;
    property CodePage: Cardinal read GetCodePage;
    property onChanged: TNotifyevent read fonChanged write fonChanged;
    property FirstTextLeft: integer read fFirstTextLeft write SetFirstTextLeft;
    property TextLeft: integer read fTextLeft write SetTextLeft;
    property TextRight: integer read fTextRight write SetTextRight;
    property WantTabs: boolean read GetWantTabs write SetWantTabs;
    property WantReturns: boolean read GetWantReturns write SetWantReturns;
    property NoComments: boolean read GetNoComments write SetNoComments;
  end;

implementation

{$R *.dfm}

{ TFormSettings }

procedure TFormSettings.cbHexClick(Sender: TObject);
begin
  DoChanged;
end;

procedure TFormSettings.cbCodePageChange(Sender: TObject);
begin
  cbCodePage.Hint := fListCodePages[cbCodePage.ItemIndex * 3 + 2];
  DoChanged;
end;

constructor TFormSettings.Create(AOwner: TComponent);
var I: integer;
begin
  inherited;
{$IFDEF DELPHI_6_UP}cbHex.WordWrap := true; {$ENDIF}
  fListCodePages := TStringList.Create;
  StringToList(CodePages, '|', fListCodePages);
  cbCodePage.Clear;
  for I := 0 to (fListCodePages.Count div 3) - 1 do
    cbCodePage.Items.AddObject(fListCodePages[I * 3 + 1], TObject(StrToInt(fListCodePages[I * 3])));
  cbCodePage.ItemIndex := 0;
  fFirstTextLeft := 0;
  fTextLeft := 0;
  fTextRight := 79;
  UpdateText;
  AutoSize := true;
end;

destructor TFormSettings.Destroy;
begin
  FreeAndNil(fListCodePages);
  inherited;
end;

procedure TFormSettings.DoChanged;
begin
  if (not (csUpdating in ComponentState)) and (@fonChanged <> nil) then fonChanged(self);
end;

function TFormSettings.GetCodePage: Cardinal;
begin
  result := Cardinal(cbCodePage.Items.Objects[cbCodePage.ItemIndex]);
end;

procedure TFormSettings.UpdateText;
begin
  eFirstLeft.Text := inttostr(fFirstTextLeft);
  eLeft.Text := inttostr(fTextLeft);
  eRight.Text := inttostr(fTextRight);
end;

procedure TFormSettings.SetFirstTextLeft(const Value: integer);
begin
  if (fFirstTextLeft <> Value) then
  begin
    if (Value <= 99) and (Value >= 0) then
    begin
      fFirstTextLeft := Value;
      DoChanged;
    end
    else
    begin
      DoBeep;
      Abort;
    end;
    UpdateText;
  end;
end;

procedure TFormSettings.SetTextLeft(const Value: integer);
begin
  if (fTextLeft <> Value) then
  begin
    if (Value <= fTextRight - 30) and (Value >= 0) then
    begin
      fTextLeft := Value;
      DoChanged;
    end
    else
    begin
      DoBeep;
      Abort;
    end;
    UpdateText;
  end;
end;

procedure TFormSettings.SetTextRight(const Value: integer);
begin
  if (fTextRight <> Value) then
  begin
    if (Value <= 199) and (Value >= 40) then
    begin
      fTextRight := Value;
      DoChanged;
    end
    else
    begin
      DoBeep;
      Abort;
    end;
    UpdateText;
  end;
end;

procedure TFormSettings.SetCheckValue(ACheckBox: TCheckBox; const Value: Boolean);
var
  oldClick: TNotifyEvent;
begin
  if Value <> ACheckBox.Checked then
  begin
    oldClick := ACheckBox.OnClick;
    ACheckBox.OnClick := nil;
    try
      ACheckBox.Checked := Value;
      DoChanged;
    finally
      ACheckBox.OnClick := oldClick;
    end;
  end;
end;

procedure TFormSettings.cbWantClick(Sender: TObject);
begin
  DoChanged;
end;

procedure TFormSettings.SetWantTabs(const Value: boolean);
begin
  SetCheckValue(cbWantTabs, Value);
end;

function TFormSettings.GetWantTabs: boolean;
begin
  result := cbWantTabs.Checked;
end;

procedure TFormSettings.SetWantReturns(const Value: boolean);
begin
  SetCheckValue(cbWantReturns, Value);
end;

function TFormSettings.GetWantReturns: boolean;
begin
  result := cbWantReturns.Checked;
end;

procedure TFormSettings.eKeyPress(Sender: TObject; var Key: Char);
begin
  if not ((Key = #8) or ((Key >= '0') and (Key <= '9'))) then
  begin
    Key := #0;
    MessageBeep(0);
  end;
end;

function TFormSettings.SenderValue(Sender: TObject): integer;
var V, Code: integer;
begin
  result := 0;
  if Sender is TMaskEdit then
  begin
    Val(TMaskEdit(Sender).Text, V, Code);
    if (V >= 0) and (Code = 0) then result := V
    else
    begin
      DoBeep;
      Abort;
    end;
  end
  else Abort;
end;

procedure TFormSettings.eFirstLeftChange(Sender: TObject);
begin
  FirstTextLeft := SenderValue(Sender);
end;

procedure TFormSettings.eExit(Sender: TObject);
begin
  UpdateText;
end;

procedure TFormSettings.eLeftChange(Sender: TObject);
begin
  TextLeft := SenderValue(Sender);
end;

procedure TFormSettings.EndUpdate;
begin
  if fUpdateCount = 1 then Updated;
  if fUpdateCount > 0 then Dec(fUpdateCount);
end;

procedure TFormSettings.DoBeep;
begin
  if not (csUpdating in componentState) then
    MessageBeep(0);
end;

procedure TFormSettings.BeginUpdate;
begin
  if fUpdateCount = 0 then Updating;
  inc(fUpdateCount);
end;

procedure TFormSettings.eRightChange(Sender: TObject);
begin
  TextRight := SenderValue(Sender);
end;

function TFormSettings.GetNoComments: boolean;
begin
  result := cbNoComments.Checked;
end;

procedure TFormSettings.SetNoComments(const Value: boolean);
begin
  SetCheckValue(cbNoComments, Value);
end;

procedure TFormSettings.cbNoCommentsClick(Sender: TObject);
begin
  DoChanged;
end;

end.
