unit ufSpecInst;
(*
 * Module:      " "
 * Version:    1.1.0.0 (24.09.2008)
 * Author:     Cep Po
 *
 * Copyright:   2008 Cep Po
 *
 * Version:    1.1.0.0 (24.09.2008)
 * Realized:     Delphi 2009
 *                
 *)
interface
{$I Directives.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, RICHEDIT, ImgList, ActnList, ExtCtrls, Buttons,
  {$IFDEF DELPHI_6_UP}{$ELSE}GraphUtilEx, {$ENDIF}
  ToolWin, Consts, ComStrs, ufSettings,
  SHFolder, IniFiles, Math, uBaseForm, ImageListCep, SpeedButtonCep,
  TextUtlCep;

resourcestring
  textAnd = '%s' + #13#10 + ' %s';
const
  FormSettingsName = 'FormSettings';

type

  TFormSpecIns = class(TBaseForm)
    Panel1: TPanel;
    Panel2: TPanel;
    RichEdit1: TRichEdit;
    Splitter1: TSplitter;
    RichEdit2: TRichEdit;
    actInsert: TAction;
    actCopyToCB: TAction;
    actPasteFromCB: TAction;
    actSettings: TAction;
    ToolBar1: TToolBar;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    procedure actCopyToCBExecute(Sender: TObject);
    procedure actPasteFromCBUpdate(Sender: TObject);
    procedure actPasteFromCBExecute(Sender: TObject);
    procedure actDefaultUpdate(Sender: TObject);
    procedure actCancelExecute(Sender: TObject);
    procedure actDefaultExecute(Sender: TObject);
    procedure actSettingsUpdate(Sender: TObject);
    procedure actSettingsExecute(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure actInsertUpdate(Sender: TObject);
    procedure actInsertExecute(Sender: TObject);
    procedure RichEdit1Change(Sender: TObject);
    procedure RichEdit2Change(Sender: TObject);
    procedure RichEdit1Enter(Sender: TObject);
    procedure RichEdit2Enter(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    fCodeChadge: boolean;
    fFormSettings: TFormSettings;
    fSelLength: integer;
    fSelStart: integer;
    fSelColumn: integer;
    fLiteralText: TLiteralText;
    fStripText: WideString;
    fTokenText: WideString;
    fonChangeSelection: TNotifyEvent;
    fonInsert: TNotifyEvent;
    fNeedUpdate: boolean;
    fInCode: boolean;
    fCopyAfterShow: boolean;
    fOldNoComments: boolean;
    fOldStart: integer;
    fOldEnd: integer;
    fOldOffset: integer;
    fOldStripText: WideString;
    fOldResult: boolean;
    fWait: TObject;

    function GetInCode: boolean;
    procedure SetInCode(const Value: boolean);
    procedure SettingsChange(Sender: TObject);
    function GetCodeText(const TextW: WideString): WideString;
    procedure SetSelLength(const Value: integer);
    procedure SetSelStart(const Value: integer);
    procedure SetLiteralText(const Value: TLiteralText);
    function UpdateSelectedText: boolean;
    procedure ChangeSelection; virtual;
    procedure UpdateInCode;
    function GetStripText(const Offset: Integer; var StartPos,
      EndPos: Integer; var S: WideString): boolean;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AfterConstruction; override;
    destructor Destroy; override;
    procedure DoShow; override;
    procedure DoInit;
    property InCode: boolean read GetInCode write SetInCode;
    property LiteralText: TLiteralText read fLiteralText write SetLiteralText;
    property SelStart: integer read fSelStart write SetSelStart;
    property SelLength: integer read fSelLength write SetSelLength;
    procedure SetSel(SelStart, SelLength: integer; Column: integer = 0) overload;
    property SelColumn: integer read fSelColumn;
    property CopyAfterShow: boolean read fCopyAfterShow write fCopyAfterShow;

    procedure BeginWait;
    procedure StopWait;

    property TokenText: WideString read fTokenText;
    property StripText: WideString read fStripText;
    property onChangeSelection: TNotifyEvent read fonChangeSelection write fonChangeSelection;
    property onInsert: TNotifyEvent read fonInsert write fonInsert;
  end;

implementation

{$R *.dfm}
uses WindowThread, CWindowThread;


{ TFormEditText }

procedure TFormSpecIns.ChangeSelection;
begin
  if @fonChangeSelection <> nil then
    fonChangeSelection(self);
end;


constructor TFormSpecIns.Create(AOwner: TComponent);
begin
  inherited;
  fOldOffset := -2;
  fFormSettings := TFormSettings.Create(self);
  fFormSettings.Name := FormSettingsName;
  RichEdit1.OnChange := RichEdit1Change;
  RichEdit2.OnChange := RichEdit2Change;
end;

procedure TFormSpecIns.AfterConstruction;
begin
  inherited;
  RichEdit1.OnChange := RichEdit1Change;
  RichEdit2.OnChange := RichEdit2Change;
  fFormSettings.OnChanged := SettingsChange;
end;

destructor TFormSpecIns.Destroy;
begin
  FreeAndNil(fWait);
  FreeAndNil(fLiteralText);
  inherited;
end;

const
  cLeft = 'Left';
  cTop = 'Top';
  cWidth = 'Width';
  cHeight = 'Height';
  cVisible = 'Visible';

procedure TFormSpecIns.FormClose(Sender: TObject;
  var Action: TCloseAction);
var
  Ini: TIniFile;
  procedure WriteRect(Ini: TIniFile; Form: TForm);
  var
    R: TRect;
  begin
    with Form do
    begin
      R := BoundsRect;
      Ini.WriteInteger(Name, cLeft, R.Left);
      Ini.WriteInteger(Name, cTop, R.Top);
      Ini.WriteInteger(Name, cWidth, R.Right - R.Left);
      Ini.WriteInteger(Name, cHeight, R.Bottom - R.Top);
    end;
  end;
begin
  BeginWait;
  try
    fOldOffset := -2;
    fOldStripText := '';
    if self.Visible then
    begin
      Ini := TIniFile.Create(IniFileName);
      try
        WriteRect(Ini, Self);
        WriteRect(Ini, fFormSettings);
        with fFormSettings do
        begin
          Ini.WriteBool(Name,
            cbHex.Name,
            cbHex.Checked);
          Ini.WriteBool(Name,
            cbNoComments.Name,
            cbNoComments.Checked);
          Ini.WriteInteger(Name,
            cbCodePage.Name,
            cbCodePage.ItemIndex);
          Ini.WriteInteger(Name, eRight.Name, TextRight);
          Ini.WriteBool(Name,
            cbWantReturns.Name,
            WantReturns);
          Ini.WriteBool(Name,
            cbWantTabs.Name,
            WantTabs);
          Ini.WriteBool(Name, cVisible, Visible);
          Ini.WriteInteger(Name, Panel2.Name + cHeight, Panel2.Height);
        end;
      finally
        FreeAndNil(Ini);
      end;
    end;
  finally
    StopWait;
  end;
  inherited;
  if fFormSettings <> nil then
    fFormSettings.Visible := false;
end;

procedure TFormSpecIns.DoInit;
var
  Ini: TIniFile;
  R: TRect;
  X1, X2: LongInt;
  W: WideString;
  H, MaxH: Integer;
  function ReadRect(Ini: TIniFile; Form: TForm): TRect;
  var
    W, H: integer;
  begin
    with Form do
    begin
      R.Left := Ini.ReadInteger(Name, cLeft, -1);
      R.Top := Ini.ReadInteger(Name, cTop, -1);
      W := Max(Ini.ReadInteger(Name, cWidth, Width), Constraints.MinWidth);
      H := Max(Ini.ReadInteger(Name, cHeight, Height), Constraints.MinHeight);
      if (R.Left = -1) or (R.Top = -1) then
      begin
        R.Left := (Screen.Width - W) div (2);
        R.Top := (Screen.Height - H) div (2);
      end;
      R.Right := R.Left + W;
      R.Bottom := R.Top + H;
      BoundsRect := R;
    end;
  end;
begin
  BeginWait;
  try
    fNeedUpdate := true;
    Panel1.Height := ClientHeight;
    fFormSettings.Caption := actSettings.Caption;
    try
      Ini := TIniFile.Create(IniFileName);
      try
        ReadRect(Ini, Self);
        ReadRect(Ini, fFormSettings);
        with fFormSettings do
        begin
          try
            cbCodePage.ItemIndex := Ini.ReadInteger(Name,
              cbCodePage.Name,
              cbCodePage.ItemIndex);
          except
            cbCodePage.ItemIndex := 0;
          end;
          cbHex.Checked := Ini.ReadBool(Name,
            cbHex.Name,
            cbHex.Checked);
          fOldNoComments := Ini.ReadBool(Name,
            cbNoComments.Name,
            cbNoComments.Checked);
          cbNoComments.Checked := fOldNoComments;
          TextRight := Ini.ReadInteger(Name, eRight.Name, TextRight);
          WantReturns := Ini.ReadBool(Name,
            cbWantReturns.Name,
            WantReturns);
          WantTabs := Ini.ReadBool(Name,
            cbWantTabs.Name,
            WantTabs);
          Visible := (self.Visible) and (Ini.ReadBool(Name, cVisible, false));
          if Visible then
            SetWindowPos(Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
          H := Ini.ReadInteger(Name, Panel2.Name + cHeight, Panel2.Height);
          MaxH := ((Panel1.ClientHeight * 3) div (4)) - Panel2.Top;
          if MaxH < 0 then
            MaxH := 0;
          if H > MaxH then
            H := MaxH;
          Panel2.Height := H;
        end;
      finally
        FreeAndNil(Ini);
      end;
    except
      //    ,     
    end;
    UpdateInCode;
    fCodeChadge := true;
    try
      W := '';
      if (fSelStart >= 0) and (fSelLength > 0) then
        GetStripText(fSelStart, X1, X2, W)
      else
        RichEdit1.TextW := '';
      RichEdit2.TextW := W;
      RichEdit2.TextW := RichEdit2.TextW;
      fStripText := W;
    finally
      fCodeChadge := false;
    end;
    if CopyAfterShow then
    begin
      if InCode then
        actPasteFromCB.Execute
      else
        actCopyToCB.Execute;
    end;
  finally
    StopWait;
  end;
end;

procedure TFormSpecIns.DoShow;
begin
  inherited;
  DoInit;
end;

function TFormSpecIns.GetCodeText(const TextW: WideString): WideString;
begin
  if fFormSettings.WantReturns then
    actDefault.ShortCut := 10
  else
    actDefault.ShortCut := 13;
  RichEdit1.WantTabs := fFormSettings.WantTabs;
  RichEdit1.WantReturns := fFormSettings.WantReturns;
  RichEdit2.WantTabs := fFormSettings.WantTabs;
  RichEdit2.WantReturns := fFormSettings.WantReturns;
  result := fLiteralText.MakeText(TextW,
    fFormSettings.cbHex.Checked,
    fFormSettings.FirstTextLeft,
    fFormSettings.TextLeft,
    fFormSettings.TextRight,
    fFormSettings.CodePage,
    fFormSettings.NoComments);
  //           
  fTokenText := Trim(Result);
  if (fLiteralText <> nil) and (fLiteralText.ConcateChar <> '') then
    fLiteralText.TruncRight(Result);
end;

procedure TFormSpecIns.RichEdit1Change(Sender: TObject);
begin
  if InCode then
  begin
    //           
    fTokenText := Trim(RichEdit1.TextW);
    if (fLiteralText <> nil) and (fLiteralText.ConcateChar <> '') then
      fLiteralText.TruncRight(fTokenText);
    Timer1.Enabled := True;
  end;
end;

procedure TFormSpecIns.RichEdit1Enter(Sender: TObject);
begin
  InCode := true;
end;

procedure TFormSpecIns.RichEdit2Change(Sender: TObject);
begin
  if fCodeChadge then
    Exit;
  Timer1.Enabled := True;
end;

procedure TFormSpecIns.RichEdit2Enter(Sender: TObject);
begin
  InCode := False;
end;

procedure TFormSpecIns.Timer1Timer(Sender: TObject);
var
  LT: TLiteralText;
  X1, X2: LongInt;
  W, CodeText: WideString;
begin
  if Timer1 <> nil then
    Timer1.Enabled := False;
  if fLiteralText = nil then
    Exit;

  if InCode then
  begin
    LT := nil;
    fCodeChadge := true;
    try
      W := RichEdit1.TextW;
      LT := TLiteralText.Create(PWideChar(W));
      LT.CodePage := fLiteralText.CodePage;
      CodeText := '';
      LT.GetStripText(0, X1, X2, CodeText);
      RichEdit2.TextW := CodeText;
      RichEdit2.TextW := RichEdit2.TextW;
      fStripText := CodeText;
    finally
      fCodeChadge := false;
      FreeAndNil(LT);
    end;
  end
  else
  begin
    RichEdit1.TextW := GetCodeText(RichEdit2.TextW);
  end;
end;

procedure TFormSpecIns.SettingsChange(Sender: TObject);
begin
  RichEdit1.RightLine := fFormSettings.TextRight;
  RichEdit1.LeftLine := fFormSettings.TextLeft;
  RichEdit1.FirstLeftLine := fFormSettings.FirstTextLeft;
  if fOldNoComments <> fFormSettings.NoComments then
  begin
    fOldNoComments := fFormSettings.NoComments;
    RichEdit1.TextW := GetCodeText(RichEdit2.TextW);
  end;
  if fCodeChadge then
    Exit;
  RichEdit1.TextW := GetCodeText(RichEdit2.TextW);
end;

function TFormSpecIns.GetInCode: boolean;
begin
  result := fInCode;
end;

procedure TFormSpecIns.UpdateInCode;
begin
  if Visible then
  begin
    if fInCode then
    begin
      if RichEdit1.Enabled then
        RichEdit1.SetFocus;
    end
    else
    begin
      if RichEdit2.Enabled then
        RichEdit2.SetFocus;
    end;
  end;
end;

procedure TFormSpecIns.SetInCode(const Value: boolean);
begin
  fInCode := Value;
  fNeedUpdate := true;
  UpdateInCode;
end;

function TFormSpecIns.GetStripText(const Offset: Integer; var StartPos,
  EndPos: Integer; var S: WideString): boolean;
begin
  result := false;
  if (fLiteralText <> nil) then
  begin
    if (Offset <> fOldOffset) then
    begin
      result := fLiteralText.GetStripText(Offset, StartPos, EndPos, S);
      fOldStart := StartPos;
      fOldEnd := EndPos;
      fOldOffset := Offset;
      fOldStripText := S;
      fOldResult := result;
    end
    else
    begin
      StartPos := fOldStart;
      EndPos := fOldEnd;
      S := fOldStripText;
      result := fOldResult;
    end;
  end
  else
  begin
    fOldStart := 0;
    fOldEnd := 0;
    fOldOffset := -1;
    fOldStripText := '';
    fOldResult := false;
  end;
end;

function TFormSpecIns.UpdateSelectedText: boolean;
var
  X1, X2, FirstLeft, Left, Right: Longint;
  W: WideString;
  OldSelStart, OldSelLength: integer;
begin
  result := false;
  if fLiteralText <> nil then
  begin
    OldSelStart := fSelStart;
    OldSelLength := fSelLength;
    try
      X1 := fSelStart;
      if not fLiteralText.InToken(SelStart, X1, X2) then
      begin
        X2 := SelStart;
        X1 := fLiteralText.FindNext(X2);
      end;
      W := '';
      if (X1 >= 0) and ((X1 < SelStart + SelLength) and (SelStart >= 0)) then
      begin
        Result := true;
        fSelStart := X1;
        fSelLength := X2 - X1;
        SetLength(W, fSelLength);
        Move(fLiteralText.WideText[fSelStart], W[1], fSelLength * SizeOf(W[1]));
        fLiteralText.CodePage := fFormSettings.CodePage;
        fLiteralText.DefaultHex := fFormSettings.cbHex.Checked;
        fLiteralText.RightMargin := fFormSettings.TextRight;
        GetStripText(SelStart, X1, X2, W);
      end;
      fLiteralText.Column := fSelColumn;
      if fLiteralText.GetMargins(fSelStart, fSelStart + fSelLength, FirstLeft, Left, Right) then
        fSelColumn := 0;
      fFormSettings.FirstTextLeft := FirstLeft;
      fFormSettings.TextLeft := Left;
      fCodeChadge := True;
      try
        RichEdit2.TextW := W;
        RichEdit2.TextW := RichEdit2.TextW;
        RichEdit1.TextW := GetCodeText(W);
        RichEdit1.TextW := RichEdit1.TextW;
        fStripText := W;
        Timer1.Enabled := False;
      finally
        fCodeChadge := False;
      end;
    finally
      //MessageBox(0, 'UpdateSelectedText' ,nil, 0);
      if (fSelStart <> OldSelStart) or (fSelLength <> OldSelLength) then
        ChangeSelection;
    end;
  end;
end;

procedure TFormSpecIns.SetLiteralText(const Value: TLiteralText);
var
  OldSelStart, OldSelLength: integer;
begin
  if Value <> fLiteralText then
  begin
    fLiteralText := Value;
    if fLiteralText <> nil then
    begin
      OldSelStart := fSelStart;
      OldSelLength := fSelLength;
      fSelStart := -1;
      fSelLength := 0;
      if (fSelStart <> OldSelStart) or (fSelLength <> OldSelLength) then
        ChangeSelection;
    end;
  end;
end;

procedure TFormSpecIns.SetSel(SelStart, SelLength: integer; Column: integer = 0);
begin
  if (fSelStart <> SelStart) or (fSelLength <> SelLength) then
  begin
    fSelStart := SelStart;
    fSelLength := SelLength;
    if fSelLength = 0 then
      fSelColumn := Column
    else
      fSelColumn := 0;
    UpdateSelectedText;
  end;
end;

procedure TFormSpecIns.SetSelLength(const Value: integer);
begin
  fSelLength := Value;
  UpdateSelectedText;
end;

procedure TFormSpecIns.SetSelStart(const Value: integer);
begin
  fSelStart := Value;
  UpdateSelectedText;
end;

procedure TFormSpecIns.actPasteFromCBUpdate(Sender: TObject);
var
  Formats: TClipboardFormats;
  Need: boolean;
begin
  Need := (not RichEdit2.ReadOnly) and
    (RichEdit2.Enabled) and
    (RichEdit2.Visible);
  if Need then
  begin
    Need := OpenClipBoard(RichEdit2.Handle);
    if Need then
    try
      Need := (RichEdit2.IsEnablePaste(Formats) <> 0);
    finally
      CloseClipboard;
    end;
  end;
  actPasteFromCB.Enabled := Need;
end;

procedure TFormSpecIns.actSettingsExecute(Sender: TObject);
begin
  fFormSettings.Visible := not fFormSettings.Visible;
  if fFormSettings.Visible then
    SetWindowPos(fFormSettings.Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
end;

procedure TFormSpecIns.actSettingsUpdate(Sender: TObject);
begin
  actSettings.Checked := (fFormSettings <> nil) and (fFormSettings.Visible);
end;

procedure TFormSpecIns.actDefaultExecute(Sender: TObject);
begin
  if Timer1.Enabled then
    Timer1Timer(nil);
  if (RichEdit1.Focused) or ((not Visible) and (fInCode)) then
  begin
    if actInsert.Enabled then
    begin
      actInsert.Execute;
      Close;
      ModalResult := IdOk;
    end;
  end
  else
  begin
    if actCopyToCB.Enabled then
    begin
      actCopyToCB.Execute;
      Close;
      ModalResult := IdOk;
    end;
  end;
end;

procedure TFormSpecIns.actDefaultUpdate(Sender: TObject);
begin
  if (fNeedUpdate) then
  begin
    if fInCode then
    begin
      RichEdit1.Color := clWindow;
      RichEdit2.Color := clBtnFace;
      actDefault.Enabled := actInsert.Enabled;
      actDefault.ImageIndex := 0;
      SpeedButton1.Gliphs.ImageIndexNormalHot := 0;
      actDefault.Caption := actInsert.Caption;
      actDefault.Hint := Format(textAnd, [actInsert.Hint, ansiLowerCase(actCancel.Hint)]);
    end
    else
    begin
      RichEdit1.Color := clBtnFace;
      RichEdit2.Color := clWindow;
      actDefault.Enabled := actCopyToCB.Enabled;
      actDefault.ImageIndex := 1;
      SpeedButton1.Gliphs.ImageIndexNormalHot := 1;
      actDefault.Caption := actCopyToCB.Caption;
      actDefault.Hint := Format(textAnd, [actCopyToCB.Hint, ansiLowerCase(actCancel.Hint)]);
    end;
    {$IFDEF DELPHI_2006_UP}
    ToolBar1.DrawingStyle := ComCtrls.dsGradient;
    ToolBar1.GradientStartColor := Color;
    ToolBar1.GradientEndColor := RichEdit1.Color;
    {$ENDIF}
  end;
end;

procedure TFormSpecIns.actInsertExecute(Sender: TObject);
begin
  ChangeSelection;
  if @fonInsert <> nil then
    fonInsert(self);
end;

procedure TFormSpecIns.actInsertUpdate(Sender: TObject);
begin
  actInsert.Enabled := (@fonInsert <> nil) and
    (fLiteralText <> nil) and
    (fSelStart >= 0) and
    (fSelLength >= 0);
end;

procedure TFormSpecIns.actPasteFromCBExecute(Sender: TObject);
begin
  fCodeChadge := true;
  try
    RichEdit2.PasteAll;
    if fLiteralText <> nil then
    begin
      RichEdit1.TextW := GetCodeText(RichEdit2.TextW);
    end;
    Timer1.Enabled := false;
  finally
    fCodeChadge := false;
  end;
end;

procedure TFormSpecIns.actCancelExecute(Sender: TObject);
begin
  Close;
  ModalResult := idCancel;
end;

procedure TFormSpecIns.actCopyToCBExecute(Sender: TObject);
begin
  RichEdit2.CopyAll;
end;

procedure TFormSpecIns.BeginWait;
const
  Width = 250;
  Height = 48;
var
  R: TRect;
  FontInfo: tagLOGFONT;
  HFont: THandle;
begin
  if fWait = nil then
  begin
    fWait := TThreadWindow.Show('');
    with TThreadWindow(fWait) do
    begin
      AlphaShow := true;
      FontInfo := GetFontIndirect(0, dStatusFont);
      HFont := CreateFontIndirect(FontInfo);
      try
        R := BoundsRect;
        R.Left := (Screen.Width - Width) div (2);
        R.Top := (Screen.Height - Height) * 2 div (5);
        R.Bottom := R.Top + Height;
        R.Right := R.Left + Width;
        Font := HFont;
        BoundsRect := R;
      finally
        DeleteObject(HFont);
      end;
    end;
  end;
end;

procedure TFormSpecIns.StopWait;
begin
  FreeAndNil(fWait);
end;

end.

