unit ufWixGenerator;

interface

uses
  Windows, Messages, SysUtils,
  {$IFDEF VER130}
  OldVersions,
  {$ELSE}
  Variants,
  {$ENDIF}
  SHFolder,
  Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ActnList, ImgList, ImageListCep,
  SpeedButtonCep, uWixGenerator, IniFiles, ComCtrls, WindowThread, uBaseForm;
resourcestring
  DefCaption = '  ';
  WaitText = '  ';
  OkText = '   ';
  ErrorText = '    ';
type
  TfrmWixGenerator = class(TBaseForm)
    Label1: TLabel;
    cbWixSource: TComboBox;
    SpeedButtonCep1: TSpeedButtonCep;
    rgMakePart: TRadioGroup;
    eExeName: TEdit;
    Label2: TLabel;
    GroupBox1: TGroupBox;
    cbStrartup: TCheckBox;
    Label3: TLabel;
    eGUID: TEdit;
    eVersion: TEdit;
    Label4: TLabel;
    actFind: TAction;
    ImageListCep1: TImageListCep;
    bMake: TButton;
    actMake: TAction;
    cbMidas: TCheckBox;
    cbMidasSource: TComboBox;
    SpeedButtonCep2: TSpeedButtonCep;
    Label5: TLabel;
    lStatus: TLabel;
    procedure actFindExecute(Sender: TObject);
    procedure actMakeUpdate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure actMakeExecute(Sender: TObject);
    procedure ActionList1Update(Action: TBasicAction; var Handled: Boolean);
  private
    { Private declarations }
    fList: TStringList;
    fExeName: string;
    fWixGenerator: TCustomWixGenerator;
    fStream: TStream;
    fSoGood: boolean;
    function onFunc(const FileName: string; const FileTime: TFileTime; const Size: int64): boolean;
    procedure SetExeName(const Value: string);
    procedure SetWixGenerator(const Value: TCustomWixGenerator);
    procedure SetWixSource(const Value: string);
    function GetWixSource: string;
    function GetMidasSource: string;
    procedure SetMidasSource(const Value: string);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoSaveToIni(Ini: TIniFile); override;
    procedure DoLoadFromIni(Ini: TIniFile); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure UpdateProps;
    property ExeName: string read fExeName write SetExeName;
    property WixGenerator: TCustomWixGenerator read fWixGenerator write SetWixGenerator;
    property WixSource: string read GetWixSource write SetWixSource;
    property MidasSource: string read GetMidasSource write SetMidasSource;
    property Stream: TStream read fStream write fStream;
    property SoGood: boolean read fSoGood;
  end;

var
  frmWixGenerator: TfrmWixGenerator;

implementation
uses CepUtils;

{$R *.dfm}


procedure TfrmWixGenerator.actFindExecute(Sender: TObject);
var FolderName, AltFolderName: string;
  function GetFolderName(CSIDL: integer): string;
  begin
    Result := '';
    SetLength(Result, MAX_PATH);
    FillChar(Result[1], SizeOf(Result[1]) * MAX_PATH, 0);
    SHGetFolderPath(0, CSIDL, 0, 0, PChar(Result));
    SetLength(Result, Length(PChar(Result)));
  end;
  procedure ClearList(AltFileName: string);
  var I, j: integer;
  begin
    I := 0;
    while I < fList.Count do
    begin
      if (AltFileName <> '') and (not FileExists(IncludeTrailingPathDelimiter(fList[I]) + AltFileName)) then
        fList.Delete(I)
      else
      begin
        for j := fList.Count - 1 downto i do
          if (j <> i) and (AnsiLowercase(fList[i]) = AnsiLowercase(fList[j])) then fList.Delete(j);
        inc(I);
      end
    end;
  end;
  procedure DoFind(cb: TComboBox; FileName, AltFileName, DirName, AltDirName: string);
  var S: string;
      i: integer;
  begin
      S := ExcludeTrailingPathDelimiter(Trim(cb.Text));
      try
        fList.Assign(cb.Items);
        if (cb.ItemIndex = -1) and (S <> '') then
          fList.Insert(0, S);

        if EnumerateFiles(DirName, FileName, faAnyFile, onFunc) then
        begin
          ClearList(AltFileName);
          if (fList.Count = 0) and (AltDirName <> '') then
          begin
            EnumerateFiles(AltDirName, FileName, faAnyFile, onFunc);
            ClearList(AltFileName);
          end;
        end;
        cb.Items := fList;
      finally
        cb.ItemIndex := -1;
        for I := 0 to fList.Count - 1 do
          if AnsiUpperCase(ExcludeTrailingPathDelimiter(fList[I])) = AnsiUpperCase(S) then
          begin
            cb.ItemIndex := I;
            break;
          end;
        if (cb.ItemIndex = -1) and (fList.Count > 0) then
        begin
          cb.Text := fList[fList.Count - 1];
          cb.ItemIndex := fList.Count - 1;
        end;
      end;
  end;
begin
  fList := TStringList.Create;
  try
    if cbWixSource.Focused then
    begin
      FolderName := GetFolderName(CSIDL_PROGRAM_FILES);
      DoFind(cbWixSource, exe2, exe1, FolderName, ExtractFileDrive(FolderName));
    end;
    if cbMidasSource.Focused then
    begin
      FolderName := GetFolderName(CSIDL_WINDOWS);
      AltFolderName := GetFolderName(CSIDL_PROGRAM_FILES);
      DoFind(cbMidasSource, midas, '', FolderName, '');
      DoFind(cbMidasSource, midas, '', AltFolderName, '');
    end;
  finally
    fList.Free;
  end;
end;

constructor TfrmWixGenerator.Create(AOwner: TComponent);
begin
  inherited;
  ExeName := ParamStr(0);
  UpdateProps;
  bMake.Top := ClientHeight - bMake.Height;
  lStatus.Caption := '';
end;

function TfrmWixGenerator.onFunc(const FileName: string;
  const FileTime: TFileTime; const Size: int64): boolean;
begin
  fList.Add(ExtractFileDir(FileName));
  result := True;
end;

procedure TfrmWixGenerator.SetExeName(const Value: string);
begin
  if fExeName <> Value then
  begin
    fExeName := Value;
    eExeName.Text := fExeName;
  end;
end;

function TfrmWixGenerator.GetMidasSource: string;
begin
  result := ExcludeTrailingPathDelimiter(Trim(cbMidasSource.Text))
end;

procedure TfrmWixGenerator.SetMidasSource(const Value: string);
var S: string;
    OldChange: TNotifyEvent;
    i: integer;
begin
  OldChange := cbMidasSource.OnChange;
  try
    cbMidasSource.OnChange := nil;
    S := ExcludeTrailingPathDelimiter(Trim(Value));
    if S <> '' then
      for i := 0 to cbMidasSource.Items.Count - 1 do
        if AnsiLowerCase(S) = AnsiLowerCase(cbMidasSource.Items[i]) then
        begin
          cbMidasSource.ItemIndex := i;
          break;
        end;
    if cbMidasSource.ItemIndex = -1 then cbMidasSource.ItemIndex := cbMidasSource.Items.Count - 1;
    if cbMidasSource.ItemIndex = -1 then cbMidasSource.Text := S;
  finally
    cbMidasSource.OnChange := OldChange;
  end;
end;

function TfrmWixGenerator.GetWixSource: string;
begin
  result := ExcludeTrailingPathDelimiter(Trim(cbWixSource.Text))
end;

procedure TfrmWixGenerator.SetWixSource(const Value: string);
var S: string;
    OldChange: TNotifyEvent;
    i: integer;
begin
  OldChange := cbWixSource.OnChange;
  try
    cbWixSource.OnChange := nil;
    S := ExcludeTrailingPathDelimiter(Trim(Value));
    if S <> '' then
      for i := 0 to cbWixSource.Items.Count - 1 do
        if AnsiLowerCase(S) = AnsiLowerCase(ExcludeTrailingPathDelimiter(Trim(cbWixSource.Items[i]))) then
        begin
          cbWixSource.ItemIndex := i;
          break;
        end;
    if cbWixSource.ItemIndex = -1 then cbWixSource.ItemIndex := cbWixSource.Items.Count - 1;
    if cbWixSource.ItemIndex = -1 then cbWixSource.Text := S;
  finally
    cbWixSource.OnChange := OldChange;
  end;
end;

procedure TfrmWixGenerator.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then
  begin
    if AComponent = fWixGenerator then
    begin
      fWixGenerator := nil;
      UpdateProps;
    end;
  end;
end;

procedure TfrmWixGenerator.SetWixGenerator(
  const Value: TCustomWixGenerator);
begin
  if fWixGenerator <> Value then
  begin
    if fWixGenerator <> nil then
      fWixGenerator.RemoveFreeNotification(self);
    fWixGenerator := Value;
    if fWixGenerator <> nil then
      fWixGenerator.FreeNotification(self);
    UpdateProps;
  end;
end;

procedure TfrmWixGenerator.UpdateProps;
var I: integer;
begin
  for I := 0 to GroupBox1.ControlCount - 1 do
    GroupBox1.Controls[I].Enabled := fWixGenerator <> nil;
  if fWixGenerator = nil then
  begin
    cbStrartup.Enabled := False;
    rgMakePart.Enabled := False;
    cbMidas.Enabled := False;
    eGUID.Text := '';
    eVersion.Text := '';
    Caption := DefCaption;
  end
  else
  begin
    ExeName := fWixGenerator.ExeName;
    if fWixGenerator.WixSource <> '' then
      WixSource := fWixGenerator.WixSource;
    if fWixGenerator.MidasSource <> '' then
      MidasSource := fWixGenerator.MidasSource;
    eGUID.Text := GuidToString(fWixGenerator.GUID);
    eVersion.Text := fWixGenerator.Version;
    cbStrartup.Enabled := True;
    rgMakePart.Enabled := True;
    cbMidas.Enabled := True;
    if fWixGenerator.Name <> '' then
      Caption := DefCaption + ' [' + fWixGenerator.Name + ']'
    else
      Caption := DefCaption;
  end;
end;

procedure TfrmWixGenerator.DoLoadFromIni(Ini: TIniFile);
begin
  inherited;
  LoadItems(Ini, Name, cbWixSource.Name, cbWixSource.Items);
  LoadItems(Ini, Name, cbMidasSource.Name, cbMidasSource.Items);
  cbWixSource.Text := Ini.ReadString(Name, cbWixSource.Name, cbWixSource.Text);
  cbMidasSource.Text := Ini.ReadString(Name, cbMidasSource.Name, cbMidasSource.Text);
  cbStrartup.Checked := Ini.ReadBool(Name, cbStrartup.Name, cbStrartup.Checked);
  cbMidas.Checked := Ini.ReadBool(Name, cbMidas.Name, cbMidas.Checked);
  SpeedButtonCep1.Left := ClientWidth - SpeedButtonCep1.Width;
  cbWixSource.Width := SpeedButtonCep1.Left - cbWixSource.Left;
  SpeedButtonCep2.Left := ClientWidth - SpeedButtonCep2.Width;
  cbMidasSource.Width := ClientWidth - SpeedButtonCep2.Width;
  bMake.Left := ClientWidth - bMake.Width;
  GroupBox1.Width := ClientWidth - GroupBox1.Left;
  rgMakePart.Width := ClientWidth - rgMakePart.Left;
  eExeName.Width := ClientWidth - eExeName.Left;
end;

procedure TfrmWixGenerator.DoSaveToIni(Ini: TIniFile);
begin
  inherited;
  Ini.WriteBool(Name, cbStrartup.Name, cbStrartup.Checked);
  Ini.WriteString(Name, cbWixSource.Name, cbWixSource.Text);
  Ini.WriteString(Name, cbMidasSource.Name, cbMidasSource.Text);
  SaveItems(Ini, Name, cbWixSource.Name, cbWixSource.Items);
  SaveItems(Ini, Name, cbMidasSource.Name, cbMidasSource.Items);
end;

procedure TfrmWixGenerator.actMakeUpdate(Sender: TObject);
begin
  inherited;
  actMake.Enabled := (WixSource <> '') and (WixGenerator <> nil);
end;

procedure TfrmWixGenerator.ActionList1Update(Action: TBasicAction; var Handled: Boolean);
begin
  inherited;
  if Action = actFind then
  begin
    cbStrartup.Enabled := rgMakePart.ItemIndex > 0;
    cbMidas.Enabled := rgMakePart.ItemIndex > 0;
  end;
end;

procedure TfrmWixGenerator.actMakeExecute(Sender: TObject);
var WS: WideString;
    OldCursor: TCursor;
begin
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    fSoGood := False;
    lStatus.Caption := WaitText;
    try
      lStatus.Repaint;
      WixGenerator.WixSource := WixSource;
      WixGenerator.MidasSource := MidasSource;
      if Stream <> nil then
      begin
        WS := 'I. ' + WaitText + #13#10;
        if Stream.Size = 0 then
          Stream.Write(SignW, SizeOf(SignW))
        else
          WS := #13#10 + WS;
        Stream.Write(WS[1], SizeOf(WS[1]) * Length(WS));
      end;
      if not FileExists(ChangeFileExt(fExeName, '.' + DefaultFileExt)) then
      begin
        rgMakePart.ItemIndex := 2;
      end;
      Repaint;
      Enabled := False;
      try
        case rgMakePart.ItemIndex of
          0: begin
                WixGenerator.RefreshByWix(True);
                WixGenerator.RefreshByVersionInfo(False);
             end;
          1: begin
                WixGenerator.RefreshByWix(True);
                WixGenerator.RefreshByVersionInfo(True);
             end;
          2: begin
                if fWixGenerator <> nil then fWixGenerator.RefreshByVersionInfo(True);
             end;
        else
        end;
        WixGenerator.WixSource := WixSource;
        WixGenerator.MidasSource := MidasSource;
        WixGenerator.Strartup := cbStrartup.Checked;
        WixGenerator.Midas := cbMidas.Checked;
        try
          case rgMakePart.ItemIndex of
            0: WixGenerator.Save(mpVersion);
            1: WixGenerator.Save(mpPropertyes);
            2: WixGenerator.Save(mpAll);
          else
            Exit;
          end;
          UpdateProps;
          Repaint;
          fSoGood := WixGenerator.MakeMSI(Stream);
          if (cbWixSource.ItemIndex = -1) and (Trim(cbWixSource.Text) <> '') then
          begin
            cbWixSource.Items.Insert(0, Trim(cbWixSource.Text));
            cbWixSource.ItemIndex := 0;
          end;
        finally
        end;
      finally
        Enabled := True;
      end;
    finally
      if SoGood then WS := OkText
                else WS := ErrorText;
      lStatus.Caption := WS;
      if Stream <> nil then
      begin
        WS := 'I. ' + WS + #13#10;
        Stream.Write(WS[1], SizeOf(WS[1]) * Length(WS));
      end;
    end;
  finally
    Screen.Cursor := OldCursor;
  end;
end;

procedure TfrmWixGenerator.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  inherited;
  SaveToIni;
end;

end.
